#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include "helpers.h"
#include <sicstus/sicstus.h>
#include "lib/include/zmq.h"
#include "lib/include/czmq.h"


static SP_pred_ref init_predicate(const char *name, int arity, const char *module);

SP_pred_ref get_successors;
SP_pred_ref next_update;
SP_pred_ref get_init;
SP_pred_ref get_state_label;
SP_pred_ref get_state_label_group;


zctx_t *ctx;


static SP_pred_ref init_predicate(const char *name, int arity, const char *module) {
    SP_pred_ref pred = SP_predicate(name,arity,module);
    if (pred == NULL) {
        fprintf(stderr, "Unable to find callback module %s:%s/%i for ltsmin, exiting.\n",
                module, name, arity);
        exit(1);
    }
    return pred;
}

void call_prolog(SP_pred_ref query, zmsg_t *req, zmsg_t *response) {
    char *f = zmsg_popstr(req); // transition group or state label
    SP_term_ref name = SP_new_term_ref();
    SP_put_address(name,f);

    State state = get_state(req);

    SP_term_ref tail;
    SP_term_ref list;
    tail = SP_new_term_ref();
    list = SP_new_term_ref();

    int i;
    for(i=state.size; i > 0; i--) {
        SP_term_ref val = SP_new_term_ref();
        assert(state.chunks[i-1].size > 0);
        if (state.chunks[i-1].data[0] != 'D') {
            puts("encountered weird frame:");
            int j;
            for (j = 0; j < state.chunks[i-1].size; j++) {
                printf("%d ", state.chunks[i-1].data[j]);
            }
            puts("");
            for (j = 0; j < state.chunks[i-1].size; j++) {
                printf("%c", state.chunks[i-1].data[j]);
            }
            puts("");
            assert(0);
        }

        SP_put_address(val, state.chunks[i-1].data);
        int rc = SP_cons_list(list, val, tail);
        assert(rc!=0);
        tail = list;
    }

    // Call Prolog
    //fprintf(stderr, "Calling %s.\n", query);
    SP_term_ref msg_ptr = SP_new_term_ref();
    SP_put_address(msg_ptr, response);
    int status = SP_query_cut_fail(query, msg_ptr, name, list);
    assert(status == SP_SUCCESS);

    // Free resource
    free(f);
    destroy_state(&state);
}


void init(zmsg_t *req, zmsg_t *rep) {
    SP_term_ref msg_ptr = SP_new_term_ref();
    SP_put_address(msg_ptr, rep);

    SP_term_ref por = SP_new_term_ref();
    SP_put_integer(por, 0);
    if (zmsg_size(req)) {
        char *s = zmsg_popstr(req);
        int is_por = atoi(s);
        free(s);
        SP_put_integer(por, is_por);
    }

    int status = SP_query_cut_fail(get_init, msg_ptr, por);
    assert(status == SP_SUCCESS);
}

static int count_next_state = 0;
static int count_next_action = 0;
static int count_state_label = 0;
static int count_state_label_group = 0;


int h_zocket2(void *zocket) {
    zmsg_t *request = zmsg_recv(zocket);
    zmsg_t *response = zmsg_new();

    // echo back the message type
    char *type = zmsg_popstr(request);
    //printf("Received Request: %s\n",type);


    zmsg_addstr(response,type);

    // echo back the id
    zframe_t *f = zmsg_pop(request);
    zmsg_append(response, &f);

    if(strcmp(type, "get-next-state") == 0) {
        //fprintf(stderr, "Calling get_successors.\n");
        call_prolog(get_successors, request, response);
        count_next_state++;
    } else if(strcmp(type, "init") == 0) {
        init(request, response);
    } else if(strcmp(type, "get-state-label") == 0) {
        //fprintf(stderr, "Calling get_state_label.\n");
        call_prolog(get_state_label, request, response);
        count_state_label++;
    } else if(strcmp(type, "next-update") == 0) {
        call_prolog(next_update, request, response);
        count_next_action++;
    } else if(strcmp(type, "get-state-label-group") == 0) {
        call_prolog(get_state_label_group, request, response);
        count_state_label_group++;
    } else if(strcmp(type, "terminate") == 0) {
        printf("get-next-state:  %d\n", count_next_state);
        printf("get-next-action: %d\n", count_next_action);
        printf("get-state-label: %d\n", count_state_label);
        printf("get-label-group: %d\n", count_state_label_group);
        zmsg_send(&response,zocket);
        zmsg_destroy(&response);
        zmsg_destroy(&request);
        free(type);
        return -1;
    } else {
        assert (0);
    }
    //puts("Sending Response.");
    zmsg_send(&response,zocket);
    zmsg_destroy(&response);
    zmsg_destroy(&request);
    free(type);
    //puts("Request done.");
    //puts("---------------------------------------------------");
    return 0;
}

int h_zocket (zloop_t *loop, zmq_pollitem_t *poller, void *zocket) {
    return h_zocket2(zocket);
}

void put_number(zmsg_t *msg, int nr) {
    zmsg_addstrf(msg,"%d",nr);
}

void put_chunk(zmsg_t *msg, char *term, int length) {
    assert(term[0] == 'D');
    assert(length > 0);
    zframe_t *frame = zframe_new(term,length);
    zmsg_append(msg, &frame);
}


void ltsmin_loop(void *zocket) {
    zmq_pollitem_t poller_zocket;
    poller_zocket = (zmq_pollitem_t) { zocket, 0, ZMQ_POLLIN };

    // reactor init
    zloop_t *reactor = zloop_new();
    zloop_poller(reactor, &poller_zocket, h_zocket, zocket);
    zloop_start(reactor);
    zloop_destroy(&reactor);
}


void *ltsmin_initialize(char const * endpoint_path) {
// Initialize Query
    get_successors        = init_predicate("get_successors",3,"ltsmin");
    next_update           = init_predicate("next_update",3,"ltsmin");
    get_init              = init_predicate("get_init",2,"ltsmin");
    get_state_label       = init_predicate("get_state_label",3,"ltsmin");
    get_state_label_group = init_predicate("get_state_label_group",3,"ltsmin");

    // poller init
    ctx = zctx_new();
    void *zocket = zsocket_new (ctx, ZMQ_REP);

    char *endpoint;
    int rc = asprintf(&endpoint, "ipc://%s", endpoint_path);
    assert(rc != -1);
    zsocket_bind(zocket, "%s", endpoint);
    free(endpoint);

    return zocket;
}


void ltsmin_teardown(void *zocket, char const * endpoint_path) {
    char *endpoint;
    int rc = asprintf(&endpoint, "ipc://%s", endpoint_path);
    assert(rc != -1);

    zsocket_unbind(zocket, "%s", endpoint);
    free(endpoint);
    zsocket_destroy(ctx, zocket);
    zctx_destroy(&ctx);
}

zmsg_t *create_message(void) {
    return zmsg_new();
}

void append_message(zmsg_t *main_message, zmsg_t *sub_message) {
    zframe_t *header = zmsg_pop(sub_message);
    zmsg_append(main_message, &header);

    int len = zmsg_size(sub_message);
    int rc = zmsg_addstrf(main_message, "%d", len);
    assert(rc == 0);
    
    zframe_t *frame;
    while ((frame = zmsg_pop(sub_message)) != NULL) {
        zmsg_append(main_message, &frame);
    }
}
