:- module(pltables_export_csv, [
                                  csv_header/2, csv_footer/2,
                                  csv_table_header/2, csv_table_footer/2,
                                  csv_table_rows/2
                               ]).

:- use_module(pltables_export_tools).
:- use_module(pltables).

:- use_module(probsrc(module_information)).

:- module_info(group,pltables).
:- module_info(description,'This module is used to export coverage tables to csv.').

% ----
% table headers
% ----
csv_header(_S, _TableName).

% ----
% table footers
% ----
csv_footer(_S, _TableName).

% ----
% table header / footer
% ----
csv_table_header(S, TableName) :-
    write(S, '<table'),
    pltable(TableName, TableAttributes),
    pltable_header(TableName, Columns, _Attributes),
    (
      member(description(XDU, _DescAttrs), TableAttributes)
      -> csv_escape(XDU,XD),
         format(S, ' description="~w"', [XD])
      ; true
    ),
    write(S, '>\n'),
    write_csv_table_header(S, Columns).

write_csv_table_header(S, Columns) :-
    write_csv_table_header2(S, Columns).

write_csv_table_header2(S, [Column]) :-
    csv_escape(Column, EscapedColumn),
    format(S, '~w\n', [EscapedColumn]).
write_csv_table_header2(S, [Column | Columns]) :-
    csv_escape(Column, EscapedColumn),
    format(S, '~w,', [EscapedColumn]),
    write_csv_table_header2(S, Columns).

csv_table_footer(_S, _TableName).

% ----
% table rows
% ----
csv_table_rows(S, TableName) :-
    findall(Entries, pltable_row(TableName, Entries, _Attributes), ListOfEntries),
    write_csv_table_rows(S, ListOfEntries).

write_csv_table_rows(_S, []).
write_csv_table_rows(S, [Entries | MoreEntries]) :-
    write_csv_table_row(S, Entries),
    write_csv_table_rows(S, MoreEntries).

write_csv_table_row(S, [Entry]) :-
    csv_escape(Entry, EscapedEntry),
    format(S, '~w\n', [EscapedEntry]).
write_csv_table_row(S, [Entry | Entries]) :-
    csv_escape(Entry, EscapedEntry),
    format(S, '~w,', [EscapedEntry]),
    write_csv_table_row(S, Entries).

% ----
% escaping characters ?
% ----
csv_escape(ToEscape, ToEscape).

%csv_escape(ToEscape, Escaped) :-
%  atom(ToEscape), !,
%  atom_codes(ToEscape, CodeList),
%  csv_escape2(CodeList, EscapedCodeList),
%  atom_codes(Escaped, EscapedCodeList).
%csv_escape(ToEscape, ToEscape).

%csv_escape2([],[]).
%csv_escape2([95|T], [92,95|EscT]) :-
%  csv_escape2(T, EscT).
%csv_escape2([35|T], [92,35|EscT]) :-
%  csv_escape2(T, EscT).
%csv_escape2([36|T], [92,36|EscT]) :-
%  csv_escape2(T, EscT).
%csv_escape2([37|T], [92,37|EscT]) :-
%  csv_escape2(T, EscT).
%csv_escape2([H|T], [H|EscT]) :-
%  H \= 95,
%  csv_escape2(T, EscT).
