#include "../include/zmq.h"
#include "../include/czmq.h"

#include "CuTest.h"
#include "../worker/whash_send.h"
#include "../common/hashtrie.h"

extern zctx_t *ctx;
static void *recv_hashes;
extern int port;

static void init(void) {
    init_send_hashes(ctx, "localhost", port-1);
    recv_hashes = zsocket_new(ctx, ZMQ_PULL);
    zsocket_bind (recv_hashes, "tcp://*:%d", port);
    port++;
}

static void teardown(void) {
    teardown_send_hashes(ctx);
    zsocket_destroy(ctx, recv_hashes);
}

void Test_single_processed(CuTest *cu) {
    init();
    init_successors();
    send_successors("foobartest1234567890");

    int rc = zsocket_poll(recv_hashes, 5000);
    CuAssertTrue(cu, rc);
    zmsg_t *msg = zmsg_recv(recv_hashes);
    CuAssertIntEquals(cu, 3, zmsg_size(msg));

    char *label = zmsg_popstr(msg);
    CuAssertStrEquals(cu, "Results:", label);
    free(label);

    label = zmsg_popstr(msg);
    CuAssertStrEquals(cu, "Hashes:", label);
    free(label);

    char *digest = zmsg_popstr(msg);
    CuAssertIntEquals(cu, 1, (int) digest[0]);
    CuAssertStrEquals(cu, "foobartest1234567890", digest + 1);
    free(digest);
    zmsg_destroy(&msg);
    teardown();
}

void Test_successors(CuTest *cu) {
    init();
    init_successors();
    add_queued_digest("blablaablaaaaaaaaah!");
    add_queued_digest("foobaztest9876543210");
    send_successors("foobartest1234567890");
    
    int rc = zsocket_poll(recv_hashes, 5000);
    CuAssertTrue(cu, rc);
    zmsg_t *msg = zmsg_recv(recv_hashes);
    CuAssertIntEquals(cu, 5, zmsg_size(msg));
    char *label = zmsg_popstr(msg);
    CuAssertStrEquals(cu, "Results:", label);
    free(label);

    label = zmsg_popstr(msg);
    CuAssertStrEquals(cu, "Hashes:", label);
    free(label);

    char *digest = zmsg_popstr(msg);
    CuAssertIntEquals(cu, 0, (int) digest[0]);
    CuAssertStrEquals(cu, "blablaablaaaaaaaaah!", digest + 1);
    free(digest);

    digest = zmsg_popstr(msg);
    CuAssertIntEquals(cu, 0, (int) digest[0]);
    CuAssertStrEquals(cu, "foobaztest9876543210", digest + 1);
    free(digest);

    digest = zmsg_popstr(msg);
    CuAssertIntEquals(cu, 1, (int) digest[0]);
    CuAssertStrEquals(cu, "foobartest1234567890", digest + 1);
    free(digest);

    zmsg_destroy(&msg);
    teardown();
}

void Test_results(CuTest *cu) {
    init();
    init_successors();
    put_result("DAeverything_is_broken", 23);
    send_successors("foobartest1234567890");
    
    int rc = zsocket_poll(recv_hashes, 5000);
    CuAssertTrue(cu, rc);
    zmsg_t *msg = zmsg_recv(recv_hashes);
    CuAssertIntEquals(cu, 4, zmsg_size(msg));
    char *label = zmsg_popstr(msg);
    CuAssertStrEquals(cu, "Results:", label);
    free(label);

    char *result = zmsg_popstr(msg);
    CuAssertStrEquals(cu, "DAeverything_is_broken", result);
    free(result);

    label = zmsg_popstr(msg);
    CuAssertStrEquals(cu, "Hashes:", label);
    free(label);

    char *digest = zmsg_popstr(msg);
    CuAssertIntEquals(cu, 1, (int) digest[0]);
    CuAssertStrEquals(cu, "foobartest1234567890", digest + 1);
    free(digest);

    zmsg_destroy(&msg);
    teardown();

}

CuSuite *WHashSendGetSuite() {
    CuSuite *suite = CuSuiteNew();
    SUITE_ADD_TEST(suite, Test_single_processed);
    SUITE_ADD_TEST(suite, Test_successors);
    SUITE_ADD_TEST(suite, Test_results);
    return suite;
}
