#include "../include/zmq.h"
#include "../include/czmq.h"

#include "CuTest.h"
#include "../worker/wwork.h"
#include "../worker/wtok.h"
#include "../common/queue.h"
#include "../common/hashtrie.h"

extern zctx_t *ctx;
static void *tmp_socket;
static void *work_recv;
extern int port;

extern wQueue *local_queue;
/*
 *wQueue *local_queue;
 * in fact, we should use our own queue for this
 * but the symbol is defined in master.c and we compile every .c file anyway
 * TODO: this should probably be changed...?
 */
static void assert_queue_items(CuTest *tc);

static void get_hello(CuTest *tc) {
    zmsg_t *msg = zmsg_recv(tmp_socket);
    CuAssertIntEquals(tc, 2, zmsg_size(msg));
    char *str = zmsg_popstr(msg);
    CuAssertStrEquals(tc, "1", str);
    free(str);
    char *hello = zmsg_popstr(msg);
    CuAssertStrEquals(tc, "hello", hello);
    free(hello);
    zmsg_destroy(&msg);
}

static void init(CuTest *tc, zmq_pollitem_t *tok) {
    zmq_pollitem_t ignored;
    init_work_handler(ctx, &ignored, &ignored, "localhost", "1", 0, port-2);
    init_tok(ctx, tok, NULL);
    init_hashtrie();
    local_queue = init_queue(0);
    work_recv = zsocket_new(ctx, ZMQ_ROUTER);
    zsocket_bind (work_recv, "tcp://*:%d", port+1);
    tmp_socket = zsocket_new(ctx, ZMQ_ROUTER);
    zsocket_set_identity(tmp_socket, "0");
    zsocket_bind (tmp_socket, "tcp://*:%d", port);
    get_hello(tc);
    port += 2;
}

static void send_work_message(void) {
    zmsg_t *msg = zmsg_new();
    // a workpackage would at least look similar to this
    char *wp1 = "foobartest1234567890DSterm\0\3DAa\0DAb\0DI3\0";
    char *wp2 = "foobartest0000000000DSterm\0\3DAb\0DAa\0DI5\0";
    char *wp3 = "foobartest1111111111DSterm\0\3DAa\0DAa\0DI7\0";
    char *wp4 = "foobartest2222222222DSterm\0\3DAb\0DAb\0DI9\0";
    zmsg_addmem(msg, wp1, 40);
    zmsg_addmem(msg, wp2, 40);
    zmsg_addmem(msg, wp3, 40);
    zmsg_addmem(msg, wp4, 40);
    zmsg_pushstr(msg, "-1"); // test stuff sent it
                             // probably needs refactoring
    zmsg_pushstr(msg, "1");
    zmsg_send(&msg, tmp_socket);
    zmsg_destroy(&msg);
    sleep(1);
}

static void teardown() {
    teardown_work_handler(ctx);
    teardown_tok(ctx);
    teardown_hashtrie();
    teardown_queue(&local_queue);
    zsocket_destroy(ctx, tmp_socket);
    zsocket_destroy(ctx, work_recv);
}

static void freecell(tCell *cell) {
    free(cell->digest);
    free(cell->term);
    free(cell);
}

void Test_h_work_recv(CuTest *tc) {
    zmq_pollitem_t polltok;
    init(tc, &polltok);
    send_work_message();
    h_work_recv(NULL, NULL, NULL);
    assert_queue_items(tc);
    int rc = zsocket_poll(polltok.socket, 5000);
    CuAssertTrue(tc, rc);
    char *tock = zstr_recv(polltok.socket);
    CuAssertStrEquals(tc, "TOCK", tock);
    free(tock);

    teardown();
}


static void assert_work_answer(CuTest *tc, zmsg_t *msg) {
    CuAssertIntEquals(tc, 3, zmsg_size(msg));
    char *str = zmsg_popstr(msg);
    CuAssertStrEquals(tc, "1", str);
    free(str);
    char *act_wp1 = zmsg_popstr(msg);
    char *act_wp2 = zmsg_popstr(msg);
    char *exp_wp1 = "foobartest1234567890DSterm\0\3DAa\0DAb\0DI3\0";
    char *exp_wp2 = "foobartest0000000000DSterm\0\3DAb\0DAa\0DI5\0";
    CuAssertMemEquals(tc, exp_wp1, act_wp1, 40);
    CuAssertMemEquals(tc, exp_wp2, act_wp2, 40);
    free(act_wp1);
    free(act_wp2);
}

static void send_work_request(void) {
    zmsg_t *req = zmsg_new();
    zmsg_addstr(req, "1");
    zmsg_addstr(req, "giev_work");
    zmsg_send(&req, work_recv);
    zmsg_destroy(&req);
}

static void assert_queue_items(CuTest *tc) {
    CuAssertIntEquals(tc, 4, local_queue->c);
    tCell *cell = dequeue(local_queue);
    CuAssertMemEquals(tc, "foobartest1234567890", cell->digest, 20);
    CuAssertMemEquals(tc, "DSterm\0\3DAa\0DAb\0DI3\0", cell->term, 20);
    CuAssertIntEquals(tc, 20, cell->term_length);
    freecell(cell);
    // TODO: external model checking file has method teardown_tcell()
    // use this one as soon it is merged with queue
    cell = dequeue(local_queue);
    CuAssertMemEquals(tc, "foobartest0000000000", cell->digest, 20);
    CuAssertMemEquals(tc, "DSterm\0\3DAb\0DAa\0DI5\0", cell->term, 20);
    freecell(cell);
    cell = dequeue(local_queue);
    CuAssertMemEquals(tc, "foobartest1111111111", cell->digest, 20);
    CuAssertMemEquals(tc, "DSterm\0\3DAa\0DAa\0DI7\0", cell->term, 20);
    freecell(cell);
    cell = dequeue(local_queue);
    CuAssertMemEquals(tc, "foobartest2222222222", cell->digest, 20);
    CuAssertMemEquals(tc, "DSterm\0\3DAb\0DAb\0DI9\0", cell->term, 20);
    freecell(cell);
    cell = dequeue(local_queue);
    CuAssertPtrEquals(tc, NULL, cell);
}

CuSuite *WWorkGetSuite() {
    CuSuite *suite = CuSuiteNew();
    SUITE_ADD_TEST(suite, Test_h_work_recv);
    return suite;
}
