% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(absint, [plugin_info/1]).

plugin_info([
             name = 'Abstract Interpretation'
            ,author = 'René Bartelmus, Sebastian Krings'
            ,status = stable
            ,transition = do_transition/4
            ,initialisation = do_initialisation/3
            ,state_property = find_state_property/2
            ,prettyprint_transition = pp_trans/2
            ,prettyprint_property = pp_prop/2
            ,prettyprint_value = pp_value/2
            ,plugin_for_modes = [b]
            ,plugin_init=plugin_init/0
            ,new_transition_notification=notify_new_transition/4
            ,check_invariant=do_check_invariant/2
            ,preferences=preference/4
            %,output=output/3
            %,internal_representation=internal_representation/1
            ]).

:- use_module(probsrc(module_information)).
:- module_info(group,plugin_absint).
:- module_info(description,'This is an experimental plugin that applies abstract interpretation on B models.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(probsrc(tools)).
:- use_module(probsrc(state_space)).
:- use_module(probsrc(specfile)).
:- use_module(probsrc(bmachine)).
:- use_module(probsrc(preferences),[get_preference/2]).

:- use_module(b_abstract_interpreter).
:- use_module(b_abstract_interpreter_helpers).

preference(abstract_domain_module, 'interval', 'Module name with abstract domain and operations',string).

plugin_init :-
    state_space:state_space_initialise.

do_initialisation(OpSetup, concrete_constants(ConstantsStateOut), [initialisation]) :-
    bmachine:bmachine_is_precompiled,
    b_machine_has_constants_or_properties, !,
    b_get_machine_constants(Constants),
    absint_build_initial_state(Constants,ConstantsState),
    b_get_properties_from_machine(Properties),
    absint_check_boolean_expression(Properties,[],[],ConstantsState,ConstantsStateOut,true),
    ((get_preference(show_initialisation_arguments,true),extract_value_list(ConstantsStateOut,Vals))
    -> safe_univ(OpSetup,['$setup_constants'|Vals])
       ; (OpSetup = '$setup_constants')
     ).

do_initialisation(Operation,StateOut,[initialisation]) :-
    bmachine:bmachine_is_precompiled,
    \+ b_machine_has_constants_or_properties, !,
    b_get_machine_variables(Variables),
    absint_build_initial_state(Variables,VariablesState),
    b_get_initialisation_from_machine(Init,_OType),
    absint_execute_statement(Init,[],VariablesState,StateOut,_Path),
    specfile:create_initialisation_operation(StateOut,Operation).

do_transition(concrete_constants(StateIn),Operation,StateOut,[op(Operation)]) :-
    bmachine:bmachine_is_precompiled, !,
    b_get_machine_variables(Variables),
    absint_build_initial_state(Variables,VariablesState),
    append(StateIn,VariablesState,State),
    b_get_initialisation_from_machine(Init,_OType),
    absint_execute_statement(Init,[],State,UnsortedStateOut,_Path),
    sort(UnsortedStateOut,StateOut),
    specfile:create_initialisation_operation(StateOut,Operation).

do_transition(State,OpSpec,NewState,[op(Operation)]) :-
    bmachine:bmachine_is_precompiled,
    b_top_level_operation(Operation),
    find_abstract_transition(Operation, State, NewState, Paras, Results),
    ( Paras \= [] ->
        OpSpec1 =.. [Operation|Paras]
    ; OpSpec1 = Operation
    ),
    (Results \= [] ->
        OpSpec = '-->'(OpSpec1,Results)
    ; OpSpec = OpSpec1
    ).

notify_new_transition(_FromID,_TransID,_DestID,_Exists).

find_state_property(State,Property):-
    member(bind(Var,Val),State),
    Property = '='(Var,Val).

pp_trans(Transition,Pretty) :-
    ajoin([Transition], Pretty).

pp_prop(Property,Pretty) :-
    ajoin([Property], Pretty).

pp_value(Value,Pretty) :-
    ajoin([Value], Pretty).

do_check_invariant(ID,_CurState) :-
    invariant_violated(ID), !.
do_check_invariant(ID,CurState) :-
    state_corresponds_to_initialised_b_machine(CurState,CurBState),
    b_get_invariant_from_machine(I),
    absint_check_boolean_expression(I,[],[],CurBState,_UpdatedCurState,false),
    set_invariant_violated(ID).
