% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(b_enumerate, [
  b_type_values_in_store/3,
  b_enumerate_values_in_store/5,
  b_enumerate_values/2,
  b_enumerate_typed_values/3,
  %b_tighter_enumerate_values_in_store/4,
  b_tighter_enumerate_values/2, b_tighter_enumerate_single_value/4,
  b_trace_tighter_enumerate_values/2,
  
  b_extract_typedvalc/4
  
  %enumerate_integer_with_min_range_wf/4 % no longer used
  ]).
  
  
:- use_module(self_check).
:- use_module(debug).
:- use_module(kernel_objects).
:- use_module(kernel_waitflags).
:- use_module(error_manager,[add_error/3]).
:- use_module(clpfd_interface,[clpfd_degree/2]).
:- use_module(kernel_waitflags,[large_finite_priority/1]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,interpreter).
:- module_info(description,'This module takes care of enumerating B variables.').

  /* ------------------------------------------ */
    
	
/* can be used to apply the result of 	
	b_getVarTypes_from_boolean_expression to a store */
	
:- use_module(store,[lookup_value_for_existing_id/3]).
	
:- use_module(typechecker,[type_check/2]).
:- assert_pre(b_enumerate:b_type_values_in_store(V,T,S), 
                  (typechecker:type_check(V,list(variable_id)),
                   typechecker:type_check(T,list(basic_type_descriptor)),typechecker:type_check(S,store))).
:- assert_post(b_enumerate:b_type_values_in_store(_,_,_), true).

b_type_values_in_store([],[],_).
b_type_values_in_store([Variable|VT],[Type|TT],Store) :-
   b_type_value_in_store(Variable,Type,Store),
   b_type_values_in_store(VT,TT,Store).
	
b_type_value_in_store(Var,Type,Store) :- 
   (Var = 'Identifier'(VarID)
      -> print(b_type_value_in_store(Var,Type)),print(' variable not in proper form'),nl
       ; VarID = Var),
   store:lookup_value_for_existing_id(VarID,Store,Val),
   (Val==fail -> add_error(b_type_value_in_store,'Variable has not been given value: ',VarID)
              ; kernel_objects:basic_type(Val,Type)).




/* ------------------------------------------ */
    

:- assert_pre(b_enumerate:b_enumerate_values_in_store(V,T,_,S,_WF), 
                  (typechecker:type_check(V,list(variable_id)),
                   typechecker:type_check(T,list(basic_type_descriptor)),typechecker:type_check(S,store))).
:- assert_post(b_enumerate:b_enumerate_values_in_store(_,_,_,_,_), true).
	
b_enumerate_values_in_store(Variables,Types,Values,Store,WF) :-
    generate_typed_var_list(Variables,Types,Store,Values,TypedValues),
    b_enumerate_values(TypedValues,WF).
    
:- block generate_typed_var_list(-,?,?,?,?).
generate_typed_var_list([],[],_,[],[]).
generate_typed_var_list([VarID|VT],[Type|TT],Store,Values,Res) :-
    ((store:lookup_value_for_existing_id(VarID,Store,Val), Val \== fail)
     ->  Res = [typedval(Val,Type,VarID,trigger_true(VarID))|Rest], Values=[Val|ValT] % always trigger enumeration warning
     ;   add_error(b_enumerate_values_in_store,'Variable has not been given value: ',VarID),
          Res=Rest,Values=ValT
   ),
   generate_typed_var_list(VT,TT,Store,ValT,Rest).
 
b_enumerate_sorted_values([],_).
b_enumerate_sorted_values([typedvalc(Val,Type,VarID,EnumWarning,Card)|Rest],WF) :-
    get_wait_flag(Card,enum(Val,Type),WF,LWF),
    b_enumerate_sorted_value_and_continue(LWF,Val,Type,VarID,EnumWarning,Rest,WF).

:- block b_enumerate_sorted_value_and_continue(-,?,?,?,?,?,?).
b_enumerate_sorted_value_and_continue(_LWF,Val,Type,_VarID,EnumWarning,Rest,WF) :-
    %%  print(enum_sorted_value(_VarID,Val,Type)),nl, %%
    enum_basic_with_wf(Val,Type,EnumWarning,WF),
    b_enumerate_sorted_values(Rest,WF).


b_enumerate_values(TypedValues,WF) :-
    ~~mnf(sort_typed_values(TypedValues,[],SortedValues)),
   % print(sort_vals(SortedValues)),nl,
    b_enumerate_sorted_values(SortedValues,WF).
/* ------------------------------------------ */

b_enumerate_typed_values(Values,Types,WF) :-
   b_generate_typed_values(Values,Types,TypedVals),
   b_enumerate_values(TypedVals,WF).

% generated typed values trigger the enumeration warning
b_generate_typed_values([],[],[]).
b_generate_typed_values([Val|Vrest],[Type|Trest],[typedval(Val,Type,'$UNKNOWN_ID',trigger_true(unknown))|TVrest]) :-
   b_generate_typed_values(Vrest,Trest,TVrest).

/*
:- assert_pre(b_enumerate:b_tighter_enumerate_values_in_store(V,T,S,_WF), 
                  (typechecker:type_check(V,list(variable_id)),
                   typechecker:type_check(T,list(basic_type_descriptor)),typechecker:type_check(S,store))).
:- assert_post(b_enumerate:b_tighter_enumerate_values_in_store(_,_,_,_), true).
b_tighter_enumerate_values_in_store(Vars,Types,Store,WF) :-
    ~~mnf(b_extract_typedvalc(Vars,Types,Store,TypedVals)),
    b_tighter_enumerate_values(TypedVals,WF).
*/
   
b_extract_typedvalc([],[],_,[]).
b_extract_typedvalc([Variable|VT],[Type|TT],Store,Res) :-
   (Variable = '_lambda_result_' % TO DO: use lambda_result info field ?; Note the check is repeated below in sort_typed_values
    -> b_extract_typedvalc(VT,TT,Store,Res) % lambda result does not need to be enumerated; will be computed from rest
    % TO DO: should we also deal specially with _lambda_result_ below in b_tighter_enumerate_values;
    %  or even infer in ANY statements, ... which variables are simply defined by equalities from the others
    ;  Res=[ExtractedH|ET],
       b_extract_typedvalc_var(Variable,Type,Store,ExtractedH),
       b_extract_typedvalc(VT,TT,Store,ET)).
 
b_extract_typedvalc_var(Var,Type,Store,typedval(Val,Type,VarID,trigger_true(VarID))) :- 
   (Var = 'Identifier'(VarID)
      -> print(b_tighter_enumerate_value_in_store(Var,Type)),print(' variable not in proper form'),nl
       ; VarID = Var),
   ((store:lookup_value_for_existing_id(VarID,Store,Val), Val \== fail)
     ->  true
     ;   add_error(b_tighter_enumerate_value_in_store,'Variable has not been given value: ',VarID)
   ).

% update priority for certain types
% TO DO: we should probably group all infinite integer values together and take the variable with the smallest size

:- use_module(kernel_tools,[ground_value/1]).
b_tighter_enumerate_sorted_values([],_WF).
b_tighter_enumerate_sorted_values([typedvalc(Val,Type,VarID,EnumWarning,Card)|Rest],WF) :- 
    (ground_value(Val)
      -> check_reached(Val,Type,VarID,Card,det,WF), %% <-----
         b_tighter_enumerate_sorted_values(Rest,WF) %% no need to enumerate this value
      ;  treat_next_typed_valc(Type,Val,VarID,EnumWarning,Card,Rest,WF)).

:- use_module(library(lists)).
:- use_module(library(clpfd),[fd_size/2]).
treat_next_typed_valc(integer,Val,VarID,EnumWarning,_Card,Rest,WF) :-
         nonvar(Val), Val=int(FD), 
         clpfd:fd_size(FD,Size), % will return sup if no bounds + also works in non-CLPFD mode
         number(Size),!,
         treat_next_typed_valc_normal(integer,Val,VarID,EnumWarning,Size,Rest,WF). 
treat_next_typed_valc(integer,Val,VarID,EnumWarning,Card,Rest,WF) :- 
         !,
         kernel_waitflags:large_finite_priority(LP),
         treat_next_integer(first_iteration,[LP],Val,VarID,EnumWarning,Card,Rest,WF).  % TO DO: maybe do several iterations 1000,LP
         % however: test 1003 with plavis-TransData_SP_13.prob fails in this case; to do: investigate
treat_next_typed_valc(Type,Val,VarID,EnumWarning,Card,Rest,WF) :- % TO DO: also look for finite values in Rest if Type is infinite but not integer
         treat_next_typed_valc_normal(Type,Val,VarID,EnumWarning,Card,Rest,WF).

% compute current priority of typedvalc; to be compared with @<
get_typedvalc_priority(typedvalc(Val,Type,_,_,_),Res) :- %Type=integer,
          (get_typedvalc_finite_size(Type,Val,Size)
             -> Res=Size % we have a finite value: use size
             ; degree_value(Type,Val,Dg) -> NDg is -Dg, Res=sup(NDg) % use negative number of degree, so that larger degrees get priority
                % relevant for e.g. test 1484:  res=x*x+4*x+14
             ; Res = sup(1)
          ).
get_typedvalc_finite_size(Type,Val,Size) :- 
         finite_value(Type,Val,Size),
         (Size=1 -> true ; \+ currently_useless_variable(Val)). % ,print(finite(Val,Size)),nl.

:- use_module(library(clpfd),[fd_degree/2]).
% try and get degree of a value
degree_value(integer,Val,Dg) :-  %translate:print_value_variable(Val),nl,
    nonvar(Val), Val=int(FD), clpfd:fd_degree(FD,Dg), number(Dg).

:- use_module(b_global_sets,[b_global_set_cardinality/2]).
finite_value(integer,Val,Size) :- nonvar(Val), Val=int(FD), clpfd:fd_size(FD,Size), number(Size).
finite_value(boolean,Val,Size) :- (ground(Val) -> Size = 1 ; Size =2).
finite_value(string(_),Val,Size) :- ground(Val), Size=1.
finite_value(global(G),Val,Size) :- (ground(Val) -> Size = 1 ; b_global_sets:b_global_set_cardinality(G,Size)).
finite_value(set(T),Val,Size) :- nonvar(Val), l_finite_value(T,Val,Size).
finite_value(couple(T1,T2),Val,Size) :- nonvar(Val), Val = (V1,V2),
    finite_value(T1,V1,Size1),
    finite_value(T2,V2,Size2),
    kernel_objects:safe_mul(Size1,Size2,Size), number(Size).
finite_value(record(Fields),Val,Size) :- %print(fv(Fields,Val)),nl,
    nonvar(Val), Val=rec(FVals),
    finite_fields(Fields,FVals,Size), nl,print(finite_field(Val,Size)),nl,nl.
% TO DO: missing   freetypes         

l_finite_value(_Type,Val,_) :- var(Val),!,fail.
l_finite_value(_,avl_set(_),1).
l_finite_value(_Type,[],1).
l_finite_value(Type,[H|T],Size) :- l_finite_value(Type,T,SizeT), finite_value(Type,H,SizeH),
    kernel_objects:safe_mul(SizeT,SizeH,Size), number(Size).

finite_fields(_Type,Val,_) :- var(Val),!,fail.
finite_fields([],[],1).
finite_fields([HType|TTypes],[H|T],Size) :- finite_fields(TTypes,T,SizeT), finite_value(HType,H,SizeH),
    kernel_objects:safe_mul(SizeT,SizeH,Size), number(Size).

% find the typedvalc (Selected) with the smallest value (Prio) and return other (Rest) typedvalc-values
smallest_value([H|T],Selected,Rest,Prio) :-
   get_typedvalc_priority(H,HPrio),
   smallest_value_aux(T,H,HPrio,Selected,Rest,Prio).

smallest_value_aux([],H,HPrio,H,[],HPrio).
smallest_value_aux([H|T],Acc,AccPrio,Selected,Rest,SelPrio) :-
   get_typedvalc_priority(H,HPrio),
   %print(compare(HPrio,AccPrio)),nl, print((H,Acc)),nl,
   (HPrio=1 -> smallest_value_aux(T,Acc,AccPrio,Selected,Rest,SelPrio) % remove this TC from list; does not need to be enumerated anymore
    ; HPrio @< AccPrio
      -> Rest = [Acc|TRest],
         smallest_value_aux(T,H,HPrio,Selected,TRest,SelPrio)
      ;  Rest = [H|TRest],
         smallest_value_aux(T,Acc,AccPrio,Selected,TRest,SelPrio)
    ).

% we try again and see if the domain is now finite
% ideally we should register all integer variables in the kernel_waitflag FD list; maybe we should have a finite and infinite list? or use our own labeling get-next function clpfd_get_next_variable_to_label_ffc adapted for infinite domains
:- block treat_next_integer(-,?,?,?,?,?,?,?).
treat_next_integer(LWF,_,Val,VarID,EnumWarning,_Card,Rest,WF) :- LWF \== first_iteration,  % print(treat(Val,VarID,Rest)),nl,
         nonvar(Val), Val=int(FD), 
         clpfd:fd_size(FD,Size), % will return sup if no bounds + also works in non-CLPFD mode
         %print(treat_next_integer(VarID,Val,Size,_Card,Rest)),nl,
         number(Size),!,
         (Size = 1 % for Size=1 no enumeration is necessary anyway
          -> b_tighter_enumerate_sorted_values(Rest,WF) % no enumeration required
          ;  treat_next_typed_valc_normal(integer,Val,VarID,EnumWarning,Size,Rest,WF)).
treat_next_integer(_,Prios,Val,VarID,EnumWarning,Card,OtherTCs,WF) :- % ideally we should move this above the clause with NextPrio; but then test 412 fails
         CurTC = typedvalc(Val,integer,VarID,EnumWarning,Card),
         smallest_value([CurTC|OtherTCs],TC,Rest,Size),
         % print(found_smallest(TC,Size,Rest)),nl,
         (number(Size) % finite domain found
           ; 
           Prios = [] % no more delaying; we have to enumerate an unbounded value
         ),
         !,
          % print(prioritising(Size,TC,VarID,Card,Rest)),nl,
         (Size == 1 % no enumeration required for this variable
          -> b_tighter_enumerate_sorted_values(Rest,WF)
           ; treat_next_typed_valc_normal4(TC,Size,Rest,WF)
            ).
treat_next_integer(_,[NextPrio|Prios],Val,VarID,EnumWarning,Card,Rest,WF) :- !, % for test 1161 this clause causes a considerable slowdown (VarID == dz1 -> trace ; true),
         % print(get_wait_flag_for(VarID,Val,NextPrio,Rest)),nl,
         get_wait_flag(NextPrio,treat_next_integer(VarID,Val),WF,LWF),
         treat_next_integer(LWF,Prios,Val,VarID,EnumWarning,Card,Rest,WF).
treat_next_integer(_,_,Val,VarID,EnumWarning,Card,Rest,WF) :-
         treat_next_typed_valc_normal(integer,Val,VarID,EnumWarning,Card,Rest,WF). % we have already delayed; no need to delay further

treat_next_typed_valc_normal4(typedvalc(Val,Type,VarID,EnumWarning,Card),Size,Rest,WF) :-
        (number(Size) -> NewCard=Size ; NewCard=Card), % use narrowed down card if possible
        treat_next_typed_valc_normal(Type,Val,VarID,EnumWarning,NewCard,Rest,WF).
treat_next_typed_valc_normal(Type,Val,VarID,EnumWarning,Card,Rest,WF) :-
         get_wait_flag(Card,tighter_enum(VarID,Val,Type),WF,LWF),
         b_tighter_enumerate_sorted_value_and_continue(LWF,Val,Type,VarID,EnumWarning,Card,Rest,WF).

:- block b_tighter_enumerate_sorted_value_and_continue(-,?,?,?,?,?,?,?).
b_tighter_enumerate_sorted_value_and_continue(_LWF,Val,Type,VarID,EnumWarning,Card,Rest,WF) :- 
         (Rest = [] -> true ; get_minimum_waitflag_prio(WF,Prio,_), Prio < 20000),
          % only check for useless if there are either other variables to chose from or other WF goals
         currently_useless_variable(Val),
         !,
         b_tighter_enumerate_useless(Val,Type,VarID,EnumWarning,Card,Rest,WF).
b_tighter_enumerate_sorted_value_and_continue(_LWF,Val,Type,VarID,EnumWarning,Card,Rest,WF) :- 
         check_reached(Val,Type,VarID,Card,WF), %% <-----
         %% tools:print_bt_message(enumerate(VarID,Val)), %%
         enum_tight_with_wf(Val,Type,EnumWarning,WF),
         %% tools:print_bt_message(after_enumerate(VarID,Val)),
         %(VarID=='VS_NOERRORCOND' -> print(done_enum(VarID,Val)),nl ; true),
         b_tighter_enumerate_sorted_values(Rest,WF).

b_tighter_enumerate_useless(Val,Type,VarID,EnumWarning,Card,Rest,WF) :- 
         get_next_useful_variable(Rest,RVal,RType,RVarID,REnumWarning,RCard,RRest), % ,print(using(RVarID,RVal,RType)),nl
          !,
          %debug_println(9,skipping(Val,VarID,Card,new(RVal,RType,RVarID,RCard))),
         (RCard =< Card /* we can enumerate straight away */
           -> check_reached(RVal,RType,RVarID,RCard,WF), %% <------
              enum_tight_with_wf(RVal,RType,EnumWarning,WF),
              b_tighter_enumerate_sorted_values([typedvalc(Val,Type,VarID,EnumWarning,Card)|RRest],WF)
            ; b_tighter_enumerate_sorted_values([typedvalc(RVal,RType,RVarID,REnumWarning,RCard),
                                                 typedvalc(Val,Type,VarID,EnumWarning,Card)|RRest],WF)
         ).
b_tighter_enumerate_useless(Val,Type,VarID,EnumWarning,Card,Rest,WF) :- 
         %debug_println(19,all_variables_useless(Val,Card)),
         get_minimum_waitflag_prio(WF,Prio,Info),
         Prio < 20000, % avoid infinite enumeration ... TO DO: improve check
         debug_println(19,all_variables_useless_info(Prio,Info)),
         !,
         ground_wait_flag_to(WF,Prio),
         % Now try again:
         b_tighter_enumerate_sorted_value_and_continue(1,Val,Type,VarID,EnumWarning,Card,Rest,WF).
b_tighter_enumerate_useless(Val,Type,VarID,EnumWarning,Card,Rest,WF) :- 
         check_reached(Val,Type,VarID,Card,WF), %% <-----
         %% tools:print_bt_message(enumerate(VarID,Val)), %%
         enum_tight_with_wf(Val,Type,EnumWarning,WF),
         %% tools:print_bt_message(after_enumerate(VarID,Val)),
         %(VarID=='VS_NOERRORCOND' -> print(done_enum(VarID,Val)),nl ; true),
         b_tighter_enumerate_sorted_values(Rest,WF).

% --------------------

% add WF information; so that we can provide better enumeration warning messages
add_wf_context(trigger_true(Info),WF,trigger_true(enum_wf_context(WF,Info))).
add_wf_context(trigger_false(Info),WF,trigger_false(enum_wf_context(WF,Info))).
add_wf_context(trigger_throw(Info),WF,trigger_throw(enum_wf_context(WF,Info))).

% TO DO: check whether we should not call enumerate_basic_type_wf instead

:- use_module(kernel_objects,[enumerate_tight_type/3]).
enum_tight_with_wf(Val,Type,EnumWarning,WF) :- 
   (add_wf_context(EnumWarning,WF,NewWarning) -> enumerate_tight_type(Val,Type,NewWarning)
     ; enumerate_tight_type(Val,Type,EnumWarning)).

:- use_module(kernel_objects,[enumerate_basic_type/3]).
enum_basic_with_wf(Val,Type,EnumWarning,WF) :- 
   (add_wf_context(EnumWarning,WF,NewWarning) -> enumerate_basic_type(Val,Type,NewWarning)
     ; enumerate_basic_type(Val,Type,EnumWarning)).


% ---------------------

get_next_useful_variable([typedvalc(Val,Type,VarID,EnumWarning,Card)|Rest],RVal,RType,RVarID,REnumWarning,RCard,RRest) :-
  (currently_useless_variable(Val)
    ->  RRest=[typedvalc(Val,Type,VarID,EnumWarning,Card)|TRRest],
        get_next_useful_variable(Rest,RVal,RType,RVarID,REnumWarning,RCard,TRRest)
    ;   RRest=Rest, RVal=Val, RType=Type, RVarID=VarID, REnumWarning = EnumWarning, RCard=Card
  ).
  

/* purpose: check if an enumeration is useful or not and give priority to useful variables ? */
%:- use_module(library(clpfd),[fd_dom/2]).
currently_useless_variable(Val) :- var(Val),!,frozen(Val,Goal),  % print(enum(Val,Goal)),nl,
      (enumeration_only_goal(Goal,Val) -> true
       ;  % print(useful_val(Val,Goal)),nl,nl, %trace, enumeration_only_goal(Goal,Val),
        fail).
currently_useless_variable(int(Val)) :- currently_useless_variable(Val).
currently_useless_variable(fd(Val,_)) :- currently_useless_variable(Val).
currently_useless_variable(string(Val)) :- currently_useless_variable(Val).
  %fd_dom(Val,Dom),print(uselessfd(Val,Dom)),nl.

enumeration_only_goal(true,_).
%enumeration_only_goal(fd_utils_clpfd:delayed_enum_fd(V,_,_,_),Val) :- V==Val.
enumeration_only_goal(kernel_objects:call_enumerate_int(V,_,_,_),Val) :- V==Val.
enumeration_only_goal(kernel_objects:enumerate_int_wf(V,_,_),Val) :- V==Val.
enumeration_only_goal(kernel_objects:enumerate_natural(V,_,_),Val) :- V==Val.
enumeration_only_goal((A,B),Val) :- 
  (enumeration_only_goal(A,Val) -> enumeration_only_goal(B,Val)
    ; % we have not detected A as enumeration only: check if we can find an annotation not to enumerate
      force_enumeration_only_goal(A,Val) % Note: we only check first component; on the assumption that these co-routines are added early and to avoid useless lookups; TO DO: is there a better solution ?
  ).
enumeration_only_goal(clpfd:(X in _),Var) :- Var==X,
  clpfd_interface:clpfd_degree(X,0). % only goal is the enumerator itself; but CLPFD constraints are ***NOT*** shown in frozen(...) so we need to call clpfd_degree
enumeration_only_goal(bsets_clp:is_sequence_wf_ex(V,GS,_,_),Val) :- V==Val, nonvar(GS),GS=global_set(_). % only constraint: we should generate a sequence
enumeration_only_goal(external_functions:string_to_int2(_,V,_,_),Val) :- V==Val.
enumeration_only_goal(external_functions:printf(_,V,_,_),Val) :- V==Val.
enumeration_only_goal(prolog:dif(A,B),Var) :- (Var==A ; Var==B).

enumeration_only_goal(prolog:trig_ground(A1,_,_,_,Trigger),Val) :- Val==A1,
      triggers_better_enumeration(Trigger,Val,3). % if this only triggers an enumeration of Val itself; it is better to let the other variable trigger the enumeration

force_enumeration_only_goal((A,B),Val) :- 
  (force_enumeration_only_goal(A,Val) -> true  ; force_enumeration_only_goal(B,Val) ).
force_enumeration_only_goal(external_functions:'DO_NOT_ENUMERATE'(V,_,_),Val) :- V==Val.

% check if Trigger either triggers a better enumerator or just debugging code
triggers_better_enumeration(Trigger,Val,Depth) :- Depth>0, D1 is Depth-1,
     frozen(Trigger,TrigGoal),
      %print(trig(TrigGoal,Val)),nl,trace,
     ( better_enum(TrigGoal,Trigger,Val,D1) -> true).

better_enum((A,B),Trigger,Val,Depth) :- better_enum(A,Trigger,Val,Depth), better_enum(B,Trigger,Val,Depth).
better_enum(prolog:trig_and(_,_,_,_,Trigger1),Trigger,Val,Depth) :- !,
   (Trigger1==Trigger -> true ; triggers_better_enumeration(Trigger1,Val,Depth)).
better_enum(prolog:trig_or(_,Trigger1,_),Trigger,Val,Depth) :- !,
   (Trigger1==Trigger -> true ; triggers_better_enumeration(Trigger1,Val,Depth)).
better_enum(prolog:trig_ground(VV,_,_,Trigger1,_),Trigger,Val,Depth) :- !,
   (Trigger1==Trigger -> true ; VV=Val -> true ; triggers_better_enumeration(Trigger1,Val,Depth)).
better_enum(prolog:when(WT,_,G), Trigger,Val,Depth) :- !,
      Trigger==WT, better_enum(G,Trigger,Val,Depth).
better_enum(custom_explicit_sets:couple_element_of_avl_set(_X,_Y,_AVL,_WF1,_WF),_,_Val,_Depth) :- !,
 % TO DO: add more enumerators of Val
 % TO DO: VAL occurs in _X !!
      true. %print(avl_trig(X,Y,Val,Depth)),nl.
better_enum(external_functions:G,_,_Val,_Depth) :- external_debug_function(G).

external_debug_function(fprintf2(_,_,_,_)). % TO DO: add more



b_tighter_enumerate_single_value(Val,Type,VarID,WF) :-
    b_tighter_enumerate_values([typedval(Val,Type,VarID,trigger_true(VarID))],WF).

b_tighter_enumerate_values(TypedValues,WF) :-
    ~~mnf(sort_typed_values(TypedValues,[],SortedValues)),  % no longer required ??
    b_tighter_enumerate_sorted_values(SortedValues,WF).

:- if(debug:global_debug_flag).
:- dynamic reached/4.
my_spy('P_Env_Output_TargetSpeed_Speed').
check_reached(Val,Type,VarID,Card,WF) :- 
  (my_spy(VarID) -> print(' SPY POINT : '), print(VarID), print(' = '), print(Val),nl,
    portray_waitflags(WF), trace ; true),
  (ground_value(Val) -> check_reached(Val,Type,VarID,Card,det,WF)
     ; check_reached(Val,Type,VarID,Card,enum,WF)).
% findall(r(Card,VarID,Info),b_enumerate:reached(VarID,Card,Info),_A),sort(_A,_AS),reverse(_AS,RAS).
check_reached(Val,Type,VarID,Card,Info,WF) :- 
    (retract(reached(VarID,RR,Info)) -> true ; RR=0),
    R1 is RR+1,
    (RR mod 1000 < 1 -> (RR=0 -> print('REACHED : ') ; print(' AGAIN  : ')), print(Info), print(' '),
                 print(VarID), print(' : '), print(Card),
                 print(' : '), print(Type), print(' : '), print(Val), print(' :: '), print(RR), nl,
                 (my_spy(VarID)  %(RR>8000,RR<40001)
                  -> portray_waitflags(WF),trace ; portray_waitflags(WF))
                 %, (RR=1000 -> trace ; true)
           ; true),
    assert(reached(VarID,R1,Info)).
b_trace_tighter_enumerate_values(TypedValues,WF) :-
    ~~mnf(sort_typed_values(TypedValues,[],SortedValues)), 
     print('ENUMERATING VALUES:'),nl,print_values(SortedValues), %%
     nl, portray_waitflags(WF),nl,
    b_trace_tighter_enumerate_values2(SortedValues,WF),
    print('FINISHED SETTING UP ENUMERATION'),nl, portray_waitflags(WF),nl.
 b_trace_tighter_enumerate_values2(SortedValues,WF) :-   retractall(reached(_,_,_)),
    b_tighter_enumerate_sorted_values(SortedValues,WF).

print_values([]).
print_values([typedvalc(Val,Type,VarID,_EnumWarning,Card)|Rest]) :-
  print_value(VarID,Card,Type,Val),
  (ground_value(Val) -> true ; when(ground(Val),(nl,print('HAS BECOME GROUND: '),print_value(VarID,Card,Type,Val)))),
  print_values(Rest).
  
print_value(VarID,Card,Type,Val) :-
  print(' '),
  print(VarID), print(' :card='), print(Card), print(' ('), print(Type), print(') = '), print(Val),nl.
:- elif(1=2).
:- dynamic reached/4, enumerated/2,enum_edge/3.
check_reached(_Val,_Type,_VarID,_Card,_WF) :- check_reached(_Val,_Type,_VarID,_Card,enum,_WF).
check_reached(_Val,_Type,VarID,_Card,Info,_WF) :- 
    (retract(reached(VarID,RR,Info)) -> true ; RR=0),
    R1 is RR+1,
    assert(reached(VarID,R1,Info)),
    (enumerated(TopVar,Level) -> add_edge(TopVar,VarID,Level),L1 is Level+1 ; L1=0),
    asserta(enumerated(VarID,L1)).
check_reached(_Val,_Type,VarID,_Card,_Info,_WF) :-
    retract(enumerated(VarID,_)),fail.
add_edge(TopVar,VarID,Level) :-
    (enum_edge(TopVar,VarID,Level) -> true ; assert(enum_edge(TopVar,VarID,Level))).

% b_enumerate:print_enum_stats.
print_enum_stats :- findall( enum(Count,VarID,Info), reached(VarID,Count,Info),L),
    sort(L,SL),
    print('Enumeration Frequency'),nl,
    member(enum(C,ID,I),SL), format(' ~w ~w ~w~n',[C, ID, I]),
    fail.
print_enum_stats :-
    enum_edge(TopVar,VarID,0), format('~w ---> ~w~n',[TopVar,VarID]),
    print_edges(VarID,1),
    fail.
print_enum_stats :- nl.
print_edges(VarID,Level) :- enum_edge(VarID,NxtId,Level),
    format('  ~w -(~w)--> ~w~n',[VarID,Level,NxtId]),
    fail.
print_edges(_,_).
b_trace_tighter_enumerate_values(TypedValues,WF) :- b_tighter_enumerate_values(TypedValues,WF).
:- else.
check_reached(_Val,_Type,_VarID,_Card,_WF).
check_reached(_Val,_Type,_VarID,_Card,_Info,_WF).
b_trace_tighter_enumerate_values(TypedValues,WF) :- b_tighter_enumerate_values(TypedValues,WF).
:- endif.

% dummy call to register variable name with free Prolog variables:
% can be obtained using frozen()
:- public register_prolog_variable_name/3. % Debugging tool
:- block register_prolog_variable_name(-,?,?).
register_prolog_variable_name(int(Val),Name,Type) :- !, %print(register_int(Val,Name,Type)),nl,
   register_prolog_variable_name(Val,Name,Type).
register_prolog_variable_name(fd(Val,_),Name,Type) :- !,
   register_prolog_variable_name(Val,Name,Type).
register_prolog_variable_name(string(Val),Name,Type) :- !,
   register_prolog_variable_name(Val,Name,Type).
register_prolog_variable_name((A,B),Name,Type) :- !,
   register_prolog_variable_name(A,left(Name,mapto(B)),Type),
   register_prolog_variable_name(B,right(Name,mapfrom(A)),Type).
register_prolog_variable_name([A|T],Name,Type) :- !,
   register_prolog_variable_name(A,el(Name),Type),
   register_prolog_variable_name(T,Name,Type).
register_prolog_variable_name(_,_,_).
% TO DO: something that pretty prints the path information for the average user

% a tool to get variable name if it was registered
:- public get_prolog_variable_name/3. % Debugging tool
get_prolog_variable_name(X,Name,Type) :- var(X),
   frozen(X,Goal), print(goal(X,Goal)),nl,
   goal_member(Goal,b_enumerate:register_prolog_variable_name(_,ID,T)),
   Name=ID, Type=T.
goal_member(X,X).
goal_member((A,B),X) :- goal_member(A,X) ; goal_member(B,X).

sort_typed_values([],R,R).
sort_typed_values([typedval(Val,Type,VarID,EnumWarning)|Rest],SortedSoFar,Res) :-
   %% comment in next line to keep track of variable names for enumeration warnings, etc...
   %% register_prolog_variable_name(Val,VarID,Type), % dummy call to register variable name with free Prolog variables
   (VarID = '_lambda_result_' -> NewSorted=SortedSoFar
    ; ground_value(Val) -> NewSorted=SortedSoFar
    ;  get_max_cardinality_as_priority(Type,CardPrio),
       %% TO DO: maybe also take shape of Val into account; e.g. if cardinality of a set already fixed...
       insert_val(SortedSoFar,Val,Type,VarID,EnumWarning,CardPrio,NewSorted)
   ),
   sort_typed_values(Rest,NewSorted,Res).

get_max_cardinality_as_priority(integer,Prio) :- !, integer_priority(Prio).
get_max_cardinality_as_priority(Type,Prio) :-
  ~~mnf(kernel_objects:max_cardinality(Type,Card)),
  (Card \= inf -> Prio=Card
    ; inf_type_prio(Type,Prio)
  ).

% Try to give priority to simpler infinite types (such as integer over seq(_) ,...)
% TO DO: to a full-blown analysis ...
inf_type_prio(integer,Prio) :- !, integer_priority(Prio).
inf_type_prio(seq(X),Prio) :- !, inf_type_prio(X,P1), Prio is P1+11.
inf_type_prio(set(X),Prio) :- !, inf_type_prio(X,P1), Prio is P1+10.
%inf_type_prio(couple(_,_),Prio) :- !, TO DO check if both args are infinite or not ...
inf_type_prio(_,Prio) :- !, integer_priority(P), Prio is P+1.


%typedvalc = typedval + cardinality information
insert_val([],Val,Type,VarID,EnumWarning,Card,[typedvalc(Val,Type,VarID,EnumWarning,Card)]).
insert_val([TC|T], Val,Type, VarID, EnumWarning, Card, Res) :-
   TC = typedvalc(V1,_T1,_Var1,_W1,C1),
   ((%cardval_greater_than(C1,Card)
     Card \== inf, (C1=inf -> true ; C1>Card)
     ; Card=C1, var(V1), nonvar(Val)) % if card equal give priority to nonvariable values
     -> Res = [typedvalc(Val,Type,VarID,EnumWarning,Card),TC|T]
     ;  Res = [TC|RT],
        insert_val(T,Val,Type,VarID,EnumWarning,Card,RT)
   ).

%cardval_greater_than(X,Y) :- Y \== inf, (X=inf -> true ; X>Y).

/*
% call to enumerate an integer ensuring that you go at least from MinFrom to MinTo
%:- block enumerate_integer_with_min_range_wf(-,?,?,?).
enumerate_integer_with_min_range_wf(int(X),MinFrom,MinTo,WF) :-
   integer_priority(Prio), P1 is Prio-1,
   ... removed ...
*/
