% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(csp_sequences,[is_null_list/2, is_empty_list/2,
                         is_elem_list/2,
                         tail_list/2, head_list/2,
                         concat_lists/2, append_list/3,
                         length_list/2,
                         
                         expand_sequence/2,
                         expand_listcomprehension/3,
                         
                         convert_seq_to_set/2, convert_set_to_seq/2]).

:- use_module(probsrc(module_information)).
:- module_info(group,csp).
:- module_info(description,'Operations on CSP sequences.').

/*********** PROB modules **********/
:- use_module(probsrc(error_manager),[add_error/3,add_internal_error/2]).
:- use_module(probsrc(self_check)).
:- use_module(probsrc(tools),[remove_variables/3,ajoin/2]).
:- use_module(probsrc(translate),[translate_csp_value/2]).
%------- CSP modules:
:- use_module(probcspsrc(csp_sets)).
:- use_module(probcspsrc(haskell_csp),[force_evaluate_argument/2,evaluate_argument/2]).
/*********** ----------- **********/

/* ------------------------------------- */
/*                SEQUENCES              */
/* ------------------------------------- */

:- assert_must_succeed(csp_sequences:is_null_list(listFromTo(22,21),true) ).
:- assert_must_succeed(csp_sequences:is_null_list(list([]),true) ).
:- assert_must_succeed((csp_sequences:is_null_list(listFromTo(22,22),R), R==false )).
:- assert_must_fail(csp_sequences:is_null_list(listFrom(22),true) ).
:- assert_must_fail(csp_sequences:is_null_list(listFromTo(22,22),true) ).
:- assert_must_fail(csp_sequences:is_null_list(listFromTo(22,23),true) ).
:- assert_must_fail(csp_sequences:is_null_list(list([int(22),int(21)]),true) ).

:- block is_null_list(-,?).
is_null_list(list(Y),R) :- !, is_empty_list(Y,R).
is_null_list(listFrom(_),R) :- !,R=false.
is_null_list(listFromTo(F,T),R) :- number(F),number(T), !, 
   (F>T -> R=true ; R=false).
is_null_list(X,_) :- add_error(csp_sequences,'Unexpected argument to null: ',X),fail.

:- block is_empty_list(-,?).
is_empty_list([],R) :- !, R=true.
is_empty_list(_,false).

:- assert_must_fail(csp_sequences:is_elem_list(int(22),listFromTo(22,21)) ).
:- assert_must_succeed(csp_sequences:is_elem_list(int(22),listFromTo(22,22)) ).
:- assert_must_succeed(csp_sequences:is_elem_list(int(22),listFrom(22)) ).
:- assert_must_succeed(csp_sequences:is_elem_list(int(22),list([int(23),int(22)])) ).

is_elem_list(E,X) :-
   (X=list(Y) 
    -> elem(Y,E)
    ; X=listFrom(F) -> E=int(EV), mygeq(EV,F)
    ; X=listFromTo(F,T) -> E=int(EV), mygeq(EV,F), mygeq(T,EV)
    ; add_error(csp_sequences,'Unexpected argument to elem: ',X),fail
   ).

:- block mygeq(-,?), mygeq(?,-).
mygeq(EV,F) :- EV>=F.

:- block elem(-,?).
elem([H|T],E) :- (E=H ; (/*dif(E,H),*/elem(T,E))).

:- assert_must_succeed((csp_sequences:tail_list(listFrom(22),R), R=listFrom(23) )).
:- assert_must_succeed((csp_sequences:tail_list(listFromTo(22,22),R), R=list([]) )).
:- assert_must_succeed((csp_sequences:tail_list(listFromTo(22,23),R), R=listFromTo(23,23) )).
:- assert_must_succeed((csp_sequences:tail_list(list([int(22),int(21)]),R), R=list([int(21)]) )).

tail_list(listFrom(F),Res) :- number(F),!,
   F1 is F+1, Res= listFrom(F1).
tail_list(listFromTo(F,T),Res) :- number(F),number(T), T >=F, !,
   (T>F -> F1 is F+1, Res=listFromTo(F1,T) ; Res=list([])).
tail_list(List,Res) :-
   (List=list([_H|T]) -> Res = list(T)
     ; (add_error(csp_sequences,'Unexpected argument to tail of sequence: ',List),fail)).
     
:- assert_must_succeed((csp_sequences:head_list(listFrom(22),R), R=int(22) )).
:- assert_must_succeed((csp_sequences:head_list(listFromTo(22,22),R), R=int(22) )).
:- assert_must_succeed((csp_sequences:head_list(list([int(22),int(21)]),R), R=int(22) )).

head_list(listFrom(F),Res) :- !, Res=int(F).
head_list(listFromTo(F,T),Res) :- number(F),number(T), T >=F, !, Res=int(F).
head_list(List,Res) :-
   (List=list([H|_T]) -> Res = H
     ; (add_error(csp_sequences,'Unexpected argument to head of sequence: ',List),fail)).

:- assert_must_succeed((csp_sequences:append_list(listFromTo(1,2),listFromTo(4,6),R), R==list([int(1),int(2),int(4),int(5),int(6)]))).
:- assert_must_succeed((csp_sequences:append_list(listFrom(2),listFromTo(4,6),R), R==listFrom(2) )).
:- assert_must_succeed((csp_sequences:append_list(list([int(1),int(2)]),listFromTo(4,6),R), R==list([int(1),int(2),int(4),int(5),int(6)]) )).
:- assert_must_succeed((csp_sequences:append_list(list([int(1)]),list([int(1)]),R), R==list([int(1),int(1)]) )).

append_list(listFrom(F),_,Res) :- !, Res=listFrom(F).
append_list(listFromTo(F,T),Y,Res) :- expand_sequence(Y,list(EY)),!, Res=list(ZL),
     expand_seq_from_to(F,T,EX), app1(EX,EY,ZL).
append_list(list(XL),Y,list(ZL)) :-  expand_sequence(Y,list(YL)),!, 
     app1(XL,YL,ZL).
append_list(X,Y,Z) :-
     translate_csp_value(X,TX),translate_csp_value(Y,TY),
     (var(Z) -> ajoin([TX,' ^ ',TY],Text)
      ; translate_csp_value(Z,TZ), ajoin([TX,' ^ ',TY,' = ', TZ],Text)),
     add_error(csp_sequences,'Illegal arguments for sequence catenation ^ : ',Text),fail.
% Note: we do not allow mixed explicit lists and symbolic listFrom
:- block app1(-,?,?).
%app1(listFrom(F),_,R) :- !, R= listFrom(F).
app1([],Y,R) :- !,Y=R.
app1([H|X],Y,R) :- !,R=[H|Z], app1(X,Y,Z).
app1(X,Y,Z) :- add_internal_error(/*csp_sequences,*/'Unexpected arguments to append: ',app1(X,Y,Z)),fail.


:- assert_must_succeed((csp_sequences:concat_lists(list([list([int(1),int(2)]),listFromTo(4,6)]),R), R==list([int(1),int(2),int(4),int(5),int(6)]) )).
:- assert_must_succeed((csp_sequences:concat_lists(list([list([]),list([int(1)]),list([]),list([int(1)]),list([])]),R), R==list([int(1),int(1)]) )).

concat_lists(X,Y) :-
   ((X=list(XL),Y=list(YL)) 
     -> conc(XL,YL)
     ; add_error(csp_sequences,'Unexpected arguments to concat of sequences: ',concat(X,Y)),fail
   ).

:- block conc(-,?).     
conc([],Res) :- !, Res=[].
conc([HL|X],Res) :- expand_sequence(HL,list(H)),!,
     conc(X,RX), app1(H,RX,Res).
conc(X,Y) :-
     add_error(csp_sequences,'Unexpected arguments to conc:',conc(X,Y)),fail.


:- assert_must_succeed((csp_sequences:length_list(list([]),R), R==int(0) )).
:- assert_must_succeed((csp_sequences:length_list(list([int(22)]),R), R==int(1) )).
:- assert_must_succeed((csp_sequences:length_list(list([int(22),int(22)]),R), R==int(2) )).
:- assert_must_succeed((csp_sequences:length_list(listFromTo(1,10),R), R==int(10) )).
:- assert_must_succeed((csp_sequences:length_list(listFromTo(2,100001),R), R==int(100000) )).
:- assert_must_succeed((csp_sequences:length_list(listFromTo(3,1),R), R==int(0) )).


length_list(listFrom(X),R) :- !,
   add_internal_error('Internal Error: Trying to get length of infinite list: ',length_list(listFrom(X),R)),fail.
length_list(list(Y),int(R)) :- !, length2(Y,R).
length_list(listFromTo(F,T),int(R)) :- !, (T<F -> R=0 ; R is T+1-F).
length_list(X,Res) :-
   add_error(csp_sequences,'Unexpected arguments to length: ',length(X,Res)),
   fail.
 
:- block length2(-,?).
length2([],0).
length2([_H|T],R) :-length3(T,1,R). %when(nonvar(T),(length2(T,RT), R is RT+1)).

:- block length3(-,?,?).
length3([],R,R).
length3([_|T],SF,R) :- SF1 is SF+1, length3(T,SF1,R).



/* ------------------------------------- */
/*       SEQUENCE EXPRESSIONS            */
/* ------------------------------------- */

:- assert_must_succeed((csp_sequences:expand_sequence(listFromTo(3,1),R), R==list([]) )).
:- assert_must_succeed((csp_sequences:expand_sequence(listFromTo(33,33),R), R==list([int(33)]) )).
:- assert_must_succeed((csp_sequences:expand_sequence(list([int(33)]),R), R==list([int(33)]) )).
:- assert_must_succeed((csp_sequences:expand_sequence(listExp(rangeEnum([_x5]),[comprehensionGenerator(_x5,listExp(rangeEnum([int(1),int(2),int(3)])))]),R), R == list([int(1),int(2),int(3)]))).
:- assert_must_succeed((csp_sequences:expand_sequence(listExp(rangeEnum([_x5]),[comprehensionGenerator(_x5,listExp(rangeEnum([int(1),int(2),int(1)])))]),R), R == list([int(1),int(2),int(1)]))).

expand_sequence(list(X),Res) :- !, Res=list(X).
expand_sequence(listFromTo(X,Y),Res) :- !, Res=list(R),expand_seq_from_to(X,Y,R).
expand_sequence(listExp(RangeExpr,GeneratorList),R) :- !,
   expand_listcomprehension(RangeExpr,GeneratorList,R).
expand_sequence(listFrom(X),_R) :- !,
   translate_csp_value(X,TX),
   add_error(csp_sequences,'Trying to expand infinite list: ',TX),fail.
expand_sequence(X,_Y) :-
   translate_csp_value(X,TX),
   add_error(csp_sequences,'Illegal argument, this is not a list: ',TX),fail.
    
expand_seq_from_to(X,Y,R) :- X>Y,!, R=[].
expand_seq_from_to(X,Y,[int(X)|T]) :- X1 is X+1, expand_seq_from_to(X1,Y,T).


:- assert_must_succeed((csp_sequences:convert_seq_to_set(listFrom(333),R), R==setFrom(333) )).
:- assert_must_succeed((csp_sequences:convert_seq_to_set(listFromTo(33,33),R), R==setFromTo(33,33) )).
:- assert_must_succeed((csp_sequences:convert_seq_to_set(list([int(1),int(3),int(2),int(1)]),R), R==setValue([int(1),int(2),int(3)]) )).
:- assert_must_succeed((csp_sequences:convert_seq_to_set(listExp(rangeEnum([_x5]),[comprehensionGenerator(_x5,listExp(rangeEnum([int(1),int(2),int(3)])))]),R), R == setValue([int(1),int(2),int(3)]))).

convert_seq_to_set(listExp(E,GeneratorList),Res) :- !,
    expand_sequence(listExp(E,GeneratorList),list(ESeq)),
    evaluate_set(ESeq,Res).%evaluate_set(ESeq,Res,evaluate_argument).
convert_seq_to_set(listFromTo(X,Y),Res) :- !, Res = setFromTo(X,Y).
convert_seq_to_set(listFrom(X),Res) :- !, Res = setFrom(X).
convert_seq_to_set(list(L),Res) :- !, evaluate_set(L,Res).%evaluate_set(L,Res,evaluate_argument).
convert_seq_to_set(X,_R) :- add_error(csp_sequences,'Unexpected argument for the seq(-) built-in function: ',X),fail.

:- assert_must_succeed((csp_sequences:convert_set_to_seq(setFromTo(1,10),R), R == listFromTo(1,10))).
:- assert_must_succeed((csp_sequences:convert_set_to_seq(setFrom(1),R), R == listFrom(1))).
:- assert_must_succeed((csp_sequences:convert_set_to_seq(setValue([int(1),int(3),int(2)]),R), R == list([int(1),int(3),int(2)]))).
:- assert_must_succeed((csp_sequences:convert_set_to_seq(setExp(rangeEnum([_x6]),[comprehensionGenerator(_x6,setExp(rangeEnum([int(1),int(2),int(3)])))]), R),
            R == list([int(1),int(2),int(3)]))).

convert_set_to_seq(setExp(E,Generators),Res) :- !,
    expand_set_comprehension(E,Generators,setValue(R)),
    %print(expand_set(Generators,E,R)),nl,
    Res=list(R).
convert_set_to_seq(setFromTo(X,Y),Res) :- !, Res = listFromTo(X,Y).
convert_set_to_seq(setFrom(X),Res) :- !, Res = listFrom(X).
convert_set_to_seq(setValue(S),Res) :- !, Res = list(S).
convert_set_to_seq(X,_R) :- add_error(csp_sequences,'Unexpected argument for the set(-) built-in function: ',X),fail.

/* ------------------- */
/* LIST COMPREHENSIONS */
/* ------------------- */
:- if(current_prolog_flag(dialect,sicstus)).
:- if((current_prolog_flag(version_data,sicstus(4,X,_,_,_)),X<3)).
:- use_module(library(terms),[term_variables/2]). % is built-in in SICSTUS 4.3
:- endif.
:- endif.

expand_listcomprehension(RangeExpr,GeneratorList,Res) :-
   %%  print(expand_listComp(RangeExpr,GeneratorList)),nl, %%
   csp_sets:extract_variables_from_generator_list(GeneratorList,Vars),
   term_variables(GeneratorList,GVars),
   remove_variables(GVars,Vars,WaitVars),
   when(ground(WaitVars), (
    %print(waitVars(WaitVars)),nl,
          findall(EExpr,(treat_list_generators(GeneratorList,Vars,Seqs,Guard),
                         %% print(treated_generators(Vars,Sets,Guard)),nl, %%
                         haskell_csp:check_boolean_expression(Guard),
                         generator_list_sol(Vars,Seqs), 
                         csp_sets:member_range_expr(RangeExpr,EExpr)),Expressions),
           Res=list(Expressions)
           %,print(expanded(Res)),nl
           )).
           

generator_list_sol([],[]).
generator_list_sol([Var|VT],[Seq|ST]) :- 
     haskell_csp:force_evaluate_argument(Seq,EvSeq),expand_sequence(EvSeq,ESeq),
    (ground(ESeq) -> true ; add_error(expand_set_comprehension,'Non-ground Seq: ',Seq)),
    is_elem_list(Var,ESeq), generator_list_sol(VT,ST).

treat_list_generators([],V,S,G) :- !, V=[], S=[], G=true.
treat_list_generators([comprehensionGenerator(Var,Seq)|T],Vars,Seqs,TG) :- !, Vars=[Var|VT], Seqs=[Seq|ST],
   treat_list_generators(T,VT,ST,TG).
treat_list_generators([comprehensionGuard(Guard)|T],VT,ST,bool_and(Guard,TGuard)) :- !,
   treat_list_generators(T,VT,ST,TGuard).
treat_list_generators(G,_,_,_) :- add_internal_error('Could not treat Seq Comprehension Generator List: ',G),fail.
           
