% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Pragma handling

:- module(pragmas, [ global_pragma/2
                   , apply_pragmas_to_event_b_machine/3
                   , extract_pragmas_from_event_b_extensions/2
                   ]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module handles pragmas loaded with a b machine').

:- use_module(library(lists)).

:- use_module(bmachine_structure).

:- dynamic global_pragma/2.

:- use_module(probsrc(eventhandling),[register_event_listener/3]).

init :-
    register_event_listener(start_initialising_specification,retractall(global_pragma(_,_)),'clean stored pragmas').
init.

% pragma rules: pragma(Pragma, Expr, Type, Info)
%pragma_rule(X,E,T,I) :- print(pragma_rule(X,E,T,I)),nl,fail.
pragma_rule(label,IdList, _, _, label(IdList)) :- IdList = [_|_].
pragma_rule(unit,Unit, _, _, unit(Unit)).
pragma_rule(inferred_unit,Unit, _, _, inferred_unit(Unit)).
pragma_rule(conversion,[], _, _, conversion).
pragma_rule(symbolic,[],_,_,prob_annotation('SYMBOLIC')).

% pragma which apply to the entire file; they do not need a location
global_pragma(generated).
global_pragma(unit_alias).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Event-B
extract_pragmas_from_event_b_extensions(Ext,Pragmas) :-
    findall(P,
            (member(P,Ext), P=pragma(_T,_S,_A,_C)),
            Pragmas).

% currently possible pragmas:
% - attach a unit to a variable
% - attach a unit to a constant
% - mark a concrete constant as abstract/symbolic -> this moves the constant!
apply_pragmas_to_event_b_machine(Pragmas,Machine,ResultingMachine) :-
    get_section(abstract_constants,Machine,AbsConstants),
    maplist(apply_event_b_pragmas(Pragmas),AbsConstants,NewAbsConstants),
    get_section(concrete_constants,Machine,ConcConstants),
    maplist(apply_event_b_pragmas(Pragmas),ConcConstants,NewConcConstants),
    get_section(abstract_variables,Machine,AbsVars),
    maplist(apply_event_b_pragmas(Pragmas),AbsVars,NewAbsVars),
    get_section(concrete_variables,Machine,ConcVars),
    maplist(apply_event_b_pragmas(Pragmas),ConcVars,NewConcVars),
    write_section(abstract_constants,NewAbsConstants,Machine,TmpMachine1),
    write_section(concrete_constants,NewConcConstants,TmpMachine1,TmpMachine2),
    write_section(abstract_variables,NewAbsVars,TmpMachine2,TmpMachine3),
    write_section(concrete_variables,NewConcVars,TmpMachine3,ResultingMachine).

apply_event_b_pragmas(Pragmas,b(identifier(Name),Type,Infos),b(identifier(Name),Type,Infos2)) :-
    member(pragma(PragmaType,_Machine,Name,Content),Pragmas)
    -> pragma_rule(PragmaType,Content,b(identifier(Name),Type,Infos),Type,AdditionalArgument),
       Infos2 = [AdditionalArgument|Infos]
    ;  Infos2 = Infos.
