% (c) 2014-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(smtlib2_cli,[smtlib2_file/1]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,smtlib).
:- module_info(description,'Command line interface and REPL for SMT-LIB 2 files.').

:- use_module(smtlib_solver(smtlib2_interpreter)).
:- use_module(smtlib_solver(smtlib2_parser)).
:- use_module(smtlib_solver(smtlib2_environment)).

:- use_module(probsrc(debug),[debug_println/2]).
:- use_module(probsrc(preferences), [set_preference/2]).
:- use_module(probsrc(error_manager),[add_error/3]).

read_file(Stream,Out) :-
  get_code(Stream,Code),
  read_file(Code,Stream,Out).
read_file(-1,_Stream,[]) :- !.
read_file(Code,Stream,[Code|L]) :-
  read_file(Stream,L).

smtlib2_file(Filename) :-
  debug_println(10,processing_smt_file(Filename)),
  % enable disprover mode to trigger modified treatment of wd conditions, etc.
  % some default preferences
  set_preference(disprover_mode,true),
  set_preference(use_clpfd_solver,true),
  set_preference(use_smt_mode,true),

  debug_println(25,opening(Filename)),nl,
  open(Filename,read,S),
  read_file(S,Content),
  close(S),
  debug_println(25,parsing(Filename)),nl,
  parse_smtlib2(Content,ListOfCommands),
  !,
  empty_env(EnvIn),
  debug_println(25,interpreting_commands(ListOfCommands)),
  on_exception(halt(0),
       (interpret_l(ListOfCommands,EnvIn,EnvOut),
       debug_println(25,finished_processing_and_starting_repl(Filename)),
       smtlib2_repl(EnvOut)),
       debug_println(25,finished_processing_exit(Filename))).
smtlib2_file(Filename) :-
  add_error(smtlib2_cli,'Could not process file: ',Filename).

smtlib2_repl(EnvIn) :-
  read_line(Line),
  (Line = end_of_file -> halt(0) ; append(RealLine,".",Line)),
  (parse_smtlib2(RealLine,Commands)
  -> interpret_l(Commands,EnvIn,EnvOut)
  ;  add_error(smtlib2_cli,'Parser error on line: ',RealLine)
  ),
  smtlib2_repl(EnvOut).
