:- module(location_vars_to_program,[location_vars_to_program/9]).

:- use_module(synthesis(constraints)).
:- use_module(library(lists),[append/2,nth0/3,remove_dups/2,maplist/3]). 

% convert location values to a program represented as an ast
% collect all location variables that are part of the solution (no deadcode)
location_vars_to_program(Library,Type,CurrentVars,LocationVars,M1,Parameter,solution(Solution),UsedLVars,AST) :- 
    % split by temporary input and output variables
    findall(In,(member(In,LocationVars) , In = b(identifier(Name),_,_) , 
         atom_concat(li,_,Name)),LocationVarsIn) , 
    findall(Out,(member(Out,LocationVars) , Out = b(identifier(Name),_,_) , 
         atom_concat(lo,_,Name)),LocationVarsOut) , 
    % get output location variable defined in the last program line
    (Type \= action
    ->  member(binding(Root,int(M1),_),Solution) , atom_concat(lo,_,Root) , 
        location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,Root,UsedLVars,AST)
    ;   length(CurrentVars,OutputAmount) , 
        M0 is M1 - OutputAmount , 
        get_program_for_output_var(Library,M0,M1,CurrentVars,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,TempUsedLVars,ASTList) , 
        remove_dups(TempUsedLVars,UsedLVars) , 
        bsyntaxtree:conjunct_predicates(ASTList,AST)) , !.

% for synthesis of actions
% if we have more than one output we need to collect the partial programs and insert 
% assigments with primed variables
get_program_for_output_var(_,M0,M0,_,_,_,_,_,_,[],[]).
get_program_for_output_var(Library,M0,ProgramLine,CurrentVars,[Var|Vars],LocationVarsIn,LocationVarsOut,Parameter,Solution,UsedLVars,[Assignment|T]) :- 
    member(binding(Root,int(ProgramLine),_),Solution) , atom_concat(lo,_,Root) ,  
    location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,Root,TempUsedLVars,AST) , 
    bmachine_eventb:prime_variable(Var,PrimedVar) , 
    Assignment = b(equal(PrimedVar,AST),pred,[]) , 
    NProgramLine is ProgramLine - 1 , 
    get_program_for_output_var(Library,M0,NProgramLine,CurrentVars,Vars,LocationVarsIn,LocationVarsOut,Parameter,Solution,R,T) , 
    append(TempUsedLVars,R,UsedLVars).

set_type(b(member(b(Elm,EType,EInfo),Set),NodeType,[]),b(member(b(Elm,EType,EInfo),NSet),NodeType,[])) :- 
    set_type_aux(EType,Set,NSet).
set_type(b(Node,Type,Info),b(NewNode,Type,Info)) :- 
    Node =.. [Head,Arg] , Head \= identifier , 
    set_type(Arg,NArg) , 
    NewNode =.. [Head,NArg].
set_type(b(Node,Type,Info),b(NewNode,Type,Info)) :- 
    Node =.. [Head,Arg1,Arg2] , 
    set_type(Arg1,NArg1) , 
    set_type(Arg2,NArg2) , 
    NewNode =.. [Head,NArg1,NArg2].
set_type(Node,Node).

set_type_aux(Type,b(set_extension(Set),_,SInfo),b(set_extension(NSet),set(Type),SInfo)) :- 
    maplist(set_type_aux(Type),Set,NSet).
set_type_aux(Type,b(Node,_,Info),b(Node,Type,Info)).

% constant assignment
location_to_ast(_Library,_CurrentVars,_LocationVarsIn,LocationVarsOut,_Parameter,_Solution,LocationVarName,[LocationVarNode],AST) :- 
    LocationVarNode = b(identifier(LocationVarName),_Type,Info) , 
    member(LocationVarNode,LocationVarsOut) , 
    sets:subset([synthesis(ConstantName,output),synthesis(type,Type)],Info) , 
    delete_numbers_from_atom(ConstantName,OpName) ,  
    OpName = constant_set , ! , 
    %member(binding(ConstantName,Value,_),Solution) , 
    %value_to_ast(Value,AST). 
    % export as infinite set
    get_single_operator_constraint_aux1(Type,_,AST). 
location_to_ast(_Library,_CurrentVars,_LocationVarsIn,LocationVarsOut,_Parameter,Solution,LocationVarName,[LocationVarNode],AST) :- 
    LocationVarNode = b(identifier(LocationVarName),_Type,Info) , 
    member(LocationVarNode,LocationVarsOut) , 
    member(synthesis(ConstantName,output),Info) , 
    delete_numbers_from_atom(ConstantName,OpName) ,  
    OpName = constant , ! , 
    % get constant solution, either integer or boolean 
    member(binding(ConstantName,Value,_),Solution) , 
    value_to_ast(Value,AST). 
% output 
% one input parameter
location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,LocationVarName,[LocationVarNode|RUsedLVars],b(Node,Type,[])) :- 
    LocationVarNode = b(identifier(LocationVarName),_Type,InfoOut) , 
    member(LocationVarNode,LocationVarsOut) , 
    sets:subset([synthesis(Operator,output),synthesis(type,Type)],InfoOut) , 
    delete_numbers_from_atom(Operator,Op) , 
    member(Op,[negation,card,size,max,min,general_union,general_intersection]) , ! , 
    % find input location variables 
    member(b(identifier(InName),_Type,InfoIn),LocationVarsIn) , 
    member(synthesis(Operator,input),InfoIn) , 
    location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,InName,RUsedLVars,InputAST) ,  
    Node =.. [Op,InputAST]. 
% two input parameters
location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,LocationVarName,UsedLVars,AST) :- 
    LocationVarNode = b(identifier(LocationVarName),_Type,InfoOut) , 
    member(LocationVarNode,LocationVarsOut) , ! , 
    member(synthesis(Operator,output),InfoOut) , 
    % find input location variables 
    member(b(identifier(InName1),_,Info1),LocationVarsIn) , 
    member(synthesis(Operator,left_input),Info1) , 
    member(b(identifier(InName2),_,Info2),LocationVarsIn) , 
    member(synthesis(Operator,right_input),Info2) , 
    location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,InName1,LeftUsedLVars,LeftInputAST) , 
    location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,InName2,RightUsedLVars,RightInputAST) , 
    delete_numbers_from_atom(Operator,OpName) , 
    % get the type from the output nodes information
    member(synthesis(type,Type),InfoOut) , 
    Node =.. [OpName,LeftInputAST,RightInputAST] , 
    append([[LocationVarNode],LeftUsedLVars,RightUsedLVars],UsedLVars) , 
    set_type(b(Node,Type,[]),AST). 
% input
location_to_ast(_Library,CurrentVars,_,_,Parameter,Solution,LocationVarName,[],b(identifier(InputName),Type,[])) :- 
    member(binding(LocationVarName,int(Value),_),Solution) , 
    Value < Parameter ,
    % input parameters are enumerated, so we can pick the right variable 
    % from our currently observed variables 
    nth0(Value,CurrentVars,IDNode) , 
    IDNode = b(identifier(InputName),Type,_).
location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,LocationVarName,UsedLVars,AST) :- 
    member(binding(LocationVarName,int(Value),_),Solution) , 
    Value >= Parameter , 
    % temporary output variable 
    LocationOut = b(identifier(LocationOutName),_,_) , 
    % get output defined at corresponding location 
    member(binding(LocationOutName,int(Value),_),Solution) , 
    LocationVarName \= LocationOutName , 
    member(LocationOut,LocationVarsOut) , 
    location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,LocationOutName,UsedLVars,AST).