:- module(user_interaction,[user_interaction/8,
                            user_interaction_distinguishing_input/3,
                            user_interaction_distinguishing_transition/4,
                            user_interaction_add_missing_transition/4,
                            example_bindings_to_ast_nodes/2,
                            remove_prime/2,
                            check_examples_from_user/9,
                            pretty_example_to_ast_nodes/3]).

:- use_module(synthesis(constraints)).
:- use_module(probsrc(specfile),[expand_const_and_vars_to_full_store/2]).
:- use_module(probsrc(bmachine),[b_get_machine_variables/1]).
:- use_module(probsrc(translate),[translate_bexpression/2,translate_bvalue/2]).
:- use_module(probsrc(bsyntaxtree),[add_texpr_infos/3]).
:- use_module(library(lists),[append/2,maplist/3,is_list/1,nth0/3]).
:- use_module(library(avl),[avl_to_list/2]).

% general user interaction
user_interaction(Vars,TExamplesAssumingNo,TExamplesAssumingYes,PositiveIn,NegativeIn,PositiveOut,NegativeOut,SynthesisType) :- 
    format("Please select positive and negative examples~n~n",[]) , 
    maplist(expand_const_and_vars_to_full_store,TExamplesAssumingNo,ExamplesAssumingNo) , 
    maplist(expand_const_and_vars_to_full_store,TExamplesAssumingYes,ExamplesAssumingYes) , 
    user_interaction_graduate_examples(Vars,ExamplesAssumingNo,no,TempPosIn1,TempNegIn1,TempPosOut1,TempNegOut1,Type1) , nl , 
    user_interaction_graduate_examples(Vars,ExamplesAssumingYes,yes,TempPosIn2,TempNegIn2,TempPosOut2,TempNegOut2,Type2) , nl , 
    user_interaction_add_further_examples(Vars,TempPosIn3,TempNegIn3,TempPosOut3,TempNegOut3) , 
    % type is either action or a variable 
    (nonvar(Type1) -> Type = Type1 ; Type = Type2) , 
    append([TempPosIn1,TempPosIn2,TempPosIn3],PosInput) , 
    append([TempNegIn1,TempNegIn2,TempNegIn3],NegInput) ,
    append([TempPosOut1,TempPosOut2,TempPosOut3],PosOutput) , 
    append([TempNegOut1,TempNegOut2,TempNegOut3],NegOutput) , 
    check_examples_from_user(Vars,PosInput,NegInput,PosOutput,NegOutput,TPositiveIn,NegativeIn,TPositiveOut,NegativeOut) , 
    user_interaction_aux(Type,Vars,ExamplesAssumingNo,TPositiveIn,NegativeIn,TPositiveOut,NegativeOut,PositiveIn,NegativeIn,PositiveOut,NegativeOut,SynthesisType).

% when synthesizing an invariant or guard we have to delete all negative inputs from
% the positive ones, because we replace outputs either by 'TRUE' or 'FALSE' transitions can be equal
user_interaction_aux(Type,Vars,ExamplesAssumingNo,TTPositiveIn,NegativeIn,TTPositiveOut,NegativeOut,PositiveIn,NegativeIn,PositiveOut,NegativeOut,SynthesisType) :- 
    var(Type) , ! , 
    % either type is set during user_interaction (action) or we need to set it (guard or invariant)
    determine_synthesis_type(ExamplesAssumingNo,TTPositiveIn,TTPositiveOut,SynthesisType) , 
    post_select_transitions(NegativeIn,TTPositiveIn,TTPositiveOut,TPositiveIn,TPositiveOut) , 
    check_examples_from_user(Vars,TPositiveIn,NegativeIn,TPositiveOut,NegativeOut,PositiveIn,NegativeIn,PositiveOut,NegativeOut).
user_interaction_aux(action,_,_,PositiveIn,NegativeIn,PositiveOut,NegativeOut,PositiveIn,NegativeIn,PositiveOut,NegativeOut,action).

% if an example from model checking was edited by the user we synthesize actions
determine_synthesis_type(_,_,_,Type) :- 
    % so the type is already set
    Type == action , !.
determine_synthesis_type(ExamplesAssumingNo,PosInput,PosOutput,Type) :- 
    determine_synthesis_type_aux(ExamplesAssumingNo,NegInput,NegOutput) , 
    % else if an example assumed no is graduated as valid by the user we synthesize invariants
    ( ((member(In,NegInput) , member(In,PosInput)) , 
       (nth0(IndexNeg,NegInput,In) , nth0(IndexPos,PosInput,In) , 
        nth0(IndexNeg,NegOutput,Out) , nth0(IndexPos,PosOutput,Out)))
    ->  Type = invariant
    % else we synthesize guards
    ;   Type = guard).

determine_synthesis_type_aux([],[],[]).
determine_synthesis_type_aux([(AssumingNoIn,AssumingNoOut)|T],[NegInput|NTIn],[NegOutput|NTOut]) :- 
    findall(Node,(member(bind(_,Val),AssumingNoIn) , value_to_ast(Val,b(N,Type,_)) , Node = b(N,Type,_)),NegInput) , 
    findall(Node,(member(bind(_,Val),AssumingNoOut) , value_to_ast(Val,b(N,Type,_)) , Node = b(N,Type,_)),NegOutput) , 
    determine_synthesis_type_aux(T,NTIn,NTOut).

% delete negative inputs from positive ones
post_select_transitions(NegativeIn,TPositiveIn,TPositiveOut,PositiveIn,PositiveOut) :- 
    findall(Pos,(member(Pos,TPositiveIn) , \+member(Pos,NegativeIn)),PositiveIn) , 
    % get all matching outputs 
    findall(Index,(member(Pos,PositiveIn) , nth0(Index,TPositiveIn,Pos)),Indices) , 
    findall(PosOut,(member(PosOut,TPositiveOut) , nth0(Index,TPositiveOut,PosOut) , 
                    member(Index,Indices)),PositiveOut).

% expect at least one positive and negative example
check_examples_from_user(Vars,[],NegInput,[],NegOutput,[PosInput],NegInput,[PosOutput],NegOutput) :- 
    format("Add one positive example~n",[]) , 
    read(Example) , 
    Example = (TPosInput-->TPosOutput) , 
    pretty_example_to_ast_nodes(TPosInput,Vars,PosInput) , 
    pretty_example_to_ast_nodes(TPosOutput,Vars,PosOutput).
check_examples_from_user(Vars,PosInput,[],PosOutput,[],[PosInput],NegInput,[PosOutput],NegOutput) :- 
    format("Add one negative example~n",[]) , 
    read(Example) , 
    Example = (TNegInput-->TNegOutput) , 
    pretty_example_to_ast_nodes(TNegInput,Vars,NegInput) , 
    pretty_example_to_ast_nodes(TNegOutput,Vars,NegOutput).
check_examples_from_user(_,PosInput,NegInput,PosOutput,NegOutput,PosInput,NegInput,PosOutput,NegOutput).

user_interaction_add_missing_transition(Vars,Input,Output,VarNodes) :- 
    format("Do you want to add a transition? [example/no] ",[]) , 
    read(Example) , 
    Example = (TPosIn-->TPosOut) , ! , 
    TPosIn \= [] , TPosOut \= [] , 
    findall(VarName,member(b(identifier(VarName),_,_),Vars),TempVarNames) , 
    b_get_machine_variables(MachineVars) , 
    length(MachineVars,VarAmount) , 
    length(TPosIn,L1) , length(TPosOut,L2) , 
    ((L1 \= VarAmount ; L2 \= VarAmount)
    ->  format("Please name the variables you refer to (list expected):",[]) , 
        read(VarNames) , 
        findall(b(identifier(VarName),Type,[]),(member(VarName,VarNames) , member(b(identifier(VarName),Type,_),MachineVars)),VarNodes) , 
        length(VarNodes,L1) 
    ;   VarNodes = MachineVars , 
        VarNames = TempVarNames) , 
    pretty_example_to_ast_nodes(TPosIn,VarNames,PosIn) , 
    pretty_example_to_ast_nodes(TPosOut,VarNames,PosOut) , 
    user_interaction_add_missing_transition(Vars,FPosIn,FPosOut,_FVarNodes) , % same var nodes expected
    Input = [PosIn|FPosIn] , Output = [PosOut|FPosOut].
user_interaction_add_missing_transition(_,[],[],[]).

% user interaction to optionally add further examples
user_interaction_add_further_examples(Vars,PosIn,NegIn,PosOut,NegOut) :- 
    format("Do you want to add any other example? [example/no] ",[]) , 
    read(Example) , 
    Example = (_-->_) , ! , 
    user_interaction_graduate_examples(Vars,[Example],none,TempPosIn1,TempNegIn1,TempPosOut1,TempNegOut1,_) , 
    user_interaction_add_further_examples(Vars,TempPosIn2,TempNegIn2,TempPosOut2,TempNegOut2) , 
    append(TempPosIn1,TempPosIn2,PosIn) , 
    append(TempNegIn1,TempNegIn2,NegIn) , 
    append(TempPosOut1,TempPosOut2,PosOut) , 
    append(TempNegOut1,TempNegOut2,NegOut).
user_interaction_add_further_examples(_,[],[],[],[]).

% user interaction to select positive and negative examples 
% Assuming is either 'yes'/'no' or 'none' if no information is given
% if the user changes an example unify Type with action, else Type will stay a variable
user_interaction_graduate_examples(_,[],_,[],[],[],[],_).
user_interaction_graduate_examples(Vars,[Example|R],Assuming,PosIn,NegIn,PosOut,NegOut,Type) :- 
    % examples can be bindings from the solver 
    (Example = (In,Out)
    ->  maplist(example_bindings_to_ast_nodes,[In,Out],[InputNodes,OutputNodes]) , 
        maplist(translate_bexpression,InputNodes,RawInput) , 
        maplist(translate_bexpression,OutputNodes,RawOutput) 
    ;   Example = (RawInput-->RawOutput) , 
        findall(Name,member(b(identifier(Name),_,_),Vars),VarNames) , 
        pretty_example_to_ast_nodes(RawInput,VarNames,InputNodes) , 
        pretty_example_to_ast_nodes(RawOutput,VarNames,OutputNodes)) , 
    % only give the opportunity to change output of examples from model checking, 
    % if there is no assuming it's an example from the user
    (Assuming = none
    ->  format("Is ~w-->~w a valid example? [yes/no] ",[RawInput,RawOutput]) 
    ;   format("Is ~w-->~w a valid example? (assuming ~w) [yes/no/edit] ",[RawInput,RawOutput,Assuming])) , 
    read(Answer) , 
    (Answer = edit
    ->  Type = action , 
        user_interaction_change_output(RawInput,RawOutput,TNewOutput) , 
        pretty_example_to_ast_nodes(TNewOutput,Vars,NewOutput) , 
        user_interaction_graduate_examples_aux([InputNodes|OldPosIn],OldNegIn,[NewOutput|OldPosOut],OldNegOut,PosIn,NegIn,PosOut,NegOut)
    ;   ((Answer = yes
        ->  user_interaction_graduate_examples_aux([InputNodes|OldPosIn],OldNegIn,[OutputNodes|OldPosOut],OldNegOut,PosIn,NegIn,PosOut,NegOut)
        ;   Answer = no , 
            user_interaction_graduate_examples_aux(OldPosIn,[InputNodes|OldNegIn],OldPosOut,[OutputNodes|OldNegOut],PosIn,NegIn,PosOut,NegOut))) , 
        user_interaction_graduate_examples(Vars,R,Assuming,OldPosIn,OldNegIn,OldPosOut,OldNegOut,Type)).

% unify variables
user_interaction_graduate_examples_aux(PosInput,NegInput,PosOutput,NegOutput,PosInput,NegInput,PosOutput,NegOutput).

user_interaction_change_output(Input,Output,NewOutput) :- 
    format("New output of ~w:~n",[Input]) , 
    read(TempNewOutput) , 
    ((length(Input,L) , length(TempNewOutput,L))
    ->  NewOutput = TempNewOutput
    ;   format("Wrong amount of variables. Please try again",[]) , nl , 
        user_interaction_change_output(Input,Output,NewOutput)).

% user decides between to non-equivalent solutions by graduating a distinguishing 
% input/state 
user_interaction_distinguishing_input(Type,solution(Distinguishing),Answer) :- 
    findall(Value,member(binding(_,Value,_),Distinguishing),Values) , 
    maplist(translate_bvalue,Values,PrettyValues) , 
    (Type = invariant
    ->  format("Is ~w a valid state? [yes/no]",[PrettyValues])
    ;   format("Is ~w a valid input? [yes/no]",[PrettyValues])) , 
    read(TAnswer) , 
    (\+member(TAnswer,[yes,no])
    ->  user_interaction_distinguishing_input(solution(Distinguishing),Answer)
    ;   Answer = TAnswer).
% or transition as valid or invalid
user_interaction_distinguishing_transition(solution(Distinguishing),CurrentVars,Answer,OutputNodes) :- 
    findall(Value,(member(binding(Var,Temp,_),Distinguishing) , \+bmachine_eventb:is_primed_id(Var) , 
                   translate_bvalue(Temp,Value)),In) , 
    findall(Node,(member(binding(Var,Temp,_),Distinguishing) , bmachine_eventb:is_primed_id(Var) , 
                   value_to_ast(Temp,TNode) , add_texpr_infos(TNode,[synthesis(machinevar,Var)],Node)),TOutputNodes) , 
    maplist(translate_bexpression,TOutputNodes,Out) , 
    format("Is ~w-->~w a valid transition? [yes/no]",[In,Out]) , 
    read(TAnswer) , 
    (TAnswer = no 
     ->  user_interaction_distinguishing_transition_aux(CurrentVars,Answer,OutputNodes)
     ;   Answer = yes , 
         OutputNodes = TOutputNodes).
user_interaction_distinguishing_transition_aux(CurrentVars,Answer,DistOut) :- 
    format("Please give the desired output: [example/none]~n",[]) ,  
    read(Answer) , 
    (Answer = none 
    ->  DistOut = none
    ;   findall(Name,member(b(identifier(Name),_,_),CurrentVars),CurrentVarNames) , 
        pretty_example_to_ast_nodes(Answer,CurrentVarNames,DistOut)).

example_bindings_to_ast_nodes(BindList,NodeList) :- 
    member(binding(_,_,_),BindList) , 
    findall(NodeIn,(member(binding(TVarName,In,_),BindList) , value_to_ast(In,TNodeIn) , remove_prime(TVarName,VarName) , 
                   add_texpr_infos(TNodeIn,[synthesis(machinevar,VarName)],NodeIn)),TempNodeList) , 
    maplist(set_type_if_necessary,TempNodeList,NodeList).
example_bindings_to_ast_nodes(BindList,NodeList) :- 
    member(bind(_,_),BindList) , 
    findall(NodeIn,(member(bind(TVarName,In),BindList) , value_to_ast(In,TNodeIn) , remove_prime(TVarName,VarName) , 
                   add_texpr_infos(TNodeIn,[synthesis(machinevar,VarName)],NodeIn)),TempNodeList) , 
    maplist(set_type_if_necessary,TempNodeList,NodeList).

remove_prime(b(identifier(VarName),Type,Info),b(identifier(UnPrimed),Type,Info)) :- 
    remove_prime(VarName,UnPrimed).
remove_prime(VarName,UnPrimed) :- 
    atom_concat(UnPrimed,'\'',VarName).
remove_prime(VarName,VarName).

pretty_example_to_ast_nodes([],[],[]).
pretty_example_to_ast_nodes([Value|T],[VarName|VT],[Node|NT]) :- 
    value_to_ast(Value,TempNode) , 
    add_texpr_infos(TempNode,[synthesis(machinevar,VarName)],TNode) , 
    set_type_if_necessary(TNode,Node) , 
    pretty_example_to_ast_nodes(T,VT,NT).

% we have to set the types of untyped ast nodes like empty sets
set_type_if_necessary(b(Node,Type,Info),b(Node,GroundType,Info)) :- 
    untyped(Type) , 
    b_get_machine_variables(MachineVars) , 
    member(synthesis(machinevar,VarName),Info) , 
    member(b(identifier(VarName),GroundType,_),MachineVars).
set_type_if_necessary(b(Node,Type,Info),b(Node,Type,Info)).

untyped(Type) :- 
    var(Type).
untyped(set(Type)) :- 
    untyped(Type).
untyped(seq(Type)) :- 
    untyped(Type).