% (c) 2014-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(table_tools, [expand_and_translate_to_table/3,
                        print_table/4,  print_table/2,
                        print_value_as_table/4,  print_value_as_table/2
                        ]).

:- use_module(translate).
:- use_module(tools_strings,[ajoin/2]).
:- use_module(library(lists)).
:- use_module(self_check).
:- use_module(error_manager).

print_value_as_table(Expr,Value) :- print_value_as_table(user_output,Expr,Value,[]).

print_value_as_table(Stream,Expr,Value,Opts) :-
    expand_and_translate_to_table(Value,Header,Table,Opts),
    print_value_as_table2(Stream,Expr,Header,Table,Opts).

print_value_as_table2(Stream,Expr,Header,Table,Opts) :-
    get_header_from_expression(Expr,ExprHeader0,Opts),
    % check if ExprHeader is valid:
    (member('no-row-numbers',Opts) -> ExprHeader = ExprHeader0 ; ExprHeader = ['Nr'|ExprHeader0]),
    same_length(Header,ExprHeader),
    !,
    print_table(Stream,ExprHeader,Table,Opts).
print_value_as_table2(Stream,_Expr,Header,Table,Opts) :-
       print_table(Stream,Header,Table,Opts).

:- use_module(tools,[latex_escape_atom/2]).
:- use_module(bsyntaxtree,[get_texpr_ids/2]).
get_header_from_expression(b(comprehension_set(IDs,_),_,_),Header,Opts) :-
   get_texpr_ids(IDs,Hs),
   (member(latex,Opts) -> maplist(latex_escape_atom,Hs,Header) ; Header=Hs).

% expand a SetValue into a Header List and a Table List

:- assert_must_succeed(( expand_and_translate_to_table([rec([field(a,int(1)),field(b,int(2))])],H,T),
                          H == ['Nr',a,b], T==[list([1,'1','2'])] )).
:- assert_must_succeed(( expand_and_translate_to_table([(int(1),int(2)),(int(33),int(44))],H,T),
                          H == ['Nr',prj1,prj2], T==[list([1,'1','2']),list([2,'33','44'])] )).
:- assert_must_succeed(( expand_and_translate_to_table([(int(1),rec([field(f,int(2)),field(g,int(3))]))],H,T),
                          H==['Nr',prj1,'prj2\'f','prj2\'g'], T==[list([1,'1','2','3'])] )).
expand_and_translate_to_table(Value,Header,Table) :-  
    expand_and_translate_to_table(Value,Header,Table,[]).
expand_and_translate_to_table(Value,Header,Table,Opts) :-  
    expand(Value, ExpVal),
    translate_to_table(ExpVal,1,Table,Header,Opts). %, print(table(Res)),nl.

:- use_module(custom_explicit_sets,[is_custom_explicit_set/1]).
expand(Var,Res) :- var(Var),!,Res=[Var].
expand((A,B),[(A,B)]) :- !. % wrap into a singleton set
expand(rec(Fields),[rec(Fields)]) :- !. % wrap into a singleton set
expand(Value,ExpVal) :- is_custom_explicit_set(Value),!,
      custom_explicit_sets:try_expand_custom_set(Value, ExpVal).
expand([H|T],V) :- !, V=[H|T].
expand([],V) :- !, V=[].
expand(Val,[Val]).

translate_to_table(A,RowNr,[list(L)],Header,Opts) :- var(A),!,
    (member('no-row-numbers',Opts) -> Header = ['Value'],L=AT ; Header = ['Nr','Value'],L=[RowNr|AT]),
    translate:translate:translate_bvalue(A,AT).
translate_to_table([],_RowNr,[],Header,Opts) :- !,
    (member('no-row-numbers',Opts) -> Header = ['Elements'] ; Header = ['Nr','Elements']).
translate_to_table([A|T],RowNr,[list(LA)|TT],Header,Opts) :- member('no-row-numbers',Opts),
    !,
    get_table_entry(A,LA,Header,Opts),
    R1 is RowNr+1,
    translate_to_table(T,R1,TT,_,Opts).
translate_to_table([A|T],RowNr,[list([RowNr|LA])|TT],['Nr'|Header],Opts) :- !,
    get_table_entry(A,LA,Header,Opts),
    R1 is RowNr+1,
    translate_to_table(T,R1,TT,_,Opts).
translate_to_table(A,_RowNr,[list(AT)],['Value'],Opts) :- member('no-row-numbers',Opts),!,
    translate:translate:translate_bvalue(A,AT).
translate_to_table(A,RowNr,[list([RowNr|AT])],['Nr','Value'],_) :- !,
    translate:translate:translate_bvalue(A,AT).

get_table_entry(Var,LA,Header,_) :- var(Var),!, LA = [VT], Header = ['prj1'],
    translate:translate:translate_bvalue(Var,VT).
get_table_entry(rec(Fields),Vals,FieldNames,Opts) :- !,
    maplist(get_record_field_val_name(Opts),Fields,Vals,FieldNames).
get_table_entry(A,[AT],['Elements'],_) :- A \= (_,_),!,
    translate:translate:translate_bvalue(A,AT).
get_table_entry(A,AT,HA,Opts) :- %A=(A1,A2), (A1 = (_,_) ; A2=(_,_)),
    !,
    get_table_entry2(A,[],AT,HA,Opts).
%get_table_entry((A,B),[AT,BT],['prj1','prj2']) :- !,
%    translate:translate:translate_bvalue(A,AT),
%    translate:translate:translate_bvalue(B,BT).

get_record_field_val_name(Opts,field(Name,Value),ValS,ColName) :-
    (member(latex,Opts) -> latex_escape_atom(Name,ColName) ; ColName=Name),
    translate:translate:translate_bvalue(Value,ValS).

get_table_entry2(A,Path,[AT],[Header],_Opts) :- var(A),!,
    reverse(Path,RP), ajoin(['prj'|RP],Header), % to do: do all this more efficiently
    translate:translate:translate_bvalue(A,AT).
get_table_entry2((A,B),Path,ABT,HAB,Opts) :- !,
    get_table_entry2(A,[1|Path],AT,HA,Opts),
    get_table_entry2(B,[2|Path],BT,HB,Opts),
    append(AT,BT,ABT), append(HA,HB,HAB).
get_table_entry2(rec([field(Name,Val)|Rest]),Path,Values,HAB,Opts) :- !,
    (member(latex,Opts) -> latex_escape_atom(Name,ColName) ; ColName=Name),
    get_table_entry2(Val,[ColName,'\''|Path],AT,HA,Opts),
    (Rest=[] -> Values=AT, HAB=HA
      ; get_table_entry2(rec(Rest),Path,BT,HB,Opts),
        append(AT,BT,Values), append(HA,HB,HAB)
    ).
get_table_entry2(A,Path,[AT],[Header],_Opts) :- !,
    reverse(Path,RP), ajoin(['prj'|RP],Header), % to do: do all this more efficiently
    translate:translate:translate_bvalue(A,AT).


valid_option(latex).
valid_option('no-tabular').
valid_option('no-hline').
valid_option('no-headings').
valid_option('no-row-numbers').
valid_option(argument_value('max-table-size',_)).
check_option(O) :- (valid_option(O) -> true ; add_error(table_tools,'Illegal option: ',O)).


print_table(Header,Table) :- print_table(user_output,Header,Table,[]).

print_table(Stream,Header,Table,Opts) :-
   maplist(check_option,Opts),
   (member(argument_value('max-table-size',Sze),Opts)
     -> split_list(Table,Sze,Tables), % split into individual tables, each with own header; useful for Latex
         maplist(print_table_aux_nl(Stream,Header,Opts),Tables)
     ;  print_table_aux(Stream,Header,Opts,Table)
   ).
print_table_aux_nl(Stream,Header,Opts,Table) :- print_table_aux(Stream,Header,Opts,Table),nl(Stream).

print_table_aux(Stream,Header,Opts,Table) :-
   print_table_header(Stream,Header,Opts),
   (member('no-headings',Opts) -> true ; 
      print_row(Stream,Opts,list(Header)),
      print_header_terminator(Stream,Opts)
   ),
   maplist(print_row(Stream,Opts),Table),
   print_table_footer(Stream,Opts).


print_row(Stream,Opts,list(Row)) :-
    print_row_aux(Row,Stream,Opts), 
    print_row_terminator(Stream,Opts).

print_row_aux(V,Stream,_Opts) :- var(V),!, format(Stream,'~w ... ',[V]).
print_row_aux([],_,_) :- !.
print_row_aux([H|T],Stream,Opts) :- !, print_entry(Stream,H,T,Opts), 
      print_row_aux(T,Stream,Opts).
print_row_aux(Other,Stream,_Opts) :- format(Stream,'~w ... ',[Other]).

get_latex_col(_,108). % 99 is lower-case l
%get_latex_col(_,99). % 99 is lower-case c

print_table_header(Stream,Header,Opts) :- member(latex,Opts),
   nonmember('no-tabular',Opts),
   !,
   maplist(get_latex_col,Header,LatexColSpec),
   format(Stream,'\\begin{tabular}{~s}~n',[LatexColSpec]),
   hline(Stream,Opts).
print_table_header(_,_,_).

print_row_terminator(Stream,Opts) :- member(latex,Opts),!, format(Stream,'\\\\~n',[]).
print_row_terminator(Stream,_) :- nl(Stream).

print_header_terminator(Stream,Opts) :- member(latex,Opts),!,
   format('\\\\~n',[]), 
   hline(Stream,Opts).
print_header_terminator(_,_).

print_table_footer(Stream,Opts) :- member(latex,Opts),
     nonmember('no-tabular',Opts),
     !,
     hline(Stream,Opts),
     format(Stream,'\\end{tabular}~n',[]).
print_table_footer(_,_).

hline(_,Opts) :- member('no-hline',Opts),!.
hline(Stream,_) :- format(Stream,'\\hline ~n',[]).

% TO DO: determine column length dynamically
print_entry(Stream,E,T,Opts) :- member(latex,Opts),!,
     print(Stream,' $'),print(Stream,E), (T==[] -> print(Stream,'$ ') ; print(Stream,'$ & ')).
print_entry(Stream,E,_,_) :- print(Stream,E), my_atom_length(E,Len), WS is 10-Len, print_ws(Stream,WS),!.
print_entry(Stream,E,_,_) :- add_internal_error('Failed: ',print_entry(Stream,E,_,_)),
     print(Stream,print_entry_failed(E)),nl(Stream).

print_ws(Stream,X) :- X>1, !, print(Stream,' '), X1 is X-1, print_ws(Stream,X1).
print_ws(Stream,_) :- print(Stream,' ').

my_atom_length(AtomOrNumber,Len) :- atom(AtomOrNumber),!, atom_length(AtomOrNumber,Len).
my_atom_length(Var,Len) :- var(Var),!, Len=6. % no idea how long a variable is
my_atom_length(C,Len) :- compound(C),!, Len=6. % no idea how long a compound is
my_atom_length(A,Len) :- name(A,List), length(List,Len).

% split lists into sub-lists of at most length Maxlen
split_list([],_,[]).
split_list([H|T],Maxlen,[ [H|T1] |Rest]) :- get_list(T,Maxlen,T1,RT), split_list(RT,Maxlen,Rest).

get_list([],_,[],[]).
get_list([H|T],Max,[H|T1],Rest) :- Max>1, !, M1 is Max-1, get_list(T,M1,T1,Rest).
get_list(L,_,[],L).
