% (c) 2009-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(tools_commands,
    [ edit_file/1,edit_file/2, diff_files_with_editor/2,
      show_errors_with_bb_results/0, show_errors_with_bb_results/1 ]).
    

:- use_module(preferences).
:- use_module(debug).
:- use_module(tools).
:- use_module(error_manager).
:- use_module(system_call,[system_call/4]).

edit_file(F) :- edit_file(F,none).
edit_file(FILE,LINE) :-
     get_preference(path_to_text_editor_launch,EDITOR),
     edit_file_with_editor(EDITOR,FILE,LINE).

edit_file_with_editor(EDITOR,FILE,LINE) :- number(LINE),
     tools:get_tail_filename(EDITOR,EdName),
     editor_understands_plus_line(EdName),!, % bbedit/vim can handle +LINE arg
     format('Opening file "~w" at LINE ~w using EDITOR "~w"~n',[FILE,LINE,EDITOR]),
     tools:string_concatenate('+',LINE,LinePosArg),
     system_call:system_call(EDITOR,[LinePosArg,FILE],ErrorTextAsCodeList,ExitCode),
     debug_println(9,exit(ExitCode,ErrorTextAsCodeList)).
edit_file_with_editor(EDITOR,FILE,LINE) :- number(LINE),
     tools:get_tail_filename(EDITOR,atom),!, % atom can handle FILE:LINE
     format('Opening file "~w" at LINE ~w using EDITOR "~w"~n',[FILE,LINE,EDITOR]),
     ajoin([FILE,':',LINE],FILE_LINE),
     system_call:system_call(EDITOR,[FILE_LINE],ErrorTextAsCodeList,ExitCode),
     debug_println(9,exit(ExitCode,ErrorTextAsCodeList)).
edit_file_with_editor(EDITOR,FILE,_) :-
     format('Opening file "~w" using EDITOR "~w"~n',[FILE,EDITOR]),
     system_call:system_call(EDITOR,[FILE],ErrorTextAsCodeList,ExitCode),
     debug_println(9,exit(ExitCode,ErrorTextAsCodeList)).


editor_understands_plus_line(bbedit).
editor_understands_plus_line(vim).



diff_files_with_editor(FILE1,FILE2) :- 
     DIFFER = '/usr/local/bin/bbdiff', % TO DO: provide preference
     system_call:system_call(DIFFER,[FILE1,FILE2],ErrorTextAsCodeList,ExitCode),
     debug_println(9,exit(ExitCode,ErrorTextAsCodeList)).


:- use_module(library(process)).
:- use_module(system_call,[system_call_keep_open/7]).
show_errors_with_bb_results :- show_errors_with_bb_results([]).
show_errors_with_bb_results(Options) :-
    stored_error(_,_,_,Span,Options),
    extract_file_number_and_name(Span,_,_Filename),
    extract_line_col(Span,_,_,_,_),
    !,
    Command = '/usr/local/bin/bbresults', % TO DO: put in preference
    BBOptions = ['-p', 'flake8'], Env = [],
    on_exception(error(existence_error(_,_),E),
         (system_call_keep_open(Command,BBOptions,Process,STDIn,_STDOut,_STDErr,Env),
          write_errors_for_bbresults(STDIn,Options),
          close(STDIn),
          process_release(Process)),
          add_error(show_errors_with_bb_results,'Cannot start error viewer: ',Command:E)).
show_errors_with_bb_results(_). % no errors to show

stored_error(Source,S,TContext,Span,Options) :-
    (member(current,Options) -> logged_error(Source,S,TContext,Span)
     ; backed_up_error(Source,S,_TContext,Span)).

% ---------------------
% write errors in bbresults style (flake8)
% filename:line:column: W/E1 message

write_errors_for_bbresults(Stream,Options) :- 
   stored_error(Source,S,_TContext,Span,Options),
   extract_file_number_and_name(Span,_,Filename),
   extract_line_col(Span,SL,SC,_EL,_EC),
   StartCol is SC+1,
   (Source = warning(_) -> Type='W1' ; Type='E1'),
   format(Stream,'~w:~w:~w: ~w ~w~n',[Filename,SL,StartCol,Type,S]),
   %format(user_output,'~w:~w:~w: ~w ~w~n',[Filename,SL,SC,Type,S]),
   fail.
write_errors_for_bbresults(_,_).

