% (c) 2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(translate_keywords,[classical_b_keyword/1, translate_keyword_id/2,
           raw_backtranslate_keyword_ids/2,
           % info about translated identifiers when pretty-printing Event-B to classical B:
           keyword_to_id_cache/2, id_to_keyword_cache/2]).


:- use_module(module_information).
:- module_info(group,tools).
:- module_info(description,'This module is responsible for translating identifiers which happen to be keywords').

:- use_module(library(lists)).

% info about translated identifiers when pretty-printing Event-B to classical B:
:- dynamic keyword_to_id_cache/2, id_to_keyword_cache/2.

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(clear_specification,reset_translate_keywords,
                           'Reset Keyword Translation.').
reset_translate_keywords :-
   retractall(keyword_to_id_cache(_,_)),
   retractall(id_to_keyword_cache(_,_)).


% translate a keyword
translate_keyword_id(ID,R) :- keyword_to_id_cache(ID,TID),!, R=TID.
translate_keyword_id(ID,R) :- 
   atom_concat(ID,'__',R), % TO DO: gensym with assert
   assert(keyword_to_id_cache(ID,R)),
   assert(id_to_keyword_cache(R,ID)). % store that if an external tool gives us infos using ids we can translate back

% conjunct(pos(2,-1,1,1,1,44),conjunct(pos(3,-1,1,1,1,26),equal(pos(4,-1,1,1,1,13),identifier(pos(5,-1,1,1,1,6),left__),boolean_true(pos(6,-1,1,10,1,13))),equal(pos(7,-1,1,17,1,26),identifier(pos(8,-1,1,17,1,21),rel__),empty_set(pos(9,-1,1,25,1,26)))),equal(pos(10,-1,1,30,1,44),identifier(pos(11,-1,1,30,1,36),right__),boolean_false(pos(12,-1,1,40,1,44))))
raw_backtranslate_keyword_ids(RawExpr,Res) :- \+ keyword_to_id_cache(_,_),!, Res=RawExpr.
raw_backtranslate_keyword_ids(RawExpr,Res) :- map_over_raw_expr(RawExpr,backtranslate_id,Res).

backtranslate_id(identifier(Pos,ID),identifier(Pos,NewID)) :-
   (id_to_keyword_cache(ID,KeywordID) -> NewID=KeywordID ; NewID = ID).

map_over(P,RawExpr,Res) :- map_over_raw_expr(RawExpr,P,Res).
map_over_raw_expr(RawExpr,P,Res) :- call(P,RawExpr,NewRawExpr),!,
   Res=NewRawExpr.
map_over_raw_expr(conjunct(Pos,A,B),P,conjunct(Pos,MA,MB)) :- !,
   map_over_raw_expr(A,P,MA),
   map_over_raw_expr(B,P,MB).
map_over_raw_expr(disjunct(Pos,A,B),P,disjunct(Pos,MA,MB)) :- !,
   map_over_raw_expr(A,P,MA),
   map_over_raw_expr(B,P,MB).
map_over_raw_expr(equal(Pos,A,B),P,equal(Pos,MA,MB)) :- !,
   map_over_raw_expr(A,P,MA),
   map_over_raw_expr(B,P,MB).
map_over_raw_expr(A,_,R) :- atomic(A),!,R=A.
map_over_raw_expr(F,P,MF) :- F =.. [Functor,Pos|Args],
   maplist(map_over(P),Args,NewArgs),
   MF =.. [Functor,Pos|NewArgs].

classical_b_keyword('FIN').
classical_b_keyword('FIN1').
classical_b_keyword('POW').
classical_b_keyword('POW1').
%classical_b_keyword('__first_of_pair').
%classical_b_keyword('__second_of_pair').
classical_b_keyword(arity).
classical_b_keyword(bin).
classical_b_keyword(bool).
classical_b_keyword(btree).
classical_b_keyword(card).
classical_b_keyword(closure).
classical_b_keyword(closure1).
classical_b_keyword(conc).
classical_b_keyword(const).
classical_b_keyword(dom).
classical_b_keyword(father).
classical_b_keyword(finite).
classical_b_keyword(first).
classical_b_keyword(fnc).
classical_b_keyword(front).
classical_b_keyword(id).
classical_b_keyword(infix).
classical_b_keyword(inter).
classical_b_keyword(iseq).
classical_b_keyword(iseq1).
classical_b_keyword(items).
classical_b_keyword(iterate).
classical_b_keyword(last).
classical_b_keyword(left).
classical_b_keyword(max).
classical_b_keyword(min).
classical_b_keyword(mirror).
classical_b_keyword(not).
classical_b_keyword(perm).
classical_b_keyword(postfix).
classical_b_keyword(pred).
classical_b_keyword(prefix).
classical_b_keyword(prj1).
classical_b_keyword(prj2).
classical_b_keyword(ran).
classical_b_keyword(rank).
classical_b_keyword(rel).
classical_b_keyword(rev).
classical_b_keyword(right).
classical_b_keyword(seq).
classical_b_keyword(seq1).
classical_b_keyword(size).
classical_b_keyword(sizet).
classical_b_keyword(son).
classical_b_keyword(sons).
classical_b_keyword(subtree).
classical_b_keyword(succ).
classical_b_keyword(tail).
classical_b_keyword(top).
classical_b_keyword(tree).
classical_b_keyword(union).
% facts above can be generated by:
% findall(ID,translate:function_like(_,ID),L),sort(L,SL), member(ID,SL),  portray_clause(classical_b_keyword(ID)),fail.

classical_b_keyword('BOOL').
classical_b_keyword('FALSE').
classical_b_keyword('MAXINT').
classical_b_keyword('MININT').
classical_b_keyword('STRING').
classical_b_keyword('TRUE').
% facts above can be generated by:
% findall(ID,translate:constants(_,ID),L),sort(L,SL), member(ID,SL),  portray_clause(classical_b_keyword(ID)),fail.

classical_b_keyword('INT').
classical_b_keyword('INTEGER').
classical_b_keyword('NAT').
classical_b_keyword('NAT1').
classical_b_keyword('NATURAL').
classical_b_keyword('NATURAL1').
% facts above can be generated by:
% findall(ID,(translate:eventb_integer_mapping(A,B),(ID=A;ID=B)),L),sort(L,SL), member(ID,SL),  portray_clause(classical_b_keyword(ID)),fail.

% quantified operators:
classical_b_keyword('INTER').
classical_b_keyword('PI').
classical_b_keyword('SIGMA').
classical_b_keyword('UNION').

% additional substitution/structuring keywords
classical_b_keyword('ABSTRACT_CONSTANTS').
classical_b_keyword('ABSTRACT_VARIABLES').
classical_b_keyword('ANY').
classical_b_keyword('ASSERT').
classical_b_keyword('ASSERTIONS').
classical_b_keyword('BE').
classical_b_keyword('BEGIN').
classical_b_keyword('CASE').
classical_b_keyword('CHOICE').
classical_b_keyword('CONCRETE_CONSTANTS').
classical_b_keyword('CONCRETE_VARIABLES').
classical_b_keyword('CONSTANTS').
classical_b_keyword('CONSTRAINTS').
classical_b_keyword('DEFINITIONS').
classical_b_keyword('DO').
classical_b_keyword('EITHER').
classical_b_keyword('ELSE').
classical_b_keyword('ELSIF').
classical_b_keyword('END').
classical_b_keyword('EVENT').
classical_b_keyword('EXTENDS').
classical_b_keyword('IF').
classical_b_keyword('IMPLEMENTATION').
classical_b_keyword('IMPORTS').
classical_b_keyword('IN').
classical_b_keyword('INCLUDES').
classical_b_keyword('INITIALISATION').
classical_b_keyword('INITIALIZATION').
classical_b_keyword('INVARIANT').
classical_b_keyword('LET').
classical_b_keyword('LOCAL_OPERATIONS').
classical_b_keyword('MACHINE').
classical_b_keyword('MODEL').
classical_b_keyword('OF').
classical_b_keyword('OPERATIONS').
classical_b_keyword('OR').
classical_b_keyword('PRE').
classical_b_keyword('PROMOTES').
classical_b_keyword('PROPERTIES').
classical_b_keyword('REFINEMENT').
classical_b_keyword('REFINES').
classical_b_keyword('SEES').
classical_b_keyword('SELECT').
classical_b_keyword('SETS').
classical_b_keyword('skip').
classical_b_keyword('SYSTEM').
classical_b_keyword('THEN').
classical_b_keyword('USES').
classical_b_keyword('VALUES').
classical_b_keyword('VARIABLES').
classical_b_keyword('VARIANT').
classical_b_keyword('WHERE').
classical_b_keyword('WHILE').
classical_b_keyword('WITH').

