% (c) 2009 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

% --------------------------------------------------------
%                       SETS
% --------------------------------------------------------


:- use_module(sets).

% Possible sets:
%  set([])  
%  set([el1,...,eln])  sorted, without duplicates
%  setFromTo(int(Low),int(Up))
% ... intType,....

%is_a_set(set(_)).
is_a_set(tset(_)).  /* a tree compiled set */
%is_a_set(closure(_)).
is_a_set(tclosure(_)).  /* a tree compiled set */
is_a_set(setFromTo(_,_)).
is_a_set(intType).
is_a_set(dataType(_)).
is_a_set(setComp(_,_)).
%is_a_set(DT) :- dataTypeDef(DT,_).


:- assert_must_succeed(( haskell_csp:evaluate_set([int(3)],R),
                       R == tset(t(int(3),empty,empty)) )).

evaluate_set(List,tset(Tree)) :- evaluate_list(List,EL), list_to_tree(EL,Tree).

evaluate_closure(X,tclosure(RS)) :- evaluate_set(X,tset(RS)).

is_empty_set(tset(X)) :- !, empty_set(X).
is_empty_set(tclosure(X)) :- !, empty_set(X).
is_empty_set(setFromTo(int(Low),int(Up))) :- !,
  when((ground(Low),ground(Up)), Low>Up). 
is_empty_set(X) :- add_error(csp_set,'Could not evaluate: ',is_empty_set(X)),fail.

:- assert_must_succeed(( haskell_csp:is_member_set(X,set([int(1),int(9)])), X=int(9) )).
:- assert_must_fail(( haskell_csp:is_member_set(X,set([int(1),int(9)])), X=int(3) )).

is_member_set(El,tset(T)) :- generate_elements_of_set(El,T).
is_member_set(int(_),intType) :- !.
is_member_set(int(X),setFromTo(int(Low),int(Up))) :- !,
  when(ground(X), (when(ground(Low),X>=Low), when(ground(Up),X=<Up))).
% add ... further types Bool,... closure,...
is_member_set(C,dataType(DT)) :- dataTypeDef(DT,L),  /* to do: precompute this */
   member(constructor(C),L),!.
is_member_set(C,setComp(GeneratorSet,Tuple)) :- !, 
    is_member_comprehension_set(C,Tuple,GeneratorSet),
    print(is_member_comprehension_set(C,Tuple,GeneratorSet)),nl.
is_member_set(X,R) :- add_error(csp_set,'Could not evaluate: ',member(X,R)),fail.
 
 /* the same as above but will not non-deterministically instantiate */
lazy_is_member_set(int(X),setFromTo(int(Low),int(Up))) :- !,
  when(ground(X), (when(ground(Low),X>=Low), when(ground(Up),X=<Up))).
lazy_is_member_set(El,tset(T)) :- when(ground(El),is_element_of_set(El,T)).
lazy_is_member_set(C,setComp(GeneratorSet,Tuple)) :- !, 
    when(ground(C),is_member_comprehension_set(C,Tuple,GeneratorSet)),
    print(is_member_comprehension_set(C,Tuple,GeneratorSet)),nl.
lazy_is_member_set(El,S) :- is_member_set(El,S).


%:- assert_must_succeed(( haskell_csp:cardinality(set([int(1),int(9)]),R), R==int(2) )).
cardinality(tset(S),R) :- !, set_cardinality(S,R).
cardinality(setFromTo(int(Low),int(Up)),R) :- !, R=int(C), when((ground(Low),ground(Up)), C is Up-Low+1).
% to do: add other sets, closure,...
cardinality(X,_R) :- add_error(csp_set,'Could not compute card: ',X),fail.


union_set(S1,S2,Res) :- Res = tset(R), expand_symbolic_set(S1,tset(ES1)),
  expand_symbolic_set(S2,tset(ES2)),
  when(ground((ES1,ES2)),union_set(ES1,ES2,R)) , print(union(S1,S2,Res)),nl.

 
 
diff_set(S1,S2,Res) :- Res = tset(R), expand_symbolic_set(S1,tset(ES1)),
  expand_symbolic_set(S2,tset(ES2)),
  when(ground((ES1,ES2)),diff_set(ES1,ES2,R)).
  %, print(diff(S1,S2,Res,ES1,ES2)),nl.


inter_set(S1,S2,Res) :- Res = tset(R), expand_symbolic_set(S1,tset(ES1)),
  expand_symbolic_set(S2,tset(ES2)),
  when(ground((ES1,ES2)),inter_set(ES1,ES2,R)).
  


equal_element(X,Y) :- var(X),!,X=Y.
equal_element(int(X),R) :- !, (R=int(Y) -> X=Y ; (add_error(haskell_csp,'Type error in equality: ',int(X)=R),fail)).
equal_element(set(X),R) :- !, (R=set(Y) -> X=Y ; (add_error(haskell_csp,'Type error in equality: ',set(X)=R),fail)).
% To do: Improve this predicate
equal_element(X,X).


not_equal_element(X,Y) :- var(X),!,dif(X,Y).
not_equal_element(int(X),R) :- !, (R=int(Y) -> X\=Y ; (add_error(haskell_csp,'Type error in disequality: ',int(X)\=R),fail)).
not_equal_element(set(X),R) :- !, (R=set(Y) -> X\=Y ; (add_error(haskell_csp,'Type error in disequality: ',set(X)\=R),fail)).
% To do: Improve this predicate
not_equal_element(X,Y) :- dif(X,Y).


/* expand_symbolic_set */
/*  force expansion of symbolic sets into explicit sets */

:- assert_must_succeed(( haskell_csp:expand_symbolic_set(setComp(
 [comprehensionGenerator(_x,set([int(1),int(3),int(5)])),
  comprehensionGenerator(_y,set([int(1),int(2),int(4)]))],tuple(['*'(_x,_y)])),_R) )).
expand_symbolic_set(dataType(T),R) :- !, R=tset(TR),
   (dataTypeDef(T,Def)
     -> %print(dt(T,Def,SET)),nl,
        expand_datatypedefbody(Def,SET),list_to_set(SET,TR)
     ;  add_error(csp_sets,'Could not expand dataType. No datatype definition for: ',T),fail
   ).
expand_symbolic_set(closure(X),Res) :- !, closure_expand(X,R),list_to_set(R,TR),
   Res= tset(TR).
expand_symbolic_set(tset(X),R) :- !, R=tset(X).
%expand_symbolic_set(set(X),R) :- !, R=set(X).
expand_symbolic_set(setFromTo(int(Low),int(Up)),R) :- !, 
    when((ground(Low),ground(Up)),
          (expand_from_to(Low,Up,RS),list_to_set(RS,TRS),R=tset(TRS))).
expand_symbolic_set(setComp(GeneratorList,tuple(ExprList)),R) :- !,
   expand_set_comprehension(ExprList,GeneratorList,R).
expand_symbolic_set(Set,R) :- add_error(haskell_csp,'Could not expand set: ',Set),R=Set.
    
expand_from_to(X,Y,R) :- X>Y,!, R=[].
expand_from_to(X,Y,[int(X)|T]) :- X1 is X+1, expand_from_to(X1,Y,T).

expand_datatypedefbody([],R) :- !, R=[].
expand_datatypedefbody([constructor(C)|T],R) :- !, R=[C|ET],
   expand_datatypedefbody(T,ET).
expand_datatypedefbody(L,R) :-
  add_error(csp_sets,'Could not expand datatype body: ',L),
  R=L.


/* ------------------ */
/* EXPANDING CLOSURES */
/* ------------------ */

/* haskell_csp:closure_expand([pat([],outf)],R),print(R),nl */

closure_expand(ListOfEls,ExpandedList) :-
  %print(closure_expand(ListOfEls)),nl,
  when(ground(ListOfEls),
   (findall(EEl,(member(El,ListOfEls), 
                 closure_expand_single_element(El,EEl)),EEls),
          %print(expanded(EEls)),nl,
    sort(EEls,ExpandedList)  /* is sort ok wrt to @< used by various set operations?? */
   )).
   
closure_expand_single_element(pat(List,Ch),R) :-
   channel_type_list(Ch,ChannelTypeList),!, R = pat(NewList,Ch),
   gen_expanded_list(ChannelTypeList,List,NewList,Ch).
closure_expand_single_element(X,R) :-
   add_error(csp_set,'Cannot closure expand: ',X),
   X=R.
   
gen_expanded_list([],List,[],Channel) :-
 (List=[] -> true ; add_error(csp_sets,'Patter List too long for channel:',(List,Channel))).
gen_expanded_list([Type|TT],List,Res,Channel) :-
   (List=[H|LT]  -> Res=[H|RT]
     ;  (Res=[NewEl|RT],LT=List,  /* is it ok not to put dot(NewEl) here ? */
         %print(enum(Type,Channel)),nl,
         enumerate_channel_input_value(Type,NewEl))
   ),
   gen_expanded_list(TT,LT,RT,Channel).

/* ------------------ */
/* SET COMPREHENSIONS */
/* ------------------ */

expand_set_comprehension(ExprList,GeneratorList,Res) :-
    treat_generators(GeneratorList,Vars,Sets,Guard),
    when(ground(Sets),
          (findall(Expr,(check_boolean_expression(Guard),
                         generator_sol(Vars,Sets), 
                        member(Expr,ExprList)),Expressions),
           evaluate_set(Expressions,Res))).
   
generator_sol([],[]).
generator_sol([Var|VT],[Set|ST]) :- is_member_set(Var,Set), generator_sol(VT,ST).

treat_generators([],V,S,G) :- !, V=[], S=[], G=true.
treat_generators([comprehensionGenerator(Var,Set)|T],Vars,Sets,TG) :- !, Vars=[Var|VT], Sets=[ESet|ST],
   evaluate_argument(Set,EvSet),expand_symbolic_set(EvSet,ESet),
   treat_generators(T,VT,ST,TG).
treat_generators([comprehensionGuard(Guard)|T],VT,ST,bool_and(Guard,TGuard)) :- !,
   treat_generators(T,VT,ST,TGuard).
treat_generators(G,_,_,_) :- add_error(csp_set,'Could not treat Set Comprehension Generator List: ',G),fail.
           
 :- assert_must_succeed(( haskell_csp:is_member_comprehension_set(int(12), 
              tuple(['*'(_x,_y)]),
              [comprehensionGenerator(_x,set([int(1),int(3)])),
               comprehensionGenerator(_y,set([int(1),int(2),int(4)]))])  )).           
 :- assert_must_fail(( haskell_csp:is_member_comprehension_set(int(11), 
              tuple(['*'(_x,_y)]),
              [comprehensionGenerator(_x,set([int(1),int(3)])),
               comprehensionGenerator(_y,set([int(1),int(2),int(4)]))])  )).       
 :- assert_must_succeed(( haskell_csp:is_member_comprehension_set(X, 
              tuple(['*'(_x,_y)]),
              [comprehensionGenerator(_x,set([int(1),int(3)])),
               comprehensionGenerator(_y,set([int(1),int(2),int(4)]))]),X=int(12)  )).
       
is_member_comprehension_set(X,tuple(ExprList),GeneratorList) :-
    treat_generators(GeneratorList,Vars,Sets,Guard),
     when(ground(Sets),
        (check_boolean_expression(Guard),
         generator_sol(Vars,Sets),
         member(Expr,ExprList),
         evaluate_argument(Expr,X))).
