% (c) 2009-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

% each cli_testcase consists of: id, list of categories, arguments to probcli, comment

:- module(testcases, [cli_testcase/5,
     cli_testcase_diff_check_output/3, cli_testcase_do_not_delete/2,
     extra_testcase_file/2]).

%:- use_module('../extensions/profiler/profiler.pl').
%:- use_module('../extensions/profiler/profiler_te.pl').
%:- enable_profiling(cli_testcase/5).

:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'This module defines the (integration) tests.').

cli_testcase(Id,Groups,TestConfiguration,Arguments,Description) :-
    cli_testcase(Id,Groups,ArgumentsTemp,Description),
    options_all_tests(Id,AdditionalArguments),
    append(ArgumentsTemp,AdditionalArguments,Arguments),
    (additional_configuration(Id,Conf) -> TestConfiguration = Conf ; TestConfiguration = []).

% options that should be used in all testcases
options_all_tests(1303,['-strict']) :- !.
options_all_tests(1351,['-strict']) :- !. % tests generate warnings and we still want to load the B machines
options_all_tests(1658,['-strict']) :- !.
options_all_tests(_,['-strict','-p','STRICT_RAISE_WARNINGS','TRUE']).

additional_configuration(37,   [conditional_skip(\+(platform_is_64_bit))]). /* CLP(FD) buffer overflow on win32 */
additional_configuration(43,   [conditional_skip(\+(platform_is_64_bit))]). /* integer too large for 32-bit systems */
additional_configuration(50,   [conditional_skip(host_platform(windows))]).
additional_configuration(161,  [conditional_skip(host_platform(windows))]).
additional_configuration(271,  [conditional_skip(host_platform(windows))]).
%additional_configuration(324,  [skip]). /* because only works if SMT mode is allowed to instantiate */
additional_configuration(539,  [conditional_skip(host_platform(windows))]).
additional_configuration(469,  [conditional_skip(\+(platform_is_64_bit))]). /* timeout */
additional_configuration(767,  [conditional_skip(\+(platform_is_64_bit))]). /* integer overflow on 32-bit systems */
additional_configuration(865,  [conditional_skip(host_platform(windows))]).
additional_configuration(875,  [skip]). % The type checker temporarily also allows {x|->y | x=y+1 } (added to support UNION/INTER with Prolog BParser)
additional_configuration(1101, [conditional_skip(\+(platform_is_64_bit))]). % not enough memory
additional_configuration(1102, [conditional_skip(\+(platform_is_64_bit))]). % not enough memory
additional_configuration(1103, [conditional_skip(\+(platform_is_64_bit))]). % not enough memory
additional_configuration(1131, [conditional_skip(\+(platform_is_64_bit))]). % not enough memory
additional_configuration(1157, [conditional_skip(\+(platform_is_64_bit))]). % not enough memory
additional_configuration(1162, [conditional_skip(host_platform(linux))]). /* linux32 is too slow for this testcase Timeout when posting constraint: clpfd_interface:(list_to_fdset([506,506],_374556577),_374485321 in_set _374556577) -> timeout; ditto for linux 64?! */
additional_configuration(1207, [conditional_skip(\+(platform_is_64_bit))]). /* integer overflow on 32-bit systems */
additional_configuration(1360,   [conditional_skip(\+(platform_is_64_bit))]). /* funnily enough: no clpfd overflow */
additional_configuration(810,   [conditional_skip(host_platform(windows))]). /* java heap space */
additional_configuration(967,  [skip]). % the proof of the invariant relies on the theorem, which is proven but has wd-problem
additional_configuration(844,  [skip]). % repl option is not anymore available
additional_configuration(988,  [skip]). % skip test case until merging 'feature/avl_sets_csp' with 'develop'
additional_configuration(987,  [skip]). % no reductions can be gained in this test case
additional_configuration(1086, [skip]).
additional_configuration(1316,  [conditional_skip(host_platform(windows))]).
additional_configuration(1441,  [conditional_skip(\+(platform_is_64_bit))]). /* timeout */
additional_configuration(1447,  [conditional_skip(\+(platform_is_64_bit))]). /* timeout */
additional_configuration(1448,  [skip]). /* PO 'axm2/WD' no longer provable due to change of cardinality3 block declaration; should be fixed by merging kernel_cardinality */
additional_configuration(1452,  [conditional_skip(\+(platform_is_64_bit))]). /* timeout */
additional_configuration(1654,  [conditional_skip(host_platform(windows))]). % file contains unicode and can not be checked out properly on windows
additional_configuration(1716,  [conditional_skip(\+(platform_is_64_bit))]). /* timeout */
additional_configuration(1771,  [conditional_skip(host_platform(windows))]).


% platoon_2_mch and platoon_3_0_mch platoon_4_0_mch are not animatable (any more) with more restrictive enumeration warning treatment; they require infinite total functions (xpos: NATURAL ** .... --> )
additional_configuration(129, [skip]).
additional_configuration(130, [skip]).
additional_configuration(131, [skip]).
additional_configuration(132, [skip]).
additional_configuration(333, [skip]).
additional_configuration(334, [skip]).
additional_configuration(1308, [skip]). % Kodkod test, now fails with more conservative card treatment, 20th April 2016, git sha: a422afcd2949b8b0a40821f0029b41240935e695
additional_configuration(1552, [skip]). % removed unsafe function application treatment in b_compiler, 22nd July 2016
additional_configuration(1705,  [skip]). % requires feature b_write_eventb_machine_to_classicalb_to_file to be added to cli_start_mc_with_tlc
additional_configuration(1755, [skip]). % requires solving test 614, 1101 for new UNION translation

% -----------------------------------

cli_testcase(0, [unit], ['-selfcheck', '-strict', '-p', 'CLPFD', 'FALSE'], 'Unit tests').
cli_testcase(1, [unit], ['-selfcheck', '-strict', '-p', 'CLPFD', 'TRUE'], 'Unit tests (with CLPFD)').
cli_testcase(2, [unit], ['-selfcheck', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'SYMBOLIC', 'TRUE'], 'Unit tests (with CLPDFD and SYMBOLIC)').
cli_testcase(3, [b_test], ['-t', '-strict', '../prob_examples/public_examples/B/FeatureChecks/DEFINITIONS/SETPREF.mch', '-p', 'MAXINT', '1024', '-card', 'ID', '3'], 'check set MAXINT & card').
cli_testcase(4, [b_test], ['-t', '-strict', '../prob_examples/public_examples/B/FeatureChecks/DEFINITIONS/DefinitionScopeOk.mch'], 'DEFINITION scope test').
cli_testcase(5, [b_test], ['-t', '-strict', '../prob_examples/public_examples/B/FeatureChecks/ForAll_VariousVersions.mch'], 'For all test').
cli_testcase(6, [b_test], ['../prob_examples/public_examples/B/FeatureChecks/NotMemberCheck.mch', '-t', '-strict', '-assertions', '-expcterr', virtual_time_out], 'Not member closures test').
cli_testcase(7, [bench], ['../prob_examples/public_examples/B/Benchmarks/CarlaTravelAgencyErr.mch', '-t', '-strict'], 'Carla Travel Agency trace check').
cli_testcase(8, [bench], ['../prob_examples/public_examples/B/Benchmarks/RouteIsSeq.mch', '-t', '-strict'], '').
cli_testcase(9, [bench], ['../prob_examples/public_examples/B/Benchmarks/Sets2.mch', '-t', '-strict'], '').
cli_testcase(10, [bench], ['../prob_examples/public_examples/B/Benchmarks/Doors.mch', '-t', '-strict'], '').
cli_testcase(11, [bench], ['../prob_examples/public_examples/B/Benchmarks/DSP0.mch', '-t', '-mc', '100', '-noinv', '-strict', '-nodead', '-expcterr', 'model_check_incomplete'], '').
cli_testcase(12, [bench], ['../prob_examples/public_examples/B/Benchmarks/DSP0_complicated_initialisation.mch', '-t', '-mc', '100', '-noinv', '-strict', '-nodead', '-expcterr', 'model_check_incomplete'], '').
cli_testcase(13, [bench], ['../prob_examples/public_examples/B/Benchmarks/CSM.mch', '-mc', '1000', '-strict', '-check_complete', '-cc', '341', '1229'], '').
cli_testcase(14, [bench], ['../prob_examples/public_examples/B/Benchmarks/scheduler.mch', '-t', '-mc', '100', '-strict', '-check_complete', '-cc', '36', '121'], '').
cli_testcase(15, [bench], ['../prob_examples/public_examples/B/Benchmarks/phonebook7.mch', '-t', '-mc', '100', '-strict'], '').
cli_testcase(16, [bench], ['../prob_examples/public_examples/B/Benchmarks/tictac.mch', '-mc', '100', '-noinv', '-nodead', '-strict'], '').
cli_testcase(17, [bench,case], ['../prob_examples/public_examples/B/Benchmarks/Teletext_bench.mch', '-mc', '100', '-strict'], '').
cli_testcase(18, [bench], ['../prob_examples/public_examples/B/Benchmarks/Cruise_finite1.mch', '-t', '-strict'], '').
cli_testcase(19, [bench], ['../prob_examples/public_examples/B/BZTT/TestBZTT_Data.mch', '-mc', '100', '-check_complete', '-cc', '26', '176', '-strict', '-l', 'log/TestBZTT_Data.log'], '').
cli_testcase(20, [performance_tests], ['../prob_examples/public_examples/B/Daniel/Kodkod/Power2Nr.mch', '-strict', '-t', '-p', 'CLPFD', 'TRUE', '-aa', '4', '0', '0'], '').
cli_testcase(21, [performance_tests], ['../prob_examples/public_examples/B/Daniel/Kodkod/Power2Nr.mch', '-strict', '-t', '-p', 'CLPFD', 'TRUE', '-ma', '4', '0', '0'], '').
cli_testcase(22, [performance_tests], ['../prob_examples/public_examples/B/Daniel/Kodkod/StarTopologySingleSol.mch', '-strict', '-t', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(23, [performance_tests,codespeed], ['../prob_examples/public_examples/B/PerformanceTests/conc_seq_test.mch', '-strict', '-t'], 'concatenation test').
cli_testcase(24, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/LargeSets/PartialFunCheck.mch', '-strict', '-t'], '').
cli_testcase(25, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/ForAll/ForAll_NATURAL.mch', '-strict', '-t'], '').
cli_testcase(26, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/ForAll/ForAllExpansion.mch', '-strict', '-t'], '').
cli_testcase(27, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/ForAll/ForAllExpansian_Cartesian.mch', '-strict', '-t', '-assertions'], '').
cli_testcase(28, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/ForAll/ForAllExpansion_Cartesian2.mch', '-strict', '-t', '-assertions', '-p', 'CLPFD', 'FALSE'], '').
cli_testcase(29, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/ForAll/ForAllExpansion_Cartesian2.mch', '-strict', '-t', '-assertions', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(30, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/ForAll/ForAllExpansion_Cartesian3.mch', '-strict', '-t', '-assertions'], '').
cli_testcase(31, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/LargeSets/SeqCheck.mch', '-strict', '-t'], '').
cli_testcase(32, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/SetLogInefficiencies.mch', '-t', '-strict'], '').
cli_testcase(33, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/SetLogInefficiencies2.mch', '-t', '-strict'], '').
cli_testcase(34, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/SetLogInefficiencies3.mch', '-t', '-strict'], '').
cli_testcase(35, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/InfiniteClosures.mch', '-strict', '-t'], '').
cli_testcase(36, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/InfiniteClosures2.mch', '-strict', '-t'], '').
cli_testcase(37, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/InfiniteAbsSqrt.mch', '-strict', '-t'], '').
cli_testcase(38, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/InfCardChecks.mch', '-p', 'CLPFD', 'TRUE', '-strict', '-t'], '').
cli_testcase(39, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/ID_Checks.mch', '-strict', '-t', '-assertions'], '').
cli_testcase(40, [performance_tests,codespeed], ['../prob_examples/public_examples/B/PerformanceTests/DirectProductPerformance.mch', '-strict', '-init', '-p', 'TIME_OUT', '20000'], 'Test the performance of the direct product operator.').
cli_testcase(41, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/DirectProductPerformance.mch', '-strict', '-init', '-p', 'SYMBOLIC', 'TRUE', '-p', 'TIME_OUT', '15000'], '').
cli_testcase(42, [performance_tests], ['../prob_examples/public_examples/B/Mathematical/Sieve.mch', '-mc', '1000', '-strict', '-expcterr', 'goal_found', '-l', 'log/Sieve.log', '-expcterr', virtual_time_out], '').
% StartSieve generates virtual_time_out
cli_testcase(43, [performance_tests], ['../prob_examples/public_examples/B/Mathematical/BinomialCoefficients.mch', '-t', '-strict'], '').
cli_testcase(44, [performance_tests], ['../prob_examples/public_examples/B/Mathematical/SieveAlternate.mch', '-mc', '1000', '-strict', '-expcterr', 'goal_found', '-l', 'log/SieveAlternate.log', '-expcterr', virtual_time_out], '').
cli_testcase(45, [performance_tests], ['../prob_examples/public_examples/B/Mathematical/GraphIso/CheckGraphIsomorphism.mch', '-t', '-strict', '-l', 'log/CheckGraphIsomorphism.log'], '').
cli_testcase(46, [performance_tests], ['../prob_examples/public_examples/B/Mathematical/GraphIso/CheckLargeGraphIsomorphismAsConstants.mch', '-t', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'TIME_OUT', '5000'], '').
cli_testcase(47, [performance_tests], ['../prob_examples/public_examples/B/Mathematical/Theories/FiniteSequences1.mch', '-t', '-mc', '100', '-strict'], '').
cli_testcase(48, [performance_tests], ['../prob_examples/public_examples/B/Mathematical/Theories/FiniteSequences2.mch', '-t', '-mc', '100', '-strict'], '').
cli_testcase(49, [performance_tests,sat], ['../prob_examples/public_examples/B/PerformanceTests/SATLIB/flat200-90.mch', '-t', '-p', 'STATIC_ORDERING', 'TRUE', '-strict', '-p', 'TIME_OUT', '15000', '-l', 'log/SAT_flat200-90.log'], 'SAT test').
cli_testcase(50, [parser,performance_tests,sat], ['../prob_examples/public_examples/B/PerformanceTests/SATLIB/sudoku.mch', '-init', '-p', 'STATIC_ORDERING', 'TRUE', '-strict', '-p', 'TIME_OUT', '60000'], 'Large machine to test the parser/JVM, skipped on Windows because the Windows build bot can not reserve enough memory.'). % used to require LARGE_JVM
cli_testcase(51, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/ConstraintPropagation/MaxPropagation.mch', '-strict', '-t'], '').
cli_testcase(52, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/ConstraintPropagation/MaxPropagation2.mch', '-strict', '-t'], '').
cli_testcase(53, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/ConstraintPropagation/SetMembershipInference.mch', '-strict', '-t'], '').
cli_testcase(54, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/ConstraintPropagation/SetMembershipReification.mch', '-strict', '-t'], '').
cli_testcase(55, [performance_tests], ['../prob_examples/public_examples/B/Demo/SortByPermutation_v2.mch', '-strict', '-t','-p', 'TIME_OUT', '7000'], '').
cli_testcase(56, [performance_tests], ['../prob_examples/public_examples/B/IPD/colouring/GraphColouringNAT2.mch', '-t', '-strict'], '').
cli_testcase(57, [performance_tests], ['../prob_examples/public_examples/EventBPrologPackages/Puzzles/Instance_small_60_400_9_ctx.eventb', '-init', '-strict', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(58, [performance_tests], ['../prob_examples/public_examples/B/BZTT/PO_Scheduler.mch', '-mc', '100', '-strict', '-t', '-nodead'], '').
cli_testcase(59, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/SMTLIB/IDL_SimpleUnsat.mch', '-strict', '-init', '-expcterr', 'setup_constants_fails'], '').
cli_testcase(60, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/SMTLIB/int_completeness1.mch', '-strict', '-init', '-p', 'CLPFD', 'TRUE', '-expcterr', 'setup_constants_fails'], '').
cli_testcase(61, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/SMTLIB/int_completeness1_event.mch', '-strict', '-p', 'CLPFD', 'TRUE', '-t'], '').
cli_testcase(62, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/SMTLIB/SquareEquivalenceProblem.mch', '-strict', '-t', '-p', 'CLPFD', 'TRUE', '-mc', '10', '-cc', '2', '2'], '').
cli_testcase(63, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/clpfd_constraint_checks/IntegerComprSets.mch', '-strict', '-t'], '').
cli_testcase(64, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/SeqEnumWithLargeMAXINT.mch', '-strict', '-t'], '').
cli_testcase(65, [performance_tests,sequences], ['../prob_examples/public_examples/B/PerformanceTests/ConstraintPropagation/SequenceSolving.mch', '-strict', '-t'], '').
cli_testcase(66, [performance_tests], ['../prob_examples/public_examples/B/PerformanceTests/SAT_Tests/SubsetReificationTest.mch', '-strict', '-t'], '').
cli_testcase(67, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/Demo/monitor2.mch', '-mc', '100', '-cc', '15', '92', '-p', 'SYMMETRY_MODE', 'nauty', '-strict'], '').
cli_testcase(68, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/Demo/Symetric.mch', '-mc', '100', '-cc', '2', '6', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'nauty', '-strict'], '').
cli_testcase(69, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/Demo/scheduler.mch', '-mc', '100', '-cc', '11', '47', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'nauty', '-strict'], '').
cli_testcase(70, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/Demo/scheduler.mch', '-mc', '100', '-cc', '36', '156', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'off', '-strict'], '').
cli_testcase(71, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/Demo/scheduler.mch', '-mc', '100', '-cc', '11', '47', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'hash', '-strict'], '').
cli_testcase(72, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/Demo/scheduler.mch', '-mc', '100', '-cc', '36', '73', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'flood', '-strict'], '').
cli_testcase(73, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/scheduler1.ref', '-mc', '150', '-cc', '145', '447', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'off', '-strict'], '').
cli_testcase(74, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/scheduler1.ref', '-mc', '100', '-cc', '29', '92', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'nauty', '-strict'], 'scheduler1.ref with nauty (3)'). % used to be 94 instead of 92 transitions; change of INITIALISATION argument treatment April 2016
cli_testcase(75, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/scheduler1.ref', '-mc', '100', '-cc', '29', '92', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'hash', '-strict'], 'scheduler1.ref with hash (3)'). % used to be 94 instead of 92 transitions; change of INITIALISATION argument treatment April 2016
cli_testcase(76, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/scheduler1.ref', '-mc', '150', '-cc', '145', '210', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'flood', '-strict'], '').
cli_testcase(77, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/scheduler1.ref', '-mc', '100', '-cc', '51', '208', '-p', 'DEFAULT_SETSIZE', '4', '-p', 'SYMMETRY_MODE', 'nauty', '-p', 'MAX_INITIALISATIONS', '20', '-p', 'MAX_OPERATIONS', '120', '-strict'], 'scheduler1.ref with nauty (4)'). % used to be 211 instead of 208 transitions; change of INITIALISATION argument treatment April 2016
cli_testcase(78, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/scheduler1.ref', '-mc', '100', '-cc', '51', '208', '-p', 'DEFAULT_SETSIZE', '4', '-p', 'SYMMETRY_MODE', 'hash', '-p', 'MAX_INITIALISATIONS', '20', '-p', 'MAX_OPERATIONS', '120', '-strict'], 'scheduler1.ref with hash (4)'). % used to be 211 instead of 208 transitions; change of INITIALISATION argument treatment April 2016
cli_testcase(79, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/scheduler1.ref', '-mc', '1000', '-cc', '81', '401', '-p', 'DEFAULT_SETSIZE', '5', '-p', 'SYMMETRY_MODE', 'nauty', '-p', 'MAX_INITIALISATIONS', '20', '-p', 'MAX_OPERATIONS', '120', '-strict'], 'scheduler1.ref with nauty (5)'). % used to be 405 transitions, see above
cli_testcase(80, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/scheduler1.ref', '-mc', '1000', '-cc', '81', '401', '-p', 'DEFAULT_SETSIZE', '5', '-p', 'SYMMETRY_MODE', 'hash', '-p', 'MAX_INITIALISATIONS', '20', '-p', 'MAX_OPERATIONS', '120', '-strict'], 'scheduler1.ref with hash (5)'). % used to be 405 transitions, see above
cli_testcase(81, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/scheduler1.ref', '-mc', '1000', '-cc', '120', '696', '-p', 'DEFAULT_SETSIZE', '6', '-p', 'SYMMETRY_MODE', 'nauty', '-p', 'MAX_INITIALISATIONS', '20', '-p', 'MAX_OPERATIONS', '120', '-strict'], 'scheduler1.ref with nauty (6)'). % used to be 701 transitions, see above
cli_testcase(82, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/scheduler1.ref', '-mc', '1000', '-cc', '120', '696', '-p', 'DEFAULT_SETSIZE', '6', '-p', 'SYMMETRY_MODE', 'hash', '-p', 'MAX_INITIALISATIONS', '20', '-p', 'MAX_OPERATIONS', '120', '-strict'], 'scheduler1.ref with hash (6)'). % used to be 701 transitions, see above
cli_testcase(83, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/scheduler1.ref', '-mc', '1000', '-cc', '169', '1121', '-p', 'DEFAULT_SETSIZE', '7', '-p', 'SYMMETRY_MODE', 'nauty', '-p', 'MAX_INITIALISATIONS', '20', '-p', 'MAX_OPERATIONS', '120', '-strict'], 'scheduler1.ref with nauty (7)'). % used to be 1127 transitions, see above
cli_testcase(84, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/scheduler1.ref', '-mc', '1000', '-cc', '169', '1121', '-p', 'DEFAULT_SETSIZE', '7', '-p', 'SYMMETRY_MODE', 'hash', '-p', 'MAX_INITIALISATIONS', '20', '-p', 'MAX_OPERATIONS', '120', '-strict'], 'scheduler1.ref with hash (7)'). % used to be 1127 transitions, see above
cli_testcase(85, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/PairRelation.mch', '-mc', '100', '-cc', '39', '173', '-p', 'DEFAULT_SETSIZE', '3', '-strict'], '').
cli_testcase(86, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/PairRelation.mch', '-mc', '100', '-cc', '11', '47', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'nauty', '-strict'], '').
cli_testcase(87, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/PairRelation.mch', '-mc', '100', '-cc', '39', '75', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'flood', '-strict'], '').
cli_testcase(88, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/SetRelation.mch', '-mc', '100', '-cc', '38', '172', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'off', '-strict'], '').
cli_testcase(89, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/SetRelation.mch', '-mc', '100', '-cc', '10', '46', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'nauty', '-strict'], '').
cli_testcase(90, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/SetRelation.mch', '-mc', '100', '-cc', '38', '74', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'flood', '-strict'], '').
cli_testcase(91, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/RussianPostalPuzzle.mch', '-mc', '1000', '-cc', '81', '177', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'off', '-nogoal', '-strict'], '').
cli_testcase(92, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/RussianPostalPuzzle.mch', '-mc', '1000', '-cc', '48', '105', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'nauty', '-nogoal', '-strict'], '').
cli_testcase(93, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/RussianPostalPuzzle.mch', '-mc', '1000', '-cc', '48', '105', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'hash', '-nogoal', '-strict', '-expcterr', 'model_check_incomplete'], '').
cli_testcase(94, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/RussianPostalPuzzle.mch', '-mc', '1000', '-cc', '81', '138', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'flood', '-nogoal', '-strict'], '').
cli_testcase(95, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/TicTacToe_SimplerSym.mch', '-mc', '1500', '-cc', '1450', '4547', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'hash', '-nogoal', '-nodead', '-strict', '-expcterr', 'model_check_incomplete'], '').
cli_testcase(96, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/TicTacToe_SimplerSym.mch', '-mc', '1500', '-cc', '1450', '4547', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'nauty', '-nogoal', '-nodead', '-strict'], '').
cli_testcase(97, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/TicTacToe_SimplerSym.mch', '-mc', '6000', '-cc', '5480', '8577', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'flood', '-nogoal', '-nodead', '-strict'], '').
cli_testcase(98, [symmtest], ['../prob_examples/public_examples/B/SymmetryReduction/USB_4Endpoints.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'nauty', '-cc', '355', '11220', '-strict'], '').
cli_testcase(99, [symmtest,codespeed], ['../prob_examples/public_examples/B/SymmetryReduction/USB_4Endpoints.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'hash', '-cc', '355', '11220', '-strict'], 'USB with Hash symmetry').
cli_testcase(100, [symmtest], ['../prob_examples/public_examples/B/SymmetryReduction/USB_4Endpoints.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'flood', '-cc', '694', '11572', '-strict'], '').
cli_testcase(101, [symmtest], ['../prob_examples/public_examples/B/SymmetryReduction/USB_4Endpoints.mch', '-mc', '10000', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'hash', '-cc', '3013', '97730', '-strict', '-expcterr', 'model_check_incomplete'], '').
cli_testcase(102, [symmtest,proz], ['../prob_examples/public_examples/Z/Derrick/video_club_abz08.fuzz', '-strict', '-card', 'TITLE', '2', '-card', 'PERSON', '2', '-mc', '100000', '-p', 'SYMMETRY_MODE', 'hash', '-cc', '113', '596', '-expcterr', 'model_check_incomplete', '-expcterr', virtual_time_out], '').
cli_testcase(103, [symmtest], ['../prob_examples/public_examples/B/Puzzles/Hanoi6Sym.mch', '-mc', '10000', '-p', 'SYMMETRY_MODE', 'off', '-cc', '731', '2187', '-nogoal', '-strict'], '').
cli_testcase(104, [symmtest], ['../prob_examples/public_examples/B/Puzzles/Hanoi6Sym.mch', '-mc', '10000', '-p', 'SYMMETRY_MODE', 'hash', '-cc', '367', '1095', '-nogoal', '-strict'], 'Hanoi6Sym with hash'). % used to be 1096 transitions; changed in April 2016 due to INITIALISATION arg treatment change
cli_testcase(105, [symmtest], ['../prob_examples/public_examples/B/Puzzles/Hanoi6Sym.mch', '-mc', '10000', '-p', 'SYMMETRY_MODE', 'nauty', '-cc', '367', '1095', '-nogoal', '-strict'], 'Hanoi6Sym with nauty'). % was also 1096 transitions, see above
cli_testcase(106, [symmtest,tickets], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/Demo/DiningCryptographers.mch', '-mc', '100', '-cc', '65', '96', '-p', 'MAX_INITIALISATIONS', '100', '-p', 'SYMMETRY_MODE', 'off', '-strict'], '').
cli_testcase(107, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/Demo/DiningCryptographers.mch', '-mc', '100', '-cc', '65', '96', '-p', 'MAX_INITIALISATIONS', '100', '-p', 'SYMMETRY_MODE', 'nauty', '-strict'], '').
cli_testcase(108, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/Demo/DiningCryptographers.mch', '-mc', '100', '-cc', '65', '96', '-p', 'MAX_INITIALISATIONS', '100', '-p', 'SYMMETRY_MODE', 'hash', '-strict'], '').
cli_testcase(109, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/Fraikin/Library_FullSym.mch', '-mc', '10000', '-cc', '1158', '12064', '-p', 'MAX_INITIALISATIONS', '100', '-p', 'SYMMETRY_MODE', 'hash', '-strict', '-expcterr', 'model_check_incomplete'], '').
cli_testcase(110, [symmtest], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-p', 'DEFAULT_SETSIZE', '2', '-mc', '10000', '-strict', '-nodead', '-cc', '149', '248', '-expcterr', 'model_check_incomplete'], '').
cli_testcase(111, [symmtest], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-p', 'DEFAULT_SETSIZE', '2', '-mc', '10000', '-strict', '-nodead', '-cc', '82', '136', '-p', 'SYMMETRY_MODE', 'nauty', '-expcterr', 'model_check_incomplete'], ''). % used to be 137 transitions; changed in April 2016 due to INITIALISATION arg treatment change
cli_testcase(112, [symmtest], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-p', 'DEFAULT_SETSIZE', '2', '-mc', '10000', '-strict', '-nodead', '-cc', '82', '136', '-p', 'SYMMETRY_MODE', 'hash', '-expcterr', 'model_check_incomplete'], ''). % ditto
cli_testcase(113, [symmtest], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-p', 'DEFAULT_SETSIZE', '2', '-mc', '10000', '-strict', '-nodead', '-cc', '149', '204', '-p', 'SYMMETRY_MODE', 'flood', '-expcterr', 'model_check_incomplete'], '').
cli_testcase(114, [symmtest], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-p', 'DEFAULT_SETSIZE', '3', '-mc', '10000', '-strict', '-nodead', '-cc', '144', '331', '-p', 'SYMMETRY_MODE', 'nauty', '-expcterr', 'model_check_incomplete'], '').% used to be 332 transitions; changed in April 2016 due to INITIALISATION arg treatment change
cli_testcase(115, [symmtest], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-p', 'DEFAULT_SETSIZE', '3', '-mc', '10000', '-strict', '-nodead', '-cc', '144', '331', '-p', 'SYMMETRY_MODE', 'hash', '-expcterr', 'model_check_incomplete'], ''). % ditto
cli_testcase(116, [symmtest], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-p', 'DEFAULT_SETSIZE', '4', '-mc', '10000', '-strict', '-nodead', '-cc', '360', '1093', '-p', 'SYMMETRY_MODE', 'nauty', '-expcterr', 'model_check_incomplete'], '').
cli_testcase(117, [symmtest], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-p', 'DEFAULT_SETSIZE', '4', '-mc', '10000', '-strict', '-nodead', '-cc', '360', '1093', '-p', 'SYMMETRY_MODE', 'hash' , '-expcterr', 'model_check_incomplete'], '').
cli_testcase(118, [symmtest], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M1_mch.eventb', '-p', 'DEFAULT_SETSIZE', '3', '-mc', '10000', '-strict', '-nodead', '-cc', '1339', '3786', '-p', 'SYMMETRY_MODE', 'hash' , '-expcterr', 'model_check_incomplete'], ''). % used to be 3787 transitions; changed in April 2016 due to INITIALISATION arg treatment change
cli_testcase(119, [symmtest], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M1_mch.eventb', '-p', 'DEFAULT_SETSIZE', '3', '-mc', '10000', '-strict', '-nodead', '-cc', '1339', '3786', '-p', 'SYMMETRY_MODE', 'nauty', '-expcterr', 'model_check_incomplete'], ''). % used to be 3787 transitions; changed in April 2016 due to INITIALISATION arg treatment change
cli_testcase(120, [symmtest], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M1_mch.eventb', '-p', 'DEFAULT_SETSIZE', '3', '-mc', '10000', '-strict', '-nodead', '-cc', '3893', '10970', '-p', 'SYMMETRY_MODE', 'off', '-expcterr', 'model_check_incomplete'], '').
cli_testcase(121, [symmtest], ['../prob_examples/public_examples/B/SymmetryReduction/PhilRing.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'hash', '-cc', '8', '18', '-strict' , '-expcterr', 'model_check_incomplete'], '').
cli_testcase(122, [symmtest], ['../prob_examples/public_examples/B/SymmetryReduction/PhilRing.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SYMMETRY_MODE', 'nauty', '-cc', '8', '18', '-strict'], '').
cli_testcase(123, [symmtest], ['../prob_examples/public_examples/B/SymmetryReduction/PhilRing.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'nauty', '-cc', '13', '46', '-strict'], '').  % used to be 47 transitions; changed in April 2016 due to INITIALISATION arg treatment change
cli_testcase(124, [symmtest], ['../prob_examples/public_examples/B/SymmetryReduction/PhilRing.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'hash', '-cc', '13', '46', '-strict' , '-expcterr', 'model_check_incomplete'], ''). % ditto
cli_testcase(125, [symmtest], ['../prob_examples/public_examples/B/SymmetryReduction/PhilRing.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '5', '-p', 'SYMMETRY_MODE', 'nauty', '-cc', '53', '342', '-strict', '-expcterr', 'model_check_incomplete'], '').  % used to be 345 transitions; changed in April 2016 due to INITIALISATION arg treatment change
cli_testcase(126, [symmtest], ['../prob_examples/public_examples/B/SymmetryReduction/PhilRing.mch', '-mc', '1000', '-p', 'DEFAULT_SETSIZE', '5', '-p', 'SYMMETRY_MODE', 'hash', '-cc', '53', '342', '-strict' , '-expcterr', 'model_check_incomplete'], ''). % ditto
cli_testcase(127, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/TestValidation_ctx.eventb', '-init', '-aa', '11', '0', '0', '-strict'], '').
cli_testcase(128, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/platoon/platoon_1_mch.eventb', '-t', '-strict', '-expcterr', virtual_time_out], '').
cli_testcase(129, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/platoon/platoon_2_mch.eventb', '-t', '-p', 'CLPFD', 'FALSE', '-strict'], '').
cli_testcase(130, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/platoon/platoon_2_mch.eventb', '-t', '-p', 'CLPFD', 'TRUE', '-strict'], '').
cli_testcase(131, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/platoon/platoon_3_0_mch.eventb', '-t', '-strict'], '').
cli_testcase(132, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/platoon/platoon_4_0_mch.eventb', '-t', '-strict'], '').
cli_testcase(133, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Records/SampleThreeRecordsTranslated.mch', '-t', '-strict', '-mc', '100', '-cc', '4', '6'], 'Check record translation').
cli_testcase(134, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Records/INTEGER_Record.mch', '-t', '-strict', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE'], '').
cli_testcase(135, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Test1.0/Testid_mch.eventb', '-t', '-strict', '-pref_group', integer, int32, '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(136, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/sieve.eventb', '-t', '-strict', '-mc', '1000', '-strict', '--hash64', 142217803204665821], 'Complete model check of Sieve').
cli_testcase(137, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/TrafficLight/tlm_0.eventb', '-t', '-strict'], '').
cli_testcase(138, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/TrafficLight/tlm_1.eventb', '-t', '-strict'], '').
cli_testcase(139, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/TrafficLight/tlm_3.eventb', '-t', '-strict'], '').
cli_testcase(140, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Alexei/ttrain_fifo_mch.eventb', '-t', '-strict'], '').
cli_testcase(141, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/SET_Game/SET_Game_mch.eventb', '-t', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'MAX_OPERATIONS', 93], '').
cli_testcase(142, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Stefan_SecureBuilding/M21_mch.eventb', '-mc', '10000', '-t', '-strict'], '').
cli_testcase(143, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Abrial_Train_Ch17/train_4_mch_lukas.eventb', '-mc', '200', '-t', '-strict', '-noass'], '').
cli_testcase(144, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Test1.0/Testid_mch.eventb', '-t', '-strict', '-p', 'MAXINT', '99999', '-p', 'MININT', '-99999', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(145, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Test1.0/Testprj_mch.eventb', '-t', '-strict', '-p', 'MAXINT', '99999', '-p', 'MININT', '-99999'], '').
cli_testcase(146, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Test1.0/TestPartition_mch.eventb', '-t', '-strict'], '').
cli_testcase(147, [eventb_test], ['../prob_examples/public_examples/B/EventB_AtelierB/Lift.sys', '-t', '-strict'], '').
cli_testcase(148, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/MultiLevel/TestMC_m1_mch.eventb', '-mc', '1000', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-expcterr', 'event_error:evt1:simulation_error'], '').
cli_testcase(149, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Convergence/countdown_r_mch.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-cc', '7', '7', '--hash64', 714084921427841469], '').
cli_testcase(150, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Convergence/anticipated_grd_error.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-expcterr', 'event_error:decr:variant_negative'], '').
cli_testcase(151, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Convergence/anticipated_var_error.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-expcterr', 'event_error:inc:invalid_variant'], '').
cli_testcase(152, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Convergence/anticipated_svar_error.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-expcterr', 'event_error:add:invalid_variant'], '').
cli_testcase(153, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Convergence/convergent_grd_error.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-expcterr', 'event_error:decr:variant_negative'], '').
cli_testcase(154, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Convergence/convergent_var_error.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-expcterr', 'event_error:decr:invalid_variant'], '').
cli_testcase(155, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Convergence/convergent_svar_error.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-expcterr', 'event_error:do_nothing:invalid_variant'], '').
cli_testcase(156, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Convergence/variant_integer.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-cc', '8', '19'], '').
cli_testcase(157, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Convergence/variant_set.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-cc', '17', '97'], '').
cli_testcase(158, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Merging/merging_R2.eventb', '-mc', '100', '-strict', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', '20', '-nodead', '-cc', '12', '11'], '').
cli_testcase(159, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/MultiLevel/TheoremsInGuardsOK.eventb', '-mc', '10', '-nodead', '-strict'], '').
cli_testcase(160, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/MultiLevel/TheoremsInGuardsKO.eventb', '-mc', '10', '-nodead', '-expcterr', 'event_error:inc:invalid_theorem_in_guard', '-strict'], '').
cli_testcase(161, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/CompareWithFDR/GenPrime.csp', '-t', '-strict'], '').
cli_testcase(162, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/demo/buses.csp', '-t', '-strict'], '').
cli_testcase(163, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/demo/Peterson_v2.csp', '-t', '-strict', '-mc',1000, '-cc', 215, 429], '').
cli_testcase(164, [csp_test], ['-c', '../prob_examples/public_examples/CSP/mydemos/demo/mbuff.csp', '-t', '-strict'], '').
cli_testcase(165, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/demo/crossing.csp', '-t', '-strict', '-mc', '2000'], '').
cli_testcase(166, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/SetCompTests.csp', '-t', '-strict'], '').
cli_testcase(167, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/SimpleCompLinkedPar.csp', '-t', '-strict'], '').
cli_testcase(168, [csp_test], ['-c', '../prob_examples/public_examples/CSP/mydemos/SequenceComprTests2.csp', '-t', '-strict', '-v'], '').
cli_testcase(169, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/BigUnionInterTests.csp', '-t', '-strict'], '').
cli_testcase(170, [csp_test], ['../prob_examples/public_examples/B/Tickets/299/faust_deadlock.csp', '-mc', '100000', '-strict', '-expcterr', 'deadlock'], '').
cli_testcase(171, [csp_test], ['../prob_examples/public_examples/CSP/FDRFeatureTests/LargeStateSpace.csp', '-mc', '999999', '-cc', '101102', '101101', '-strict', '-expcterr', 'deadlock'], '').
cli_testcase(172, [csp_test], ['../prob_examples/public_examples/CSP/FDRFeatureTests/RenameRecordPattern.csp', '-mc', '999999', '-t', '-strict'], '').
cli_testcase(173, [csp_test], ['../prob_examples/public_examples/CSP/FDRFeatureTests/RenameRecordPattern2.csp', '-t', '-assertions', '-strict'], '').
cli_testcase(174, [csp_test], ['../prob_examples/public_examples/CSP/FDRFeatureTests/SimpleRefTest.csp', '-assertions', '-strict'], '').
cli_testcase(175, [csp_test], ['../prob_examples/public_examples/CSP/fontaine/RecDataType.csp', '-t', '-strict'], '').
cli_testcase(176, [csp_test], ['../prob_examples/public_examples/CSP/fontaine/RecDataTypeFunctions.csp', '-t', '-strict'], '').
cli_testcase(177, [csp_test], ['../prob_examples/public_examples/CSP/simple/InSituRefChecks.csp', '-csp_assertion', 'Q [F= MAIN', '-csp_assertion', 'MAIN [F= Q', '-csp_assertion', 'Q [T= R', '-csp_assertion', 'R [F= Q', '-cc', 9, 20, '-strict'], '').
cli_testcase(178, [csp_test], ['../prob_examples/public_examples/CSP/simple/InSituRefChecks.csp', '-csp_assertion', 'Q [F= MAIN', '-csp_assertion', 'Q [F= R', '-strict', '-expcterr', 'refinement_check_fails'], '').
cli_testcase(179, [csp_test], ['../prob_examples/public_examples/CSP/UML_Compass/Microwave.csp', '-assertions', '-strict'], '').
cli_testcase(180, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/demo/roscoe_section2-2.csp', '-assertions', '-strict'], '').
cli_testcase(181, [csp_test,csp_bench], ['../prob_examples/public_examples/CSP/ref_benchmarks/SimpleDeadlockCheck.csp', '-assertions', '-strict'], '').
cli_testcase(182, [csp_test,csp_bench], ['../prob_examples/public_examples/CSP/simple/DivChecks.csp', '-assertions', '-strict'], '').
cli_testcase(183, [csp_test,csp_bench], ['../prob_examples/public_examples/CSP/other/Ivo/Failure_Divergence_Tests.csp', '-assertions', '-strict'], '').
cli_testcase(184, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/Deadlock_Skip_Tests.csp', '-assertions', '-strict'], '').
cli_testcase(185, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/Failures_Tests.csp', '-assertions', '-strict'], '').
cli_testcase(186, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/AgentErrorsLetWithIn.csp', '-assertions', '-strict'], '').
cli_testcase(188, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/Deterministic_simple_v2.csp', '-assertions', '-strict'], '').
cli_testcase(189, [csp_test,codespeed], ['../prob_examples/public_examples/CSP/ref_benchmarks/abp_chapter5_roscoe_bench.csp', '-assertions', '-strict'], 'Roscoe chapter 5 tests').
cli_testcase(190, [csp_test], ['../prob_examples/public_examples/CSP/ref_benchmarks/ucs1.csp', '-assertions', '-strict'], '').
cli_testcase(191, [csp_test], ['../prob_examples/public_examples/CSP/simple/check_print.csp', '-t', '-strict'], '').
cli_testcase(192, [csp_test], ['../prob_examples/public_examples/CSP/simple/check_set_compr_pat.csp', '-t', '-mc', '1000', '-nodead', '-cc', '6', '10', '-strict'], '').
cli_testcase(193, [csp_test], ['../prob_examples/public_examples/CSP/ref_benchmarks/failsim.csp', '-assertions', '-strict'], '').
cli_testcase(194, [typechecker_test], ['../prob_examples/public_examples/B/Tickets/305/Environment.mch', '-t', '-strict'], '').
cli_testcase(195, [typechecker_test], ['../prob_examples/public_examples/B/Tickets/308/Basic_Controller.mch', '-t', '-mc', '100', '-strict'], '').
cli_testcase(196, [typechecker_test,unit], ['-typecheckertest','../prob_examples/public_examples/B/Typecheckertests/expressions.mch', '-strict'], '').

% SBMF 2010
cli_testcase(197, [counterex], ['../prob_examples/public_examples/B/Counterexamples/CarlaTravelAgencyErr.mch', '-mc', '1000', '-expcterr', 'invariant_violation', '-strict'], 'SBMF TravelAgency').
cli_testcase(198, [counterex], ['../prob_examples/public_examples/EventBPrologPackages/Stefan_SecureBuilding/M21_err3_mch.eventb', '-mc', '1000', '-expcterr', 'invariant_violation', '-strict'], 'SBMF SecureBldg_M21_err3').
cli_testcase(199, [counterex], ['../prob_examples/public_examples/B/PerformanceTests/dfcheck_houseset.mch', '-mc', '1000', '-df', '-expcterr', 'invariant_violation', '-expcterr', 'invariant_violation', '-strict'], 'SBMF Houseset').
cli_testcase(200, [counterex], ['../prob_examples/public_examples/B/PerformanceTests/BreadthFirstTest.mch', '-mc', '100000', '-expcterr', 'invariant_violation', '-strict'], 'SBMF BFTest').
cli_testcase(201, [counterex], ['../prob_examples/public_examples/B/PerformanceTests/DepthFirstTest.mch', '-mc', '100000', '-expcterr', 'invariant_violation', '-strict'], 'SBMF DFTest').
cli_testcase(202, [counterex], ['../prob_examples/public_examples/B/SymmetryReduction/Peterson_err.mch', '-mc', '10000', '-expcterr', 'invariant_violation', '-strict'], 'SBMF Peterson_err').
cli_testcase(203, [counterex], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-mc', '10000', '-strict', '-expcterr', 'deadlock'], '').
cli_testcase(204, [counterex], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M3_mch.eventb', '-mc', '10000', '-strict', '-expcterr', 'deadlock'], '').
cli_testcase(205, [counterex,infinite,total_function], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M0_infinite_int.eventb', '-p', 'CLPFD', 'TRUE',
 '-p', 'DEFAULT_SETSIZE', 3,
 '-t', '-expcterr', virtual_time_out ], 'Requires total function from NATURAL to TC_Type; now succeeds again'). % was  '-expcterr', setup_constants_fails

cli_testcase(206, [log_performance_tests], ['../prob_examples/public_examples/EventBPrologPackages/sieve.eventb', '-mc', '1000', '-l', 'log/sieve.log', '-strict', '--hash64', 142217803204665821], '').
cli_testcase(207, [puzzles_test], ['../prob_examples/public_examples/B/Demo/RussianPostalPuzzle.mch', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(208, [puzzles_test], ['../prob_examples/public_examples/B/Demo/TrainTorchPuzzle.mch', '-mc', '10000', '-nodead', '-expcterr', 'goal_found'], '').
cli_testcase(209, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/BlocksWorldGeneric.mch', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(210, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/Farmer.mch', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(211, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/Hanoi.mch', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(212, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/Hanoi6.mch', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(213, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/Hanoi8.mch', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(214, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/Hanoi6Sym.mch', '-p', 'SYMMETRY_MODE', 'hash', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(215, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/Hanoi6Sym.mch','-p','SYMMETRY_MODE','hash','-mc','10000','-card','Stakes','4','-expcterr','goal_found'], '').
cli_testcase(216, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/Hanoi8Sym.mch', '-p', 'SYMMETRY_MODE', 'hash', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(217, [puzzles_test], ['-mc', '10000', '../prob_examples/public_examples/XTL/Hanoi.P', '-nodead', '-expcterr', 'xtl_error'], '').
cli_testcase(218, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/water.mch', '-mc', '10000', '-expcterr', 'goal_found'], '').
cli_testcase(219, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/LatinSquares.mch', '-mc', '10000', '-p', 'MAXINT', 5, '-expcterr', deadlock], '').
cli_testcase(220, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/CrewAllocationConstants.mch', '-t', '-p', 'TIME_OUT', '5000', '-p', 'CLPFD', 'FALSE'], '').
cli_testcase(221, [puzzles_test,codespeed], ['../prob_examples/public_examples/B/Puzzles/CrewAllocationConstants.mch', '-t', '-p', 'TIME_OUT', '8000', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(222, [b_test,cruise], ['../prob_examples/public_examples/B/Benchmarks/Cruise_finite1.mch', '-mc', '1400', '-cc', '1361', '25696', '-strict', '--hash64', 203058745203935766, '-p', 'MAX_INITIALISATIONS', 5], 'Complete Model Check of Cruise Controller'). % 8.12.2015: adapted hash from 212678735433640605 for new order of variables
cli_testcase(223, [error_checks,wd], ['../prob_examples/public_examples/B/ErrorMachines/CardInfOverflow.mch', '-mc', '100', '-strict', '-expcterr', 'well_definedness_error', '-expcterr', 'invariant_violation', '-p', 'CLPFD', 'FALSE'], '').
cli_testcase(224, [error_checks], ['../prob_examples/public_examples/B/ErrorMachines/TwoOpsSameName.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(225, [error_checks], ['../prob_examples/public_examples/B/ErrorMachines/PreCondViolated.mch', '-mc', '10', '-strict', '-expcterr', 'precondition_error'], '').
cli_testcase(226, [error_checks], ['../prob_examples/public_examples/B/ErrorMachines/AssertCondViolated.mch', '-mc', '10', '-strict', '-expcterr', 'assert_error'], '').
cli_testcase(227, [error_checks,while], ['../prob_examples/public_examples/B/ErrorMachines/WhileLoopOk.mch', '-t', '-mc', '30', '-p', 'MAX_OPERATIONS', 30, '-strict'], '').
cli_testcase(228, [error_checks,while], ['../prob_examples/public_examples/B/ErrorMachines/WhileLoopVariantError.mch', '-mc', '10', '-strict', '-expcterr', 'while_variant_error'], '').
cli_testcase(229, [error_checks,while], ['../prob_examples/public_examples/B/ErrorMachines/WhileLoopVariantErr2.mch', '-mc', '10', '-strict', '-expcterr', 'while_variant_error'], '').
cli_testcase(230, [error_checks,while], ['../prob_examples/public_examples/B/ErrorMachines/WhileLoopInvariantError.mch', '-mc', '10', '-strict', '-expcterr', 'while_invariant_violation'], '').
cli_testcase(231, [error_checks,wd], ['../prob_examples/public_examples/B/ErrorMachines/FunLaws_wderror.mch', '-mc', '2500', '-strict', '-expcterr', 'well_definedness_error', '-expcterr', 'well_definedness_error'], '').
cli_testcase(232, [error_checks,wd], ['../prob_examples/public_examples/B/ErrorMachines/UndefinedArithmetic/DivByZero.mch', '-init', '-strict', '-expcterr', 'well_definedness_error'], '').
cli_testcase(233, [error_checks,wd], ['../prob_examples/public_examples/B/ErrorMachines/TestWDProblems_mch.eventb', '-init', '-animate', '1', '-strict', '-expcterr', 'well_definedness_error', '-expcterr', 'invariant_violation', '-expcterr', 'event_error:divby0:action_not_executable'], '').
cli_testcase(234, [error_checks,wd], ['../prob_examples/public_examples/B/ErrorMachines/UndefinedArithmetic/DivByZeroInsideNotExists.mch', '-strict', '-init', '-expcterr', 'well_definedness_error', '-expcterr', 'setup_constants_inconsistent'], '').
cli_testcase(235, [error_checks,wd], ['../prob_examples/public_examples/B/ErrorMachines/UndefinedArithmetic/PowerOfNegativeNumber.mch', '-strict', '-t', '-expcterr', 'well_definedness_error'], '').
cli_testcase(236, [error_checks,wd], ['../prob_examples/public_examples/B/ErrorMachines/UndefinedInter.mch', '-init', '-animate', '1', '-strict', '-expcterr', 'well_definedness_error','-expcterr', 'animate'], '').
cli_testcase(237, [error_checks,wd,quantified], ['../prob_examples/public_examples/B/ErrorMachines/UndefinedBigINTER.mch', '-init', '-animate', '1', '-strict', '-expcterr', 'well_definedness_error','-expcterr', 'animate'], '').
cli_testcase(238, [error_checks,wd], ['../prob_examples/public_examples/B/ErrorMachines/UndefinedArithmetic/MaxOfEmptySet.mch', '-init', '-strict', '-expcterr', 'well_definedness_error'], '').
cli_testcase(239, [error_checks,wd], ['../prob_examples/public_examples/B/ErrorMachines/UndefinedArithmetic/MinOfEmptySet.mch', '-init', '-strict', '-expcterr', 'well_definedness_error'], '').
cli_testcase(240, [error_checks,wd], ['../prob_examples/public_examples/B/ErrorMachines/UndefinedSeqOperations/LastEmptySeq.mch', '-init', '-strict', '-expcterr', 'well_definedness_error'], '').
cli_testcase(241, [error_checks,wd], ['../prob_examples/public_examples/B/ErrorMachines/UndefinedSeqOperations/FrontEmptySeq.mch', '-init', '-strict', '-expcterr', 'well_definedness_error'], '').
cli_testcase(242, [error_checks], ['../prob_examples/public_examples/B/ErrorMachines/OpCalls/Caller.mch', '-strict', '-expcterr', 'precondition_error', '-mc', '100'], '').
cli_testcase(243, [error_checks], ['../prob_examples/public_examples/B/ErrorMachines/MultiAssignInit_bug.mch', '-t', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], 'Check that we detect when we assign a variable twice in the INITIALISATION').
cli_testcase(244, [error_checks], ['../prob_examples/public_examples/B/ErrorMachines/TypeError1.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(245, [error_checks], ['../prob_examples/public_examples/B/ErrorMachines/TypeError2.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(246, [error_checks], ['../prob_examples/public_examples/B/ErrorMachines/TypeError3.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(247, [error_checks], ['../prob_examples/public_examples/B/ErrorMachines/TypeError4.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(248, [error_checks], ['../prob_examples/public_examples/B/ErrorMachines/CstDefError1.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(249, [error_checks], ['../prob_examples/public_examples/B/ErrorMachines/CstDefError2.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(250, [error_checks], ['../prob_examples/public_examples/B/ErrorMachines/ParameterError1.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(251, [error_checks], ['../prob_examples/public_examples/B/ErrorMachines/TicTacToe_ErrOp.mch','-strict','-expcterr','type_error','-expcterr','load_main_file','-expcterr',bmachine_static_checks], '').
cli_testcase(252, [error_checks], ['../prob_examples/public_examples/B/ErrorMachines/AssignmentToParameterErr.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(253, [error_checks], ['../prob_examples/public_examples/B/ErrorMachines/NoTypingForPara.mch', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], '').
cli_testcase(254, [error_checks,parser], ['../prob_examples/public_examples/B/ErrorMachines/RecursiveDefinition.mch', '-strict', '-expcterrpos', 'parse_error', 10, 0, '-expcterr', 'load_main_file'], '').
cli_testcase(255, [tickets,codespeed], ['../prob_examples/public_examples/B/Tickets/184/ModeProtocolMachine_mch.eventb', '-mc', '20000', '-noass', '-card', 'MANAGERS', '1', '-p', 'MAX_INITIALISATIONS', '10', '-p', 'CLPFD', 'TRUE', '-strict' , '-expcterr', 'model_check_incomplete'], '').
cli_testcase(256, [tickets], ['../prob_examples/public_examples/B/Tickets/184/ModeProtocolMachine_mch.eventb', '-mc', '200000', '-noass', '-p', 'MAX_INITIALISATIONS', '10', '-c', '--goal', 'card(incoming(MnAOC))>0 & card(outgoing(MnAOC))>0', '-strict', '-expcterr', 'goal_found'], '').
cli_testcase(257, [tickets], ['../prob_examples/public_examples/B/Tickets/102/Bijections_mch.eventb', '-t', '-mc', '1000', '-nodead', '-strict'], '').
cli_testcase(258, [tickets], ['../prob_examples/public_examples/B/Tickets/102/PartialInjections_mch.eventb', '-t', '-mc', '1000', '-nodead', '-strict'], '').
cli_testcase(259, [tickets], ['../prob_examples/public_examples/B/Tickets/102/PartialFunctions_mch.eventb', '-mc', '1000', '-nodead', '-strict'], '').
cli_testcase(260, [tickets], ['../prob_examples/public_examples/B/Tickets/102/Relations_mch.eventb', '-mc', '1000', '-nodead', '-strict'], '').
cli_testcase(261, [tickets], ['../prob_examples/public_examples/B/Tickets/102/TotalFunctions_mch.eventb', '-mc', '1000', '-nodead', '-strict'], '').
cli_testcase(262, [tickets], ['../prob_examples/public_examples/B/Tickets/102/TotalInjections_mch.eventb', '-mc', '1000', '-nodead', '-strict'], '').
cli_testcase(263, [tickets], ['../prob_examples/public_examples/B/Tickets/102/TotalSurjections_mch.eventb', '-mc', '1000', '-t', '-nodead', '-strict'], '').
cli_testcase(264, [tickets], ['../prob_examples/public_examples/B/Tickets/102/PartialSurjections_mch.eventb', '-mc', '1000', '-t', '-nodead', '-strict'], '').
cli_testcase(265, [tickets], ['../prob_examples/public_examples/B/Tickets/102/TotalRelations_mch.eventb', '-mc', '1000', '-t', '-nodead', '-strict'], '').
cli_testcase(266, [tickets], ['../prob_examples/public_examples/B/Tickets/Sven1/AbsIntern.mch', '-t', '-strict'], '').
cli_testcase(267, [tickets], ['../prob_examples/public_examples/B/Tickets/Sven1/AbsInternCst.mch', '-t', '-strict'], '').
cli_testcase(268, [tickets], ['../prob_examples/public_examples/B/Tickets/Valerio1/MAX_BIT_VECTOR_DEFINITION_v2.mch', '-t', '-strict', '-properties'], '').
cli_testcase(269, [tickets], ['../prob_examples/public_examples/B/Tickets/Valerio1/MAX_BIT_VECTOR_DEFINITION_v2.mch', '-t', '-strict', '-p', 'SYMBOLIC', 'TRUE'], '').
cli_testcase(270, [tickets], ['../prob_examples/public_examples/B/Tickets/Valerio1/MAX_BIT_VECTOR_DEFINITION_v2.mch', '-t', '-strict', '-p', 'MAXINT', '100', '-properties', '-p', 'SYMBOLIC', 'TRUE'], ''). % , '-p', 'EXPAND_CLOSURES_FOR_STATE', 'FALSE'
cli_testcase(271, [tickets], ['../prob_examples/public_examples/B/Tickets/Valerio1/MAX_BIT_VECTOR_DEFINITION_v2.mch', '-t', '-strict', '-p', 'MAXINT', '100', '-properties', '-p', 'SYMBOLIC', 'TRUE'], 'Valerio1 Ticket. Can not be run on windows due to memory restrictions.'). % , '-p', 'EXPAND_CLOSURES_FOR_STATE', 'TRUE'
cli_testcase(272, [tickets], ['../prob_examples/public_examples/B/Tickets/Butler1/M2_mch.eventb', '-mc', '1000', '-strict', '-expcterr', 'invariant_violation', '-expcterr', 'event_error:Enter:simulation_error'], '').
cli_testcase(273, [tickets], ['../prob_examples/public_examples/B/Tickets/Valerio3/BV16_DEFINITION.mch', '-t', '-strict', '-properties', '-p', 'SYMBOLIC', 'TRUE', '-p', 'CLPFD', 'TRUE'], ''). % , '-p', 'EXPAND_CLOSURES_FOR_STATE', 'FALSE'
cli_testcase(274, [tickets], ['../prob_examples/public_examples/B/Tickets/Valerio3/POWER2_v2.mch', '-init', '-strict', '-aa', '19', '0', '0'], '').
cli_testcase(275, [tickets], ['../prob_examples/public_examples/B/Tickets/Valerio3/POWER2.mch', '-init', '-strict', '-ma', '18', '0', '0'], '').
cli_testcase(276, [tickets], ['../prob_examples/public_examples/B/Tickets/Valerio4/FailingInit1.mch', '-init', '-strict', '-expcterr', 'initialisation_fails', '-expcterr', bmachine_static_checks], '').
cli_testcase(277, [tickets,wd], ['../prob_examples/public_examples/B/Tickets/Valerio4/FailingInit2.mch', '-init', '-strict', '-expcterr', 'initialisation_fails', '-expecterr', 'well_definedness_error', '-expcterr', bmachine_static_checks], '').
cli_testcase(278, [tickets,wd], ['../prob_examples/public_examples/B/Tickets/Valerio4/FailingInit3.mch', '-init', '-strict', '-expcterrpos', 'initialisation_fails', 10, 27, '-expcterr', 'well_definedness_error', '-expcterr', bmachine_static_checks], '').
cli_testcase(279, [tickets], ['../prob_examples/public_examples/B/Tickets/Valerio4/FailingInit4.mch', '-init', '-strict', '-expcterr', 'initialisation_fails', '-expcterr', bmachine_static_checks], '').
cli_testcase(280, [tickets], ['../prob_examples/public_examples/B/Tickets/Valerio4/FailingInit5.mch', '-init', '-strict', '-expcterr', 'initialisation_fails', '-expcterr', 'initialisation_fails', '-expcterr', bmachine_static_checks], '').
cli_testcase(281, [tickets], ['../prob_examples/public_examples/B/Tickets/Valerio6/Z80_Simple.mch', '-strict', '-t', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(282, [tickets], ['../prob_examples/public_examples/B/Tickets/Alstom1/TestBecomesElementOf.mch', '-strict', '-t', '-p', 'CLPFD', 'TRUE'], 'becomes element of cartesian product').
cli_testcase(283, [tickets], ['../prob_examples/public_examples/B/Tickets/Alstom1/TestPFBecomeSuch.mch', '-strict', '-t', '-p', 'CLPFD', 'TRUE'], 'becomes element of partial function').
cli_testcase(284, [tickets], ['../prob_examples/public_examples/B/Tickets/Alstom1/MAXINT_M1.mch', '-strict', '-t', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(285, [tickets], ['../prob_examples/public_examples/B/Tickets/93/lift.eventb', '-strict', '-t', '-expcterr', 'event_error:ScheduleGoingUp:invalid_theorem_in_guard', '-expcterr', 'event_error:ScheduleGoingUp:invalid_modification'], '').
cli_testcase(286, [tickets], ['../prob_examples/public_examples/B/Tester/TestExistsLetExtraction.mch', '-t', '-strict'], '').
cli_testcase(287, [tickets,imp], ['../prob_examples/public_examples/B/Tickets/177/loop_invariant.imp', '-strict', '-t'], '').
cli_testcase(288, [tickets], ['../prob_examples/public_examples/B/Tickets/181/alstom_memory.mch', '-strict', '-t', '-mc', '100'], '').
cli_testcase(289, [laws,rel_fnc], ['../prob_examples/public_examples/EventBPrologPackages/Laws/ExplicitComputationsEventB_ctx.eventb', '-init', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'TIME_OUT', '9500'], 'Many explicit computations which check that all operators work as expected on various concrete data (for Event-B).'). % increased time-out from 7 sec to 9.5 sec on 18th of May; TO DO: investigate whether we can reduce it again;  % was slowed down by commit ab8c8667c43324c15b337c213c4140af6ec7d110 add propagate_result_to_input for domain operator
cli_testcase(290, [laws], ['../prob_examples/public_examples/EventBPrologPackages/Laws/LawsEventBSpecific.eventb', '-t', '-mc', '1000', '-nodead', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE', '-p', 'CLPFD', 'FALSE', '-expcterr', virtual_time_out], '').
cli_testcase(291, [laws], ['../prob_examples/public_examples/B/Laws/EqualityLaws.mch', '-t', '-mc', '100000', '-nodead', '-c', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE'], '').
cli_testcase(292, [laws,rel_fnc], ['../prob_examples/public_examples/B/Laws/ExplicitComputations.mch', '-t', '-mc', '100000', '-nodead', '-c', '-strict', '-p', 'CLPFD', 'FALSE', '-p', 'TIME_OUT', '7000'], 'Many explicit computations which check that all operators work as expected on various concrete data.').
cli_testcase(293, [laws,rel_fnc,cbc], ['../prob_examples/public_examples/B/Laws/ExplicitComputations.mch', '-t', '-mc', '100000', '-nodead', '-c', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'TIME_OUT', '9000'], 'Many explicit computations which check that all operators work as expected on various concrete data (CLPFD mode).').
cli_testcase(294, [laws], ['../prob_examples/public_examples/B/Laws/BoolLaws.mch', '-mc', '100000', '-nodead', '-c', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE'], 'Various laws about BOOL datatype and boolean connectives.').
cli_testcase(295, [laws], ['../prob_examples/public_examples/B/Laws/ArithmeticLaws.mch', '-mc', '100000', '-c', '-nodead', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE'], 'Various arithmetic laws.').
cli_testcase(296, [laws], ['../prob_examples/public_examples/B/Laws/NatRangeLaws.mch', '-mc', '100000', '-c', '-nodead', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE'], 'Laws about intervals etc.').
cli_testcase(297, [laws,rel_fnc], ['../prob_examples/public_examples/B/Laws/RelLaws.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE'], 'Use model checking to check a variety of laws on relations.').
cli_testcase(298, [laws], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
'../prob_examples/public_examples/B/Laws/SetLaws.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'FALSE'], 'Mathematical Laws about Sets').
cli_testcase(299, [laws], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
'../prob_examples/public_examples/B/Laws/SetLaws.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict', '-p', 'SYMBOLIC', 'TRUE', '-p', 'CLPFD', 'FALSE'], 'Mathematical Laws about Sets').
cli_testcase(300, [laws], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
'../prob_examples/public_examples/B/Laws/SetLaws.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE'], 'Mathematical Laws about Sets').
cli_testcase(301, [laws], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
'../prob_examples/public_examples/B/Laws/SetLaws.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'SYMBOLIC', 'TRUE'], 'Mathematical Laws about Sets').
cli_testcase(302, [laws], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
'../prob_examples/public_examples/B/Laws/SetLawsNat.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'FALSE'], 'Mathematical Laws about Sets (with Natural Numbers)').
cli_testcase(303, [laws], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
'../prob_examples/public_examples/B/Laws/SetLawsNat.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE','-p', 'DOUBLE_EVALUATION', 'TRUE'], 'Mathematical Laws about Sets (with Natural Numbers)').
cli_testcase(304, [laws], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
'../prob_examples/public_examples/B/Laws/SetLawsNat.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict', '-p', 'SYMBOLIC', 'TRUE'], 'Mathematical Laws about Sets (with Natural Numbers)').
cli_testcase(305, [laws,repl], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
  '../prob_examples/public_examples/B/Laws/REPL_UNIT_TESTS0.mch', '-t', '-mc', '100000', '-c', '-strict', '-p', 'CLPFD', 'TRUE', '-assertions', '-p', 'TIME_OUT', '28000'], 'Some REPL Unit Tests').
cli_testcase(306, [laws,repl], ['../prob_examples/public_examples/B/Laws/REPL_UNIT_TESTS1.mch', '-t', '-mc', '100000', '-c', '-strict', '-p', 'CLPFD', 'TRUE', '-assertions', '-p', 'TIME_OUT', '28000'], 'Some REPL Unit Tests').
cli_testcase(307, [laws,repl], ['../prob_examples/public_examples/B/Laws/REPL_UNIT_TESTS.mch', '-t', '-mc', '100000', '-c', '-strict', '-p', 'CLPFD', 'TRUE', '-assertions', '-p', 'TIME_OUT', '45000'], 'Some REPL Unit Tests').
cli_testcase(308, [laws,rel_fnc], ['../prob_examples/public_examples/B/Daniel/rel_fnc.mch', '-t', '-mc', '20000', '-strict', '-p', 'CLPFD', 'TRUE'], 'A small machine to check that rel/fnc work as expected.').
cli_testcase(309, [laws], ['../prob_examples/public_examples/B/Laws/CardinalityLaws.mch', '-mc', '2500', '-c', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE'], 'Cardinality Laws').
cli_testcase(310, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-cbc_deadlock', '-expcterr', 'cbc_deadlock_check', '-strict'], '').
cli_testcase(311, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M1_mch.eventb', '-cbc_deadlock', '-expcterr', 'cbc_deadlock_check', '-strict'], '').
cli_testcase(312, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M2_mch.eventb', '-cbc_deadlock', '-expcterr', 'cbc_deadlock_check', '-strict'], '').
cli_testcase(313, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M3_mch.eventb', '-cbc_deadlock', '-expcterr', 'cbc_deadlock_check', '-strict'], '').
cli_testcase(314, [cbc_deadlock], ['../prob_examples/public_examples/B/Demo/Lift.mch', '-cbc_deadlock', '-strict'], '').
cli_testcase(315, [cbc_deadlock], ['../prob_examples/public_examples/B/Benchmarks/Cruise_finite1.mch', '-cbc_deadlock', '-strict'], '').
cli_testcase(316, [cbc_deadlock], ['../prob_examples/public_examples/B/Demo/RussianPostalPuzzle.mch', '-cbc_deadlock', '-strict'], '').
cli_testcase(317, [cbc_deadlock], ['../prob_examples/public_examples/B/CBC/NoDeadlock.mch', '-cbc_deadlock', '-strict'], '').
cli_testcase(318, [cbc_deadlock], ['../prob_examples/public_examples/B/CBC/SimpleDeadlock1.mch', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(319, [cbc_deadlock], ['../prob_examples/public_examples/B/CBC/DeadlockIntBools.mch', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(320, [cbc_deadlock], ['../prob_examples/public_examples/B/CBC/ProcessIDStarvation.mch', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(321, [cbc_deadlock], ['../prob_examples/public_examples/B/CBC/ProcessIDStarvation_InvErr.mch', '-cbc_deadlock', '-strict'], '').
cli_testcase(322, [cbc_deadlock], ['../prob_examples/public_examples/B/CBC/MinSet_Deadlock.mch', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(323, [cbc_deadlock,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/Vizing_Breitner/m12_Deadlock_Freedom_mch.eventb', '-p','TIME_OUT','200','-cbc_deadlock', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-strict', '-p','DEFAULT_SETSIZE',1], ''). /* with size 2 only works if SMT mode is allowed to instantiate */
cli_testcase(324, [cbc,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/Vizing_Breitner/m11_Conv_Stages_mch.eventb', '-cbc_deadlock', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-strict','-p', 'DEFAULT_SETSIZE',2,
 '-p', 'TIME_OUT', '7500'], 'Vizing_Breitner challenging Deadlock Test'). % increased time-out for jenkins; locally the test runs in 22-23 seconds (MBA, Sept. 2015); time-out factor is 10 for -cbc_deadlock
cli_testcase(325, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/Deadlock/BrakeSystem_mch.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(326, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/Deadlock/QueenEventsTF_mch.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check', '-p', 'CLPFD', 'FALSE'], '').
cli_testcase(327, [cbc_deadlock,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/Deadlock/QueenEventsTF_mch.eventb', '-cbc_deadlock_pred', 'n=15', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(328, [cbc_deadlock,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/Deadlock/QueenEventsTF_mch.eventb', '-cbc_deadlock_pred', 'n=32', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(329, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/Alexei/ttrain_mch.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(330, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/Alexei/ttrain_mch_unproven.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(331, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/Alexei/ttrain_fifo_mch.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(332, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/platoon/platoon_1_mch.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check' ,'-p', 'CLPFD', 'TRUE' % TO DO: investigate why we now need CLPFD or SMT
], '').
cli_testcase(333, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/platoon/platoon_2_mch.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check', '-p', 'CLPFD', 'FALSE'], '').
cli_testcase(334, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/platoon/platoon_2_mch.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(335, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/Abrial_Teaching/ch2_car/m0_mch_deadlock.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(336, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/Abrial_Teaching/ch2_car/m0_mch_deadlock.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(337, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/Abrial_Teaching/ch2_car/m0_mch.eventb', '-cbc_deadlock', '-strict'], '').
cli_testcase(338, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/Abrial_Teaching/ch2_car/m1_mch.eventb', '-cbc_deadlock', '-strict'], '').
cli_testcase(339, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/Abrial_Teaching/ch2_car/m2_mch.eventb', '-cbc_deadlock', '-strict'], '').
cli_testcase(340, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/Abrial_Teaching/ch2_car/m3_mch.eventb', '-cbc_deadlock', '-strict'], '').
cli_testcase(341, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/Abrial_Teaching/search/m_1b_mch.eventb', '-cbc_deadlock', '-strict'], '').
cli_testcase(342, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/TrafficLight/tlm_1.eventb', '-cbc_deadlock', '-strict', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(343, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/TrafficLight/tlm_2.eventb', '-cbc_deadlock', '-strict', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(344, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/TrafficLight/tlm_3.eventb', '-cbc_deadlock', '-strict', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(345, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/ProofDirected/benchmarks/siemens_mch_0.eventb', '-cbc_deadlock', '-strict'], '').
cli_testcase(346, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/ProofDirected/benchmarks/scheduler.eventb', '-cbc_deadlock', '-strict'], '').
cli_testcase(347, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/ProofDirected/benchmarks/earley_2.eventb', '-cbc_deadlock', '-strict', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(348, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/ProofDirected/benchmarks/mondex_m2.eventb', '-cbc_deadlock', '-strict', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(349, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/ProofDirected/benchmarks/cxcc0.eventb', '-cbc_deadlock', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-expcterr', 'cbc_deadlock_check', '-strict'], '').
cli_testcase(350, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/ProofDirected/benchmarks/cxcc0.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check', '-p', 'USE_RECORD_CONSTRUCTION', 'FALSE'], '').
cli_testcase(351, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/ProofDirected/benchmarks/FMCH02.eventb', '-p', 'DEFAULT_SETSIZE', '1', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(352, [cbc_deadlock,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_mch_WithInit.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check', '-p', 'SMT', 'TRUE'], '').
cli_testcase(353, [cbc_deadlock,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check', '-p', 'SMT', 'TRUE'], '').
cli_testcase(354, [cbc_deadlock,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v2.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check', '-p', 'SMT', 'TRUE'], '').
cli_testcase(355, [cbc_deadlock,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v3.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check', '-p', 'SMT', 'TRUE'], '').
cli_testcase(356, [cbc_deadlock,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v4.eventb', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check', '-p', 'SMT', 'TRUE'], '').
cli_testcase(357, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v5.eventb', '-cbc_deadlock', '-strict', '-p', 'DEFAULT_SETSIZE', '1'], '').
cli_testcase(358, [cbc_deadlock,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v5.eventb', '-cbc_deadlock', '-strict', '-p', 'DEFAULT_SETSIZE', '2', '-expcterr', 'cbc_deadlock_check', '-p', 'SMT', 'TRUE', '-p', 'CLPFD', 'FALSE'], '').
cli_testcase(359, [cbc_deadlock,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v5.eventb', '-cbc_deadlock', '-strict', '-p', 'DEFAULT_SETSIZE', '2', '-expcterr', 'cbc_deadlock_check', '-p', 'SMT', 'TRUE', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(360, [cbc_deadlock,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v6.eventb', '-cbc_deadlock', '-strict', '-p', 'DEFAULT_SETSIZE', '1', '-p', 'SMT', 'TRUE'], '').
cli_testcase(361, [cbc_deadlock,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v6.eventb', '-cbc_deadlock', '-strict', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SMT', 'TRUE'], '').
cli_testcase(362, [cbc_deadlock,smt_test], ['../prob_examples/public_examples/B/LTL/Turcanu/machineEco_mch4.eventb', '-cbc_deadlock', '-p', 'SMT', 'TRUE', '-strict', '-expcterr', 'cbc_deadlock_check'], '').
cli_testcase(363, [cbc_deadlock], ['../prob_examples/public_examples/B/Mathematical/GraphIso/HardGraph.mch', '-cbc_deadlock', '-expcterr', 'cbc_deadlock_check_time_out', '-p', 'TIME_OUT', '80', '-p', 'MAX_INITIALISATIONS', '0', '-strict'], 'Test that time-out triggered within CBC Deadlock check').
cli_testcase(364, [cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/BMS/Poker_v2/FiveCardSolo_mch.eventb', '-cbc_deadlock', '-strict'], '').
cli_testcase(365, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/StableMarriages.mch', '-t', '-strict', '-cc', '7', '15', '-mc', '100', '--hash64', 192740982584039033], '').
cli_testcase(366, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/StableMarriagesLarger.mch', '-t', '-strict', '-cc', '7', '46', '-mc', '100', '--hash64', 13443197532344872], '').
cli_testcase(367, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/GolombRuler.mch', '-t', '-p', 'CLPFD', 'TRUE', '-strict'], '').
cli_testcase(368, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/Tischordnung.mch', '-t', '-strict', '-cc', '3', '7', '-mc', '100', '--hash64', 168992916868770688], '').
cli_testcase(369, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/Wegenetz.mch', '-mc', '1000', '-strict', '-nodead', '-expcterr', 'goal_found'], '').
cli_testcase(370, [puzzles_test], ['../prob_examples/public_examples/B/Demo/RussianPostalPuzzle.mch', '-mc', '1000', '-strict', '-nodead', '-expcterr', 'goal_found'], '').
cli_testcase(371, [puzzles_test], ['../prob_examples/public_examples/B/Demo/TrainTorchPuzzle.mch', '-mc', '1000', '-strict', '-nodead', '-expcterr', 'goal_found'], '').
cli_testcase(372, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/BlocksWorldGeneric.mch', '-mc', '1000', '-strict', '-nodead', '-expcterr', 'goal_found'], '').
cli_testcase(373, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/Farmer.mch', '-mc', '1000', '-strict', '-nodead', '-expcterr', 'goal_found'], '').
cli_testcase(374, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/Hanoi.mch', '-mc', '1000', '-strict', '-nodead', '-expcterr', 'goal_found'], '').
cli_testcase(375, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/water.mch', '-mc', '1000', '-strict', '-nodead', '-expcterr', 'goal_found'], '').
cli_testcase(376, [puzzles_test], ['-p', 'CLPFD', 'TRUE', '../prob_examples/public_examples/EventBPrologPackages/Puzzles/CopyPastAsEvent_mch.eventb', '-t', '-strict'], '').
cli_testcase(377, [puzzles_test], ['../prob_examples/public_examples/EventBPrologPackages/Puzzles/Sudoku_ctx.eventb', '-t', '-strict'], '').
cli_testcase(378, [puzzles_test,queens], ['-p', 'CLPFD', 'TRUE', '-p', 'MAX_OPERATIONS', 5000,
 '../prob_examples/public_examples/EventBPrologPackages/Puzzles/NQueensAsEvent_mch.eventb', '-t', '-strict'], '').
cli_testcase(379, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/PexSolve.mch', '-t', '-strict'], '').
cli_testcase(380, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/KnightsTour.mch', '-t', '-strict'], '').
cli_testcase(381, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/KnightsTour_v2.mch', '-t', '-strict'], '').
cli_testcase(382, [puzzles_test], ['-p', 'CLPFD', 'TRUE', '../prob_examples/public_examples/B/Puzzles/SudokuAsConstant.mch', '-t', '-strict'], '').
cli_testcase(383, [puzzles_test,cbc], ['-p', 'CLPFD', 'TRUE', '../prob_examples/public_examples/B/Puzzles/SudokuHexAsConstant.mch', '-t', '-properties', '-sptxt', '../prob_examples/public_examples/B/Puzzles/SudokuHexAsConstant_sptxt.txt', '-strict'], '').
cli_testcase(384, [puzzles_test], ['-p', 'CLPFD', 'TRUE', '../prob_examples/public_examples/B/Puzzles/SudokuAsConstantNoSolution.mch', '-init', '-strict', '-expcterr', 'setup_constants_fails'], '').
cli_testcase(385, [puzzles_test], ['-p', 'CLPFD', 'TRUE', '../prob_examples/public_examples/B/Puzzles/LatinSquaresBig.mch', '-t', '-strict'], '').
cli_testcase(386, [puzzles_test,queens], ['-p', 'CLPFD', 'TRUE', '../prob_examples/public_examples/B/Puzzles/NQueens50.mch', '-t', '-properties', '-strict'], '').
cli_testcase(387, [puzzles_test,cbc,codespeed], ['-p', 'CLPFD', 'TRUE', '-p', 'MAX_INITIALISATIONS', 19,
'../prob_examples/public_examples/B/Puzzles/TwoQueensSevenKnights.mch', '-t', '-assertions', '-strict'], '').
cli_testcase(388, [puzzles_test], ['-p', 'CLPFD', 'TRUE', '../prob_examples/public_examples/B/Puzzles/SendMoreMoney.mch', '-t', '-strict'], '').
cli_testcase(389, [puzzles_test], ['-p', 'CLPFD', 'TRUE', '../prob_examples/public_examples/B/Puzzles/KissPassion.mch', '-t', '-strict'], '').
cli_testcase(390, [puzzles_test], ['-p', 'CLPFD', 'TRUE', '../prob_examples/public_examples/B/Puzzles/CopySavePasteTools.mch', '-t', '-strict'], '').
cli_testcase(391, [puzzles_test], ['-p', 'CLPFD', 'TRUE', '../prob_examples/public_examples/B/Puzzles/NoNeighboursProblem.mch', '-cbc_sequence', 'SolAbove', '-strict'], 'Neighbour differ by more than one puzzle').
cli_testcase(392, [puzzles_test], ['-p', 'CLPFD', 'TRUE', '../prob_examples/public_examples/B/Puzzles/SubsetSum.mch', '-t', '-strict'], '').
cli_testcase(393, [puzzles_test], ['-p', 'CLPFD', 'TRUE', '../prob_examples/public_examples/B/Puzzles/SubsetSum2.mch', '-t', '-strict'], '').
cli_testcase(394, [puzzles_test], ['-p', 'CLPFD', 'TRUE', '../prob_examples/public_examples/B/Puzzles/WhoKilledAgatha.mch', '-t', '-mc', '100', '-p', 'MAX_INITIALISATIONS', '100', '-strict'], '').
cli_testcase(395, [puzzles_test,codespeed], ['../prob_examples/public_examples/EventBPrologPackages/SET_Game/SET_GAM_Sym_NoSet20_mch.eventb', '-mc', '1000', '-p', 'SYMMETRY_MODE', 'hash', '-p', 'TIME_OUT', '7000', '-p', 'CLPFD', 'TRUE', '-df', '-goal', 'n=18', '-p', 'MAX_OPERATIONS', '20', '-strict', '-expcterr', 'goal_found'], 'SET Game Test').
cli_testcase(396, [time_out_tests], ['../prob_examples/public_examples/B/PerformanceTests/SMTLIB/DTP_k2_n35_c175_s1.mch', '-strict', '-init', '-expcterr', 'setup_constants_fails', '-expcterr', 'time_out', '-p', 'CLPFD', 'FALSE'], 'Without CLPFD, this test times out. 1191 is a copy of this test using CLPFD').
cli_testcase(397, [cbc,smt_test], ['../prob_examples/public_examples/B/Benchmarks/scheduler.mch', '-cbc', 'all', '-strict', '-p', 'SMT', 'TRUE'], '').
cli_testcase(398, [cbc,smt_test], ['../prob_examples/public_examples/B/CBC/NATURAL_Add.mch', '-cbc', 'all', '-strict', '-p', 'SMT', 'TRUE', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(399, [cbc], ['../prob_examples/public_examples/B/Demo/Lift.mch', '-cbc', 'all', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(400, [cbc], ['../prob_examples/public_examples/B/Tutorial/DoubleCounter.mch', '-cbc', 'Double', '-strict', '-expcterr', 'cbc', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(401, [cbc], ['../prob_examples/public_examples/B/Tutorial/DoubleCounter.mch', '-cbc', 'Double', '-strict', '-expcterr', 'cbc', '-p', 'CLPFD', 'FALSE', '-p', 'MAXINT', '150'], '').
cli_testcase(402, [cbc], ['../prob_examples/public_examples/B/Special/ConstrainedBasedChecking/mesiSet.mch', '-cbc', 'all', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(403, [cbc], ['../prob_examples/public_examples/B/Special/ConstrainedBasedChecking/mesiSet.mch', '-cbc', 'write', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(404, [cbc], ['../prob_examples/public_examples/B/CBC/ProcessIDStarvation_InvErr.mch', '-cbc', 'all', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(405, [cbc], ['../prob_examples/public_examples/B/CBC/ProcessIDStarvation_InvErr.mch', '-cbc', 'Delete', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(406, [cbc], ['../prob_examples/public_examples/B/CBC/ProcessIDStarvation_InvErr.mch', '-cbc', 'Run', '-strict'], '').
cli_testcase(407, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/Alexei/ttrain_mch.eventb', '-cbc', 'all', '-strict'], '').
cli_testcase(408, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/Alexei/ttrain_mch_unproven.eventb', '-cbc', 'bridge', '-strict'], '').
cli_testcase(409, [cbc,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/Alexei/ttrain_mch_unproven.eventb', '-cbc', 'approach', '-strict', '-p', 'SMT', 'TRUE'], '').
cli_testcase(410, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/Alexei/ttrain_mch_unproven.eventb', '-cbc', 'exit', '-strict'], '').
cli_testcase(411, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/Alexei/ttrain_fifo_mch.eventb', '-cbc', 'bridge', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(412, [cbc,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/Alexei/ttrain_fifo_mch.eventb', '-cbc', 'approach', '-strict', '-p', 'SMT', 'TRUE'], '').
cli_testcase(413, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/Alexei/ttrain_fifo_mch.eventb', '-cbc', 'exit', '-strict'], '').
cli_testcase(414, [cbc,smt_test,codespeed], ['../prob_examples/public_examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v5.eventb', '-cbc', 'all', '-p', 'DEFAULT_SETSIZE', '4', '-p', 'SMT', 'TRUE', '-strict'], '').
cli_testcase(415, [cbc,smt_test,codespeed], ['../prob_examples/public_examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v6.eventb', '-cbc', 'all', '-p', 'DEFAULT_SETSIZE', '2', '-p', 'SMT', 'TRUE', '-strict'], '').
cli_testcase(416, [cbc,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/BPEL2B_PurchaseOrder/Model_Machine_4_NoDeadlock_v6.eventb', '-cbc', 'all', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SMT', 'TRUE', '-strict'], '').
cli_testcase(417, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/Tests/Ticket167.eventb', '-cbc', 'inc', '-strict'], '').
cli_testcase(418, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/BMS/Poker_v2/FiveCardSolo_mch.eventb', '-p', 'CLPFD', 'TRUE', '-cbc', 'redeal', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(419, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/BMS/Poker_v2/FiveCardSolo_mch.eventb', '-p', 'CLPFD', 'TRUE', '-cbc', 'HasFlush', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(420, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/BMS/Poker_v2/FiveCardSolo_mch.eventb', '-p', 'CLPFD', 'TRUE', '-cbc', 'hasFourOfAKind', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(421, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/BMS/Poker_v2/FiveCardSolo_mch.eventb', '-p', 'CLPFD', 'TRUE', '-cbc', 'HasStraight', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(422, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/BMS/Poker_v2/FiveCardSolo_mch.eventb', '-p', 'CLPFD', 'TRUE', '-cbc', 'AddACredit', '-cbc', 'deal', '-strict', '-cbc', 'selectme', '-cbc', 'invselectme'], '').
cli_testcase(423, [cbc,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/BMS/Poker_v2/FiveCardSolo_mch.eventb', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-cbc', 'hasFullHouse', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(424, [cbc,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/BMS/Poker_v2/FiveCardSolo_mch.eventb', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-cbc', 'all', '-strict', '-expcterr', 'cbc'], '').
cli_testcase(425, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M0_mch.eventb', '-cbc', 'all', '-strict'], '').
cli_testcase(426, [cbc,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/SSF/Bepi_Soton/M1_mch.eventb', '-cbc', 'all', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE'], '').
cli_testcase(427, [kodkod], ['-kodkod_comparision', '2', '../prob_examples/public_examples/B/Kodkod/expressions/functions.mch', '-strict'], '').
cli_testcase(428, [kodkod], ['-p', 'CLPFD', 'true', '-kodkod_comparision', '2', '../prob_examples/public_examples/B/Kodkod/expressions/triples.mch', '-strict'], '').
cli_testcase(429, [kodkod], ['-kodkod_comparision', '3', '../prob_examples/public_examples/B/Kodkod/expressions/relations.mch', '-strict'], '').
cli_testcase(430, [kodkod], ['-kodkod_comparision', '0', '../prob_examples/public_examples/B/Kodkod/expressions/sequences.mch', '-strict'], '').
cli_testcase(431, [kodkod], ['-kodkod_comparision', '1', '../prob_examples/public_examples/B/Kodkod/NoNeighboursProblem.mch', '-strict'], '').
cli_testcase(432, [kodkod], ['-p','MAX_INITIALISATIONS',50,'-kodkod_comparision', '0', '../prob_examples/public_examples/B/Kodkod/WhoKilledAgatha.mch', '-strict'], '').
cli_testcase(433, [analysis_test], ['-pacheck', '../prob_examples/public_examples/B/predicate_analysis/basic.mch', '-strict'], '').
cli_testcase(434, [analysis_test], ['-pacheck', '../prob_examples/public_examples/B/predicate_analysis/function_cards.mch', '-strict'], '').
cli_testcase(435, [analysis_test], ['-pacheck', '../prob_examples/public_examples/B/predicate_analysis/setext_dombug.mch', '-strict'], '').
cli_testcase(436, [csp_test], ['../prob_examples/public_examples/CSP/ref_benchmarks/general4.csp', '-assertions', '-strict'], '').
cli_testcase(437, [b_test], ['../prob_examples/public_examples/B/FeatureChecks/TestPrimedBecomeSuch.mch', '-t', '-strict', '-expcterr', virtual_time_out], 'Test for $0 usage in become such substitutions.').
cli_testcase(438, [csp_test], ['../prob_examples/public_examples/CSP/ref_benchmarks/basin_olderog_bank.csp', '-assertions', '-strict'], '').
cli_testcase(439, [b_test,strings], ['../prob_examples/public_examples/B/FeatureChecks/EmptyString.mch', '-t', '-strict', '-assertions'], 'Check that we allow the empty string.').

cli_testcase(440, [b_test,animate], ['../prob_examples/public_examples/B/Simple/UpCounter.mch', '-init','-animate', '10', '-strict', '-goal', 'c=10', '-check_goal', '-p', 'MAXINT', '100'], 'Check that animate executes exactly 10 operations after INITIALISATION.').
cli_testcase(441, [b_test,animate], ['../prob_examples/public_examples/B/Simple/UpCounter.mch','-animate', '10', '-strict', '-goal', 'c=8', '-check_goal', '-p', 'MAXINT', '100'], 'Without -init we exacte SETUP_CONSTANTS, INITIALISATION and 8 operations.').
cli_testcase(442, [b_test,animate], ['../prob_examples/public_examples/B/Simple/UpCounter.mch', '-init','-animate', '10', '-strict', '-goal', 'c=11', '-check_goal', '-p', 'MAXINT', '100', '-expcterr', 'check_goal'], 'Check that -check_goal fails if GOAL false.').
cli_testcase(443, [b_test,animate], ['../prob_examples/public_examples/B/Simple/UpCounter.mch','-animate', '10', '-strict', '-p', 'MAXINT', '100', '-det_check'], 'Check that all animation steps are deterministic.').
cli_testcase(444, [b_test,animate], ['../prob_examples/public_examples/B/Simple/UpCounter_NonDet.mch','-animate', '4', '-strict', '-p', 'MAXINT', '100', '-det_check', '-expcterr', 'det_check'], 'Check that all -det_check detects non-deterministic steps.').
cli_testcase(445, [b_test,animate], ['../prob_examples/public_examples/B/Simple/UpCounter.mch', '-init','-animate', '10', '-strict', '-goal', 'c=10 & c>1 & c<mx', '-check_goal', '-p', 'MAXINT', '100'], 'Check that -check_goal can also deal with conjuncts.').
cli_testcase(446, [b_test,animate], ['../prob_examples/public_examples/B/Simple/UpCounter.mch', '-init','-animate', '30', '-strict', '-goal', 'c=20', '-check_goal', '-p', 'MAXINT', '20', '-expcterr', 'animate'], 'Check that error raised if we cannot animate all steps -animate (but that we stay in the final state).').
cli_testcase(447, [b_test,animate], ['../prob_examples/public_examples/B/Simple/UpCounter.mch', '-init','-animate_all', '-animate_stats', '-strict', '-goal', 'c=20', '-check_goal', '-p', 'MAXINT', '20'], 'Check that no error raised if we cannot animate all steps -animate_all (and that we stay in the final state).').

cli_testcase(448, [tickets], ['../prob_examples/public_examples/B/Tickets/Hansen1/Bi_func.mch', '-mc', '1000', '-t', '-strict'], 'Check issue with existential quantifier inside set comprehension.').
cli_testcase(449, [tickets], ['../prob_examples/public_examples/B/Tickets/Burdy1/TestOrApplyFun.mch', '-mc', '1000', '-t', '-strict'], 'Check issue with undefined function application in context of disjunct.').
cli_testcase(450, [tickets], ['../prob_examples/public_examples/B/Tickets/Plagge1/KnightKnaves2.mch', '-init', '-strict'], 'Check that no time-out occurs in compiled mode').

cli_testcase(451, [b_test,animate,history], ['../prob_examples/public_examples/B/Simple/UpCounter.mch', '-init','-animate', '10', '-strict', '-p', 'MAXINT', '100', '-his', '../prob_examples/public_examples/B/Simple/UpCounter_his.txt'], 'Check that history properly written.').
cli_testcase(452, [b_test,animate,history], ['../prob_examples/public_examples/B/Simple/UpCounter.mch', '-init','-animate', '10', '-strict', '-p', 'MAXINT', '100', '-his', '../prob_examples/public_examples/B/Simple/UpCounter_his.txt','-his_option', 'show_states'], 'Check that history option -his_option show_states works.').
cli_testcase(453, [b_test,animate,history], ['../prob_examples/public_examples/B/Simple/UpCounter.mch', '-init','-animate', '10', '-strict', '-p', 'MAXINT', '100', '-his', '../prob_examples/public_examples/B/Simple/UpCounter_his.txt','-his_option', 'show_states','-his_option', 'show_init'], 'Check that history option -his_option show_states & show_init works.').
cli_testcase(454, [b_test,animate,history], ['../prob_examples/public_examples/B/Simple/UpCounter.mch', '-init','-animate', '10', '-strict', '-p', 'MAXINT', '100', '-his', '../prob_examples/public_examples/B/Simple/UpCounter_his.txt','-his_option', 'show_init'], 'Check that history option -his_option show_init works.').
cli_testcase(455, [b_test,animate,sptxt], ['../prob_examples/public_examples/B/Simple/UpCounter.mch', '-init','-animate', '10', '-strict', '-p', 'MAXINT', '100', '-sptxt', '../prob_examples/public_examples/B/Simple/UpCounter_sptxt.txt'], 'Check that history option -sptxt works.').

cli_testcase(456, [puzzles_test,queens], ['../prob_examples/public_examples/B/Puzzles/Queens/QueensAllSolutions.mch', '-t', '-strict', '-check_complete', '-mc', 100], 'A version of N-Queens where all solutions are computed and we try to find N.').

cli_testcase(457, [laws,rel_fnc], ['../prob_examples/public_examples/B/Laws/FunLaws.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE'], 'Use model checking to check a variety of laws on functions.').
cli_testcase(458, [laws,rel_fnc], ['../prob_examples/public_examples/B/Laws/FunLawsWithLambda.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE'], 'Use model checking to check a variety of laws on functions, including lambda abstractions.').
cli_testcase(459, [b_test,history], ['../prob_examples/public_examples/B/Daniel/rel_fnc.mch', '-t',  '-strict', '-p', 'CLPFD', 'TRUE','-his', '../prob_examples/public_examples/B/xxdoesnotexist/rel_fnc_his.txt','-expcterr','open_file','-expcterr','history'], 'Check that we generate an error if hist file destination does not exist.').
cli_testcase(460, [b_test,sptxt], ['../prob_examples/public_examples/B/Daniel/rel_fnc.mch', '-t',  '-strict', '-p', 'CLPFD', 'TRUE','-sptxt', '../prob_examples/public_examples/B/xxdoesnotexist/rel_fnc_his.txt','-expcterr','open_file','-expcterr','sptxt'], 'Check that we generate an error if sptxt file destination does not exist.').
cli_testcase(461, [b_test,history], ['../prob_examples/public_examples/B/Daniel/rel_fnc.mch', '-t',  '-strict', '-p', 'CLPFD', 'TRUE','-his', '../prob_examples/public_examples/B/Daniel/rel_fnc_his.txt','-his_option', 'show_states','-his_option', 'show_init', '-p','SHOW_EVENTB_ANY_VALUES','TRUE'], 'A small machine to check that rel/fnc work as expected; here we check it in conjunction with -his option.').
cli_testcase(462, [b_test,history], ['../prob_examples/public_examples/B/Simple/UpCounter.mch', '-mc', '100',  '-goal', 'c=10','-strict', '-p', 'MAXINT', '100', '-his', '../prob_examples/public_examples/B/Simple/UpCounter_his.txt','-his_option', 'show_states','-his_option', 'show_init','-expcterr','goal_found'], 'Check that history option works together with model checking.').
cli_testcase(463, [cbc_deadlock,history], ['../prob_examples/public_examples/B/CBC/SimpleDeadlock1.mch', '-cbc_deadlock', '-strict', '-expcterr', 'cbc_deadlock_check','-his', '../prob_examples/public_examples/B/CBC/SimpleDeadlock1_his.txt','-his_option', 'show_states'], 'Check that -his works together with cbc_deadlock').
cli_testcase(464, [cbc,history], ['../prob_examples/public_examples/B/CBC/ProcessIDStarvation_InvErr.mch', '-cbc', 'Delete', '-strict', '-expcterr', 'cbc', '-his', '../prob_examples/public_examples/B/CBC/ProcessIDStarvation_InvErr_his.txt','-his_option', 'show_states', '-p', 'DEFAULT_SETSIZE', 6, '-p', 'TIME_OUT', 500], 'Check that -his works with cbc invariant checking.'). % set DEFAULT_SETSIZE on 28th Aug 2014 to check that partition in Invariant detected
cli_testcase(465, [b_test,animate], ['../prob_examples/public_examples/B/Simple/UpCounter.mch', '-animate', '5', '-strict', '-p', 'MAXINT', '100', '-load_state', '../prob_examples/public_examples/B/Simple/UpCounter_saved10.P', '-goal', 'c=15',  '-check_goal'], 'Check that animate works after load_state.').
cli_testcase(466, [b_test,animate,while], ['../prob_examples/public_examples/B/PerformanceTests/WhileLoopBench.mch', '-animate', '22', '-strict', '-p', 'TIME_OUT', 28000, '-his', '../prob_examples/public_examples/B/PerformanceTests/WhileLoopBench_his.txt', '-his_option' , 'show_states'], 'Check animate with operation with while loop.').
cli_testcase(467, [b_test,history], ['../prob_examples/public_examples/B/FeatureChecks/IFTHENELSE.mch', '-t', '-strict','-his', '../prob_examples/public_examples/B/FeatureChecks/IFTHENELSE_his.txt','-his_option', 'show_states'], 'Check simple if-then-else works correctly, and check -his option writes operation arguments into file.').
cli_testcase(468, [b_test,history,sptxt], ['../prob_examples/public_examples/B/Puzzles/Farmer.mch', '-mc', '1000', '-bf','-strict', '-nodead', '-expcterr', 'goal_found','-his', '../prob_examples/public_examples/B/Puzzles/Farmer_his.txt','-sptxt','../prob_examples/public_examples/B/Puzzles/Farmer_sptxt.txt'], 'Check that -his and -sptxt work when used together').
cli_testcase(469, [b_test,sptxt], ['../prob_examples/public_examples/B/PerformanceTests/LargeSets/LargeExplicitSet.mch', '-animate', '7', '-strict', '-p', 'TIME_OUT', 5000,
'-nodead','-sptxt','../prob_examples/public_examples/B/PerformanceTests/LargeSets/LargeExplicitSet_sptxt.txt'], 'Check that -sptxt expands large sets.'). % introduced a time-out on 18th of May 2014 to avoid time-out during closure expansion; cause commit on May 12 to b_enumerate.pl; reduced TIME_OUT again on 19th of May [completely removing time-out seems to still cause issues]; another slowdown was caused by commit 13d3b41ff8e1f15cea7437b3530c883c4e3bda6b on Sep 20th 2013 (improve union of avl_set with large intervals)
cli_testcase(470, [b_test,sptxt,history], ['../prob_examples/public_examples/EventBPrologPackages/Abrial_Train_Ch17/train_4_mch_lukas.eventb', '-t', '-strict', '-noass','-his', '../prob_examples/public_examples/EventBPrologPackages/Abrial_Train_Ch17/train_4_mch_lukas_his.txt','-his_option','show_states','-sptxt','../prob_examples/public_examples/EventBPrologPackages/Abrial_Train_Ch17/train_4_mch_lukas_sptxt.txt'], 'Check -his and sptxt on a larger Event-B model.').

cli_testcase(471, [b_test,sptxt,history], ['../prob_examples/public_examples/B/Benchmarks/CarlaTravelAgencyErr.mch', '-t', '-strict', '-his', '../prob_examples/public_examples/B/Benchmarks/CarlaTravelAgencyErr_his.txt','-sptxt','../prob_examples/public_examples/B/Benchmarks/CarlaTravelAgencyErr_sptxt.txt'], 'Check -his and -sptxt on a larger classical B model.').
cli_testcase(472, [b_test,sptxt,history], ['../prob_examples/public_examples/B/SchneiderBook/Chapter12/TownsRRR.ref', '-t', '-strict', '-his', '../prob_examples/public_examples/B/SchneiderBook/Chapter12/TownsRRR_his.txt','-his_option','show_states','-sptxt','../prob_examples/public_examples/B/SchneiderBook/Chapter12/TownsRRR_sptxt.txt'], 'Check -his and -sptxt on a classical B refinement.').
cli_testcase(473, [b_test,sptxt,history,records], ['../prob_examples/public_examples/B/NewSyntax/SET_Game_Rec.mch', '-t', '-strict', '-his', '../prob_examples/public_examples/B/NewSyntax/SET_Game_Rec_his.txt','-his_option','show_states','-sptxt','../prob_examples/public_examples/B/NewSyntax/SET_Game_Rec_sptxt.txt'], 'Check -his and -sptxt for classical B records.').
cli_testcase(474, [b_test,sptxt,history,dot], ['../prob_examples/public_examples/B/ExpressionViewer/TotFunAnalyseTest.mch', '-ma', 1, 4, 0, '-dot_output', '../prob_examples/public_examples/B/ExpressionViewer/dot_output/', '-dot_all'], 'Check -dot_output, -dot_all works.').
cli_testcase(475, [laws,repl], ['../prob_examples/public_examples/B/Laws/REPL_UNIT_TESTS2.mch', '-t', '-mc', '100000', '-c', '-strict', '-p', 'CLPFD', 'TRUE', '-assertions', '-p', 'TIME_OUT', '60000'], 'Check that REPL unit tests work').
cli_testcase(476, [tickets,dot], ['../prob_examples/public_examples/B/Tickets/Mejia1/TestInit_mch.eventb', '-t', '-mc', '1000', '-cc', 4, 13, '-strict', '-p', 'CLPFD', 'TRUE', '-dot', 'state_space', '../prob_examples/public_examples/B/Tickets/Mejia1/TestInit_statespace.dot', '-p', 'DOT_PROP', 'TRUE', '-pref_group', dot_colors, classic, '-p','DOT_PEN_WIDTH',1], 'Check that -dot state_space output works and that no warning generated for existential quantifier on disjunction in Event-B mode').
cli_testcase(477, [cbc,smt_test], ['-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-strict', '../prob_examples/public_examples/B/Puzzles/WhoKilledAgatha.mch', '-cbc_assertions'], 'Check CBC static assertion checking.').
cli_testcase(478, [cbc,smt_test], ['-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '../prob_examples/public_examples/B/Puzzles/WhoKilledAgatha_wrongassert.mch','-strict',  '-cbc_assertions','-expcterr',cbc_assertions], 'Check CBC static assertion checking with counter example.').
cli_testcase(479, [b_test,dot], ['-p', 'CLPFD', 'TRUE', '-mc', '100', '-p', 'MAX_INITIALISATIONS', 7, '-bf', '../prob_examples/public_examples/B/Puzzles/WhoKilledAgatha.mch','-dot', 'state_space',  '../prob_examples/public_examples/B/Puzzles/WhoKilledAgatha_statespace.dot','-strict', '-p', 'DOT_COLOR_ARC', steelblue, '-p', 'DOT_COLOR_NODE', olivedrab2, '-p', 'DOT_PEN_WIDTH', 2], 'Check Dot output with proper newline.').
cli_testcase(480, [b_test,dot], ['-p', 'CLPFD', 'TRUE', '-p', 'DOT_LOOPS', 'FALSE',
'-p', 'DOT_ROOT', 'FALSE', '-p', 'MAX_INITIALISATIONS', 5,
'../prob_examples/public_examples/B/Benchmarks/Cruise_finite1.mch','-mc', 4 ,'-bf' ,'-dot', 'state_space',  '../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_statespace.dot','-strict', '-pref_group', dot_colors, classic, '-p','DOT_PEN_WIDTH',1], 'Check Dot output with proper newline.').
cli_testcase(481, [b_test,let], ['../prob_examples/public_examples/B/FeatureChecks/LET_multiple_variables.mch', '-t', '-strict', '-expcterr','type_error', '-expcterr', 'load_main_file'], 'Check illegal LET with multiple variables.'). % added '-expcterr','type_error' : we now detect the error in the type checker
cli_testcase(482, [b_test,let], ['../prob_examples/public_examples/B/FeatureChecks/LET_multiple_variables.mch', '-mc', 100, '-expcterr','type_error', '-expcterr', 'load_main_file'], 'Check that LET with non-equation predicate does not work.'). % error now detected during type check, not during model checking
cli_testcase(483, [csp_test,csp_bench], ['../prob_examples/public_examples/CSP/ref_benchmarks/abp_chapter5_roscoe_DLK_DIV_bench.csp', '-assertions', '-strict'], 'Check performance of Deadlock & DIV checks').
cli_testcase(484, [csp_test,csp_bench], ['../prob_examples/public_examples/CSP/NastyFDRDemos/LargeSpecRefineCheck.csp', '-assertions', '-strict'], 'Check performance with Large Spec and small Impl').
cli_testcase(485, [csp_test,csp_bench], ['../prob_examples/public_examples/CSP/simple/DeadlockTimeout.csp', '-assertions', '-strict'], 'Check that deadlock checker works with timeout generated tau transitions').
cli_testcase(486, [csp_test,csp_bench], ['../prob_examples/public_examples/CSP/simple/ParallelIntChoice.csp', '-assertions', '-strict'], 'Check that deadlock checker works with internal choice generated tau transitions').
cli_testcase(487, [csp_test], ['../prob_examples/public_examples/CSP/simple/DivChecks3.csp', '-assertions', '-strict'], 'Check that ProB deals with Divergence correctly').
cli_testcase(488, [csp_test,csp_bench], ['../prob_examples/public_examples/CSP/simple/ParallelIntChoice.csp', '-assertions', '-strict', '-p',  'CSP_STRIP_SOURCE_LOC', 'TRUE'], 'Check that stripping of source location works.').
cli_testcase(489, [csp_test,csp_bench], ['../prob_examples/public_examples/CSP/ref_benchmarks/section7-1.csp', '-assertions', '-strict'], 'Check more CSP assertions from Roscoe book usc').
cli_testcase(490, [csp_test], ['../prob_examples/public_examples/CSP/simple/exception_operator.csp', '-assertions', '-strict'], 'Check that new exception operator works').
cli_testcase(491, [csp_test,csp_bench], ['../prob_examples/public_examples/CSP/ref_benchmarks/section7-2.csp', '-assertions', '-strict'], 'Check more CSP assertions from Roscoe book usc, here with exception operator').
cli_testcase(492, [csp_test],['../prob_examples/public_examples/CSP/evans/traffic_spec.csp', '-assertions', '-strict'], 'Checking CSP assertions.').
cli_testcase(493, [tickets], ['../prob_examples/public_examples/B/Tickets/Mejia2/ticket_simplified.mch', '-t', '-mc',100,'-assertions', '-strict'], 'Check cartesian product ticket').
cli_testcase(494, [laws,case], ['../prob_examples/public_examples/B/Laws/SubstitutionLaws.mch', '-mc', '100', '-c', '-t', '-nodead', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE'], 'Check basic properties of B substitutions.').

cli_testcase(495, [error_checks], ['../prob_examples/public_examples/B/ErrorMachines/SubstitutionSeqCompAssertErrCheck.mch', '-mc', '100', '-nodead', '-strict','-expcterr', 'assert_error'], 'Check that ProB finds error in Substitution assertion').
cli_testcase(496, [laws], ['../prob_examples/public_examples/B/Laws/ExplicitComputationsCLPFD.mch', '-assertions', '-mc', '100000', '-nodead', '-c', '-strict', '-p', 'TIME_OUT', '7000'], 'Many explicit computations which check that all operators work as expected on various concrete data.').
cli_testcase(497, [b_test], ['../prob_examples/public_examples/B/Benchmarks/Cruise_finite1.mch', '-pp', '../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_pp.mch', '-strict'], 'Check that pretty printing works').
cli_testcase(498, [b_test], ['../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_prettyprinted.mch', '-pp', '../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_pp.mch', '-strict'], 'Check that pretty printing works').
cli_testcase(499, [b_test], ['../prob_examples/public_examples/B/Benchmarks/Cruise_finite1.mch', '-ppf', '../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_pp.mch', '-strict'], 'Check that pretty printing works').
cli_testcase(500, [b_test], ['../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_prettyprintedf.mch', '-ppf', '../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_pp.mch', '-strict'], 'Check that pretty printing works').
cli_testcase(501, [b_test], ['../prob_examples/public_examples/B/Benchmarks/Cruise_finite1.mch', '-mc', '1400', '-cc', '1361', '25696', '-p', 'COMPRESSION', 'TRUE', '-strict', '-p', 'MAX_INITIALISATIONS', 5], 'Check that compression works').
cli_testcase(502, [laws], ['../prob_examples/public_examples/B/Abrial_BBook/BBook_Page80.mch', '-init', '-assertions', '-strict'], 'Check example from B-Book').
cli_testcase(503, [laws], ['../prob_examples/public_examples/B/Abrial_BBook/BBook_Page83.mch', '-init', '-assertions', '-strict'], 'Check example from B-Book').
cli_testcase(504, [laws], ['../prob_examples/public_examples/B/Abrial_BBook/MembershipProperty4_Page96.mch', '-init',  '-expcterr', 'setup_constants_fails', '-p', 'CLPFD', 'TRUE', '-strict'], 'Check rule from B-Book').
cli_testcase(505, [laws], ['../prob_examples/public_examples/B/Abrial_BBook/MembershipProperty5_Page96.mch', '-init', '-p', 'CLPFD', 'TRUE', '-strict'], 'Check that ProB finds counter example to erroneous rule from B-Book').
cli_testcase(506, [laws], ['../prob_examples/public_examples/B/Abrial_BBook/MembershipProperty6_Page96.mch', '-init', '-p', 'CLPFD', 'TRUE', '-strict'], 'Check that ProB finds counter example to erroneous rule from B-Book').
cli_testcase(507, [laws], ['../prob_examples/public_examples/B/Laws/BoolLaws_SetCompr.mch', '-mc', '1000', '-nodead', '-c', '-strict'], 'Various laws about boolean connectives in set comprehension form.').
cli_testcase(508, [laws], ['../prob_examples/public_examples/B/Laws/BoolLaws_SetComprCLPFD.mch', '-mc', '1000', '-nodead', '-c', '-strict'], 'Various laws about boolean connectives in set comprehension form.').
cli_testcase(509, [laws,strings], ['../prob_examples/public_examples/B/Laws/StringLaws.mch', '-t', '-mc', '500',  '-nogoal', '-nodead', '-strict'], 'Some Laws about the STRING datatype').
cli_testcase(510, [laws], ['../prob_examples/public_examples/B/Laws/RelLaws.mch', '-init','-ppf', '../prob_examples/public_examples/B/Laws/RelLaws_internal.mch', '-strict'], 'Check that Pretty Printing works on a file with many operators').
cli_testcase(511, [csp_test,csp_det], ['../prob_examples/public_examples/CSP/other/Ivo/Deterministic1.csp', '-assertions', '-strict'], 'Check more CSP deterministic assertions').
cli_testcase(512, [csp_test,csp_det], ['../prob_examples/public_examples/CSP/other/Ivo/Deterministic2_v2.csp', '-assertions', '-strict'], 'Check more CSP deterministic assertions').
cli_testcase(513, [csp_test,csp_det], ['../prob_examples/public_examples/CSP/other/Ivo/WaterBoiler.csp', '-assertions', '-strict'], 'Check CSP assertions from WaterBoiler.csp example.').
cli_testcase(514, [tickets,dot], ['../prob_examples/public_examples/B/Mathematical/Voting/NoCondorcetWinnerKodkod.mch', '-t', '-strict', '-assertions','-dot_output', '../prob_examples/public_examples/B/Mathematical/Voting/dot_output/', '-dot_all'], 'Check that we have no problem with the existential quantifier; check dot output for assertion with perm.').
cli_testcase(515, [b_test,dot], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/Demo/scheduler.mch', '-mc', '100', '-cc', '36', '156', '-p', 'DEFAULT_SETSIZE', '3', '-df', '-strict','-dot', 'signature_merge','../prob_examples/public_examples/B/Demo/scheduler_spmdot.dot','-dot', dfa_merge, '../prob_examples/public_examples/B/Demo/scheduler_spddot.dot', '-dot','state_space', '../prob_examples/public_examples/B/Demo/scheduler_spdot.dot', '-pref_group', dot_colors, classic, '-p','DOT_PEN_WIDTH',1], 'Test reduced state space dot output.').
cli_testcase(516, [symmtest], ['../prob_examples/public_examples/B/SymmetryReduction/Contrived/PartiallyEnumeratedSet.mch', '-mc', '100', '-cc', '3', '16', '-p', 'SYMMETRY_MODE', 'hash', '-strict'], 'Check that we detect unused enumerated set elements as symmetrical').
cli_testcase(517, [symmtest], ['../prob_examples/public_examples/B/SymmetryReduction/Contrived/PartiallyEnumeratedSet.mch', '-mc', '100', '-cc', '3', '16', '-p', 'SYMMETRY_MODE', 'nauty', '-strict'], 'Check that we detect unused enumerated set elements as symmetrical').
cli_testcase(518, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/Demo/monitor2.mch', '-mc', '100', '-cc', '15', '92', '-p', 'SYMMETRY_MODE', 'hash', '-strict'], 'Check that we detect unused enumerated set elements as symmetrical').
cli_testcase(519, [symmtest], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/Demo/monitor2.mch', '-mc', '100', '-cc', '81', '529', '-p', 'SYMMETRY_MODE', 'off', '-strict'], 'Check original without symmetry').
cli_testcase(520, [tickets,csp_test], ['../prob_examples/public_examples/CSP/Tickets/Shi/hanoi7.csp', '-mc', '100', '-t', '-strict'], 'Check that exception does not occur (Ticket)').
cli_testcase(521, [tickets,csp_test], ['../prob_examples/public_examples/CSP/Tickets/Shi/hanoi4.csp', '-assertions', '-strict'], 'Check that exception does not occur (Ticket)').
cli_testcase(522, [tickets,csp_test], ['../prob_examples/public_examples/CSP/Tickets/ErrManagerLoop/PrintRoom.csp', '-mc', 100, '-strict','-expcterr', 'general_error_occurred', '-expcterr', check_channel_output_value], 'Check that channel output error detected and no infinite loop in printing error').
cli_testcase(523, [b_test,infinite], ['../prob_examples/public_examples/B/PerformanceTests/InfiniteClosures4.mch', '-strict','-assertions'], 'Check that infininite function represented by union is detected and can be applied; also test setting MAXINT using 2**31-1').
cli_testcase(524, [tickets], ['../prob_examples/public_examples/B/Tickets/Beauger1/PredProblem.mch', '-t', '-strict'], 'Check that we can use relational composition ; with pred.').
cli_testcase(525, [b_test,infinite,finite], ['../prob_examples/public_examples/B/PerformanceTests/InfiniteParityFunction.mch', '-t', '-strict', '-assertions'], 'Some more tests with infinite functions.').
cli_testcase(526, [b_test,infinite], ['../prob_examples/public_examples/EventBPrologPackages/Tests/FiniteInfiniteFunctions_ctx.eventb', '-assertions', '-strict'], 'Some more tests with infinite functions and the finite check in Event-B.').
cli_testcase(527, [b_test,infinite], ['../prob_examples/public_examples/B/PerformanceTests/InfinitePartialSqrtFunction.mch', '-t', '-strict'], 'Check that we can deal with a partial infinite function.').
cli_testcase(528, [b_test,external], ['../prob_examples/public_examples/B/ExternalFunctions/UsesLibraryMath.mch', '-t', '-strict', '-assertions'], 'Check that external functions work.').
cli_testcase(529, [b_test,external], ['../prob_examples/public_examples/B/ExternalFunctions/PrologFun.mch', '-assertions', '-strict'], 'Check that external functions PROLOG_FUN work.').
cli_testcase(530, [b_test,external], ['../prob_examples/public_examples/B/ExternalFunctions/PrintExternal.mch', '-t', '-strict'], 'Check that external printing functions work.').
cli_testcase(531, [b_test,external,strings], ['../prob_examples/public_examples/B/ExternalFunctions/LibraryStrings.mch', '-assertions', '-strict'], 'Check that external String functions work.').
cli_testcase(532, [symmtest], ['../prob_examples/public_examples/B/SymmetryReduction/mondex_m1_c1_2.eventb', '-mc', 1000, '-p','SYMMETRY_MODE', hash, '-p','MAXINT',20, '-p','CLPFD', 'TRUE','-cc', 269, 4079, '-strict', '-expcterr', 'model_check_incomplete'], 'Check that symmetry reduction and detection of unused constants also works if we have too many disequality axioms.').
cli_testcase(533, [b_test,infinite], ['../prob_examples/public_examples/B/PerformanceTests/InfiniteLanguageConcatenation.mch', '-mc', 1000, '-nodead','-assertions', '-strict'], 'Check that this function is detected as infinite and not expanded.').
cli_testcase(534, [tickets], ['../prob_examples/public_examples/B/Tickets/Hansen3/TestingCircDef2.mch', '-mc', 1000, '-nodead','-assertions', '-strict'], 'Check that the erroneous circular definition problem is gone.').
cli_testcase(535, [parser,tickets], ['../prob_examples/public_examples/B/Benchmarks/scheduler.mch', '-init', '-strict', '-evalt', '1+1=3', '-expcterr', 'eval_string'], 'Test that parser for eval strings works and that erroneous result caught.').
cli_testcase(536, [parser,tickets,infinite], ['../prob_examples/public_examples/B/PerformanceTests/InfiniteParityFunction.mch', '-init', '-strict', '-evalt', 'not(finite({x|x>2}))','-evalt', 'not(finite(NATURAL))', '-evalf', 'finite({x|x>2})', '-evalt', 'finite(NAT1)', '-evalt', 'NATURAL /: FIN1(INTEGER)'], 'Test that parser for eval strings works.').
cli_testcase(537, [tickets,strings], ['../prob_examples/public_examples/B/Benchmarks/scheduler.mch', '-init', '-strict', '-p','MAXINT',10,
'-evalt', 'INTEGER/<:{1,2,3,7,9,10}',
'-evalt', 'NATURAL/<:{1,2,3,7,9,10}',
'-evalt', 'NATURAL1/<:{1,2,3,7,9,10}',
'-evalt', 'NATURAL1/<:{0,1,2,3,4}',
'-evalt', 'NATURAL1/<<:{1,2,3,7,9,10}',
'-evalt', 'NATURAL1/<<:1..100',
'-evalt', 'NATURAL1/<<:{x|x<100}',
'-evalt', 'NATURAL1/<<:{x|x>0}',
'-evalt', 'NATURAL1/<<:{x|x>=2}',
'-evalf', 'NATURAL1/<<:{x|x>=0}',
'-evalt', 'NATURAL/<<:{x|x>=0}',
'-evalt', 'NATURAL1/<<:0..MAXINT',
'-evalt', 'NATURAL1/<:0..MAXINT',
'-evalt', 'NATURAL/<:0..MAXINT',
'-evalt', 'INTEGER/<:MININT..MAXINT',
'-evalt', 'STRING/<<:{"a"}',
'-evalt', 'STRING/<:{"a"}',
'-evalf', 'STRING/<:STRING',
'-evalt', 'STRING/<<:STRING',
'-evalt', 'BOOL /<<: {TRUE,FALSE}',
'-evalf', 'BOOL /<: {TRUE,FALSE}',
'-evalt', 'BOOL /<: {TRUE}'
%, '-expcterr', 'eval_string_enum_warning'
], 'Test subset works without enum warnings.'). % used to be: 'Test that eval_string catches expansion errors.').
cli_testcase(538, [tickets], ['../prob_examples/public_examples/B/Benchmarks/scheduler.mch', '-init', '-strict', '-p','MAXINT',10, '-evalt', '1..10 <: NATURAL1'], 'Test that these predicates do not cause enumeration warnings.').
cli_testcase(539, [b_test,kodkod], ['../prob_examples/public_examples/B/Kodkod/CrewAllocationConstantsLarge.mch', '-t', '-strict', '-p','TIME_OUT',3000, '-p', 'KODKOD', 'TRUE', '-properties', '-assertions'], 'Check that Kodkod translator works and that solution is ok (via -assertions). Skipped on Windows because of Memory restrictions.').
cli_testcase(540, [cbc,cbc_tests], ['../prob_examples/public_examples/B/TestCaseGeneration/SimpleCounter.mch', '-cbc_tests', 5 ,'count=MAX-1', '../prob_examples/public_examples/B/TestCaseGeneration/SimpleCounter_results.xml', '-cbc_cover', 'Increase', '-cbc_cover',  'Reset', '-cbc_cover',  'Halfway', '-p', 'CLPFD', 'TRUE', '-strict'], 'Check that CBC test-case generation works.').
cli_testcase(541, [csp_test,ltl], ['../prob_examples/public_examples/CSP/Tickets/WilliamsLTL/SimpleLTL.csp', '-ltlformulat', 'G([a] => F(e(a)))', '-strict'],'Check LTL formula in File.').
cli_testcase(542, [csp_test,ltl], ['../prob_examples/public_examples/CSP/Tickets/WilliamsLTL/SimpleLTL.csp', '-ltlformulat', 'not (GF e(a) => GF [a]) => F [b]', '-strict'],'Check LTL formula in File.').
cli_testcase(543, [csp_test,ltl], ['../prob_examples/public_examples/CSP/other/Ivo/VendingMachine.csp', '-ltlformulat', 'G([coin] => F (e(coffee) & e(tea)))', '-strict'],'Check LTL formula in File.').
cli_testcase(544, [csp_test,ltl], ['../prob_examples/public_examples/CSP/other/Ivo/VendingMachine.csp', '-ltlformulat', 'GF(e(coin))', '-strict'],'Check LTL formula in File.').
cli_testcase(545, [csp_test,ltl], ['../prob_examples/public_examples/CSP/other/Ivo/dot_tuples_ltl_simple.csp', '-ltlformulat', 'G([d] => F(e(ch.a)))', '-strict'],'Check LTL formula in File.').
cli_testcase(546, [csp_test,ltl], ['../prob_examples/public_examples/CSP/other/Ivo/dot_tuples_ltl_simple.csp', '-ltlassertions', '-strict'],'Check LTL formulas in File.').
cli_testcase(547, [csp_test,ltl], ['../prob_examples/public_examples/CSP/other/Ivo/Simple_LTL_Tupels.csp', '-ltlassertions', '-strict'],'Check LTL formulas in File.').
cli_testcase(548, [csp_test,ltl], ['../prob_examples/public_examples/CSP/other/Ivo/Simple_LTL_Tupels.csp', '-ltlformulat', 'GF([tau] => not e(c.a.10.10))', '-strict'],'Check LTL formula in File.').
cli_testcase(549, [csp_test,ltl], ['../prob_examples/public_examples/CSP/other/Ivo/Simple_LTL_Tupels.csp', '-ltlformulat', '[e] U [c.a.1.10]', '-strict'],'Check LTL formula in File.').
cli_testcase(550, [b_test,recursion,external], ['../prob_examples/public_examples/B/ExternalFunctions/SortSet.mch', '-assertions', '-strict', '-p', 'CLPFD', 'TRUE'], 'Check that recursive sorting function works.').
cli_testcase(551, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/TestingCardinality.csp', '-assertions', '-strict', '-expcterr', virtual_time_out], '').
cli_testcase(552, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/DotTupleTestCase.csp', '-assertions', '-strict'], '').
cli_testcase(553, [b_test,recursion,external], ['../prob_examples/public_examples/B/RecursiveFunctions/RecursiveSeqLenFunction.mch', '-t', '-strict', '-p', 'SYMBOLIC', 'FALSE'], 'Test that recursive functions work in non-symbolic mode.').
cli_testcase(554, [b_test,recursion,external], ['../prob_examples/public_examples/B/RecursiveFunctions/Factorial.mch', '-t', '-strict', '-p', 'SYMBOLIC', 'FALSE'], 'Test that other recursive function works.').
cli_testcase(555, [b_test,recursion,external,choose,sigma], ['../prob_examples/public_examples/B/RecursiveFunctions/RecursiveSigmaCHOOSEv1.mch', '-t', '-strict', '-p', 'SYMBOLIC', 'FALSE', '-assertions'], 'Test that union of explicit set and recursive lambda function works.').
cli_testcase(556, [parser,tickets], ['../prob_examples/public_examples/B/Tickets/DefinitionCallSwallow/PerserDefBug.mch', '-strict', '-mc', '1'], 'Regression test for Definitions Bug in Parser').
cli_testcase(557, [b_test,recursion,external], ['../prob_examples/public_examples/B/RecursiveFunctions/Tests/NestedDef.mch', '-init', '-assertions','-strict', '-p', 'SYMBOLIC', 'FALSE'], 'Test that we do not accidentally treat nested definion of variable with same name as abstract constant.').
cli_testcase(558, [b_test,recursion,external], ['../prob_examples/public_examples/B/RecursiveFunctions/Tests/RecursiveIntegerEquation.mch', '-init', '-assertions','-strict', '-p', 'SYMBOLIC', 'FALSE', '-mc', 100, '-nodead'], 'Test that we do not accidentally treat equation over non-relation type.').
cli_testcase(559, [b_test,recursion,external], ['../prob_examples/public_examples/B/RecursiveFunctions/Tests/RecursiveEnumEquation.mch', '-init', '-assertions','-strict', '-p', 'SYMBOLIC', 'FALSE', '-mc', 100, '-nodead'], 'Test that we do not accidentally treat equation over non-relation type.').
cli_testcase(560, [b_test,csp_test], ['../prob_examples/public_examples/CSPB/Dining/Dining.mch', '-mc', '1000', '-strict', '-p', 'MAX_INITIALISATIONS', 15], 'Test that the dining philosophers without guidance by CSP have no deadlock').
cli_testcase(561, [b_test,csp_test,cspb], ['../prob_examples/public_examples/CSPB/Dining/Dining.mch', '-csp-guide', '../prob_examples/public_examples/CSPB/Dining/Dining.csp', '-mc', '1000', '-strict', '-expecterr', 'deadlock'], 'Test that the dining philosophers with guidance by CSP have a deadlock').
cli_testcase(562, [b_test,cbc_tests], ['../prob_examples/public_examples/EventBPrologPackages/Pitesti/BankAccount_mch.eventb', '-test_description', '../prob_examples/public_examples/EventBPrologPackages/Pitesti/test_description.xml', '-strict'], 'Simple constraint based test case generation').
cli_testcase(563, [b_test,mcm_tests], ['../prob_examples/public_examples/B/Daniel/TestM.mch', '-mcm_tests', '20', '1000', 'state = End', '../prob_examples/public_examples/B/Daniel/TestM_testcases.xml', '-strict'],'State based test generation').
cli_testcase(564, [b_test,csp_test,cspb], ['../prob_examples/public_examples/CSPB/Tickets/Treharne1/Test_corrected.mch', '-csp-guide', '../prob_examples/public_examples/CSPB/Tickets/Treharne1/Test-ctrl.csp', '-mc', '1000', '-strict', '-cc', 5, 6], 'Test that BOOL synchronises with CSP booleans').

cli_testcase(565, [b_test,proz],['../prob_examples/public_examples/Z/Testcases/comsets.fuzz', '-t', '-strict'],'ProZ test').
cli_testcase(566, [b_test,proz],['../prob_examples/public_examples/Z/Testcases/distributed_concatenation.fuzz', '-t', '-strict'],'ProZ test').
cli_testcase(567, [b_test,proz],['../prob_examples/public_examples/Z/Testcases/compaction.fuzz', '-t', '-strict'],'ProZ test').
cli_testcase(568, [b_test,proz],['../prob_examples/public_examples/Z/Testcases/zrm_refcard_ops.fuzz', '-t', '-strict'],'ProZ test').
cli_testcase(569, [b_test,ltl], ['../prob_examples/public_examples/B/SymmetryReduction/Fraikin/Library.mch','-p','IGNORE_HASH_COLLISIONS',
'TRUE','-p','FORGET_STATE_SPACE','FALSE','-p','SYMMETRY_MODE','hash', '-expcterr', 'model_check_incomplete','-mc','9999',
'-ltlformulat','G( e(Reserve(m1,b1)) => not({b1:dom(reservation) & m1 : ran(reservation(b1))}) )', '-strict'],'Check enabled e() with multiple parameters').
cli_testcase(571, [b_test, let], ['../prob_examples/public_examples/B/Daniel/CommonSubLetBug.mch','-mc', '100', '-nodead', '-strict'], 'Regression test regarding let predicate and common subexpression elimination').
cli_testcase(572, [b_test,ltl],['../prob_examples/public_examples/B/LTL/DerrickSmith/Abstract.mch', '-ltlassertions', '-strict'],'Check ltl assertions from DEFINITIONS').
cli_testcase(573, [b_test,ltl],['../prob_examples/public_examples/B/LTL/Mikrowelle.mch', '-ltlassertions', '-strict'],'Check ltl assertions from DEFINITIONS').
cli_testcase(574, [b_test,ltl],['../prob_examples/public_examples/CSPB/Dining/Dining.mch', '-ltlassertions', '-expcterr', ltl, '-strict'],'Check ltl assertions from DEFINITIONS in Dining.mch').
cli_testcase(575, [csp_test,ltl],['../prob_examples/public_examples/CSPB/Dining/Dining.csp', '-ltlassertions', '-strict'],'Check ltl assertions from Pragmas in Dining.csp').
cli_testcase(576, [b_test,ltl,cspb],['../prob_examples/public_examples/CSPB/Dining/Dining.mch', '-csp-guide', '../prob_examples/public_examples/CSPB/Dining/Dining.csp', '-ltlassertions', '-strict'],'Check ltl assertions from both Dining models in Csp and B.').
cli_testcase(577, [b_test,ltl],['../prob_examples/public_examples/B/LTL/DerrickSmith/Abstract.mch', '-ltlformulat', 'G([AOp2] => F(sink))', '-strict'],'Check ltl formula').
cli_testcase(578, [b_test,ltl],['../prob_examples/public_examples/B/LTL/DerrickSmith/Abstract.mch', '-ltlformulat', 'G([AOp2] => F(deadlock))', '-strict'],'Check ltl formula').
cli_testcase(579, [csp_test],['../prob_examples/public_examples/CSP/mydemos/ReplicatedSharing.csp', '-mc', '100000', '-t', '-strict'],'Model check and trace replicated sharing processes.').
cli_testcase(580, [csp_test],['../prob_examples/public_examples/CSP/mydemos/SimpleReplicatedLinkedParallel.csp', '-mc', '100000', '-t', '-strict'],'Model check and trace replicated linked parallel processes.').
cli_testcase(581, [csp_test],['../prob_examples/public_examples/CSP/mydemos/SeqTests.csp', '-mc', '100000', '-nodead', '-t', '-strict'], 'Testing CSP sequences.').
cli_testcase(582, [csp_test],['../prob_examples/public_examples/CSP/other/Ivo/Sequences.csp', '-mc', '100000', '-t', '-strict'], 'Testing CSP sequence structures.').
cli_testcase(583, [b_test,tla],['../prob_examples/public_examples/TLA/GraphIso/GraphIso.tla', '-t', '-strict', '-p','TIME_OUT',7000], 'Testing TLA2B translation and loading.'). % increased time-out for commit 884a6ece23363e6d120cc9fdb3f72da9f6348f15: TO DO Investigate
cli_testcase(584, [b_test,tla],['../prob_examples/public_examples/TLA/ErrorTests/GraphIso.tla', '-init', '-strict'
  %'-expcterr', tla2b_parser, '-expcterr', load_main_file % TLA2B no longer generates .mch file !
  ],'Check that TLA2B does not overwrite pre-existing machine file').
cli_testcase(585, [b_test],['-strict',
  '-eval_file', '../prob_examples/public_examples/Eval/SendMoreMoney.eval', '-p', 'CLPFD', 'TRUE'],'Check that we can evaluate expr/pred from file without loading another machine').
cli_testcase(586, [b_test,tla],['../prob_examples/public_examples/TLA/Login/Login_Deadlock.tla', '-init', '-strict',
  '-mc', 100, '-p', 'SYMMETRY_MODE', hash, '-expcterr', deadlock],'Check that config file found and model checking TLA works').
cli_testcase(587, [csp_test],['../prob_examples/public_examples/CSP/other/Ivo/SeqsAndSetsTest.csp', '-t', '-strict'], 'Testing the CSP-M builtin functions seq(-) and set(-).').
cli_testcase(588, [b_test],['../prob_examples/public_examples/B/Pragmas/SymbolicPragmas.mch', '-init',
 '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '-assertions', '-strict'], 'Test that symbolic annotation pragma works.').
cli_testcase(589,[csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/InputOutputChannelValues.csp', '-mc', '10000', '-t', '-strict'], 'Testing complicated CSP-M datatypes.').
cli_testcase(590,[csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/CurryAgents.csp', '-mc', '10000', '-t', '-strict'], 'Testing Curry Agents in CSP-M models.').
cli_testcase(591,[csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/replicates.csp', '-mc', '10000', '-t', '-strict'], 'Testing replicate processes.').
cli_testcase(592,[csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/replicates2.csp', '-mc', '10000', '-t', '-strict'], 'Testing replicate processes.').
cli_testcase(593, [b_test,operation_reuse], ['../prob_examples/public_examples/B/Benchmarks/Cruise_finite1.mch', '-mc', '1400', '-cc', '1361', '25696', '--hash64', 203058745203935766,
 '-strict',  '-p', 'OPERATION_REUSE', 'TRUE', '-p', 'MAX_INITIALISATIONS', 5], 'Test Operation Reuse').
cli_testcase(594, [b_test], ['../prob_examples/public_examples/B/Benchmarks/scheduler.mch', '-init', '-strict', '-evalt', '!x.(x:1..10=>x<11)', '-evalf', '!y.(y:1..10 => y<10)'], 'Test that universally quantified predicates work in eval_string.').
cli_testcase(595, [b_test],['-strict',
  '-eval_rule_file', '../prob_examples/public_examples/Eval/KnightsKnaves.eval', '-p', 'CLPFD', 'TRUE'],'Check that we can evaluate expr/pred from file with universal quantification').
cli_testcase(596, [b_test,laws],['../prob_examples/public_examples/B/Laws/TautologiesPL.mch', '-p', 'CLPFD', 'TRUE', '-mc',100, '-nodead', '-cc',17,16, '-strict', '-p', 'MAX_INITIALISATIONS', 15],'Check a list of tautologies').
cli_testcase(597,[csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/DotDatatypes.csp', '-mc', '10000', '-t', '-strict'], 'Testing complex datatypes.').
cli_testcase(598,[tickets], ['../prob_examples/public_examples/B/Tickets/DefMchSameName/DefMchSameName.mch', '-ma', 7, 0 , 0, '-strict'], 'Testing that -main_assertions work properly.').
cli_testcase(599,[tickets], ['../prob_examples/public_examples/B/Tickets/858/ubahn_mch4.eventb', '-init', '-expcterr', setup_constants_inconsistent , '-strict'], 'Testing that no call_residue.').
cli_testcase(600,[tickets], ['../prob_examples/public_examples/B/Tickets/1043/SymTest.mch', '-mc', 300, '-cc', 3, 31 , '-p', 'SYMMETRY_MODE', hash, '-strict'], 'Testing that symmetry detected for enumerated set elements which are not used.'). % used to be -cc 3 44 but now INITIALISATION arguments are derived from target state
cli_testcase(601,[tickets], ['../prob_examples/public_examples/B/Tickets/1043/SymTest.mch', '-mc', 300, '-cc', 3, 31 , '-p', 'SYMMETRY_MODE', nauty, '-strict'], 'Testing that symmetry detected for enumerated set elements which are not used.'). % used to be -cc 3 44 but now INITIALISATION arguments are derived from target state
cli_testcase(602,[tickets], ['../prob_examples/public_examples/B/Tickets/1043/SymTest.mch', '-mc', 300, '-cc', 197, 238 , '-p', 'SYMMETRY_MODE', flood, '-strict'], 'Testing that symmetry detected for enumerated set elements which are not used.').
cli_testcase(603,[b_test], ['../prob_examples/public_examples/B/Tester/DefinitionMultiInclude/Main.mch', '-init', '-aa', 3, 0,0,  '-strict'], 'Test same DEFINITION file include twice in different contexts.').
cli_testcase(604, [csp_test,ltl],['../prob_examples/public_examples/CSP/mydemos/demo/Peterson_v2.csp', '-ltlformulat', 'G(not [critical.2])', '-strict', '-expcterr' ,ltl ],'Check ltl formula is reported as false').
cli_testcase(605, [b_test], ['../prob_examples/public_examples/B/Benchmarks/scheduler.mch', '-init', '-strict', '-eval', '(A <=> (waiting={})) & A', '-p', 'BOOL_AS_PREDICATE', 'FALSE',  '-expcterr' ,parse_machine_predicate_error], 'Test that BOOL as predicate can be disabled.').
cli_testcase(606, [b_test], ['../prob_examples/public_examples/B/Benchmarks/scheduler.mch', '-init', '-strict', '-evalt', '(A <=> (waiting={})) & A', '-evalf', '(B <=> (waiting={})) & (B <=> not(waiting={}))', '-evalt', '!x.(x:1..10 => (A <=> x>z)) & z:0..11 & not(A)', '-p', 'BOOL_AS_PREDICATE', 'TRUE'], 'Test that BOOL as predicate works.').
cli_testcase(607, [b_test,sat],['-strict',
  '-eval_file', '../prob_examples/public_examples/Eval/SAT_uf20-02_pred.eval', '-p', 'CLPFD', 'TRUE', '-p', 'BOOL_AS_PREDICATE', 'TRUE'],'Test that BOOL as predicate works.').
cli_testcase(608,[tickets], ['../prob_examples/public_examples/B/Tickets/ConcatDisjUnion/SimpleGrammar.mch', '-init', '-mc', 100 , '-nodead', '-cc', 4, 3, '-strict'], 'Testing that concat does not generate wrong solutions.').
cli_testcase(609,[tickets,b_test], ['../prob_examples/public_examples/B/Compilers/SyracuseGrammar.mch', '-init', '-mc', 100 , '-expcterr', deadlock, '-strict', '-p', 'CLPFD', 'FALSE'], 'Testing that concat works efficiently enough.').
cli_testcase(610,[tickets,b_test], ['../prob_examples/public_examples/B/Compilers/SyracuseGrammar.mch', '-init', '-mc', 100 , '-expcterr', deadlock, '-strict', '-p', 'CLPFD', 'TRUE'], 'Testing that concat works efficiently enough.').
cli_testcase(611,[b_test,infinite], ['../prob_examples/public_examples/B/PerformanceTests/InfiniteRelationImage.mch', '-init', '-assertions', '-properties', '-strict', '-p', 'CLPFD', 'TRUE'], 'Testing that image for infinite relation works.').
cli_testcase(612, [b_test,recursion,external,quantified], ['../prob_examples/public_examples/B/Compilers/NFA.mch', '-t', '-mc', 100, '-cc', 6, 12, '-assertions','-strict'], 'Test recursive function with recursive call inside UNION.').
cli_testcase(613, [b_test],['../prob_examples/public_examples/B/Pragmas/AssertionPragmas.mch', '-init', '-pp', '../prob_examples/public_examples/B/Pragmas/AssertionPragmas_internal.mch', '-strict'], 'Test that labels recognised.').
cli_testcase(614, [b_test,recursion,infinite,quantified],['../prob_examples/public_examples/B/Compilers/NFA_det.ref', '-init', '-refchk', '../prob_examples/public_examples/B/Compilers/NFA_nondet_refine_spec.P', '-cc', 10, 22, '-strict'], 'Test trace refinement checking and ABSTRACT_CONSTANT reuse in NFA_det.').
cli_testcase(615,[b_test,ltl], ['../prob_examples/public_examples/B/LTL/Tests/LTL_1State.mch', '../prob_examples/public_examples/B/LTL/Tests/LTL_2States.mch', '-init', '-ltlassertions', '-strict'], 'Testing LTL formulas on simple statespaces.').
cli_testcase(616, [b_test],['../prob_examples/public_examples/B/Tester/Partition.mch', '-init',  '-assertions','-pp','../prob_examples/public_examples/B/Tester/Partition_internal.mch', '-strict'], 'Test that partition properly solved and printed.').
cli_testcase(617, [parser,b_test,tickets],['../prob_examples/public_examples/B/Tickets/RefLoadLoop/NFA_det.ref', '-expcterr', 'load_main_file', '-expcterr', invalid_machine_reference, '-expcterr', b_load_machine, '-strict'], 'Test that parser does not loop.').
cli_testcase(618,[b_test,ltl], ['../prob_examples/public_examples/B/LTL/Tests/LTL_4States.mch', '-init', '-ltlassertions', '-strict'], 'Testing LTL formulas on simple statespaces.').
cli_testcase(619,[b_test,ctl], ['../prob_examples/public_examples/B/LTL/Tests/CTL_SetMove_2States.mch', '-init',
 '-ctlformulat', 'AG(EX{xx=bb})', '-ctlformula', 'AGEFEX[Move]true', '-ctlformula', 'AGEFEX[Set(aa)]{xx=aa}',
 '-ctlformulat', 'not(EF EX[Set(aa)]{xx=bb})',
 '-ctlformulat', 'EF EG {xx=bb}',
 '-ctlformulat', '{xx=bb} => EG {xx=bb}',
 '-ctlformulat', '{xx=bb} => not AF {xx=aa}',
 '-ctlformulat', '{xx=bb} => AF {not(xx=aa)}',
 '-strict'], 'Testing CTL formulas on simple statespaces.').
cli_testcase(620,[b_test,ctl], ['../prob_examples/public_examples/B/LTL/Tests/CTL_SetMove_2States.mch', '-init', '-ctlformulat', '(EF EX[Set(aa)]{xx=bb})',  '-expcterr', ctl,'-strict'], 'Testing CTL formulas on simple statespaces.').
cli_testcase(621,[b_test,ctl], ['../prob_examples/public_examples/B/Benchmarks/scheduler.mch', '-init',
 '-ctlformulat', 'AG ( EX[swap]true => {active /= {}})',
 '-ctlformulat', 'AG ( (EX[new]true)  or (EX[del]true) or (EX[swap]true) )',
 '-ctlformulat', 'AF {card(waiting)>0}',
 '-ctlformulat', 'AG( {card(active)>0} => EG e(swap))',
  '-strict'], 'Testing CTL formulas on simple statespaces.').
cli_testcase(622,[b_test,ctl], ['../prob_examples/public_examples/B/Benchmarks/scheduler.mch', '-init','-ctlformulat', 'AG ( (EX[new]true)  or (EX[del]true) )', '-expcterr', ctl,'-strict'], 'Testing CTL formulas on simple statespaces.').
cli_testcase(623,[tickets,infinite], ['../prob_examples/public_examples/B/PerformanceTests/Closure1_Computations.mch', '-init', '-assertions','-strict'], 'Check that closure1 works well for large or infinite sets.').
cli_testcase(624,[csp_test], ['../prob_examples/public_examples/CSP/Tickets/3675/AIDCPRB_small.csp', '-mc', '10000', '-t', '-strict'], 'Testing complex datatypes.').
cli_testcase(625, [b_test,tla,choose],['../prob_examples/public_examples/TLA/Tests/ChooseTest.tla', '-mc', 100, '-strict'], 'Testing TLA2B translation and loading.').
cli_testcase(626, [b_test],['../prob_examples/public_examples/B/PerformanceTests/TestFD_Domain_Narrowing.mch', '-t', '-p', 'CLPFD', 'TRUE', '-strict'], 'Test CLPFD on enumerated sets.').
cli_testcase(627, [cbc,smt_test], ['../prob_examples/public_examples/B/Benchmarks/scheduler.mch', '-cbc_sequence', 'new;new;del;ready', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE'], 'Check cbc_sequence').
cli_testcase(628, [cbc,smt_test], ['../prob_examples/public_examples/B/Benchmarks/scheduler.mch', '-cbc_sequence', 'new;new;del;del;del', '-strict', '-p', 'SMT', 'TRUE', '-p', 'CLPFD', 'TRUE', '-expcterr', cbc_sequence_no_solution_found], 'Check cbc_sequence').
cli_testcase(629, [tickets,wd], ['../prob_examples/public_examples/B/Tickets/Burdy1/TestOrApplyFunErr.mch', '-mc', '1000', '-expcterr', well_definedness_error, '-strict'], 'Check issue with undefined function application in context of disjunct.').
cli_testcase(630, [tickets], ['../prob_examples/public_examples/B/Tickets/PROBCORE-45/NotIntervalSol.mch', '-init', '-p', 'MAX_INITIALISATIONS', 30, '-mc', '1000', '-expcterr', 'model_check_incomplete', '-nodead', '-assertions', '-cbc_assertions', '-strict'], 'Check problem with not subset for intervals.').
cli_testcase(631, [tickets], ['../prob_examples/public_examples/B/Tickets/PROBCORE-45/NotIntervalSol.mch', '-init', '-p', 'MAX_INITIALISATIONS', 30, '-p', 'CLPFD', 'TRUE', '-mc', '1000', '-expcterr', 'model_check_incomplete', '-nodead', '-assertions', '-cbc_assertions', '-strict'], 'Check problem with not subset for intervals.').
cli_testcase(632, [cbc,smt_test], ['-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-strict', '-p', 'DEFAULT_SETSIZE', 2, '../prob_examples/public_examples/B/CBC/data_structure1_ctx_theorem_error.eventb', '-cbc_assertions', '-expcterr',cbc_assertions, '-p', 'PROOF_INFO', 'TRUE'], 'Check CBC static assertion checking.').
cli_testcase(633, [tickets], ['../prob_examples/public_examples/EventBPrologPackages/ProofDirected/TestR_label_reuse.eventb', '-mc', 1000, '-noass', '-strict', '-expcterr', 'invariant_violation', '-p', 'PROOF_INFO', 'TRUE'], 'Test that PROOF_INFO not confused by two invariants with same name (inv3).').
cli_testcase(634, [b_test,kodkod,dot], ['../prob_examples/public_examples/B/Kodkod/Loop.mch', '-strict', '-mc', 20, '-nodead', '-bf', '-p', 'MAX_INITIALISATIONS', 8, '-p', 'CLPFD', 'TRUE', '-cc', 15, 14, '-dot', 'state_space', '../prob_examples/public_examples/B/Kodkod/Loop_statespace.dot'], 'Test performance on example where ProB used to be very slow').  %% HAD TO INCREASE MAX_INITIALISATION to 15 from 8 since adding in_domain_wf to element_of_closure1_wf; TO DO:  FIX
cli_testcase(635, [b_test,kodkod], ['../prob_examples/public_examples/B/Kodkod/Loop.mch', '-strict', '-mc', 20, '-nodead', '-bf', '-p', 'MAX_INITIALISATIONS', 8, '-p', 'KODKOD', 'TRUE', '-cc', 15, 14], 'Test performance on Loop example with Kodkod').
cli_testcase(636, [b_test], ['../prob_examples/public_examples/EventBPrologPackages/ConfigChooser/PetChooser_mch.eventb', '-strict', '-mc', 300, '-p', 'CLPFD', 'TRUE', '-cc', 41, 236], 'Test performance on example where ProB used to be very slow').
cli_testcase(637, [b_test,kodkod,strings], ['../prob_examples/public_examples/B/Kodkod/LoopAsStrings.mch', '-strict', '-mc', 20, '-nodead', '-bf', '-p', 'MAX_INITIALISATIONS', 8, '-p', 'CLPFD', 'TRUE', '-cc', 15, 14], 'Test performance on example where ProB used to be very slow').  %% HAD TO INCREASE MAX_INITIALISATION to 15 from 8 since adding in_domain_wf to element_of_closure1_wf; TO DO:  FIX
cli_testcase(638, [b_test,recursion,external], ['../prob_examples/public_examples/B/Tickets/TransitiveClosure/ComputeTClos_mch.eventb', '-t', '-strict', '-p', 'SYMBOLIC', 'FALSE', '-assertions'], 'Test that recursive function works in Event-B.').
cli_testcase(639, [laws], ['../prob_examples/public_examples/B/Laws/SetLawsPow.mch', '-mc', '500', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'FALSE'], 'Laws about sets of sets').
cli_testcase(640, [laws], ['../prob_examples/public_examples/B/Laws/SetLawsPow.mch', '-mc', '500', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE'], 'Laws about sets of sets').
cli_testcase(641, [laws], ['../prob_examples/public_examples/B/Laws/SetLawsPow.mch', '-mc', '500', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'SYMBOLIC', 'TRUE'], 'Laws about sets of sets').
cli_testcase(642, [laws], ['../prob_examples/public_examples/B/Laws/SetLawsPowPow.mch', '-mc', '250', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE'], 'Laws about sets of sets').
cli_testcase(643, [laws], ['../prob_examples/public_examples/EventBPrologPackages/Deploy/s1_mch1_switches_mch.eventb', '-mc', '250', '-p', 'DEFAULT_SETSIZE', 2, '-strict', '-p', 'CLPFD', 'TRUE', '-cc', 10, 30], 'Check that deferred set is set to at least 3').
cli_testcase(644, [laws], ['../prob_examples/public_examples/EventBPrologPackages/Laws/TestPartitionDeferredSetCardMch_mch.eventb', '-mc', '250', '-nodead', '-p', 'DEFAULT_SETSIZE', 2, '-strict', '-p', 'CLPFD', 'TRUE', '-cc', 5, 6], 'Check that deferred set is set to at least 6').
cli_testcase(645, [plugin_units,wd], ['../prob_examples/public_examples/B/PragmasUnits/Tests/WrongComparisonInInvariant.mch', '-plugin', 'units', '-expecterr', 'well_definedness_error', '-strict', '-mc', 5], '').
cli_testcase(646, [b_test,cache], ['../prob_examples/public_examples/B/Puzzles/CrewAllocationConstants.mch', '-t', '-strict', '-p','TIME_OUT',3000, '-cache', 'cache_does_not_exist/', '-expecterr', cache], 'Check that -cache option works on simple example.').
cli_testcase(647, [b_test,cache], ['../prob_examples/public_examples/B/Puzzles/CrewAllocationConstants.mch', '-t', '-strict', '-p','TIME_OUT',3000, '-cache', '../prob_examples/public_examples/cache/'], 'Check that -cache option works on simple example.').
cli_testcase(648, [b_test,cache], ['../prob_examples/public_examples/B/Puzzles/CrewAllocationConstants.mch', '-t', '-strict', '-p','TIME_OUT',3000, '-cache', '../prob_examples/public_examples/cache/'], 'Check that -cache option works on simple example (2nd time).').
cli_testcase(649, [tickets,infinite], ['../prob_examples/public_examples/B/Tickets/PROB-83/Time.mch', '-t', '-strict'], 'Check infinite function works. (PROB-83)').
cli_testcase(650, [tickets,cbc,override], ['../prob_examples/public_examples/B/Tickets/Berkani1/CartesianProductOverrideAssertion.mch', '-cbc_assertions', '-strict', '-expcterr',cbc_assertions], 'Check that counter example for Cartesian product found.').
cli_testcase(651, [tickets,override], ['../prob_examples/public_examples/B/Tickets/Berkani1/CartesianProductOverride.mch', '-init', '-strict', '-p', 'CLPFD', 'TRUE', '-properties', '-assertions'], 'Check that counter example for Cartesian product found.').
cli_testcase(652, [tickets,infinite,recursion], ['../prob_examples/public_examples/B/Compilers/TuringMachine_Goedelisation.mch', '-mc', 300, '-p', 'CLPFD', 'TRUE', '-expecterr', deadlock, '-strict', '-cc', 245, 244], 'Check machine with several complicated infinite & recursive functions.').
cli_testcase(653, [tickets], ['../prob_examples/public_examples/B/Tickets/Mejia5/essai_fncs.mch', '-t', '-strict'], 'Check ticket (with NAT >+> NAT)').
cli_testcase(654, [laws], ['../prob_examples/public_examples/B/Laws/SetRelLaws_NatBool.mch', '-mc', '750', '-nodead', '-c', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE', '-p', 'CLPFD', 'TRUE'], 'Various laws about relations.').
cli_testcase(655, [typechecker_test,tickets], ['-mc','1000','../prob_examples/public_examples/B/Tickets/PROB-46/Validate.mch','-expecterr',type_error ,'-expecterr', load_main_file, '-expecterr', b_read_write_info, '-strict'], 'Check type-error: Operation calls another operation of the same machine.'). % ADDED '-expecterr', btypechecker --> TO DO: ensure that we do not raise internal errors here !! <--------------
cli_testcase(656, [b_test], ['../prob_examples/public_examples/B/Mathematical/PostCorrespondence_Turing_MC.mch', '-mc', '750', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE', '-expecterr', goal_found], 'A non-trivial machine.').
cli_testcase(657, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/SimpleCorrectUsage.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/SimpleCorrectUsage_internal.mch'], '').
cli_testcase(658, [cbc,cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/Deploy/s4_mch9_schedule_mch.eventb', '-cbc_deadlock', '-p', 'SMT', 'TRUE', '-strict', '-p', 'CLPFD', 'TRUE', '-p','DEFAULT_SETSIZE',1], 'Check no typing error.').
cli_testcase(659, [cbc,cbc_deadlock], ['../prob_examples/public_examples/EventBPrologPackages/Deploy/s4_mch9_schedule_mch.eventb', '-cbc', 'all', '-p', 'SMT', 'TRUE', '-strict', '-p', 'CLPFD', 'TRUE', '-p','DEFAULT_SETSIZE',1], '').
cli_testcase(660, [b_test,external], ['../prob_examples/public_examples/B/ExternalFunctions/TestLibraryIO.mch', '-t', '-strict', '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(661, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/MultiplicationConversion.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/MultiplicationConversion_internal.mch'], '').
cli_testcase(662, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/KnownExponent.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/KnownExponent_internal.mch'], '').
cli_testcase(664, [ltl], ['../prob_examples/public_examples/B/TestCaseGeneration/SimpleCounter.mch', '-ltllimit', '50', '-ltlformula', 'G{1=1}', '-strict', '-expcterr', ltl_incomplete, '-expcterr', state_space], 'check that no error reported for incomplete model check').
cli_testcase(665, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/Car.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/Car_internal.mch'], '').
cli_testcase(666, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/TrafficLight.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/TrafficLight_internal.mch'], '').
cli_testcase(667, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/System.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/System_internal.mch'], '').
cli_testcase(668, [ltl,csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/BobAlice.csp', '-ltlassertions', '-strict'],'Check LTL formulas in File.').
cli_testcase(669, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/Tests/IncorrectUnitDefinition.mch', '-plugin', 'units', '-strict', '-expecterr', 'incorrect_unit_definition'], '').
cli_testcase(670, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/Tests/BoolAndString.mch', '-plugin', 'units', '-strict','-mc', 5], 'Bool and String are not annotated / used by the unit plugin. Just check that there is no error.').
cli_testcase(672, [plugin_units,wd], ['../prob_examples/public_examples/B/PragmasUnits/Tests/WrongOperation.mch', '-plugin', 'units', '-strict', '-expecterr', 'well_definedness_error', '-mc', 5], '').
cli_testcase(673, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/IrregularConversions.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/IrregularConversions_internal.mch'], '').
cli_testcase(675, [plugin_units,wd], ['../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/IfTypeAndClash.mch','-plugin','units','-expecterr','well_definedness_error','-strict','-pp','../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/IfTypeAndClash_internal.mch'], '').
cli_testcase(676, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/UserDefinedAlias.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/UserDefinedAlias_internal.mch'], '').
cli_testcase(677, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/CaseStudies/FuelLevel/ctx.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/CaseStudies/FuelLevel/ctx_internal.mch'], '').
cli_testcase(678, [plugin_units,imp], ['../prob_examples/public_examples/B/PragmasUnits/CaseStudies/FuelLevel/ctx_i.imp', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/CaseStudies/FuelLevel/ctx_i_internal.mch'], '').
cli_testcase(679, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/CaseStudies/FuelLevel/fuel0.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/CaseStudies/FuelLevel/fuel0_internal.mch'], '').
cli_testcase(680, [plugin_units,imp], ['../prob_examples/public_examples/B/PragmasUnits/CaseStudies/FuelLevel/fuel_i.imp', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/CaseStudies/FuelLevel/fuel_i_internal.mch'], '').
cli_testcase(681, [plugin_units,wd], ['../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/IntegerOperations.mch', '-plugin', 'units', '-strict', '-expecterr', 'well_definedness_error', '-pp', '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/IntegerOperations_internal.mch','-p', 'plugin(units,fixpoint_halt_on_error)','false'], '').
cli_testcase(682, [ltl,csp_test], ['../prob_examples/public_examples/CSP/slicing/peterson_modelcheck.csp', '-ltlassertions', '-strict'],'Check LTL formulas in File.').
cli_testcase(683, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/AdditionConversion.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/AdditionConversion_internal.mch'], '').
cli_testcase(684, [csp_test], ['../prob_examples/public_examples/CSP/Tickets/3675/AIDCPRB.csp', '-mc', '10000', '-nodead', '-strict'],'Model checking test for fixed bug number 3675.').
cli_testcase(685, [csp_test], ['../prob_examples/public_examples/CSP/Tickets/3675/SafetyPRB.csp', '-mc', '150000', '-nodead', '-strict'],'Model checking test for fixed bug number 3675.').
cli_testcase(686, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/SequenceOperations.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/SequenceOperations_internal.mch'], '').
cli_testcase(687, [b_test,csp_test,cspb,strings], ['../prob_examples/public_examples/CSPB/FeatureTests/StringMachine.mch', '-csp-guide', '../prob_examples/public_examples/CSPB/FeatureTests/StringMachine.csp', '-mc', '1000', '-strict', '-expecterr', 'deadlock'], 'Test that the machine with guidance by CSP have a deadlock').
cli_testcase(688, [plugin_units,wd], ['../prob_examples/public_examples/B/PragmasUnits/Tests/SomeUnitErrors.mch', '-plugin', 'units', '-strict', '-expecterr', 'well_definedness_error'], '').
cli_testcase(689, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/Advance/Safety_Metro_23_08_2012/GCDoor_M2_mch.eventb', '-cbc', all, '-strict'], 'Test that proof info works well and no invariant checked').
cli_testcase(690, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/ProofDirected/M1_err_mch.eventb', '-cbc', all, '-strict', '-expecterr', cbc, '-expecterr', 'event_refinement', '-p', 'SMT', 'TRUE', '-p', 'CLPFD', 'TRUE'], 'Test that error detected').
cli_testcase(691, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/ProofDirected/TestInvPreserv_M2_ko_mch.eventb', '-mc', 100, '-strict', '-expecterr', 'invariant_violation', '-p', 'SMT', 'TRUE', '-p', 'CLPFD', 'TRUE', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', 1], 'Test that error detected').
cli_testcase(692, [eventb_test,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/ProofDirected/TestInvPreserv_M2_ko_mch.eventb', '-mc', 100, '-nodead', '-strict', '-expecterr', 'event_error:setX:action_not_executable', '-p', 'SMT', 'TRUE', '-p', 'CLPFD', 'TRUE'], 'Test that error detected').
cli_testcase(693, [eventb_test,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/ProofDirected/TestInvPreserv_M2_ko_rename.eventb', '-mc', 100, '-strict', '-expecterr', 'invariant_violation', '-p', 'SMT', 'TRUE', '-p', 'CLPFD', 'TRUE', '-p', 'NUMBER_OF_ANIMATED_ABSTRACTIONS', 1], 'Test that error detected').
cli_testcase(694, [eventb_test,smt_test], ['../prob_examples/public_examples/EventBPrologPackages/ProofDirected/TestInvPreserv_M2_ko_rename.eventb', '-mc', 100, '-nodead', '-strict', '-expecterr', 'event_error:setX3:action_not_executable', '-p', 'SMT', 'TRUE', '-p', 'CLPFD', 'TRUE'], 'Test that error detected').
cli_testcase(695, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/InferenceTests/InvolvedConstraintUnits.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/InferenceTests/InvolvedConstraintUnits_internal.mch'], '').
cli_testcase(696, [plugin_units,wd], ['../prob_examples/public_examples/B/PragmasUnits/InferenceTests/InvolvedConstraintUnitsUnsat.mch', '-plugin', 'units', '-strict','-expecterr', 'well_definedness_error', '-pp', '../prob_examples/public_examples/B/PragmasUnits/InferenceTests/InvolvedConstraintUnitsUnsat_internal.mch'], '').
cli_testcase(697, [tickets,external], ['../prob_examples/public_examples/B/Tickets/Burlando1/machine_test_PRINTF.mch', '-t', '-strict'], 'Test PRINTF').
cli_testcase(698, [b_test], ['-p', 'CLPFD', 'TRUE', '../prob_examples/public_examples/B/Puzzles/SudokuAsConstant.mch', '-t','-strict', '-p', 'TRACE_INFO', 'TRUE'], 'Check TRACE_INFO flag').
cli_testcase(699, [cbc,cbc_deadlock], ['../prob_examples/public_examples/B/CBC/ExistsDeadlockTests.mch', '-cbc_deadlock', '-p', 'TRACE_INFO', 'TRUE','-expcterr', 'cbc_deadlock_check', '-strict'], 'Check that guard simplification works correctly').
cli_testcase(700, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_train/T_m0.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_train/T_m0_internal.mch'], '').
cli_testcase(701, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_train/T_m1.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_train/T_m1_internal.mch'], '').
cli_testcase(702, [b_test,tla],['../prob_examples/public_examples/TLA/Tests/SetLaws.tla', '-mc', 1000, '-nodead', '-strict'], 'Testing TLA2B translation laws.').
cli_testcase(703, [b_test,tla],['../prob_examples/public_examples/TLA/Tests/NatLaws.tla', '-mc', 1000, '-nodead', '-strict'], 'Testing TLA2B translation laws.').
cli_testcase(704, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m0.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m0_internal.mch'], '').
cli_testcase(705, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m1.mch','-expcterr',bmachine_static_checks, '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m1_internal.mch'], '').
cli_testcase(706, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m2.mch','-expcterr',bmachine_static_checks, '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m2_internal.mch'], '').
cli_testcase(707, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m3.mch','-expcterr',bmachine_static_checks, '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m3_internal.mch'], '').
cli_testcase(708, [laws],['../prob_examples/public_examples/B/Kodkod/LawsArithmeticCsts.mch', '-init', '-expcterr', 'setup_constants_fails','-strict', '-p', 'CLPFD', 'TRUE'], 'Testing setting up constants finds no counter-example to mathematical laws.').
cli_testcase(709, [laws,kodkod],['../prob_examples/public_examples/B/Kodkod/LawsArithmeticCsts.mch', '-init', '-expcterr', 'setup_constants_fails','-strict', '-p', 'CLPFD', 'FALSE', '-p', 'KODKOD', 'TRUE','-p', 'TIME_OUT', '5000'], 'Testing setting up constants finds no counter-example to mathematical laws.').
cli_testcase(710, [laws],['../prob_examples/public_examples/B/Kodkod/LawsSetsCsts.mch', '-init', '-expcterr', 'setup_constants_fails', '-p', 'CLPFD', 'TRUE', '-strict', '-p', 'TIME_OUT', 3500], 'Testing setting up constants finds no counter-example to mathematical laws.').
cli_testcase(711, [laws,kodkod],['../prob_examples/public_examples/B/Kodkod/LawsRelCsts.mch', '-init', '-expcterr', 'setup_constants_fails','-strict', '-p', 'CLPFD', 'FALSE', '-p', 'KODKOD', 'TRUE'], 'Testing setting up constants finds no counter-example to mathematical laws.').
cli_testcase(712, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sequences/AppendBuiltInFunctionTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the built-in operator ^ raises expected error.').
cli_testcase(713, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sequences/ConcatBuiltInFunctionTypeError1.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the concat(<...>) built-in function raises expected error.').
cli_testcase(714, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sequences/ConcatBuiltInFunctionTypeError2.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the concat(<...>) built-in function raises expected error.').
cli_testcase(715, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sequences/ExpandingInfiniteSequenceError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing of raising error by expanding infinite sequences.').
cli_testcase(716, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sequences/HeadBuiltInFunctionTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the head(-) built-in function raises expected error.').
cli_testcase(717, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sequences/IsElemListTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'general_error_occurred'], 'Testing if the elem(-,-) built-in function raises expected error.').
cli_testcase(718, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sequences/LengthBuiltInFunctionInfiniteSeqError.csp', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock', '-strict'], 'Testing if the length(-) built-in function raises expected error by calling it with infinite sequence.').
cli_testcase(719, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sequences/LengthBuiltInFunctionTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the length(-) built-in function raises expected error.').
cli_testcase(720, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sequences/NullBuiltInFunctionTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the null(-) built-in function raises expected error.').
cli_testcase(721, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sequences/SeqBuiltInFunctionTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the seq(-) built-in function raises expected error.').
cli_testcase(722, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sequences/SetBuiltInFunctionTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the set(-) built-in function raises expected error.').
cli_testcase(723, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sequences/TailBuiltInFunctionTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sequences', '-expcterr', 'deadlock'], 'Testing if the tail(-) built-in function raises expected error.').
cli_testcase(724, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/ReplicatedInternalChoiceEmptySetError.csp', '-strict', '-mc', 10,'-expcterr', 'cspm_trans', '-expcterr', 'deadlock'], 'Testing of error raising by defining a replicated internal choice process over an empty set.').
cli_testcase(725, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/EmptySetTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sets', '-expcterr', 'deadlock'], 'Testing if the empty(-) built-in function raises expected error.').
cli_testcase(726, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/ReplicatedOperatorsOverEmptyPatternSets.csp', '-strict', '-mc', 10], 'Testing replicated operators over over empty pattern sets (||| x:{}@ P, ; x:<>@P, || x:{} @[a] P and [|a|] x:{} @P).').
cli_testcase(727, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/EmptySetInReplicatedLinkedParallelError.csp', '-strict', '-mc', 10,'-expcterr', 'haskell_csp', '-expcterr', 'csp_sequences', '-expcterr', 'general_error_occurred'], 'Testing of error raising by defining a replicated linked parallel process over an empty set.').
cli_testcase(728, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/InterleavingSpecialCase.csp', '-strict', '-mc', 10, '-expcterr', 'deadlock'], 'Testing a special case of the |||-operator (STOP ||| STOP).').
cli_testcase(729, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/UndefinedConstructError.csp', '-strict', '-mc', 10, '-expcterr', 'general_error_occurred', '-expcterr', 'full_normalise_csp_process'], 'Testing for raising undefined constructor error.').
cli_testcase(730, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/TooManyOutputValuesOnChannelError.csp', '-strict', '-mc', 10, '-expcterr', 'enumerate_action', '-expcterr', 'general_error_occurred'], 'Testing for raising an error, when there are too many output values on channel.').
cli_testcase(731, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/NotABooleanValueErrorIfThenElse.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for raising an error, when there is no boolean value expression in the if-then-else-condition.').
cli_testcase(732, [laws,kodkod],['../prob_examples/public_examples/B/Kodkod/LawsSetsCsts.mch', '-init', '-expcterr', 'setup_constants_fails','-strict', '-p', 'KODKOD', 'TRUE'], 'Testing setting up constants finds no counter-example to mathematical laws.').
cli_testcase(733, [b_test,csp_test,cspb], ['../prob_examples/public_examples/CSPB/TestVariableProbing.mch', '-csp-guide', '../prob_examples/public_examples/CSPB/TestVariableProbing.csp', '-t', '-strict'], 'Test probing of B variables').
cli_testcase(734, [laws,kodkod], ['../prob_examples/public_examples/B/Laws/BoolLaws_SetCompr.mch', '-init', '-c', '-p', 'KODKOD','TRUE', '-strict'], 'Various laws about boolean connectives in set comprehension form.').
cli_testcase(735, [laws,kodkod], ['../prob_examples/public_examples/B/Laws/BoolLaws_SetComprCLPFD.mch', '-init', '-c', '-p', 'KODKOD','TRUE', '-strict'], 'Various laws about boolean connectives in set comprehension form.').
cli_testcase(736, [b_test,csp_test,cspb], ['../prob_examples/public_examples/CSPB/EventB/m_partner_behaviour.eventb', '-csp-guide', '../prob_examples/public_examples/CSPB/EventB/m_partner_behaviour.csp', '-t', '-strict'], 'Test Event-B+CSP + discharged calc').
cli_testcase(737, [b_test,external], ['../prob_examples/public_examples/B/ExternalFunctions/LibraryFiles.mch', '-init', '-assertions','-strict'], 'Check that external functions work.').
cli_testcase(738, [b_test,proz],['../prob_examples/public_examples/Z/Testcases/freetypes.fuzz', '-mc', '1000', '-nodead', '-strict'],'ProZ test for freetypes').
cli_testcase(739, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_tupels/ChannelOutputValuesError.csp', '-strict', '-mc', 10, '-expcterr', 'evaluate_channel_outputs', '-expcterr', 'deadlock'], 'Testing raising an error when no channel tuple is given as prefix event.').
cli_testcase(740, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/TypeErrorInChannelOutput.csp', '-strict', '-mc', 10, '-expcterr', 'check_channel_output_value', '-expcterr', 'general_error_occurred'], 'Testing raising of channel output error.').
cli_testcase(741, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_tupels/ChannelNotDeclaredError.csp', '-strict', '-mc', 10, '-expcterr', 'evaluate_channel_outputs', '-expcterr', 'general_error_occurred', '-expcterr', 'enumerate_action'], 'Testing raising of channel not declared errors.').
cli_testcase(742, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_tupels/ListProjectionError1.csp', '-strict', '-mc', 10, '-expcterr', 'list_projection','-expcterr', 'general_error_occurred'], 'Testing for list projection errors inside local definitions (let-within clauses).').
cli_testcase(743, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_tupels/ListProjectionError2.csp', '-strict', '-mc', 10, '-expcterr', 'list_projection','-expcterr', 'deadlock'], 'Testing for list projection errors inside local definitions (let-within clauses).').
cli_testcase(744, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_tupels/ListProjectionError3.csp', '-strict', '-mc', 10, '-expcterr', 'list_projection','-expcterr', 'deadlock'], 'Testing for list projection errors inside local definitions (let-within clauses).').
cli_testcase(745, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_tupels/TupleProjectionError1.csp', '-strict', '-mc', 10, '-expcterr', 'csp_tuples','-expcterr', 'deadlock'], 'Testing for tuple projection errors inside local definitions (let-within clauses).').
cli_testcase(746, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_tupels/TupleProjectionError2.csp', '-strict', '-mc', 10, '-expcterr', 'csp_tuples','-expcterr', 'deadlock'], 'Testing for tuple projection errors inside local definitions (let-within clauses).').
cli_testcase(747, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_tupels/NaTupleProjectionError1.csp', '-strict', '-mc', 10, '-expcterr', 'csp_tuples','-expcterr', 'deadlock'], 'Testing for not-associative tuple projection errors inside local definitions (let-within clauses).').
cli_testcase(748, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_tupels/NaTupleProjectionError2.csp', '-strict', '-mc', 10, '-expcterr', 'csp_tuples','-expcterr', 'deadlock'], 'Testing for not-associative tuple projection errors inside local definitions (let-within clauses).').
cli_testcase(749, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/CardinalityOfInfiniteSetError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sets', '-expcterr', 'deadlock'], 'Testing if the card(-) built-in function raises expected error by calling it with infinite set.').
cli_testcase(750, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/CardinalityTypeError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sets', '-expcterr', 'deadlock'], 'Testing if the card(-) built-in function raises expected type error by calling it with expression, which is not a set.').
cli_testcase(751, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/InterEmptySetError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sets', '-expcterr', 'deadlock'], 'Testing if the Inter(-) built-in function raises expected error by calling it with an empty set.').
cli_testcase(752, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/ExpandClosureSetError.csp', '-strict', '-mc', 10,'-expcterr', 'csp_sets', '-expcterr', 'deadlock'], 'Testing for errors by expanding closure set elements (error occurs if the element is not a channel name e.g. just constructor).').
cli_testcase(753, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/ExpandingSymbolicSetError1.csp', '-strict', '-mc', 10,'-expcterr', 'expand_symbolic_set', '-expcterr', 'deadlock'], 'Testing for errors by expanding symbolic sets.').
cli_testcase(754, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/demo/Peterson_v2.csp', '-mc', 1000, '-strict', '-p', 'COMPRESSION', 'TRUE', '-cc', 215, 429, '-statistics'], 'Test compression with CSP').
cli_testcase(755, [csp_test], ['../prob_examples/public_examples/CSPB/TestVariableProbing.csp', '-MAIN', 'OTHER', '-mc', 100, '-cc', 4, 4, '-strict'], 'Test setting main process').
cli_testcase(756, [promela], ['../prob_examples/public_examples/Promela/atomic.pml', '-t', '-mc', 100, '-strict'], 'Test Promela mode').
cli_testcase(757, [error_checks,csp_repl], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/TestingForParseAndLoadErrors.csp', '-strict', '-expcterr', 'haskell_csp'], 'Testing the error clauses of the pasrse and load predicates in the haskell_csp module.').
cli_testcase(758, [csp_repl], ['../prob_examples/public_examples/CSP/other/Ivo/Farmer.csp', '-strict', '-get_csp_assertions_as_string'], 'Testing getting Csp assertions from file.').
cli_testcase(759, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/EnumerateInfiniteChannelInputError1.csp', '-strict', '-mc', 10, '-expcterr', 'enumerate_channel_input_value', '-expcterr', 'general_error_occurred'], 'Testing enumerating infinite channel input error.').
cli_testcase(760, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/EnumerateIllegalTypeOfChannelError.csp', '-strict', '-mc', 10, '-expcterr', 'enumerate_channel_input_value', '-expcterr', 'general_error_occurred'], 'Testing enumerating infinite channel input error.').
cli_testcase(761, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/EnumerateInfiniteChannelInputError2.csp', '-strict', '-mc', 10, '-expcterr', 'enumerate_channel_input_value', '-expcterr', 'general_error_occurred'], 'Testing for raising enumerating infinite channel type error for sequences.').
cli_testcase(762, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/ChannelInputSetEnumerationResultError.csp', '-strict', '-mc', 10, '-expcterr', 'set_enumeration_tuple_result', '-expcterr', 'deadlock'], 'Testing for raising type error on channel input.').
cli_testcase(763, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/ChannelTooFewValuesOnChannelError.csp', '-strict', '-mc', 10, '-expcterr', 'enumerate_action', '-expcterr', 'general_error_occurred'], 'Testing for raising type error in channel input.').
cli_testcase(764, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/InterleavedIfThenElseBlocks.csp', '-strict', '-mc', 1000], 'Testing interleaved if-then-else blocks.').
cli_testcase(767, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/BAC9_ZC/pas_as_env_inv_ld.mch','-strict', '-t', '-p',
 'CLPFD', 'TRUE', '-aa', 27, 0, 0, '-p', 'MAX_OPERATIONS', 1, '-p', 'WARN_IF_DEFINITION_HIDES_VARIABLE', 'FALSE', '-sptxt', '../prob_examples/examples/B/Siemens/BAC9_ZC/pas_as_env_inv_ld_sptxt.txt'], 'Siemens Data Validation Test.').
cli_testcase(777, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/AdaptedBModelPropCheck/acs_as_env_ctx_aiguille.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 49, 2, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0, '-strict'], 'Siemens Data Validation Test.').
cli_testcase(778, [private,data_validation,siemens,codespeed], ['../prob_examples/examples/B/Siemens/AdaptedBModelPropCheck/acs_as_env_cfg_ipart.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 144, 4, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(779, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/AdaptedBModelPropCheck/acs_as_env_cfg_ipart.mch','-strict', '-p','CLPFD', 'TRUE', '-ma', 95, 2, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test (main assertions).').
cli_testcase(780, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/Canarsie_PAR1_PAL/cbtc_mes_as_env_inv_easitf_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 127, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(781, [private,data_validation,siemens,codespeed], ['../prob_examples/examples/B/Siemens/Canarsie_PAR1_PAL/cbtc_mes_as_env_inv_easitf_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-ma', 22, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test (main assertions).').
cli_testcase(782, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/Canarsie_PAR1_PAL/cbtc_mes_as_env_inv_etors_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 142, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(783, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/Canarsie_PAR1_PAL/cbtc_mes_as_env_inv_stors_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 147, 1, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(784, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/Canarsie_PAR1_PAL/par1_mes_as_env_inv_etors_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 125, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(785, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/Canarsie_PAR1_PAL/par1_pans_as_env_inv_etors_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 188, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(786, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/Canarsie_PAR1_PAL/par1_pans_as_env_inv_etors_dist_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 184, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(787, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/Canarsie_PAR1_PAL/par1_pans_as_env_inv_mesitf_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 182, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(788, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/Canarsie_PAR1_PAL/par1_pans_as_env_inv_pasitf_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 176, 1, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(789, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/Canarsie_PAR1_PAL/par1_pans_as_env_inv_se_dj_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 174, 2, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(790, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/Canarsie_PAR1_PAL/par1_pans_as_env_inv_ss_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 192, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(791, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/Canarsie_PAR1_PAL/par1_pans_as_env_inv_stors_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 192, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(792, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/Canarsie_PAR1_PAL/par1_pans_as_env_inv_tcs_bs.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 183, 8, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(793, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/CBTC_Sol_Line4_SaoPaulo/pas_as_env_inv_etorf.mch','-strict', '-p','CLPFD', 'TRUE', '-init', '-expcterr', 'setup_constants_inconsistent', '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(794, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/CBTC_Sol_Line4_SaoPaulo/pas_as_env_inv_etorf.mch','-strict', '-p','CLPFD', 'TRUE', '-init', '-expcterr', 'setup_constants_inconsistent', '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'WARN_IF_DEFINITION_HIDES_VARIABLE', 'FALSE','-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(795, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/CBTC_Sol_Line4_SaoPaulo/cmnb_heure_cst_gen.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 7, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'WARN_IF_DEFINITION_HIDES_VARIABLE', 'FALSE','-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(796, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/CBTC_Sol_Line4_SaoPaulo/pas_as_env_ctx_trainstop.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 6, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'WARN_IF_DEFINITION_HIDES_VARIABLE', 'FALSE','-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(797, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/CBTC_Sol_Line4_SaoPaulo/pas_as_env_inv_quai.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 50, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'WARN_IF_DEFINITION_HIDES_VARIABLE', 'FALSE','-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.'). % changed -aa 51 to 50 due to partition detection in Sept. 2014
cli_testcase(798, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/CBTC_Sol_Line4_SaoPaulo/pas_as_env_inv_zaum.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 40, 0, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 5000, '-p', 'WARN_IF_DEFINITION_HIDES_VARIABLE', 'FALSE','-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(799, [private,data_validation,siemens], ['../prob_examples/examples/B/Siemens/CBTC_Sol_Line4_SaoPaulo/pas_as_env_inv_cv_sui.mch','-strict', '-p','CLPFD', 'TRUE', '-aa', 62, 1, 0, '-p', 'MAX_OPERATIONS', 0, '-p', 'MAXINT', 2147483647, '-p', 'TIME_OUT', 25000, '-p', 'WARN_IF_DEFINITION_HIDES_VARIABLE', 'FALSE','-p', 'EXPAND_FORALL_UPTO', 0], 'Siemens Data Validation Test.').
cli_testcase(800, [private,smt_test,bosch,codespeed], ['../prob_examples/examples/RodinModels/Bosch/CrCtl_0.5.0_v9/CrCtl_Comb2_Final_mch.eventb', '-cbc_deadlock_pred', 'Counter=10 & P_CrCtl_Mode=STANDBY', '-p', 'SMT', 'TRUE',
 '-p', 'CLPFD', 'TRUE', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-p', 'MAXINT', 2, '-p', 'DEFAULT_SETSIZE', 1,  '-strict', '-expcterr', cbc_deadlock_check, '-p', 'MAX_INITIALISATIONS', 0],'Bosch CBC Test.').
cli_testcase(801, [private,smt_test,bosch], ['../prob_examples/examples/RodinModels/Bosch/CrCtl_0.5.0_v8/CrCtl_Comb2_Final_mch.eventb', '-cbc_deadlock_pred', 'Counter=10', '-p', 'SMT', 'TRUE',
 '-p', 'CLPFD', 'TRUE', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-p', 'MAXINT', 2, '-p', 'DEFAULT_SETSIZE', 1,  '-strict', '-expcterr', cbc_deadlock_check, '-p', 'MAX_INITIALISATIONS', 0],'Bosch CBC Test.').
cli_testcase(802, [private,smt_test,bosch], ['../prob_examples/examples/RodinModels/Bosch/CrCtl_0.5.0_v7/CrCtl_Comb2_Final_mch.eventb', '-cbc_deadlock_pred', 'Counter=10', '-p', 'SMT', 'TRUE',
 '-p', 'CLPFD', 'TRUE', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-p', 'MAXINT', 2, '-p', 'DEFAULT_SETSIZE', 1,  '-strict', '-expcterr', cbc_deadlock_check, '-p', 'MAX_INITIALISATIONS', 0],'Bosch CBC Test.').
cli_testcase(803, [private,smt_test,bosch,codespeed], ['../prob_examples/examples/RodinModels/Bosch/CrCtl_0.5.0_v6/CrCtl_Comb2_Final_mch.eventb', '-cbc_deadlock_pred', 'Counter=10', '-p', 'SMT', 'TRUE',
 '-p', 'CLPFD', 'TRUE', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-p', 'MAXINT', 2, '-p', 'DEFAULT_SETSIZE', 1,  '-strict', '-expcterr', cbc_deadlock_check, '-p', 'MAX_INITIALISATIONS', 0],'Bosch CBC Test.').
cli_testcase(804, [private,smt_test,bosch], ['../prob_examples/examples/RodinModels/Bosch/CrCtl_0.5.0_v5/CrCtl_Comb2_Final_mch.eventb', '-cbc_deadlock_pred', 'Counter=10', '-p', 'SMT', 'TRUE',
 '-p', 'CLPFD', 'TRUE', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-p', 'MAXINT', 2, '-p', 'DEFAULT_SETSIZE', 1,  '-strict', '-expcterr', cbc_deadlock_check, '-p', 'MAX_INITIALISATIONS', 0],'Bosch CBC Test.').
cli_testcase(805, [private,smt_test,bosch], ['../prob_examples/examples/RodinModels/Bosch/CrCtl_0.5.0_v5/CrCtl_Comb2_Final_mch.eventb', '-cbc_deadlock', '-p', 'SMT', 'TRUE',
 '-p', 'CLPFD', 'TRUE', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-p', 'MAXINT', 2, '-p', 'DEFAULT_SETSIZE', 1,  '-strict', '-expcterr', cbc_deadlock_check, '-p', 'MAX_INITIALISATIONS', 0],'Bosch CBC Test.').
cli_testcase(806, [private,smt_test,bosch], ['../prob_examples/examples/RodinModels/Bosch/CrCtl_0.5.0_v4/CrCtl_Comb2_Final_mch.eventb', '-cbc_deadlock_pred', 'Counter=10', '-p', 'SMT', 'TRUE',
 '-p', 'CLPFD', 'TRUE', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-p', 'MAXINT', 2, '-p', 'DEFAULT_SETSIZE', 1,  '-strict', '-expcterr', cbc_deadlock_check, '-p', 'MAX_INITIALISATIONS', 0],'Bosch CBC Test.').
cli_testcase(807, [tickets,csp_test], ['../prob_examples/public_examples/CSP/Tickets/Shi/milnerSched5.csp', '-assertions', '-strict'], 'Check that exception does not occur (Ticket PROB-43)').
cli_testcase(808, [b_test,dot], ['../prob_examples/public_examples/B/Benchmarks/CarlaTravelAgencyErr.mch', '-t', '-strict', '-dot', 'state_as_graph', '../prob_examples/public_examples/B/Benchmarks/CarlaTravelAgencyErr_state_graph.dot'], 'Check State Graph Generation').
cli_testcase(809, [private,data_validation], ['../prob_examples/public_examples/B/Tickets/LargeMachTypeErr/LargeValueOk.mch', '-assertions', '-strict'], 'Check loading large data').
cli_testcase(810, [private,typechecker_test], ['../prob_examples/public_examples/B/Tickets/LargeMachTypeErr/sequenceur_values.mch', '-expcterr', type_error, '-expcterr', load_main_file, '-strict'], 'Check large number of errors').
cli_testcase(811, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_tupels/ChannelNameUsedAsValueError.csp', '-strict', '-mc', 10, '-expcterr', 'valid_value', '-expcterr', 'general_error_occurred','-expcterr', 'check_channel_output_value'], 'Testing for raising channel name used as value error.').
cli_testcase(812, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_tupels/IllegalDataValueError.csp', '-strict', '-mc', 10, '-expcterr', 'valid_value', '-expcterr', 'general_error_occurred','-expcterr', 'check_channel_output_value'], 'Testing for raising illegal data value error.').
cli_testcase(813, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_tupels/MismatchNumberOfArgumentsForChannelSynchrError.csp', '-strict', '-mc', 10, '-expcterr', 'unify_tuple_args', '-expcterr', 'deadlock'], 'Testing for raising mismatch number of arguments error.').
cli_testcase(814, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality1.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'general_error_occurred'], 'Testing for type errors in equality.').
cli_testcase(815, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality2.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'general_error_occurred'], 'Testing for type errors in equality.').
cli_testcase(816, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality3.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'general_error_occurred'], 'Testing for type errors in equality.').
cli_testcase(817, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality4.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'general_error_occurred'], 'Testing for type errors in equality.').
cli_testcase(818, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality5.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(819, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality6.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(820, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality7.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(821, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality8.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(822, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality9.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(823, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality10.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(824, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality11.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(825, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality12.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(826, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality1.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(827, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality2.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'general_error_occurred'], 'Testing for type errors in equality.').
cli_testcase(828, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality3.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'general_error_occurred'], 'Testing for type errors in equality.').
cli_testcase(829, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality4.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'general_error_occurred'], 'Testing for type errors in equality.').
cli_testcase(830, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality5.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(831, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality6.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(832, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality7.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(833, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality8.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(834, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality9.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(835, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality10.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(836, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality11.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(837, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality12.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for type errors in equality.').
cli_testcase(838, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/TestingComprehensionSetGuards.csp', '-strict', '-mc', 10000], 'Testing set comprehension guards.').
cli_testcase(839, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/ExpandingDataTypeDefBody.csp', '-strict', '-mc', 10], 'Testing expanding symbolic sets.').
cli_testcase(840, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/ChannelNameUsedForLocalVariableError.csp', '-strict', '-expcterr', 'csp_sets'], 'Testing for raising channel name used as local variable error.').
cli_testcase(841, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/PatternListTooLongForChannelError.csp', '-strict', '-mc', 10, '-expcterr', 'csp_sets', '-expcterr', 'deadlock'], 'Testing raising pattern list too long error.').
cli_testcase(842, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/TooManyValuesInRenamePatternError.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'general_error_occurred'], 'Testing raising too many values in rename pattern error.').
cli_testcase(843, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/TypeErrorForChannelInputVariable.csp', '-strict', '-mc', 10, '-expcterr', 'enumerate_datatype_el', '-expcterr', 'deadlock'], 'Testing raising type error for channel input variable exception.').
cli_testcase(844, [csp_repl], ['../prob_examples/public_examples/CSP/other/Ivo/Farmer.csp', '-strict', '-get_csp_enumerated_assertions_as_string'], 'Testing getting enumerated assertions from file.').
cli_testcase(845, [csp_repl], ['../prob_examples/public_examples/CSP/other/Ivo/Farmer.csp', '-strict', '-prologOut', '../prob_examples/public_examples/CSP/other/Ivo/Farmer_temp.pl'], 'Testing translating CSP-M file to specific prolog file.').
cli_testcase(846, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/IsMemberSetCouldNotEvaluateError.csp', '-strict', '-mc', 10, '-expcterr', 'csp_sets', '-expcterr', 'deadlock'], 'Testing for raising could not evaluate error for the member(-,-) built-in function.').
cli_testcase(847, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/NotASingletonSetError.csp', '-strict', '-mc', 10, '-expcterr', 'singleSetElement', '-expcterr', 'deadlock'], 'Testing for raising not a singleton set error.').
cli_testcase(848, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality13.csp', '-strict', '-mc', 10000], 'Testing for type errors in equality.').
cli_testcase(849, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality13.csp', '-strict', '-mc', 10000], 'Testing for type errors in disequality.').
cli_testcase(850, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/NotABooleanExpressionError.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for raising not a boolean expression error.').
cli_testcase(851, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/subtype_ex.csp', '-strict', '-mc', 1000,'-expcterr', 'deadlock' ,'-df'], 'Testing complex subtype expressions.').
cli_testcase(852, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/subtype_nametype_ex.csp', '-strict', '-mc', 2500,'-expcterr', 'deadlock'], 'Testing complex subtype expressions.').
cli_testcase(853, [private], ['../prob_examples/examples/CSP/Winter/miniAlvey.csp', '-strict', '-t', '-c', '-v'], 'Testing complex CSP spec.').
cli_testcase(854, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/NoMatchingCaseForFunctionCallError.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'cspm_trans', '-expcterr', 'deadlock'], 'Testing for no matching case for function call error.').
cli_testcase(855, [b_test,cache], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/RussianPostalPuzzle.mch', '-mc', '1000', '-cc', '81', '177', '-p', 'DEFAULT_SETSIZE', '2', '-nogoal', '-strict', '-cache', '../prob_examples/public_examples/cache/'], 'Check that no error in reuse of cached operations').
cli_testcase(856, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInEquality14.csp', '-strict', '-mc', 10000], 'Testing for type errors in equality.').
cli_testcase(857, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/TypeErrorInNonEquality14.csp', '-strict', '-mc', 10000], 'Testing for type errors in disequality.').
cli_testcase(858, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/Farmer.csp', '-strict', '-mc', 1000, '-expcterr', 'deadlock', '-save', '../prob_examples/public_examples/CSP/other/Ivo/Farmer_state.txt'], 'Testing saving XSB Specification state.').
cli_testcase(859, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/NotEnoughArgumentsOnChannelError.csp', '-strict', '-mc', 20, '-bf', '-expcterr', 'l_enumerate_channel_input_value', '-expcterr', 'deadlock'], 'Testing for raising not enough arguments on channel error.').
cli_testcase(860, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/TestingIsMemberBuiltInFunction.csp', '-mc', '100000', '-nodead', '-strict', '-assertions'], 'Testing member(-,-) builtin function.').
cli_testcase(861, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/UnknownExpandChannelPatternExpressionError.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for raising unknown expand channel pattern expression error (possible when expanding sharing, synchronisation-sets/closures).').
cli_testcase(862, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/UnsupportedRefinementOperatorError.csp', '-strict', '-assertions', '-expcterr', 'unsupported_refinement_operator', '-expcterr', 'check_assertions', '-expcterr', 'checkAssertion'], 'Testing for raising unsupproted refinement operator error (e.g. [R=).').
cli_testcase(863, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/NonAtomicArgumentError.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', deadlock, '-bf'], 'Testing for raising Non-atomic argument error.').
cli_testcase(864, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/TestingUnicodeFeatures.csp', '-strict', '-csp_add_unicode','../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/TestingUnicodeFeatures_unicode.csp'], 'Testing the unicode features of the CSPM tool.').
cli_testcase(865, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/TestingUnicodeFeatures_unicode_saved.csp', '-strict', '-csp_remove_unicode', '../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/TestingUnicodeFeatures_no_unicode.csp'], 'Testing the unicode features of the CSPM tool.').
cli_testcase(866, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/CouldNotExpandInfiniteDatatypeBody.csp', '-strict', '-mc', 10, '-expcterr', 'csp_sets', '-expcterr', 'deadlock'], 'Testing for raising could not expand infinite datatype body error.').
cli_testcase(867, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/ExpandingDatatypeBodies.csp', '-strict', '-mc', 10], 'Testing expanding datatype bodies with complicated datat structures.').
cli_testcase(868, [parser,tickets], ['../prob_examples/public_examples/CSPB/Dining/Dining.mch', '-ltlformulat', 'G {taken= {} ', '-strict', '-expcterr', 'ltl'], 'Testing that LTL parser does not loop. (PARSERLIB-11)').
cli_testcase(869, [parser,tickets], ['../prob_examples/public_examples/CSPB/Dining/Dining.mch', '-ctlformula', 'AG {taken= {} ', '-strict', '-expcterr', 'ctl'], 'Testing that CTL parser does not loop. (PARSERLIB-11)').
cli_testcase(870, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/Renaming.csp', '-strict', '-mc', 10], 'Testing for raising could not expand infinite datatype body error.').
cli_testcase(871, [tickets], ['../prob_examples/public_examples/EventBPrologPackages/openETCS/ATB/Model_ctx.eventb', '-init', '-strict', '-properties'], 'Check that no time-out occurs; check C <: S21*S22 <=> C : S21 <-> S22 rule in check_subset_of0').
cli_testcase(872, [plugin_units,wd], ['../prob_examples/public_examples/B/PragmasUnits/Tests/WrongOperationTopParameter.mch', '-plugin', 'units', '-strict', '-expecterr', well_definedness_error, '-mc', 5], '').
cli_testcase(873, [plugin_units,wd], ['../prob_examples/public_examples/B/PragmasUnits/Tests/WrongConversionMultiplication.mch', '-plugin', 'units', '-strict', '-expecterr', well_definedness_error, '-mc', 5], '').
cli_testcase(874, [plugin_units,wd], ['../prob_examples/public_examples/B/PragmasUnits/Tests/WrongConversionAddition.mch', '-plugin', 'units', '-strict', '-expecterr', well_definedness_error, '-mc', 5], '').
cli_testcase(875, [tickets,error_checks], ['-eval', 'g = {x|->y | x=y+1 }', '-expecterr', type_expression_error], 'Check that a type error instead of an internal failure is reported when an expression is used where an identifier is expected (Ticket PROB-213).').
cli_testcase(876, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/demo/mbuff.csp', '-assertions', '-strict'], 'Checking CSP assertions.').
cli_testcase(877, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/demo/crossing.csp', '-assertions', '-strict'], 'Checking CSP assertions.').
cli_testcase(878, [csp_repl], ['-csp_eval', '(seq(set(<1,2,3>))==<1,2,3>) and (set(seq({1,2,3}))=={1,2,3})', '-strict'], 'Checking CSP expressions without loading a specific model first.').
cli_testcase(879, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/demo/crossing.csp', '-csp_eval', 'diff(NonTock,{|enter,leave|}) == {error,gate.down,gate.go_down,gate.go_up,gate.up,sensor.in,sensor.out}', '-strict'], 'Checking CSP expressions.').
cli_testcase(880, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/demo/crossing.csp', '-csp_eval', 'diff(NonTock,{|enter,leave|}', '-expcterr', 'parsing_csp_expression', '-strict'], 'Checking raising parse error while evaluating syntactically non correct CSP expressions.').
cli_testcase(881, [plugin_units,wd], ['../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/Choice.mch', '-plugin', 'units', '-strict','-expecterr', 'well_definedness_error','-pp','../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/Choice_internal.mch'], '').
cli_testcase(882,[private,data_validation],['../prob_examples/examples/B/Alstom/Ticket_Oct11_2012/Rule_DB_SIGAREA_0024.mch','-strict', '-t', '-assertions'], 'Issue related to cut in b_try_check_boolean_expression_lwf.').
cli_testcase(883, [csp_test], ['../prob_examples/public_examples/CSP/ucsexamples/chapter14/maxpro.csp', '-assertions', '-strict'], 'Checking CSP assertions from file').
cli_testcase(884, [b_test,proz], ['../prob_examples/public_examples/Z/Woodcock/InnerMember.fuzz', '-t', '-mc', 100, '-strict'], 'Checking Z model').
cli_testcase(885, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/TryingToApplyUndefinedFunctionError.csp', '-strict', '-mc', 10, '-expcterr', 'evaluate_agent_call_parameters', '-expcterr', 'cspm_trans', '-expcterr', 'deadlock'], 'Testing for raising trying to apply undefined process/function error.').
cli_testcase(886, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/CouldNotUnifyLambdaArgsError.csp', '-strict', '-mc', 10, '-expcterr', 'cspm_trans', '-expcterr', 'deadlock'], 'Testing for raising of could not unify lambda args error.').
cli_testcase(887, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/CallingNonAtomicFunctionArgumentError.csp', '-strict', '-mc', 10, '-expcterr', 'haskell_csp', '-expcterr', 'deadlock'], 'Testing for raising of trying to call non-atomic function error.').
cli_testcase(888, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp_analyzer/UnsupportedAppendPatternError.csp', '-expcterr', 'compile_head_para', '-strict'], 'Testing for raising of unsupported append pattern error.').
cli_testcase(889, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/TestingComplexChannelTypes.csp', '-assertions', '-strict'], 'Testing outputing channel values of complex types.').
cli_testcase(890, [error_checks,csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp_analyzer/IllegalASTCSPProcessInIllegalPositionError.csp', '-expcterr', 'haskell_csp_analyzer', '-strict'], 'Testing for raising of unsupported append pattern error.').
cli_testcase(891, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/PatternMatchingAppendPatternsAsChannelInputValues.csp', '-mc', 1000, '-nodead', '-assertions', '-strict'], 'Testing for pattern matching append patterns as input values.').
cli_testcase(892, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/CheckingPossibleInputVariablesOnChannel.csp', '-assertions', '-strict'], 'Testing for possible input patterns on channels.').
cli_testcase(893,[private],['../prob_examples/examples/B/L4MicroKernel/L4API/API.mch','-strict', '-cbc_sequence', 'TimeoutIPC','-p', 'DEFAULT_SETSIZE', 3], 'Check that no errors produced by cbc_sequence (Ticket PROB-215).').
cli_testcase(894,[tickets,enabling],['-enabling_analysis_csv','../prob_examples/public_examples/B/Daniel/minienabling_analysis.csv','../prob_examples/public_examples/B/Daniel/minienabling.mch'], 'Check CSV output of a simple enabling analysis (Ticket PROB-215).').
cli_testcase(895, [csp_test], ['../prob_examples/public_examples/CSP/Tickets/3675/SafetyPRB.csp', '-assertions', '-strict'],'Assertions check of processes with complex events.').
cli_testcase(896, [kodkod], ['-p','KODKOD','TRUE','-eval','card(10..x) = 0 & x : 0..100 & (x<10 => x+x>5)', '-strict'],'The translation to Kodkod should not fail because a helper variable has no possible values (Ticket PROB-218)').
cli_testcase(897, [kodkod], ['-p','KODKOD','TRUE','-eval','S<:1..10 & S/={}', '-strict'],'Checking the information about the integer range of the elements of an empty set (Ticket PROB-220)').
cli_testcase(898, [b_test], ['-evalt', 'x \\/ 2..n = 0..(n+10) & x /\\ 2..n = {} & x <: 0..n+20 & n=10000', '-strict'],'Checking set partition inside eval').
cli_testcase(899, [kodkod], ['-p','KODKOD','TRUE','-mc',1000,'../prob_examples/public_examples/B/Tickets/PROB-217/kodkod_overflow.mch', '-nodead', '-strict'],'The output of the predicate analysis alone does not prevent integer overflows -  (Ticket PROB-217)').
cli_testcase(900, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-csp_assertion', 'Q(3) [F= Q3', '-csp_assertion', 'Q3 [F= Q(3)', '-csp_assertion', 'Q(4) [T= Q(1,3)', '-csp_assertion', 'Q(1,3) [F= Q(4)', '-strict'], 'Checking various refinement assertions of processes with arguments.').
cli_testcase(901, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-csp_assertion', 'Q(10) :[deadlock free [FD]]', '-expcterr', 'model_check_fails', '-strict'], 'Deadlock checking of a process called with argument.').
cli_testcase(902, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-csp_assertion', 'Q2(9) :[livelock free]', '-cc', 13, 12, '-strict'], 'Livelock checking of a process called with argument.').
cli_testcase(903, [csp_test, csp_det], ['../prob_examples/public_examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-csp_assertion', 'Q2(9) :[deterministic [FD]]', '-expcterr', 'model_check_fails','-strict'], 'Determinism checking of a process called with argument (must fail).').
cli_testcase(904, [csp_test, csp_det], ['../prob_examples/public_examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-csp_assertion', 'a.1 -> a.2 -> STOP :[deadlock free [FD]]', '-expcterr', 'model_check_fails', '-strict'], 'Deadlock checking of a process called with argument.').
cli_testcase(905, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-csp_assertion', 'a.1 -> a.2 -> STOP [T= a.1 -> STOP', '-strict'], '').
cli_testcase(906, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-csp_assertion', 'a.1 -> a.2 -> STOP [F= a.1 -> STOP', '-expcterr', 'refinement_check_fails', '-strict'], '').
cli_testcase(907, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-csp_assertion', 'a.1 -> a.2 -> STOP [F= a.1 -> b.1 -> STOP', '-expcterr', 'parsing_csp_declaration', '-strict'], 'Checking expecting error to occur.').
cli_testcase(908, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-csp_assertion', 'a.1 -> a.2 -> STOP [F= a.1 -> a.1.1 -> STOP', '-expcterr', 'checkAssertion', '-expcterr', 'enumerate_action', '-strict'], '').
cli_testcase(909, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/ProcessesWithArguments.csp', '-csp_assertion', 'Q(4) [] Q(2) [FD= Q(2,2) [] Q(1,1)', '-strict'], '').
cli_testcase(910, [analysis_test], ['-pacheck','../prob_examples/public_examples/B/predicate_analysis/sets.mch', '-strict'], 'Test cases for the predicate analysis: rules about sets').
cli_testcase(911, [analysis_test,sigma], ['-pacheck','../prob_examples/public_examples/B/predicate_analysis/arithmetic.mch', '-strict'], 'Test cases for the predicate analysis: some arithmetic (mainly SIGMA)').
cli_testcase(912, [kodkod,cbc], ['../prob_examples/public_examples/B/Kodkod/TestCBCAssertionCheck.mch', '-cbc_assertions', '-strict', '-p','KODKOD', 'TRUE', '-expcterr',cbc_assertions], 'Check that Kodkod CBC Checking works.').
cli_testcase(913, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/ExpandingDataTypeDefBody.csp', '-csp_eval', 'D1 == {A,B.0,B.1,B.2,B.3,C.0.0,C.0.1,C.0.2,C.0.3,C.1.0,C.1.1,C.1.2,C.1.3,C.2.0,C.2.1,C.2.2,C.2.3,C.3.0,C.3.1,C.3.2,C.3.3}', '-strict'], 'Evaluating CSP-M expressions.').
cli_testcase(914, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/ExpandingDataTypeDefBody.csp', '-csp_eval', '{|c.C|} =={c.(C.0.0),c.(C.0.1),c.(C.0.2),c.(C.0.3),c.(C.1.0),c.(C.1.1),c.(C.1.2),c.(C.1.3),c.(C.2.0),c.(C.2.1),c.(C.2.2),c.(C.2.3),c.(C.3.0),c.(C.3.1),c.(C.3.2),c.(C.3.3)}', '-strict'], 'Evaluating CSP-M expressions').
cli_testcase(915, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/ExpandingDataTypeDefBody.csp', '-csp_eval', '{1..3} == {1,2,3}', '-strict'], 'Evaluating CSP-M expressions.').
cli_testcase(916, [csp_test], ['../prob_examples/public_examples/CSP/simple/verysimple.csp', '-csp_eval', 'Events', '-strict'], 'Evaluating CSP-M expressions.').
cli_testcase(917, [csp_test], ['../prob_examples/public_examples/CSP/simple/verysimple.csp', '-csp_eval', '{|a|}', '-strict'], 'Evaluating CSP-M expressions.').
cli_testcase(918, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/simple/verysimple.csp', '-csp_eval', '{1..}','-strict','-expcterr',expand_symbolic_set], 'Testing raising of infinite set error.').
cli_testcase(919, [b_test, csp_test,cspb], ['../prob_examples/public_examples/CSP/Tickets/WilliamsCSPB/Airports.mch', '-csp-guide', '../prob_examples/public_examples/CSP/Tickets/WilliamsCSPB/AirportsCtrl.csp', '-assertions', '-strict'], 'Checking CSP-M assertions for CSP-M || B models.').
cli_testcase(920, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/ExpandingDataTypeDefBody.csp', '-csp_eval',
             '{x | x <- D1, (B.1) != x} == {A,B.0,B.2,B.3,C.0.0,C.0.1,C.0.2,C.0.3,C.1.0,C.1.1,C.1.2,C.1.3,C.2.0,C.2.1,C.2.2,C.2.3,C.3.0,C.3.1,C.3.2,C.3.3}', '-strict'], 'Evaluating CSP-M expressions.').
cli_testcase(921, [error_checks, csp_test], ['../prob_examples/public_examples/CSP/ErrorModels/csp_sets/EmptySetNotAllowedForInterError.csp', '-mc', 10,'-expcterr', 'is_member_set2', '-expcterr', 'general_error_occurred', '-strict', '-expcterr', virtual_time_out], 'Testing raising of empty set not allowed for Inter(-) error.').
cli_testcase(922, [csp_test], ['-csp_eval', 'if member(1,{1..10}) then true else false', '-strict'], 'Evaluating CSP-M expressions.').
cli_testcase(923, [error_checks], ['-mc', 1000, '../prob_examples/public_examples/B/Tickets/PROB-226/parallel_subst.mch', '-expcterr', 'type_error', '-expcterr', 'load_main_file', '-strict'], 'Parallel assignments to the same variable should be forbidden. (Ticket PROB-226)').
cli_testcase(924, [error_checks], ['-mc', 1000, '../prob_examples/public_examples/B/ErrorMachines/UnknownPromotedOp.mch', '-expcterr', 'type_error', '-expcterr', 'load_main_file', '-strict'], 'Type-checker error when promoting an operation that does not exist.').
cli_testcase(925, [error_checks], ['-mc', 1000, '../prob_examples/public_examples/B/ErrorMachines/IncludeSeenOnlyOnce.mch', '-expcterr', 'type_error', '-expcterr', 'load_main_file', '-strict'], 'Expecting error when including a machine twice that is used as a seen/used machine.').
cli_testcase(926, [csp_test,ltl], ['../prob_examples/public_examples/CSP/simple/verysimple.csp', '-ltlformulat', 'GF [a]', '-expcterr', 'ltl', '-strict'],'Checking for raising number of arguments not correct error.').
cli_testcase(927, [typechecker_test], ['-eval', 'a-b=c & a=0', '-strict'], 'Type-checking a minus when the type of the arguments is determinded later (an integer).').
cli_testcase(928, [typechecker_test], ['-eval', 'a-b=c & a={0}', '-strict'], 'Type-checking a minus when the type of the arguments is determinded later (a set).').
cli_testcase(929, [csp_test,ltl], ['../prob_examples/public_examples/CSP/ErrorModels/ltl/ExtraArgumentsInCSPLTLPatternError.csp', '-ltlassertions', '-expcterr', 'enumerate_action', '-expcterr', 'ltl', '-strict'],'Checking for raising number of arguments not correct error.').
cli_testcase(930, [csp_test,ltl], ['../prob_examples/public_examples/CSP/ErrorModels/ltl/MissingElementsInCSPPatternError.csp', '-ltlassertions', '-expcterr', 'ltl', '-strict'],'Checking for raising number of arguments not correct error.').
cli_testcase(931, [csp_test], ['../prob_examples/public_examples/CSP/ucsexamples/chapter20/phone.csp', '-csp_eval', 'Phones', '-strict'], 'Evaluating CSP-M expressions.').
cli_testcase(932, [typechecker_test,error_checks], ['-eval', 'a-b=c & a=TRUE', '-expcterr', 'type_expression_error', '-strict'], 'Type-checking a minus when the type of the arguments is determinded later (type error).').
cli_testcase(933, [typechecker_test,error_checks], ['-eval', 'a*b=c & a=TRUE', '-expcterr', 'type_expression_error', '-strict'], 'Type-checking * when the type of the arguments is determinded later (type error).').
cli_testcase(934, [typechecker_test], ['-eval', '[a,c:=1,3](a<b&b<c)', '-strict'], 'Type-checking a substitution on a predicate').
cli_testcase(935, [typechecker_test,error_checks], ['-eval', '[a,c:=1,3,5](a<b&b<c)', '-expcterr', 'type_expression_error', '-strict'], 'Type-checking a substitution on a predicate (type error)').
cli_testcase(936, [b_test, csp_test,ltl,cspb], ['../prob_examples/public_examples/CSP/Tickets/WilliamsCSPB/Airports.mch', '-csp-guide', '../prob_examples/public_examples/CSP/Tickets/WilliamsCSPB/AirportsCtrl.csp',
  %'-p', 'csp_event_visible_to_user', 'false', % used to be csp_event probcli option
   '-ltlformulat', 'G not [link.AMS.AMS]', '-strict'], 'Checking ltl formulas for CSP-M || B models.').
cli_testcase(937, [b_test, csp_test,ltl,cspb], ['../prob_examples/public_examples/CSP/Tickets/WilliamsCSPB/Airports.mch', '-csp-guide', '../prob_examples/public_examples/CSP/Tickets/WilliamsCSPB/AirportsCtrl.csp', '-ltlformulat', 'G not [link(AMS,AMS)]', '-strict'], 'Checking ltl formulas for CSP-M || B models.').
cli_testcase(938, [error_checks], ['-mc', 1000, '-expcterr', 'type_error', '-expcterr', 'load_main_file', '-strict', '../prob_examples/public_examples/B/ErrorMachines/MachineWithParametersErr1.ref'], 'Check for error: Wrong number of parameters in refinement.').
cli_testcase(939, [error_checks], ['-mc', 1000, '-expcterr', 'type_error', '-expcterr', 'load_main_file', '-strict', '../prob_examples/public_examples/B/ErrorMachines/MachineWithParametersErr2.ref'], 'Check for error: Wrong parameters in refinement.').
cli_testcase(940, [cbc], ['-cbc', all, '-p','CLPFD', 'TRUE', '-strict', '../prob_examples/public_examples/B/Tickets/Treharne1/House_Tracker16.mch'], 'Check that we quickly see that there is no counterexample.').
cli_testcase(941, [cbc], ['-cbc', all, '-p','CLPFD', 'TRUE', '-strict', '../prob_examples/public_examples/B/CBC/ReservationLivre.mch', '-expcterr',cbc], 'Check that we find counterexample with larger scope.').
cli_testcase(942, [plugin_units,wd,case], ['../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/Case.mch', '-plugin', 'units', '-strict', '-expcterr', 'well_definedness_error', '-pp', '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/Case_internal.mch','-p','plugin(units,fixpoint_halt_on_error)','false'], '').
cli_testcase(943, [parser,typechecker_test], ['../prob_examples/public_examples/B/Pragmas/UnrecognisedPragma.mch','-strict','-expcterr',load_main_file,'-expcterrpos',parse_error, 7, 5],'Check Pragma error message').
cli_testcase(944, [dot], ['-evalt','r:1..3-->>1..3 & r(1)/=1 & r(1)/=2 & r(2)/=2', '-evaldot', '../prob_examples/public_examples/Eval/evalout.dot','-strict'],'').
cli_testcase(945, [eventb_test, theories], ['-t', '-strict', '../prob_examples/public_examples/EventBPrologPackages/Theory/TestBoolOps_mch.eventb'], 'Check BoolOps Theory.').
cli_testcase(946, [cbc,smt_test], ['../prob_examples/public_examples/B/Tickets/CBC_Refinement/M2_mch.eventb', '-p', 'CLPFD', 'TRUE', '-cbc', all, '-strict', '-expcterr', 'cbc', '-p', 'SMT', 'TRUE'], 'Check counter-example found in SMT mode').
cli_testcase(947, [cbc,smt_test], ['../prob_examples/public_examples/B/Tickets/CBC_Refinement/M2_mch.eventb', '-p', 'CLPFD', 'TRUE', '-cbc', 'RemAuth', '-strict', '-expcterr', 'cbc', '-p', 'SMT', 'TRUE'], 'Check counter-example found in SMT mode').
cli_testcase(948, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/WaterBoiler.csp', '-csp_eval', 'member(fill.3,{|fill|}) == true', '-strict'], 'Evaluating CSP-M expressions.').
cli_testcase(949, [tickets], ['-assertions', '-strict', '../prob_examples/public_examples/B/Tickets/PrjMemCheck/PrjTestsOK.mch',
 '../prob_examples/public_examples/B/Tickets/PrjMemCheck/PrjTestsOK_For_AtelierB.mch'], 'Check that we prj1/prj2 work correctly.').
cli_testcase(950, [tickets,wd], ['-assertions', '-strict', '../prob_examples/public_examples/B/Tickets/PrjMemCheck/PrjTestsKO1.mch', '-expcterr', 'well_definedness_error', '-expcterr', 'check_assertions', '-p', 'TRY_FIND_ABORT', 'TRUE'], 'Check that we prj1/prj2 detect wd errors.').
cli_testcase(951, [tickets,wd], ['-assertions', '-strict', '../prob_examples/public_examples/B/Tickets/PrjMemCheck/PrjTestsKO2.mch', '-expcterr', 'well_definedness_error', '-expcterr', 'check_assertions', '-p', 'TRY_FIND_ABORT', 'TRUE'], 'Check that we prj1/prj2 detect wd errors.').
cli_testcase(952, [tickets,wd], ['-assertions', '-strict', '../prob_examples/public_examples/B/Tickets/PrjMemCheck/PrjTestsKO3.mch', '-expcterr', 'well_definedness_error', '-expcterr', 'check_assertions', '-p', 'TRY_FIND_ABORT', 'TRUE'], 'Check that we prj1/prj2 detect wd errors.').
cli_testcase(953, [tickets,wd], ['-assertions', '-strict', '../prob_examples/public_examples/B/Tickets/PrjMemCheck/PrjTestsKO4.mch', '-expcterr', 'well_definedness_error', '-expcterr', 'check_assertions', '-p', 'TRY_FIND_ABORT', 'TRUE'], 'Check that we prj1/prj2 detect wd errors.').
cli_testcase(954, [eventb_test, theories], ['-t','-strict','../prob_examples/public_examples/EventBPrologPackages/Theory/DumbDataTheoryModel.eventb'],'Check model that uses user defined datatypes via the Theory Plugin.').
cli_testcase(955, [eventb_test, theories], ['-mc',100,'-strict','../prob_examples/public_examples/EventBPrologPackages/Theory/NonTrivialTypeParameters.eventb'],'Theory Plugin: Datatypes whose type parameters are not only type expressions').
cli_testcase(956, [eventb_test], ['-mc',1000,'-strict','../prob_examples/public_examples/EventBPrologPackages/Puzzles/Hanoi_mch_3s_6d.eventb', '-noass', '-cc', 731, 2186, '-p', 'CLPFD', 'TRUE'],'Check that constraint solver not perplexed by use of card() for last').
cli_testcase(957, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/DifferentUnitNotations.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/DifferentUnitNotations_internal.mch'], '').
cli_testcase(958, [eventb_test], ['-t', '-p', 'CLPFD', 'FALSE', '-p', 'PROOF_INFO', 'FALSE'  , '-p', 'TIME_OUT', '200',  '-expcterr', 'time_out_for_invariant',
'-strict','../prob_examples/public_examples/EventBPrologPackages/ProofDirected/ComplicatedProvenInvariant2_v2.eventb'],'Check that Time Out occurs without Proof Info usage. (CLPFD disabled because it causes an overflow on the 32 bit testbots').
cli_testcase(959, [eventb_test], ['-mc',100, '-t',  '-p', 'TIME_OUT', '200', '-strict','../prob_examples/public_examples/EventBPrologPackages/ProofDirected/ComplicatedProvenInvariant2_v2.eventb'],'Check that Proof Info works').
cli_testcase(960, [eventb_test], ['-mc',0,  '-p', 'PROOF_INFO', 'FALSE'  , '-p', 'TIME_OUT', '200', '-expcterr', 'time_out_for_assertions', '-strict','../prob_examples/public_examples/EventBPrologPackages/ProofDirected/ComplicatedProvenInvariant_INITCheck_mch.eventb'],'Check that Proof Info works also for INITIALISATION').
cli_testcase(961, [eventb_test], ['-mc',100,  '-p', 'PROOF_INFO', 'TRUE'  , '-p', 'TIME_OUT', '200', '-strict','../prob_examples/public_examples/EventBPrologPackages/ProofDirected/ComplicatedProvenInvariant_INITCheck_mch.eventb'],'Check that Proof Info works also for INITIALISATION').
cli_testcase(962, [eventb_test], ['-mc',100,  '-p', 'PROOF_INFO', 'TRUE'  , '-p', 'TIME_OUT', '200', '-strict','../prob_examples/public_examples/EventBPrologPackages/ProofDirected/ComplicatedProvenDynamicTheorem_mch.eventb'],'Check that Proof Info works also for dynamic theorems').

cli_testcase(963, [csp_test,ltl], ['../prob_examples/public_examples/CSP/other/Ivo/WaterBoiler.csp', '-ltlformulat', '[open] => ((X [fill._]) or (X [close])) ', '-strict'],'Check LTL formula in File.').
cli_testcase(964, [eventb_test,wd], ['-mc',100,  '-p', 'PROOF_INFO', 'TRUE'  , '-expcterr', well_definedness_error,  '-expcterr', invariant_violation, '-strict','../prob_examples/public_examples/EventBPrologPackages/ProofDirected/TestProofDirected_WD_Inv_mch.eventb'],'Check that Proof Info works well in conjunction with ill-defined invariants').
cli_testcase(965, [eventb_test,wd], ['-mc',100,  '-p', 'PROOF_INFO', 'TRUE'  , '-expcterr', well_definedness_error,  '-expcterr', assertion_violation, '-strict','../prob_examples/public_examples/EventBPrologPackages/ProofDirected/TestProofDirected_WD_Thm_mch.eventb'],'Check that Proof Info works well in conjunction with ill-defined theorems').
cli_testcase(966, [eventb_test], ['-mc',100,  '-noass', '-nodead', '-p', 'PROOF_INFO', 'FALSE'  , '-expcterr', invariant_violation, '-strict','../prob_examples/public_examples/EventBPrologPackages/ProofDirected/TestProofDirected_WD_Thm_mch.eventb'],'Check that Proof Info works well in conjunction with ill-defined theorems and -noass checking').
cli_testcase(967, [eventb_test], ['-mc',100,  '-noass', '-nodead', '-p', 'PROOF_INFO', 'TRUE'  , '-expcterr', invariant_violation, '-strict','../prob_examples/public_examples/EventBPrologPackages/ProofDirected/TestProofDirected_WD_Thm_mch.eventb'],'Check that Proof Info works well in conjunction with ill-defined theorems and -noass checking').
cli_testcase(968, [eventb_test], ['-mc',100,  '-noass', '-nodead', '-p', 'PROOF_INFO', 'TRUE'  , '-expcterr', 'model_check_incomplete', '-strict','../prob_examples/public_examples/EventBPrologPackages/ProofDirected/TestProofDirected_FalseThm_mch.eventb'],'Check that Proof Info works well in conjunction with ill-defined theorems and -noass checking').
cli_testcase(969, [b_test,recursion,external], ['-assertions', '-strict','../prob_examples/public_examples/B/Tickets/Bracknell1/C0.mch'],'Check that no recursion detected').
cli_testcase(970, [b_test,recursion,external], ['-init', '-strict','../prob_examples/public_examples/B/Tickets/Bracknell1/C0_ctx.eventb'],'Check that no recursion detected').
cli_testcase(971, [b_test], ['../prob_examples/public_examples/B/Tester/Prj1NATURAL.mch', '-t', '-mc', '1000', '-strict'], 'Check that prj1(NATURAL,...) not expanded').
cli_testcase(972, [cbc], ['../prob_examples/public_examples/B/CBC/plavis_MemoryLoad_SP_14_TestComp.mch', '-p', 'CLPFD', 'TRUE', '-p', 'MAXINT', 3, '-p', 'MININT', -1, '-strict'], 'Check that partitioning into components works').
cli_testcase(973, [cbc], ['../prob_examples/public_examples/B/CBC/ComponentPartitioningTest.mch', '-p', 'CLPFD', 'TRUE', '-init', '-strict'], 'Check that partitioning into components works').
cli_testcase(974, [cbc], ['../prob_examples/public_examples/B/CBC/PartitionConstraintAssertionCheck.mch', '-p', 'CLPFD', 'TRUE', '-cbc_deadlock', '-expcterr', 'cbc_deadlock_check', '-strict', '-v'], 'Check that partitioning into components works').
cli_testcase(975, [por], ['../prob_examples/public_examples/B/Ivo/PruningActionsTest1.mch', '-mc', '1000000', '-p', 'por', ample_sets, '-expcterr', invariant_violation, '-strict'], 'POR test.').
cli_testcase(976, [por], ['../prob_examples/public_examples/B/Ivo/PruningActionsTest2.mch', '-mc', '1000000', '-p', 'por', ample_sets, '-expcterr', deadlock, '-strict'], 'POR test (deadlock and invariant violation).').
cli_testcase(977, [por], ['../prob_examples/public_examples/B/Ivo/BridgeTransitions.mch', '-mc', '1000000', '-cc', 55, '_', '-check_complete_operation_coverage', '-p', 'por', ample_sets, '-get_min_max_coverage', '../prob_examples/public_examples/B/Ivo/BridgeTransitions_por.cov', '-strict'], 'POR test (no state space reductions).').
cli_testcase(978, [por], ['../prob_examples/public_examples/B/Ivo/CycleDetections.mch', '-mc', '1000000', '-p', 'por', ample_sets, '-expcterr', invariant_violation, '-strict'], 'POR test.').
% commit 484a10c2cdd8cc4b13f1be8f30e8eac7be5a30c3 from 23 Oct 2013 broke 979-982; CSE optimization should recover performance
cli_testcase(979, [private,data_validation], ['../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_2_ori.mch', '-animate_all', '-strict', '-cc', 113, 112, '-p', 'TIME_OUT', 35000, '-p', 'CLPFD', 'FALSE', '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647, '-his', '../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_2_ori.his','-expcterr', invariant_violation], 'Rule Test 4 of Alstom compilation project'). % increased time-out from 15000; added -p CLPFD FALSE
cli_testcase(980, [private,data_validation,codespeed], ['../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_SIGAREA_0024_ori.mch', '-animate_all', '-strict', '-cc', 40, 39, '-p', 'TIME_OUT', 35000, '-p', 'CLPFD', 'FALSE', '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647, '-his', '../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_SIGAREA_0024_ori.his'], 'Rule Test 5 of Alstom compilation project'). % increased time-out from 15000; added -p CLPFD FALSE
cli_testcase(981, [private,data_validation,codespeed], ['../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_Route_0001ori.mch', '-animate_all', '-strict', '-cc', 97, 96, '-p', 'TIME_OUT', 28000, '-p', 'CLPFD', 'FALSE', '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647, '-his', '../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_Route_0001ori.his'], 'Rule Test 1 of Alstom compilation project'). % increased time-out from 15000; added -p CLPFD FALSE
cli_testcase(982, [private,data_validation,alstom], ['../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_1_ori.mch', '-animate_all', '-strict', '-cc', 48, 47, '-p', 'TIME_OUT', 55000, '-p', 'CLPFD', 'FALSE', '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647, '-his', '../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_1_ori.his','-expcterr', invariant_violation], 'Rule Test 3 of Alstom compilation project'). % increased time-out from 45000; added -p CLPFD FALSE
cli_testcase(983, [private,data_validation,alstom], ['../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_ESA_TRIGGER_0009_ori.mch', '-strict', '-cc', 112, 111, '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647, '-t'], 'Rule Test 2 of Alstom compilation project').
cli_testcase(984, [private,symmetry_test], ['../prob_examples/examples/B/BossaAcadie/scheduler.mch', '-strict', '-cc', 145, 1935, '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647, '-mc', 9000, '-p', 'MAX_OPERATIONS', 20, '-p', 'SYMMETRY_MODE', hash, '-expcterr', 'model_check_incomplete'], 'Bossa example'). % was 1936 transitions; now 1935 with INITIALISATION arguments derived from target state
cli_testcase(985, [cbc], ['../prob_examples/public_examples/B/Mathematical/GraphIso/HardGraph_Nr5.mch', '-strict', '-cc', 3, 3, '-p', 'TIME_OUT', 5000,  '-mc', 20, '-p', 'MAX_OPERATIONS', 4, '-p', 'CLPFD', 'TRUE', '-expcterr', 'model_check_incomplete'], 'Hard graph; isomorphism detected means no deadlock').
cli_testcase(986, [eventb_test], ['-t','-mc',100,'-strict','../prob_examples/public_examples/EventBPrologPackages/Puzzles/MRushHour_new.eventb','-expcterr', invariant_violation],'Check that new Proof Info works with wd in guards/actions accepted').
cli_testcase(987, [por], ['../prob_examples/public_examples/EventBPrologPackages/Advance/CAN_Bus/CB3FSMM_mch_v2.eventb', '-mc', '1000000', '-noass', '-nogoal', '-p', 'por', ample_sets, '-strict'], 'Testing if coverages of the constants and values of the generated reduced state space are matching with the coverages of the constants and values of the full state space.').
cli_testcase(988, [csp_test,dot], ['../prob_examples/public_examples/CSP/simple/Copy.csp', '-df', '-mc', '1000', '-cc', '14', '24', '-strict','-dot','signature_merge','../prob_examples/public_examples/CSP/simple/Copy_spmdot.dot','-dot', 'dfa_merge','../prob_examples/public_examples/CSP/simple/Copy_spddot.dot', '-pref_group', dot_colors, classic, '-p','DOT_PEN_WIDTH',1], 'Test reduced state space dot output for CSP. (DF search to avoid different dot graphs due to order of exploration)').
cli_testcase(989, [eventb_test], ['-assertions','-strict','../prob_examples/public_examples/EventBPrologPackages/PerformanceTests/LargeConstants_ctx.eventb','-pp', '../prob_examples/public_examples/EventBPrologPackages/PerformanceTests/LargeConstants_ctx_internal.mch'],'Check that set comprehensions converted properly').
cli_testcase(990, [b_test,smt_test], ['-evalt', 'ran(%v0.(v0: 0..30001 | v0)) = 0..30001', '-p', 'SMT', 'TRUE', '-strict'],'Checking non-regresson against issue reported by Medeiros Jr on 28.1.2013').
cli_testcase(991, [tickets,csp_test], ['../prob_examples/public_examples/CSP/Tickets/RecursiveInclude/Procs.csp', '-mc', 100000, '-t', '-strict'], 'Testing of recursive and local (by means of no absolute path files must be given in order to include a CSP model to another CSP model) including for CSP models.').
cli_testcase(992, [b_test], ['-evalt', '{x|  x = x mod 33 & x = x mod 44} = 0..32', '-p', 'CLPFD', 'TRUE', '-strict'],'Ensure mod constraint propagation works').
cli_testcase(993, [csp_test], ['../prob_examples/examples/CSP/Kharmeh_PhD/4-isa-system-evaluation.csp', '-mc', 2000, '-cc', 265, 431, '-strict'],'Non trivial test case').
cli_testcase(994, [b_test,case], ['../prob_examples/public_examples/B/FeatureChecks/CaseStmt.mch', '-t', '-mc', 200, '-cc', 4, 13, '-strict'],'Check case statement with else').
cli_testcase(995, [laws,rel_fnc], ['../prob_examples/public_examples/B/Laws/FunLawsWithLambda.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict','-p', 'MEMO', 'TRUE', '-p', 'CLPFD', 'TRUE'], 'Laws this time with MEMO and CLPFD').
cli_testcase(996, [b_test,smt_test], ['-evalt', '{x|x/2 : 20..30} = 40..61', '-evalt', '{x|x/20 : 22..23} = 440..479', '-evalt', 'ran(%v0.(v0: 0..20000 | [v0 mod 2/1,v0 mod 4/2,v0 mod 8/4,v0 mod 16/8,v0 mod 32/16,v0 mod 64/32,v0 mod 128/64])) = 1..7 --> 0..1', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE', '-strict'],'Ensure division constraint propagation works').
cli_testcase(997, [csp_test,refusal_trace_refinement], ['../prob_examples/public_examples/CSP/RefusalBasedRefinement/RefusalBasedRefinement_simple.csp', '-assertions', '-strict'],'Simple test case for refusal based refinement checking (CSP).').
cli_testcase(998, [csp_test,refusal_trace_refinement], ['../prob_examples/public_examples/CSP/RefusalBasedRefinement/RefTestRoscoeBook.csp', '-assertions', '-strict'],'Test case for refusal based refinement checking (CSP) from Roscoe\'s book. Note: Difference in the results delivered by FDR and ProB of the commented out assertion "P2 [R= P1".').
cli_testcase(999, [eventb_test], ['../prob_examples/public_examples/B/Tickets/197/robot_ctx02_ctx.eventb', '-init', '-properties', '-p', 'DEFAULT_SETSIZE', 2, '-strict'],'Check WD Axiom loading').
cli_testcase(1000, [b_test], ['-evalt', 'x \\/ y \\/ z <: 1011..1013 & card(x) > card(y) & card(y)>card(z)', '-evalt', '{x,y,z|x \\/ y \\/ z <: 21011..21012 & card(x) > card(y) & card(y)>card(z)} = {(({21011,21012}|->{21011})|->{}),(({21011,21012}|->{21012})|->{})}', '-p', 'CLPFD', 'TRUE', '-strict'], 'Ensure that domains propagated to union arguments.').
cli_testcase(1001, [b_test], ['-evalt', '{x|x <| [1,2,3] = [1,2] & card(x)=2} = {{1,2}}',
 '-evalt', 'x <| s = [1,2,3] \\/ {29|->29}  & x <: 1..1000 & s = %i.(i:1..300|i)',
 '-evalt', 's |> x = [1,2,3] \\/ {29|->29}  & x <: 1..1000 & s = %i.(i:1..300|i)',
 '-evalt', '{x| [1,4,3]|>x = [1,4] & card(x)=2}={{1,4}}',
 '-evalf', 's |> x = [1,2,3] \\/ {29|->29}  & x <: 2..3000 & s = %i.(i:1..900|i) ',
 '-evalt', '{x|#s.(x <| s = [1,2,3] \\/ {29|->29}  & x <: 1..30 & s = %i.(i:1..300|i))} = {{1,2,3,29}}',
 '-evalt', '{z,x|[1,2,z] |>> x = [1,2]  & z:12..13 & x<:11..12}={(12|->{11,12}),(12|->{12})}',
 '-evalt', 'x <<| [11,12,13,14,z] = [11,12,13,14]  & z:12..13 ',
 '-evalt', 'x <| [111,222,333,444,555]  = [111,222,333,444]',
 '-evalt', 'x <<| [111,222,333,444,555]  = [111,222,333,444]',
 '-evalt', '[111,222,333,444,555] |>x  = [111,222,333,444]',
 '-evalt', '[111,222,333,444,555] |>>x  = [111,222,333,444]',
 '-evalt', '{x|{x} <<| [111,222,333,444,555,666]  = [111,222,333,444,555]} = {6}',
 '-evalt', '{x| [111,222,333,444,555,666] |>> {x}  = [111,222,333,444,555]} = {666}',
 '-evalt', '{x| {x} <| [111,222,333,444,555,666]  = {5|->555}} = {5}',
 '-evalt', '{x| [111,222,333,444,555,666] |> {x}  = {5|->555}} = {555}',
 '-evalt', '{x| [111,222,333,444,555,555] |> {x}  = {5|->555,6|->555}} = {555}',
  '-p', 'CLPFD', 'TRUE', '-strict'], 'Ensure that domain/range restriction properly treated.').
cli_testcase(1002, [cbc], ['-p', 'CLPFD', 'TRUE', '-p', 'MAXINT', 127, '-p', 'MININT', -128,
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/bank-Extraer_SP_29.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/bancoG-Withdraw_NR_22.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/qsee-LoadMemory_SP_11.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/qsee-LoadMemory_SP_12.prob'
  %,'-expcterr', 'eval_string_enum_warning'
  ], 'SetLog tests now succeeding tests').
cli_testcase(1003, [cbc], ['-p', 'CLPFD', 'TRUE', '-p', 'MAXINT', 127, '-p', 'MININT', -128,
  %sc
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/sc-SCAddCat_SP_4.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/sc-SCAddCat_SP_22.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/sc-SCSetLevel_DNF_3.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/sc-SCSetSC_DNF_3.prob',
  %bank
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/bank-CerrarCaja_SP_12.prob',
  %banco G
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/bancoG-Withdraw_NR_18.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/bancoG-Withdraw_NR_2.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/bancoG-Withdraw_NR_24.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/bancoG-Withdraw_NR_25.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/bancoG-Withdraw_NR_4.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/bancoG-Withdraw_NR_5.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/bancoG-Withdraw_NR_8.prob',
  %qsee
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/qsee-LoadMemory_SP_10.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/qsee-LoadMemory_SP_9.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/qsee-TransmitMemoryDumpAux_SP_6.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/qsee-TransmitMemoryDumpAux_SP_8.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/qsee-TransmitMemoryDumpOk21_SP_3.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/qsee-TransmitMemoryDumpOk21_SP_6.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/qsee-TransmitMemoryDumpOk21_SP_8.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/qsee-TransmitScientificData_SP_16.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/qsee-TransmitScientificData_SP_24.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/qsee-TransmitScientificData_SP_8.prob',
  %plavis ; two more missing
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/plavis-TransData_SP_45.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/plavis-TransData_SP_52.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/plavis-MemoryLoad_SP_55.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/plavis-RetrieveEData_SP_23.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/plavis-TransData_SP_13.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/plavis-TransData_SP_24.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/plavis-TransData_SP_3.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/plavis-TransData_SP_31.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/plavis-TransData_SP_38.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/plavis-TransData_SP_39.prob',
  % iae
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/iae-DetectReferenceEvent_NR_103.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/iae-DetectReferenceEvent_NR_146.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/iae-DetectReferenceEvent_NR_148.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/iae-DetectReferenceEvent_NR_149.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/iae-DetectReferenceEvent_NR_18.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/iae-DetectReferenceEvent_NR_56.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/iae-DetectReferenceEvent_NR_58.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/iae-DetectReferenceEvent_NR_59.prob',
  % sensors
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/sensors-KeepMaxReading_SP_5.prob'
  %,'-expcterr', 'eval_string_enum_warning'
  ], 'SetLog tests some originally succeeding tests').
cli_testcase(1004, [cbc_refinement], ['../prob_examples/public_examples/EventBPrologPackages/RefinementChecking/CCounterErr_mch.eventb', '-cbc_refinement', '-expcterr', cbc_refinement], '').
cli_testcase(1005, [cbc_refinement], ['../prob_examples/public_examples/EventBPrologPackages/RefinementChecking/CCounter_mch.eventb', '-cbc_refinement'], '').
cli_testcase(1006, [cbc_refinement], ['../prob_examples/public_examples/EventBPrologPackages/RefinementChecking/CEnterLeave_mch.eventb', '-cbc_refinement'], '').
cli_testcase(1007, [cbc_refinement], ['../prob_examples/public_examples/EventBPrologPackages/RefinementChecking/CEnterLeaveErr1_mch.eventb', '-cbc_refinement', '-expcterr', cbc_refinement], '').
cli_testcase(1008, [cbc_refinement], ['../prob_examples/public_examples/EventBPrologPackages/RefinementChecking/CEnterLeave_Rename_mch.eventb', '-cbc_refinement'], '').
cli_testcase(1009, [cbc_refinement], ['../prob_examples/public_examples/EventBPrologPackages/RefinementChecking/CMergeErr_mch.eventb', '-cbc_refinement', '-expcterr', cbc_refinement, '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(1010, [cbc_refinement], ['../prob_examples/public_examples/EventBPrologPackages/RefinementChecking/CMerge_mch.eventb', '-cbc_refinement'], '').
cli_testcase(1011, [cbc_refinement], ['../prob_examples/public_examples/ForDistribution/Refinement/ServerR.ref', '-cbc_refinement'], '').
cli_testcase(1012, [cbc_refinement], ['../prob_examples/public_examples/ForDistribution/Refinement/ServerR_wrong.ref', '-cbc_refinement', '-expcterr', cbc_refinement], '').
cli_testcase(1013, [cbc_refinement,imp], ['../prob_examples/public_examples/B/Implementations/ConcreteVars/MI.imp', '-cbc_refinement'], '').
cli_testcase(1014, [b_test], ['../prob_examples/public_examples/B/Mathematical/FibMemo.mch', '-mc', 120, '-p', 'CLPFD', 'TRUE', '-strict', '-expcterr', 'deadlock', '-cc', 100, 99], 'Test that no CLPFD overflow occurs'). % no warning is generated for clpfd overflows at the moment; so the test is not really that useful
cli_testcase(1015, [cbc_refinement,imp], ['../prob_examples/public_examples/B/Implementations/Copy/ReturnEqualI.imp', '-cbc_refinement'], '').
cli_testcase(1016, [cbc_refinement,imp], ['../prob_examples/public_examples/B/Implementations/Division/DivisionImplOk.imp', '-cbc_refinement'], '').
cli_testcase(1017, [cbc_refinement,imp], ['../prob_examples/public_examples/B/Implementations/Division/DivisionImplWrongCondition.imp', '-cbc_refinement', '-expcterr', cbc_refinement], '').
cli_testcase(1018, [b_test], ['../prob_examples/public_examples/B/RefinementChecking/Schedular/Schedular1.mch', '-refchk', '../prob_examples/public_examples/B/RefinementChecking/Schedular/Schedular0_refine_spec.P', '-expcterr', 'refinement_check_fails', '-strict'], 'Checking for expected refinement error. Example taken from the paper: Automatic Refinement Checking for B.').
cli_testcase(1019, [cbc_refinement], ['../prob_examples/public_examples/B/Implementations/FTP/FileTransfer1.ref', '-cbc_refinement'], '').
cli_testcase(1020, [cbc_refinement], ['../prob_examples/public_examples/B/Implementations/FTP/FileTransfer2.ref', '-cbc_refinement'], '').
cli_testcase(1021, [cbc_refinement,imp], ['../prob_examples/public_examples/B/Implementations/Mul/MulImp.imp', '-cbc_refinement'], '').
cli_testcase(1022, [cbc_refinement,imp], ['../prob_examples/public_examples/B/Implementations/Mul/MulImpErr.imp', '-cbc_refinement', '-expcterr', cbc_refinement], '').
cli_testcase(1023, [cbc_refinement,imp], ['../prob_examples/public_examples/B/Implementations/Power/AExpBImpl.imp', '-cbc_refinement'], '').
cli_testcase(1024, [cbc_refinement,imp], ['../prob_examples/public_examples/B/Implementations/Simple10/Simple10Imp.imp', '-cbc_refinement'], '').
cli_testcase(1025, [cbc_refinement,imp], ['../prob_examples/public_examples/B/Implementations/Simple10/Simple10ImpErr.imp', '-cbc_refinement', '-expcterr', cbc_refinement, '-p', 'CLPFD', 'TRUE'], '').
cli_testcase(1026, [cbc_refinement,imp], ['../prob_examples/public_examples/B/Implementations/Simple10/Simple10ImpErr2.imp', '-cbc_refinement', '-expcterr', cbc_refinement], '').
cli_testcase(1027, [cbc_refinement], ['../prob_examples/public_examples/B/RefinementChecking/ServerR.ref', '-cbc_refinement'], '').
cli_testcase(1028, [cbc_refinement], ['../prob_examples/public_examples/B/RefinementChecking/ServerR_wrong.ref', '-cbc_refinement', '-expcterr', cbc_refinement], '').
cli_testcase(1029, [cbc_refinement], ['../prob_examples/public_examples/B/RefinementChecking/SwapR.ref', '-cbc_refinement'], '').
cli_testcase(1030, [b_test,infinite,sequences, theories], ['../prob_examples/public_examples/B/FunctionalProgramming/SequenceTheory.mch', '-t', '-assertions'], 'Check infinite functions and sets').
cli_testcase(1031, [b_test], ['../prob_examples/public_examples/B/FeatureChecks/DeferredSetAndConstants.mch', '-init',
 '-evalt', 'ID3=3 & ID1:ID2 & ID0:ID2 & (xx=ID5 => not(xx=ID6)) & (xx=ID6 => not(xx=ID5))'],'Check that name clashes between deferred set ids and other ids handled gracefully').
cli_testcase(1032, [csp_test,refusal_trace_refinement], ['../prob_examples/public_examples/CSP/RefusalBasedRefinement/RefusalBasedRefinement_DIV.csp', '-assertions', '-strict'],'Simple test case for refusals-divergence based refinement checking (CSP).').
cli_testcase(1033, [dot], ['../prob_examples/public_examples/B/Benchmarks/scheduler.mch', '-bf', '-mc', 1000, '-strict', '-dotexpr', transition_diagram, 'card(active)', '../prob_examples/public_examples/B/Benchmarks/scheduler_transition_diagram.dot'],'Check Transition Diagram for Expression.').
cli_testcase(1034, [tickets], ['../prob_examples/public_examples/B/Tickets/leuschel4/SetBecomeElementOf.mch', '-t'], 'check no error with ::').
cli_testcase(1035, [tickets], ['../prob_examples/public_examples/B/Tester/SubstSetInTests.mch', '-t', '-p', 'CLPFD' ,'TRUE', '-p', 'SYMBOLIC', 'TRUE'], 'check no error with ::').
cli_testcase(1036, [tickets,smt_test], ['../prob_examples/public_examples/B/PerformanceTests/LargeRelationSetTests.mch', '-t', '-p', 'CLPFD' ,'TRUE', '-p', 'SMT', 'TRUE',  '-p', 'SYMBOLIC', 'TRUE'], 'check no residue in symbolic mode due to lambda_res projected away').
cli_testcase(1037, [b_test], ['-evalt', '{x|x**4 = 160000}={-20,20}', '-evalt', '{x|x**3 = -125000} = {-50}', '-evalt', 'x**4 = y & y=160000', '-evalt', '{x,y|x**4 = y & y:140000..160004}={(-20|->160000),(20|->160000)}', '-p', 'CLPFD', 'TRUE', '-strict'], 'Ensure exponentation constraint propagation.').
cli_testcase(1038, [b_test], ['-evalt', 'x mod 100000 >= 99999',
  '-evalf', 'x mod 100000 > 99999',
  '-evalt', '(x+2) mod 100000 >= 99999',
  '-evalt', '(x-2) mod 100000 >= 99999 & x>2',
  '-evalt', '{x,y|x mod y = x & y : 10..11 & x>0 & x+3>y}={(8|->10),(9|->10),(9|->11),(10|->11)}',
  '-evalt', 'x:NATURAL & x mod 10000 = 9999',
  '-evalt', 'x:NATURAL & x mod 10000 = 998',
  '-evalf', 'x:NATURAL & x mod 10000 = 10000',
  '-evalf', 'x:NATURAL & x mod 10000 = 10001',
 '-p', 'CLPFD', 'TRUE', '-strict'
%, '-expcterr', eval_string_enum_warning
], 'Ensure modulo constraint propagation.').
cli_testcase(1039, [cbc_refinement,imp], ['../prob_examples/public_examples/B/Implementations/MaxDevelopment/MaxImp.imp', '-cbc_refinement'], '').
cli_testcase(1040, [cbc_refinement,imp], ['../prob_examples/public_examples/B/Implementations/MaxDevelopment/MaxImpVector.imp', '-cbc_refinement'], '').
cli_testcase(1041, [cbc_refinement,wd], ['../prob_examples/public_examples/B/Implementations/MaxDevelopment/MaxR.ref', '-cbc_refinement', '-expcterr', 'well_definedness_error'], 'Max is applied to an empty set here - wd error in constraint solver').
cli_testcase(1042, [cbc_refinement], ['../prob_examples/public_examples/B/Implementations/MaxDevelopment/MaxR2.ref', '-cbc_refinement'], '').
cli_testcase(1043, [pge], ['../prob_examples/public_examples/B/Benchmarks/CAN_BUS_tlc.mch', '-mc', '1000000', '-bf', '-p', pge, full, '-get_coverage_information', '../prob_examples/public_examples/B/Benchmarks/CAN_BUS_tlc.cov', '-strict'], 'Testing for identical coverage information after the PGE optimization.').
cli_testcase(1044, [pge], ['../prob_examples/public_examples/B/EventB/EventB_Projekt/lift_solution.mch', '-mc', '1000000', '-bf', '-p', pge, full, '-p', timeout_cbc_analysis, 60000, '-get_coverage_information', '../prob_examples/public_examples/B/EventB/EventB_Projekt/lift_solution_Coverage_pge.txt', '-strict'], 'Testing for identical coverage information after the PGE optimization.').
cli_testcase(1045, [pge_fast], ['../prob_examples/examples/RodinModels/Deploy/Bosch/CruiseControl/CrsCtl_m4.eventb', '-mc', '1000000', '-bf', '-p', pge, full, '-get_coverage_information', '../prob_examples/examples/RodinModels/Deploy/Bosch/CruiseControl/CrsCtl_m4_Coverage_pge.txt', '-strict', '-expecterr', 'model_check_incomplete'], 'Testing for identical coverage information after the PGE optimization.').
cli_testcase(1046, [por], ['../prob_examples/public_examples/B/Ivo/PruningActionsTest2.mch', '-mc', '1000000', '-noinv', '-p', 'por', ample_sets, '-expcterr', deadlock, '-strict'], 'POR test (deadlock).').
cli_testcase(1047, [por], ['../prob_examples/public_examples/B/Ivo/ActivatingDepAction.mch', '-mc', '1000000', '-nodead', '-p', 'por', ample_sets, '-expcterr', invariant_violation, '-strict'], 'POR test (invariant violation).').
cli_testcase(1048, [eventb_test], ['../prob_examples/public_examples/B/Tickets/Georghe1/seq_m_mch_concrete_constants.eventb', '-mc', '100', '-expecterr', 'deadlock', '-expecterr', 'event_error:append:action_not_executable'], 'Can now also be initialized with ProB and enumeration_warning exception detection.').
cli_testcase(1049, [eventb_test], ['../prob_examples/public_examples/B/Tickets/Georghe1/seq_m_mch_abstract_constants.eventb', '-mc', '100', '-expecterr', 'deadlock', '-expecterr', 'event_error:append:action_not_executable'], 'Same as 1048, no with abstract constants - should not time out').
cli_testcase(1050, [eventb_test,cbc], ['../prob_examples/examples/EventBPrologPackages/Advance_WP2/SubstationMeter_2603_2013.eventb', '-cbc', 'all', '-strict'], 'Test that proof info correctly used; machine fully proven.').
cli_testcase(1051, [laws], ['../prob_examples/public_examples/B/Laws/FinitenessLaws.mch', '-init', '-assertions', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE', '-p', 'CLPFD', 'TRUE'], 'Various laws about finitness of sets.').
cli_testcase(1052, [eventb_test,cbc], ['../prob_examples/examples/EventBPrologPackages/Advance_WP2/SubstationMeter_finite_mch.eventb',  '-strict', '-mc', 1000, '-p', 'CLPFD', 'TRUE'], 'Test multi-level animation.').
cli_testcase(1054, [b_test, let], ['../prob_examples/public_examples/B/FeatureChecks/NestedAnys.mch', '-t', '-mc', 100, '-cc', 6, 7], 'Check LET extraction').
cli_testcase(1055, [csp_test], ['../prob_examples/public_examples/CSP/ucsexamples/chapter04/abp.csp','-assertions','-strict'], 'Checking CSP-M assertions.').
cli_testcase(1056, [csp_test,refusal_trace_refinement], ['../prob_examples/public_examples/CSP/other/Williams/vg.csp','-assertions','-strict'], 'Checking CSP-M assertions of refusal trace refinements.').
cli_testcase(1057, [tickets], ['../prob_examples/public_examples/B/Tickets/LoopPartition/bug.eventb', '-init','-expecterr', 'setup_constants_fails'], 'Check that partition(SET,{A},{B},{A,B}) does not lead to a loop in find_inequal_global_set_identifiers').
cli_testcase(1058, [csp_test,refusal_trace_refinement], ['../prob_examples/public_examples/CSP/Tickets/WilliamsCSP/protocol_short.csp','-assertions','-t','-strict'], 'Expanding partial constructors inside of closures (see declaration: Network_Data_Fwd = {| CData |} in file).').
cli_testcase(1059, [b_test], ['-evalt', '{X,F|bool(X /\\ {1} = {} & X \\/ {1} = 1..50000)=F & F:{TRUE,FALSE} & F=bool({1} <: {2,1})} = { (2..50000,TRUE) }',
 '-evalt', '{X,F|bool(X /\\ {1} = {} & X \\/ {1} = 1..50000)=F & F:{TRUE,FALSE} & F=bool({1} <: {2,1})} <<: { (2..50000,TRUE), ({},TRUE) }',
 '-evalt', 'card(%(X,F).(bool(X /\\ {1} = {} & X \\/ {1} = 1..50000)=F & F:{TRUE,FALSE} & F=bool({1} <: {2,1})|F))=1',
'-evalt', '{X| X /\\ {1} = {} & X \\/ {1} = 1..20} = {2..20}'], 'Check partition detection inside expressions').
cli_testcase(1060, [b_test], ['../prob_examples/public_examples/B/FeatureChecks/PartitionDetection.mch', '-t'], 'Check partition detection inside substitutions').
cli_testcase(1061, [b_test], ['-evalt', '#f.(f : 1..11 --> 1..10 & dom(f~) = 1..10)', '-evalt', '#f.(f:1..49 -->> 1..49)', '-evalf', '#f.(f:1..19 -->> 1..20)',
  '-evalt', '#f.(f : 1..10 --> 1..10 &  ran(f) = 1..10)'],'Check that range/domain/-->> scale properly').
cli_testcase(1062, [b_test,infinite], ['-evalt', '#(i,f).(i = %x.(x:INTEGER|x+1) & f = (i;i;i;i;i;i;i;i) & 2008 = f(2000))', '-evalt', '#(i,f).(i = %x.(x:INTEGER|x*x) & f = (i;i) & f[{1,2,4,8,101}]={1,2**4,4**4,8**4,104060401})',
'-evalt', '#(i,m,f).(i = %x.(x:INTEGER|x+1) & m = %y.(y:NATURAL|y*y) & f = (i;m;i) & 1+2001**2 = f(2000))',
'-evalt', 'i = %(x,y).(x:INTEGER&y:INTEGER|x+y) & m = %y.(y:NATURAL|y*y) & f = (i;m) & 2500 = f(20,30)',
'-evalt', 'i = %(x,y).(x:INTEGER&y:INTEGER|x+y) & m = %y.(y:NATURAL1|bool(y>100)) & f = (i;m) & TRUE = f(20,90)',
'-evalt', 'poly = %(x,y,z).(x:INTEGER&y:INTEGER&z:INTEGER|x*x+y*y+z*z) & dec = %v.(v:NATURAL1|v-1) & f = (poly ; dec) & f(1,2,3)=13',
'-evalt', 'max0 = %x.(x<:INTEGER|max(x)-1) & inc = %v.(v:INTEGER|v+1) & m = (max0;inc) & m(1..100 \\/ {4321})=4321',
'-evalt', 'max0 = %x.(x<:INTEGER|max(x)-2) & inc = %v.(v:INTEGER|v+1) & is=%s.(s<:INTEGER|inc[s]) & m = (is;max0;inc) & m(1..100 \\/ {4321})=4321'
],'Check that relational composition of infinite function works').
cli_testcase(1063, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/FeatureTests/ThmInGuard_mch.eventb', '-mc', 100, '-strict'], 'Test that theorem in guards can be loaded.').
cli_testcase(1064, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/FeatureTests/EventWithSlash_mch.eventb', '-mc', 100, '-expcterr', invariant_violation, '-strict'], 'Test that slash in Event Name does not cause problems.').

cli_testcase(1065, [b_test], ['-evalt', '{X,Y,Z,V,W,A,B,C} = {1,2,3,4,5,6,7,8} & X>Y & Y>Z & Z>V & V>W & W>A & A>B & B>C',
'-evalf', '{X,Y,Z,V,W,A,B,C} = {1,2,3,4,5,7,8} & X>Y & Y>Z & Z>V & V>W & W>A & A>B & B>C'], 'Test that constraint propagation for set extension works').

cli_testcase(1066, [b_test], ['../prob_examples/public_examples/B/Mathematical/GCD.mch', '-init', '-p', 'CLPFD', 'TRUE', '-p', 'MAXINT', 128, '-evalt', 'half(201)=100', '-eval', 'gcd({10,15})', '-evalt', 'gcdc({1000,750})=250'], 'Check that expression values enumerated'). % gcd({10,15}) requires enumeration

cli_testcase(1067, [cbc], ['../prob_examples/public_examples/B/Benchmarks/scheduler.mch', '-cbc_sequence', 'new;del', '-cbc_sequence', 'new;ready', '-cbc_sequence', 'new;ready,swap'], 'Check that constraints can be solved').
cli_testcase(1068, [cbc], ['../prob_examples/public_examples/B/Benchmarks/scheduler.mch', '-cbc_sequence', 'new;del;del', '-expcterr', cbc_sequence_no_solution_found], 'Check that constraints cannot be solved').
cli_testcase(1069, [cbc,cbc_tests], ['../prob_examples/public_examples/B/Benchmarks/scheduler.mch', '-cbc_tests', 3, '1=1', '../prob_examples/public_examples/B/Benchmarks/scheduler_cbc_tests.xml'], 'Check that full coverage can be achieved').
cli_testcase(1070, [cbc], ['../prob_examples/public_examples/B/TestCaseGeneration/InvolvedSequences.mch', '-cbc_sequence', 'BookFlight;BookHotel;BookReturnFlight', '-cbc_sequence', 'BookFlight;BookReturnFlight;BookHotel;BookHotel'], 'Check that constraints can be solved').
cli_testcase(1071, [cbc], ['../prob_examples/public_examples/B/TestCaseGeneration/InvolvedSequences2.mch', '-cbc_sequence', 'Op1;Op2;Op3;Target3', '-p', 'CLPFD', 'TRUE'], 'Check that constraints can be solved').
cli_testcase(1072, [cbc,cbc_tests], ['../prob_examples/public_examples/B/TestCaseGeneration/InvolvedSequences2.mch', '-cbc_tests', 2, '1=1', '../prob_examples/public_examples/B/TestCaseGeneration/InvolvedSequences2_cbc_tests.xml', '-p', 'CLPFD', 'TRUE', '-expcterr', 'cbc_tests'], 'Check that coverage cannot be achieved').
cli_testcase(1073, [cbc,cbc_tests], ['../prob_examples/public_examples/B/TestCaseGeneration/InvolvedSequences2.mch', '-cbc_tests', 4, '1=1', '../prob_examples/public_examples/B/TestCaseGeneration/InvolvedSequences2_cbc_tests.xml', '-p', 'CLPFD', 'TRUE'], 'Check that full coverage can be achieved').
cli_testcase(1074, [cbc,tla,smt_test], ['../prob_examples/public_examples/TLA/SumAndProduct/Merz/SumAndProductTransition.tla', '-cbc_sequence', 'Step2;Step3;Step4', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE' ], 'Check cbc sequence on TLA example').
cli_testcase(1075, [b_test], ['../prob_examples/public_examples/B/PerformanceTests/SetLogInefficiencies4.mch', '-init', '-p', 'CLPFD', 'TRUE', '-p', 'MAXINT', 128, '-assertions'], 'Check that ProB deals well with certain examples which Setlog cannot deal with').
cli_testcase(1076, [b_test], ['../prob_examples/public_examples/B/FeatureChecks/VariableCaptureComprSet.mch', '-cbc' ,'all', '-expcterr', 'cbc','-assertions', '-p', 'CLPFD', 'TRUE', '-mc', 100, '-expcterr', 'invariant_violation'], 'Check that variable capture problems do not confuse ProB').

cli_testcase(1077, [b_test], ['-evalt', '{y|#(r,r2).({x,y|x<y} = r & r2 = (r;r) &  20|->y : r2 & y|->25 : r2)} = (22..23)',
 '-evalf', '{y|#(r,r2).({x,y|x<y} = r & r2 = (r;r) &  20|->y : r2 & y|->25 : r2)} = (22..24)',
 '-evalt', 'f = %x.(x:NATURAL|x+x) & f2 = (f;f) & i = f2[1..n] & card(i)=n & min(i)=4 & max(i)=4*n & n=5000',
 '-evalt', 'f = %x.(x:NATURAL|x+x) & f2 = (f;f) & f4 = (f2;f2) & i = f4[1..n] & card(i)=n & min(i)=16 & max(i)=16*n & n=2000',
 '-evalt', 'f = %x.(x:NATURAL|x+x) & f2 = (f;f) & r1 = {x,y|x:1..n & y:{x,x+1}} & comp = {x,z|#y.(x|->y:r1 & y|->z:f2)} & comp2 = (r1 ; f2) & comp=comp2 & n=1000 & card(comp2) = 2*n & max(ran(comp2))=4*(n+1)',
 '-evalt', 'half = {x,y|y+y=x} & card(half[1024 .. xx] \\/ (closure1(half) ; half)[1024 .. xx])=10 & xx=1024',
 '-p', 'SYMBOLIC', 'TRUE', '-p', 'CLPFD', 'TRUE'
 , '-p', 'TIME_OUT', 7000 % TEMPORARILY INCREASE TIME OUT: TO DO investigate why  r1 = {x,y|x:1..n & y:{x,x+1}} & comp = {x,z|#y.(x|->y:r1 & z=y+y+y+y)} & n=1000 & max(ran(comp))=4*(n+1)  is much slower since frozen_dif_aux checks for int_plus4
 ], 'Test that symbolic relational composition works').
cli_testcase(1078, [private,data_validation], ['../prob_examples/examples/B/Systerel/PerformanceIssue_Apr13/TOP3_CONC_def7_obs_original.mch', '-init' ,'-aa', 0,1,0, '-p', 'SYMBOLIC', 'TRUE'], 'Check that performance issue related to symbolic relational composition solved').
cli_testcase(1079, [cbc,smt_test], ['../prob_examples/public_examples/TLA/SumAndProduct/Merz/SumAndProductCst.mch', '-assertions', '-init', '-det_check', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE' , '-p' ,'TIME_OUT', 12000], 'Check solving on example with nested set comprehensions').
cli_testcase(1080, [b_test], ['../prob_examples/public_examples/B/Laws/RelClosureLaws.mch', '-init' ,'-assertions', '-p', 'CLPFD', 'TRUE'], 'Check various assertions about relational composition and closure1').
cli_testcase(1081, [b_test], ['../prob_examples/public_examples/B/Laws/RelClosureLaws.mch', '-t', '-p', 'CLPFD', 'TRUE', '-mc', 20, '-nogoal', '-nodead'], 'Check various assertions about relational composition and closure1').
cli_testcase(1082, [csp_test], ['../prob_examples/public_examples/CSP/Tickets/TUBerlin/Interleaving.csp','-assertions','-strict'], 'Checking CSP-M assertions.').
cli_testcase(1083, [tickets,data_validation], ['-evalt', 'adii = %ii.(ii : INTEGER +-> {FALSE,TRUE}|(ii |>> {FALSE})) & afq = {y,z|z : BOOL & (y : {(0|->1)}[{m}] & z = FALSE)} & afs = adii(afq)', '-p', 'SYMBOLIC', 'TRUE',
 '-p', 'MEMO', 'TRUE' %, '-expcterr' , eval_string_enum_warning
 ],'ensure no avl_set(empty) is generated anymore').
cli_testcase(1084, [cbc], ['-evalt', 'r:INTEGER<->INTEGER & dom(r)=2..20',
 '-evalt', 'r:INTEGER<->BOOL & dom(r)=2..50',
 '-evalt', 'r:INTEGER<->INTEGER & ran(r)=2..50',
 %'-evalt', 'r:INTEGER<->INTEGER & ran(r)=2..16 & card(r)>=15',
 %'-evalt', 'r:INTEGER<->INTEGER & dom(r)=2..16 & card(r)>=15',
 '-evalt', 'r:INTEGER<->INTEGER & dom(r)=10002..10016',
 '-evalt', 'r:INTEGER<->INTEGER & ran(r)=10002..10016',
 '-evalt', 'card({r|r:INTEGER<->INTEGER & dom(r)=61..64 & card(r)=4 & ran(r) = 55..56})=14',
 '-evalt', 'card({r|r:INTEGER<->INTEGER & dom(r)=9961..9964 & card(r)=4 & ran(r) = 2155..2158})=24',
 '-evalt', 'card({r|r:INTEGER<->INTEGER & dom(r)=61..64 & card(r)=4 & ran(r) = 155..159})=0',
 '-p', 'CLPFD', 'TRUE' %, '-expcterr', eval_string_enum_warning
 ],'ensure proper constraint-solving for domain/range').
cli_testcase(1085, [b_test], ['../prob_examples/public_examples/EventBPrologPackages/Advance/DoorLockSystem/LOCKSR2_mch.eventb', '-mc', 100, '-cc', 24, 32, '-vacuity_check', '-expcterr', vacuity_check, '-strict'], 'Checking vacuity_check command').
cli_testcase(1086, [csp_test, cbc], ['../prob_examples/public_examples/CSP/clpfd_tests/SetComprehensionCBC.csp','-p', 'CLPFD', 'TRUE', '-assertions','-strict'], 'Checking CSP-M assertions (CBC test case).').
cli_testcase(1087, [b_test,infinite], ['../prob_examples/public_examples/B/PerformanceTests/InfiniteIntersection.mch', '-p', 'CLPFD', 'TRUE', '-init', '-assertions'], 'Check we can take intersection with infinite sets').
cli_testcase(1088, [cbc,smt_test], ['../prob_examples/public_examples/B/Benchmarks/phonebook7.mch','-p', 'CLPFD', 'TRUE',
  '-p', 'TIME_OUT', 5000, % increased time-out after change of pfun enumeration priority; TO DO: fix
  '-p', 'SMT', 'TRUE', % added SMT option after adding wait flag for propagate_result_to_input in domain_wf in bsets_clp
  '-cbc', 'all','-strict'], 'Check that we do not time-out (due to delaying normalisation of variables).').
cli_testcase(1089, [b_test], ['../prob_examples/public_examples/B/PerformanceTests/InfiniteIntersection2.mch', '-p', 'CLPFD', 'TRUE', '-p', 'SYMBOLIC', 'FALSE','-init', '-assertions'], 'Check ProB now detects even/odd automatically as symbolic').
cli_testcase(1090, [b_test,infinite,sigma], [
 '-evalt', 'f = {x,y,z| x*x + y*y = z*z} & {y|(3,4,y):f} = {-5,5}',
 '-evalt', 's = {x,y,z| z:seq(NATURAL) & x^y=z} & {v|([1111,2222,3333,4444],[99999,8888],v):s} = {[1111,2222,3333,4444,99999,8888]}',
 '-evalt', 's = {x,y| x<:POW(NATURAL) & union(x)=y & SIGMA(z).(z:y|z)>1000} & s({{890},{},{1,199}}) = res & SIGMA(v).(v:res|v)=1090',
 '-p', 'CLPFD', 'TRUE'],'ensure ProB detects certain relations as infinite automatically').
cli_testcase(1091, [csp_test, csp_det], ['../prob_examples/public_examples/CSP/Tickets/Ivo/AgentCurry.csp', '-assertions','-strict'], 'Checking CSP-M assertions (agent curry calls).').
cli_testcase(1092, [csp_test, csp_det], ['../prob_examples/public_examples/CSP/ucsexamples/chapter06/commsec.csp', '-assertions','-strict'], 'Checking CSP-M assertions (agent curry calls).').
cli_testcase(1093, [b_test,infinite,override,sigma], [
 '-evalt', 'f = %x.(x:NATURAL|x+x) & g = f <+ {0|->1} & g(0)=res0 & g(111)=res1',
 '-evalt', 'f = {x,y|x:NATURAL & y=x+x or y=x*x} & g = f <+ {0|->111,222|->1024} & union({res|g[{222}] =res})=rr222 & union({res|g[{1000}] =res}) = rr1000 & card(rr222)=1 & card(rr1000)=2 & SIGMA(z).(z:rr1000|z)=1002000',
 '-evalt', 'r1 = {x,y|x*x<y & (x+1)*(x+1)>y} & r2 = %z.(z>20|z) &  r12 = r1 <+ r2 & (21,x21):r12 & {x10|10|->x10 : r12}=r10 & SIGMA(v).(v:r10|v)+x21=2231',
 '-evalt', 'r1 = {x,y|x*x<y & (x+1)*(x+1)>y} & r2 = %z.(z>20|z) &  r21 = r2 <+ r1 & {x|21|->x : r21}=x21 & x21 = ((21*21)+1)..((22*22)-1)',
 '-p', 'CLPFD', 'TRUE'],'ensure symbolic treatment of override').
cli_testcase(1094, [tickets, proz], ['../prob_examples/public_examples/Z/Tickets/Thompson1/System.fuzz', '-init', '-eval', 'map={}','-strict'], 'Checking CSP-M assertions (agent curry calls).').
cli_testcase(1095, [b_test,cbc], ['../prob_examples/public_examples/B/BZTT/GSM_revue.mch', '-t', '-mc', 3000, '-cc', 1850, 53594, '-cbc_deadlock', '-strict', '-p', 'CLPFD', 'TRUE'], 'Checking old BZTT example.').
cli_testcase(1096, [b_test,cbc,smt_test], ['../prob_examples/public_examples/B/BZTT/GSM_revue.mch', '-cbc', all, '-cbc_deadlock', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE'], 'Checking old BZTT example.').
cli_testcase(1097, [tickets,eventb_test,'PROB-292'], ['../prob_examples/public_examples/B/Tickets/PROB-292/MTool1__04.eventb', '-mc', 100, '-expcterr', invariant_violation, '-p', 'PROOF_INFO', 'TRUE'], 'Check that proof info not confused by event extending itself').
cli_testcase(1098, [tickets], ['../prob_examples/public_examples/B/Tickets/Hansen4/Test.mch', '-state_trace', '../prob_examples/public_examples/B/Tickets/Hansen4/Test.tla.trace'], 'Check that state_trace works with constants').
cli_testcase(1099, [b_test], [ '-p', 'TIME_OUT', 1500,
 '-evalt', 'f = {x,y|x:0..2**32 & x mod 2 =0 & y=x*x} & f(x) = 10000',
 '-evalt', 'f = {x,y|x:0..2**132 & x mod 2 =0 & y=x*x} & f(x) = 10000',
 '-evalt', 'f = {x,y|x:0..2**25 & x mod 2 =0 & y=x*x} & {r|f(r) = 10000}=res & card(res)=1 & #vv.(vv:res &vv=100)',
 '-p', 'CLPFD', 'TRUE'],'ensure detection of large set comprehensions').
cli_testcase(1100, [b_test], [
 '-evalt', 'r = {(111,2222),(333,4444)} & s = (2000..2001)*{v|v>3000} & x : (r||s)',
 '-evalt', 'card(((1..100)*(50..60) || (2..30)*(20..22))) = 95700',
 '-evalt', 'card(((1..100)*(50..60) || (2..30)*(23..22)))=0',
 '-evalt', '(((1..10)*(50..60) || (2..30)*(23..24)))~ = (((50..60)*(1..10) || (23..24)*(2..30)))',
 '-evalt', 'card(((1..1000)*(50..60) || (2..30)*(20..22))) = 957000',
 '-evalt', 'card(((1..n)*(50..60) || (2..30)*(20..22))) = 957*n & n=10',
 '-p', 'CLPFD', 'TRUE' %, '-expcterr', 'eval_string_enum_warning'
 ],'ensure symbolic treatment of parallel product works').

cli_testcase(1101, [private,data_validation,infinite],[
 '../prob_examples/examples/B/Systerel/C578.EML.014/360_002.mch',
 '../prob_examples/examples/B/Systerel/C578.EML.014/380_002.mch',
 '../prob_examples/examples/B/Systerel/C578.EML.014/612_001.mch',
 '../prob_examples/examples/B/Systerel/C578.EML.014/019_100_corrected.mch',
 '../prob_examples/examples/B/Systerel/C578.EML.014/CF_TVD_8.mch',
 '../prob_examples/examples/B/Systerel/C578.EML.014/CF_CBTC_TER_1.mch',
 '../prob_examples/examples/B/Systerel/C578.EML.014/CF_ZMA_PRUD_1.mch',
 '../prob_examples/examples/B/Systerel/C578.EML.014/CF_ZMA_PRUD_7.mch',
'-p', 'TIME_OUT', '11000', '-p', 'SYMBOLIC', 'TRUE', '-p', 'DOUBLE_EVALUATION', 'FALSE', '-p', 'EXPAND_FORALL_UPTO', '0', '-p', 'CLPFD', 'TRUE', '-p', 'MEMO', 'TRUE', '-init', '-assertions', '-p', 'TRACE_INFO', 'TRUE', '-p', 'ENUMERATE_INFINITE_TYPES', 'FALSE', '-strict'],
  'check data validation tasks with infinite sets').
cli_testcase(1102, [private,data_validation,infinite],[
 '../prob_examples/examples/B/Systerel/C578.EML.014/019_100.mch',
'-p', 'TIME_OUT', '9000', '-p', 'SYMBOLIC', 'TRUE', '-p', 'DOUBLE_EVALUATION', 'FALSE', '-p', 'EXPAND_FORALL_UPTO', '0', '-p', 'CLPFD', 'TRUE', '-p', 'MEMO', 'TRUE', '-init', '-aa', 12,1,0,'-p', 'TRACE_INFO', 'TRUE', '-p', 'ENUMERATE_INFINITE_TYPES', 'FALSE', '-strict'],
  'check data validation tasks with infinite sets').
cli_testcase(1103, [private,data_validation,infinite],[
 '../prob_examples/examples/B/Systerel/C578.EML.014/435_002.mch',
'-p', 'TIME_OUT', '9000', '-p', 'SYMBOLIC', 'TRUE', '-p', 'DOUBLE_EVALUATION', 'FALSE', '-p', 'EXPAND_FORALL_UPTO', '0', '-p', 'CLPFD', 'TRUE', '-p', 'MEMO', 'TRUE', '-init', '-aa', 0,1,0,'-p', 'TRACE_INFO', 'TRUE', '-p', 'ENUMERATE_INFINITE_TYPES', 'FALSE', '-strict'],
  'check data validation tasks with infinite sets').
cli_testcase(1104, [b_test, let,quantified], [
 '-evalt', 'INTER(x , y,z ).(x=1..y & y=5&z=2|x) = {1,2,3,4,5}',
 '-evalt', 'UNION(x , y,z ).(x=1..y & y=5&z=2|x) = {1,2,3,4,5}',
 '-evalt', 'UNION(x , y ).(x=1..y & y=20|x)=1..20',
 '-evalt', 'UNION(beu).(beu=1|{2}) = {2}',
 '-evalt', 'UNION(xx).(xx<:55..60|xx) = 55..60',
 '-evalt', 'UNION(x , y ).(x=1..y & y=20&v<10|x)=r & v=20 & r={}',
 '-evalt', 'UNION(x , y ).(x=1..y & y=20&v<10|x)=r & v=5 & r=1..20',
 '-evalt', 'UNION(x).(x=1 & x>2|{1}) = {}',
 '-evalf', 'card(UNION(x).(x=1 & x>2|{1}))>0',
 '-evalt', 'UNION(x).(x=1 & (x>2 or x<3)|{x}) = {1}',
 '-evalt', 'UNION(x,y).(x=1 & (x>2 or x<y) & y:0..3|{x|->y}) = {(1|->2),(1|->3)}',
 '-evalt', 'UNION(x,y).(x=y & y=x+1 & x:1..100|{x|->y}) = {}',
 '-evalt', 'UNION(x,y).(x=1 & y:1..2|{x+y}) = {2,3}',
 '-evalt', 'INTER(x,y).(x=1 & y:1..2|{x+y}) = {}',
 '-evalt', 'INTER(x,y).(x=1 & y:1..2|{x+y,x+y+1})={3}',
 '-evalt', 'INTER(x,y,z).(x=1000 & z=x+1000 & y:1..2|{x+y+z,x+y+z+1}) = {3002}',
 '-evalf', '!(x,y).(x=1 & y:(x+1)..(x+19) => y+y<30)',
 '-evalt', '!(x,y).(x=1 & y:(x+1)..(x+19) => y+y<42)',
 '-evalt', '!x.(x=100 & x>200 => x>999)',
 '-evalt', '!x.(x=100 & x>50 => x>99)',
 '-evalf', '!x.(x=100 & x>50 => x>100)',
 '-evalt', '!(x,y,z,v).(x=10 & y=20 & z=x+y & v:x..z => v:NATURAL1)',
 '-evalt', 'f = {1|->2} & !e.(2:dom(f) & e=f(2) => e>100)',
 '-evalf', 'f = {1|->2} & !e.(1:dom(f) & e=f(1) => e>100)',
 '-evalf', 'f = {1|->2} & !e.(e=f(1) => e>100)',
 '-evalt', 'f = {1|->200} & !e.(e=f(1) => e>199)'
  ],'ensure UNION/INTER can be dealt with by typechecker, + LET introduction works correctly...').
cli_testcase(1105,[private, data_validation,alstom,codespeed],[
 '../prob_examples/examples/B/Alstom/sgd_Jun11/vital_gradient_v3/vital_gradient_optimized.mch',
  '-p', 'CLPFD', 'TRUE', '-p', 'MAXINT', '2147483647', '-p', 'MININT', '-2147483648', '-p', 'TIME_OUT', '35000', '-t'],
  'check that trace can be replayed').
cli_testcase(1106, [b_test,recursion,external], ['../prob_examples/public_examples/B/RecursiveFunctions/PrimitiveRecursiveFunctions.mch',
 '../prob_examples/public_examples/B/RecursiveFunctions/SumTLA.mch',
 '../prob_examples/public_examples/B/RecursiveFunctions/TransitiveClosureFunction.mch',
  '-assertions', '-strict', '-p', 'SYMBOLIC', 'FALSE'], 'Test that other recursive functions work.').
cli_testcase(1107, [ltl,csp_test], ['../prob_examples/public_examples/CSP/RefusalBasedRefinement/RefusalBasedRefinement_simple.csp', '-ltlassertions', '-strict'],'Check LTL formulas with the \'Av(-)\'-operator from File.').
cli_testcase(1108, [b_test,sigma], ['-p', 'CLPFD', 'TRUE',
 '-evalt', 'bool(!(x,y).(x:1..400000 & y:1..x => x>=y)) /= bool(!(x,y).(x:1..400000 & y:1..x => x>y))',
 '-evalt', 'bool(!x.(x:POW(1..100) & card(x)>1 => SIGMA(y).(y:x|y)>3)) = bool(!x.(x:POW(1..100) & card(x)>1 => SIGMA(y).(y:x|y)<200))',
 % we wrap into bool to avoid top-level rewriting of !
 '-evalt', '!(x,y).(x:1..400000 & y:1..x => x>=y)',
 '-evalf', '!(x,y).(x:1..400000 & y:1..x => x>y)',
 '-evalf', '!x.(x:POW(1..100) & card(x)>1 => SIGMA(y).(y:x|y)>3)',
 '-evalf', '!x.(x:POW(1..100) & card(x)>1 => SIGMA(y).(y:x|y)<200)',
 '-strict'], 'Check that forall can now deal with large domains').
cli_testcase(1109, [tickets], ['../prob_examples/public_examples/B/Tickets/Valerio3/BV16_DEFINITION.mch', '-strict', '-init', '-p', 'SYMBOLIC', 'TRUE', '-p', 'CLPFD', 'TRUE',
  '-evalf', '!(vec).( vec : BIT_VECTOR => size (vec)<=16)',
 '-evalt', 'bool(!(vec).( vec : BIT_VECTOR => size (vec)<=16))=bool(!(vec).( vec : BIT_VECTOR => size (vec)<=2))'
 %,'-expcterr', 'eval_string_enum_warning'
 ], 'Check that ProB now detects this quantification to be false'). % previously it was not detected due to limited enumeration
cli_testcase(1110, [ltl], ['../prob_examples/public_examples/B/Ivo/PruningActionsTest2.mch', '-ltlassertions', '-strict'], 'Checking independency of actions by means of LTL Formulas.').
cli_testcase(1111, [tickets,quantified],['../prob_examples/public_examples/B/PerformanceTests/UNION_PerfTest.mch',
 '-p', 'SYMBOLIC', 'TRUE', '-assertions' ,'-p' ,'CLPFD', 'TRUE',
 '-p', 'TIME_OUT', 1000], 'Ensure symbolic ran closure can be expanded efficiently').
cli_testcase(1112, [tickets,alstom],['../prob_examples/examples/B/Alstom/cbtc/params_cbtc.mch',
 '-init' ,'-p' ,'IGNORE_PRJ_TYPES', 'TRUE', '-p', 'TIME_OUT', '3500'], 'Ensure symbolic ran closure can be expanded efficiently').
cli_testcase(1113, [b_test], ['-p', 'CLPFD', 'TRUE',
 '-evalt', 'card({x|#y.(y:111..112 & x = {{y*y}})}) = 2',
 '-evalt', 'card({x|#y.(y:111..112 & x = {{{{{y*y}}}}})}) = 2',
 '-evalt', 'card({x|#y.(y:POW(BOOL) & card(y)=1 & x = ({(y,TRUE)},TRUE,TRUE,FALSE,TRUE,FALSE,TRUE,FALSE))}) =2',
 '-evalt', '{x|#y.(y:POW(BOOL) & card(y)=1 & x = {(y,TRUE,FALSE,TRUE)})} = {{((({FALSE}|->TRUE)|->FALSE)|->TRUE)},{((({TRUE}|->TRUE)|->FALSE)|->TRUE)}}',
 '-evalt', '{x,y|#z.(z:{1024,2048,4096} & x=z+z & y={{z*z}})} = {(2048|->{{1048576}}),(4096|->{{4194304}}),(8192|->{{16777216}})}',
 '-evalt', '{x|#z.(z:1..2**20 & x=z & x+z<100)} = 1..49',
 '-strict'], 'Check closures with top-level existential quantifiers').
cli_testcase(1114, [tickets,csp_test],['../prob_examples/public_examples/B/Tickets/315/ANN_Neuron.csp',
 '-t' , '-mc', 200, '-check_complete', '-cc', 26, 42], 'Ensure ticket remains fixed').
cli_testcase(1115, [tickets,eventb_test],['../prob_examples/public_examples/EventBPrologPackages/Ticket_ThmGuard/M2_mch.eventb',
 '-t' , '-mc', 200, '-check_complete'], 'Ensure theorem in guards work').
cli_testcase(1116, [dot,eventb_test,dot],['../prob_examples/public_examples/EventBPrologPackages/Tests/UnicodeIds_mch.eventb',
  '-mc', 200, '-nodead', '-check_complete', '-dot', 'state_space', '../prob_examples/public_examples/EventBPrologPackages/Tests/UnicodeIds_spdot.dot'], 'Ensure Unicode characters work in Dot export').
cli_testcase(1117, [tickets],['../prob_examples/public_examples/B/Tickets/Srivastava1/Total_M1.mch',
 '-t' , '-strict'], 'Ensure that expensive arg in relations check not computed').
cli_testcase(1118, [eventb_test, theories], ['../prob_examples/public_examples/EventBPrologPackages/Theory/ClosureTest.eventb','-assertions','-strict','-cc',2,1], 'Check if the closure definition works').
cli_testcase(1119, [csp_test,cspb], ['-expecterr','deadlock','../prob_examples/public_examples/CSPB/Tickets/Bauer/machine.mch','-csp-guide','../prob_examples/public_examples/CSPB/Tickets/Bauer/machine.csp','-mc',1000,'-his','../prob_examples/public_examples/CSPB/Tickets/Bauer/machine_his.txt','-his_option','show_states','-strict'], 'Check if the closure definition works').
cli_testcase(1120, [tickets,csp_test],['../prob_examples/public_examples/CSP/ErrorModels/haskell_csp/TypeErrorInChannelInput.csp', '-mc', 10, '-expcterr', 'check_channel_output_value', '-expcterr', 'general_error_occurred'], 'Checking for raising type error in channel input.').
cli_testcase(1121, [tickets,csp_test],['../prob_examples/public_examples/CSP/other/Ivo/ComplexDataTypes.csp', '-t'], 'Checking complex input/output values for channel.').

cli_testcase(1122, [private,data_validation], ['../prob_examples/examples/B/Systerel/C578/2013/machines_14082013/410_002.mch', '-init' ,'-assertions', '-p', 'SYMBOLIC', 'TRUE'], 'Check that WD issue inside negate exists solved').
cli_testcase(1123, [tickets], ['../prob_examples/public_examples/B/ErrorMachines/IllegalFunCalls/FunCall_Exists_correct.mch', '-init' ,'-assertions'], 'Check that WD issue inside negate exists solved').
cli_testcase(1124, [tickets,wd], ['../prob_examples/public_examples/B/ErrorMachines/IllegalFunCalls/FunCall_Exists_error.mch', '-init' ,'-assertions', '-expcterr', 'well_definedness_error', '-expcterr', check_assertions], 'Check that WD issue inside negate exists solved').
cli_testcase(1125, [b_test,data_validation,infinite,override], ['-p', 'SYMBOLIC', 'TRUE',
 '-evalt', 'r=INTEGER*{3333} &  s= {(1,1)} &  ov = r <+ s & r1 = ov(1) & r2 = ov(2) & r3 = ov[1..3] & r3={r1}\\/{r2} & r1000=ov(1000) & r1000>3332'],
  'Test that override works symbolically').
cli_testcase(1126, [private,data_validation,infinite,override], ['../prob_examples/examples/B/Systerel/C578/2013/machines_14082013/PB_00611_005.mch',
 '../prob_examples/examples/B/Systerel/C578/2013/machines_14082013/PS_00611_006.mch',
 '-init' ,'-assertions', '-p', 'SYMBOLIC', 'TRUE'], 'Check symbolic override issues solved').
cli_testcase(1127, [tickets], ['../prob_examples/public_examples/B/Tickets/PROB-306/OperationCallBug.mch', '-strict', '-init',
 '-expcterr', 'type_error', '-expcterr', 'load_main_file'], 'Check type checker for wrong number of arguments in operation call').
cli_testcase(1128, [tickets], ['../prob_examples/public_examples/B/Tickets/PROB-306/OperationCallBug2.mch', '-strict', '-init',
 '-expcterr', 'type_error', '-expcterr', 'load_main_file'], 'Check type checker for wrong number of results in operation call').
cli_testcase(1130, [private,data_validation], ['../prob_examples/examples/B/Systerel/C578/2013/machines_03092013/error/0050_001.mch',
 '-init' ,'-assertions', '-p', 'SYMBOLIC', 'TRUE', '-p', 'TIME_OUT', 3500], 'Check bug remains solved (unification in b_compute_expression2)').
cli_testcase(1131, [private,data_validation],[
 '../prob_examples/examples/B/Systerel/C578/2013/machines_06092013/C_02_001.mch',
'-p', 'TIME_OUT', '11000', '-p', 'SYMBOLIC', 'TRUE', '-p', 'DOUBLE_EVALUATION', 'FALSE', '-p', 'EXPAND_FORALL_UPTO', '0', '-p', 'CLPFD', 'TRUE', '-p', 'MEMO', 'TRUE', '-init', '-assertions', '-p', 'TRACE_INFO', 'FALSE', '-p', 'ENUMERATE_INFINITE_TYPES', 'FALSE', '-strict'],
  'check memory consumption ok').
cli_testcase(1132,[csp_test], ['../prob_examples/public_examples/CSP/CSPOZ/CurriedNATupleFunNested.csp', '-assertions', '-strict'], 'Nested NA-Tuples.').

cli_testcase(1133, [csp_test], ['../prob_examples/public_examples/CSP/mydemos/demo/basin_olderog_bank.csp', '-mc', 1000, '-cc', 129, 673, '-strict'], '').
cli_testcase(1134, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/AgentErrorsLetWithIn.csp', '-mc', 10, '-nodead', '-expcterr', 'general_error_occurred', '-expcterr', 'singleSetElement', '-strict'], 'Testing raising of singleSetElement error.').
cli_testcase(1135, [csp_test], ['../prob_examples/public_examples/CSP/CSPOZ/CurriedNATupleFunNested.csp', '-assertions', '-t', '-strict'], 'Testing curry functions.').
cli_testcase(1136, [csp_test], ['../prob_examples/public_examples/CSP/evans/KeyNETv2.csp', '-mc', '1000', '-nodead', '-cc', 24, 23, '-assertions', '-strict'], '').
cli_testcase(1137, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/DIV.csp', '-mc', '10', '-cc', 2, 3, '-strict'], 'Testing normalising of DIV |~| DIV.').
cli_testcase(1138, [por], ['../prob_examples/public_examples/B/Ivo/SleepSetsStudyCase.mch', '-mc', 1000, '-noinv', '-expcterr', 'deadlock', '-cc',7,7, '-p', por, 'ample_sets', '-strict'], 'Model checking with partial order reduction.').
cli_testcase(1139, [por,ltl], ['../prob_examples/public_examples/B/Ivo/SleepSetsStudyCase.mch', '-ltlformulat', 'F {x=1 & y=1}', '-cc',10,12, '-p', por, ample_sets,'-strict'], 'LTL Model checking with partial order reduction.').
cli_testcase(1140, [por], ['../prob_examples/public_examples/B/Ivo/SleepSetsStudyCase.mch', '-mc', 1000, '-nodead', '-cc',11,17, '-p', por, 'ample_sets', '-strict'], 'Model checking with partial order reduction (no state space reductions).').
cli_testcase(1141, [por], ['../prob_examples/public_examples/B/Ivo/BenchmarksPOR/BPEL2B/Model_Machine_4_NoDeadlock_v2.eventb', '-mc', 10000, '-noinv', '-p', por, 'ample_sets', '-expcterr', deadlock,'-strict'], 'Model checking with partial order reduction (deadlock check).').
cli_testcase(1142, [por], ['../prob_examples/public_examples/B/Ivo/BenchmarksPOR/BPEL2B/Model_Machine_4_NoDeadlock_v3.eventb', '-mc', 10000, '-noinv', '-p', por, 'ample_sets', '-expcterr', deadlock,'-strict'], 'Model checking with partial order reduction (deadlock check).').
cli_testcase(1143, [por], ['../prob_examples/public_examples/B/Ivo/BenchmarksPOR/BPEL2B/Model_Machine_4_NoDeadlock_v4.eventb', '-mc', 10000, '-noinv', '-p', por, 'ample_sets', '-expcterr', deadlock,'-strict'], 'Model checking with partial order reduction (deadlock check).').
cli_testcase(1144, [por], ['../prob_examples/public_examples/B/Ivo/BenchmarksPOR/BPEL2B/Model_Machine_4_NoDeadlock_v5.eventb', '-mc', 10000, '-check_complete_operation_coverage', '-p', por, 'ample_sets', '-get_min_max_coverage', '../prob_examples/public_examples/B/Ivo/BenchmarksPOR/BPEL2B/Model_Machine_4_NoDeadlock_v5_Coverage_POR.txt', '-strict', '-expcterr', 'model_check_incomplete'], 'Model checking with partial order reduction (min...max coverage check).').
cli_testcase(1145, [b_test], ['-p', 'CLPFD', 'TRUE',
 '-evalt', '{} <: (1..100000000 \\/ {-1})'],
  'Check that we do not compute <: if first arg {}').
cli_testcase(1146, [private,data_validation], ['../prob_examples/examples/B/Systerel/C578/2013/machines_06092013/T_032_002.mch',
 '-init' ,'-aa', 0,1,0, '-p', 'SYMBOLIC', 'TRUE', '-p', 'MEMO', 'TRUE', '-p', 'TIME_OUT', 11500, '-p', 'CLPFD', 'TRUE',
  '-p', 'ENUMERATE_INFINITE_TYPES', 'FALSE', '-p', 'EXPAND_FORALL_UPTO', 0 ], 'Check test eq_emtpy_set bug remains solved'). % increased time-out from 7500
cli_testcase(1147, [por], ['../prob_examples/public_examples/B/Ivo/BenchmarksPOR/Peterson/em_abs.mch', '-mc', 10000, '-p', por, 'ample_sets', '-get_min_max_coverage', '../prob_examples/public_examples/B/Ivo/BenchmarksPOR/Peterson/em_abs_Coverage_POR.txt', '-strict'], 'Model checking with partial order reduction (min...max coverage check). Peterson\'s algorithm').
cli_testcase(1148, [por], ['../prob_examples/public_examples/B/Ivo/BenchmarksPOR/Peterson/peterson_R.ref', '-mc', 10000, '-p', por, 'ample_sets', '-check_complete_operation_coverage', '-get_min_max_coverage', '../prob_examples/public_examples/B/Ivo/BenchmarksPOR/Peterson/peterson_R_Coverage_POR.txt', '-strict'], 'Model checking with partial order reduction (min...max coverage check). Peterson\'s algorithm').
cli_testcase(1149, [por], ['../prob_examples/public_examples/B/Ivo/BenchmarksPOR/Peterson/syspeterson.mch', '-mc', 10000, '-p', por, 'ample_sets', '-expcterr', invariant_violation, '-strict'], 'Model checking with partial order reduction (invariant violation is expected). Peterson\'s algorithm').
cli_testcase(1150, [eventb_test, theories], ['../prob_examples/public_examples/EventBPrologPackages/Theory/BoolOpsTest1.eventb', '-mc', 10, '-nodead', '-strict'], 'Checking the BoolOps theory').
cli_testcase(1151, [eventb_test, theories], ['../prob_examples/public_examples/EventBPrologPackages/Theory/ListTest1.eventb', '-mc', 100, '-nodead', '-strict'], 'Checking the List theory').
cli_testcase(1152, [eventb_test, theories], ['../prob_examples/public_examples/EventBPrologPackages/Theory/ClosureTest1.eventb', '-p', 'SYMMETRY_MODE', nauty, '-p', 'MAX_OPERATIONS', 100, '-mc', 1000, '-nodead', '-strict'], 'Checking the closure theory').
cli_testcase(1153, [eventb_test, theories], ['../prob_examples/public_examples/EventBPrologPackages/Theory/SumProductTest1.eventb', '-mc', 100, '-nodead', '-strict'], 'Checking the SUM and PRODUCT theory').
cli_testcase(1154, [b_test], ['-p', 'CLPFD', 'TRUE',
 '-evalt', 'x : 1..1000040 \\/ {10000061} & x mod 2 = 0',
 '-evalt', 'x : 1..10000040 \\/ {10000061} & x mod 2 = 0',
 '-evalt', 'x : 1..10000040 \\/ {10000061}'
% does not work yet:  ,'-evalt', ' x : {-33} \/ {0} \/ 1..10000040 \/ {10000061} \/ {-22}'
], 'Check union of AVL with large interval does not expand interval').
cli_testcase(1155, [eventb_test, theories], ['../prob_examples/public_examples/EventBPrologPackages/Theory/BinaryTreeTest1.eventb', '-mc', 100, '-nodead', '-strict'], 'Checking the BinaryTree theory').
cli_testcase(1156, [eventb_test, theories], ['../prob_examples/public_examples/EventBPrologPackages/Theory/SeqTest1.eventb', '-mc', 2000, '-nodead', '-strict'], 'Checking the Seq theory').
cli_testcase(1157, [private,data_validation,infinite,override],[
 '../prob_examples/examples/B/Systerel/C578.EML.014/670_002.mch',
'-p', 'TIME_OUT', '15000', '-p', 'SYMBOLIC', 'TRUE', '-p', 'DOUBLE_EVALUATION', 'FALSE', '-p', 'EXPAND_FORALL_UPTO', '0', '-p', 'CLPFD', 'TRUE', '-p', 'MEMO', 'TRUE', '-init', '-assertions', '-p', 'TRACE_INFO', 'FALSE', '-p', 'ENUMERATE_INFINITE_TYPES', 'FALSE', '-strict'],
  'check override computed explicitly for A*B*C <+ {...} for reasonable sizes'). % increased TIME_OUT from 11000
cli_testcase(1158, [b_test], ['-p', 'SYMBOLIC', 'TRUE','-p', 'CLPFD', 'TRUE',
'-evalt', '95|->93 : dom({x,y,z| x>z & y<z})',
'-evalf', '94|->93 : dom({x,y,z| x>z & y<z})' ],
 'check symbolic treatment of domain').
cli_testcase(1159, [b_test], ['-p', 'CLPFD', 'TRUE',
'-evalt', '{S1|S1 <: {"d","f","3","14","play"} & {"d"} /<<: S1 & card(S1):4..5} = {{"14","3","f","play"}}' ],
 'simple CBC check from old paper').
cli_testcase(1160, [b_test,infinite], ['-p', 'SYMBOLIC', 'TRUE','-p', 'CLPFD', 'TRUE',
'-evalt', 'f = %i.(i:1..100|i*i) & f[NATURAL1] = x & max(x)=10000',
'-evalt', 'f = %i.(i:1..100|i*i) & f[{v|v<=20}] = x & max(x)=400',
'-evalt', 'f = %i.(i:0..100|i*i) & f[NATURAL1] = x & min(x)=1'],
 'test that we can take image of infinite set').
cli_testcase(1161, [private,data_validation], ['../prob_examples/examples/B/Systerel/C578/2013/machines_24092013/07_001.mch',
 '-init' ,'-assertions', '-p', 'SYMBOLIC', 'TRUE', '-p', 'MEMO', 'TRUE', '-p', 'TIME_OUT', 39500,
  '-p', 'ENUMERATE_INFINITE_TYPES', 'FALSE', '-p', 'EXPAND_FORALL_UPTO', 0 ], 'Check performance'). % increased time-out from 9500 (16th July 2014)
cli_testcase(1162, [private,data_validation], [%'../prob_examples/examples/B/Systerel/C578/2013/machines_24092013/02_001_adapted.mch', % moved to test 1376
 '../prob_examples/examples/B/Systerel/C578/2013/machines_24092013/30_001_adapted.mch',
 '-init' ,'-assertions', '-p', 'SYMBOLIC', 'TRUE', '-p', 'MEMO', 'TRUE', '-p', 'CLPFD', 'TRUE', '-p', 'TIME_OUT', 35000,
  '-p', 'ENUMERATE_INFINITE_TYPES', 'FALSE', '-p', 'EXPAND_FORALL_UPTO', 0 ], 'Check performance').
cli_testcase(1163, [b_test], ['-p', 'CLPFD', 'TRUE',
'-evalf', '0..x <: 1..100000 & x>0',
'-evalt', '0..101001 <: 0..x & x <1001000',
'-evalt', '1..2**65 <: 1..2**66',
'-evalf', '1..2**67 <: 1..2**66',
'-evalt', '1..2**65 <: x..2**66 & x:0..10',
'-evalf', '1..2**65 <: x..2**66 & x:2..100000',
'-evalf', 'x:180..200 & y: 3000..4000 & x+y=z & x..z <: x..w & w<3500 & w<z',
'-evalt', 'x:180..200 & y: 3000..4000 & x+y=z & x..z <: x..w & w<3500 & w<=z',

'-evalf', '0..x <<: 1..100000 & x>0',
'-evalt', '0..101001 <<: 0..x & x <1001000',
'-evalt', '1..2**65 <<: 1..2**66',
'-evalf', '1..2**66 <<: 1..2**66',
'-evalt', '2..2**66 <<: 1..2**66',
'-evalt', '-10000..x <<: 1..2 & x>-(2**20)',
'-evalf', 'x:180..200 & y: 3000..4000 & x+y=z & x..z <<: x..w & w<3500 & w<=z',

'-evalt', '{x|1..x /<: 1..10000 & x<10003} = {10001,10002}',
'-evalt', '1000..x /<: 1..y & y..1010 /<: y..1009 & x..1010 /<: x..1009 & y>0',

'-evalt', '1..2**65 /<<: 1..2**65',
'-evalf', '1..2**65 /<<: 1..2**66',
'-evalf', '2..2**66 /<<: 1..2**66',
'-evalf', 'x..y /<<: 1001..1002 & x:100..200 & y:0..10',
'-evalt', '{x,y|x..y /<<: 1001..1002 & x:100..200 & y:0..100}={(100,100)}',
'-evalt', '{x,y|x..y /<<: 1001..1002 & x:1001..1030 & y<1004} = {(1001|->1002),(1001|->1003),(1002|->1003),(1003|->1003)}',
'-evalt', '{x,y|x..y /<<: 1001..1002 & x:1001..1030 & y<1005} = {(1001|->1002),(1001|->1003),(1001|->1004),(1002|->1003),(1002|->1004),(1003|->1003),(1003|->1004),(1004|->1004)}',
'-evalt', '10..y /<<: 200..x & x>100 & y:1..5',
'-evalt', '{x,y|10000..x /<<: 10000..y & y>10000 & y<150000 & x<10005} = {(10001|->10001),(10002|->10001),(10002|->10002),(10003|->10001),(10003|->10002),(10003|->10003),(10004|->10001),(10004|->10002),(10004|->10003),(10004|->10004)}'

], 'Check interval treatment in subset').
cli_testcase(1164, [csp_test], ['../prob_examples/public_examples/CSP/BPMN/intervention.csp', '-mc', '10', '-nodead', '-strict'], 'Testing the CSPM-Tool for parsing very long CSP expressions.').
cli_testcase(1165, [b_test,sequences,strings], ['-p', 'CLPFD', 'TRUE',  %'-expcterr', eval_string_enum_warning,
'-evalt', 'mdp = 5 * 43 & (yy: mdp + 1 .. ( mdp + 43 )  & yy: dom(memp)) & memp:seq(STRING)',
'-evalt', '16: dom(memp) & memp:seq(STRING)',
'-evalt', 'memp:seq(STRING) & 16: dom(memp)',
'-evalt', 'memp:seq(STRING) & mdp = 5 * 43 & (yy: mdp + 1 .. ( mdp + 43 )  & yy: dom(memp))',
'-evalt', '{x} /\\ 0..2000 = {} & x>0'
], 'Check constraint propagation of in domain + in sequence + interval intersection').
cli_testcase(1166, [b_test], ['-p', 'CLPFD', 'TRUE',
'-evalt', 'x /\\ 1000..2000 = {1001}',
'-evalt', 'x /\\ 1000..2000 = {y}',
'-evalt', '1000..2000 /\\ x = {y}',
'-evalt', '{x} /\\ 10000..999999 = r & r={y}',
'-evalt', '{x} /\\ 10000..999999 = {y}',
'-evalt', '10000..x = {y}',
'-evalf', '10000..20000000 = {y}',
'-evalf', '{x|x>2} = {y}',
'-evalt', 'x:0..999999 & y:0..99999 & {x} /\\ 10000..999999 = {y}',
'-evalt', '{x} /\\ {y} /= {} & x:1000..2000 & y>=0',
'-evalt', '{x} /\\ {y} /= {} & x:1000000..20000000 & y>=0',
'-evalt', 'a..b = {x} & x:100..1002',
'-evalt', '{a,b,x|a..b = {x} & x:1000..1002} = {((1000|->1000)|->1000),((1001|->1001)|->1001),((1002|->1002)|->1002)}',

'-evalt', '{x|x>0} = NATURAL1',
'-evalf', '{x|x>0} = NATURAL',
'-evalt', '{x|x>0} /= NATURAL',
'-evalt', '{x|x<110} /= NATURAL',
'-evalt', '{x|x>0} <: NATURAL1',
'-evalf', '{x|x>=0} <: NATURAL1',
'-evalt', '{x|x>0} <: INTEGER',
'-evalt', '{x|x>0} <: {x|x>-1}',
'-evalt', '{x|x>0} <: {x|x>0}',
'-evalf', '{x|x>0} <: {x|x>1}',
'-evalf', 'NATURAL1 <: 0..10000',
'-evalt', '{x|x>1} <<: {x|x>0}',
'-evalf', '{x|x>0} <<: {x|x>0}',
'-evalf', '{x|x>0} <<: NATURAL1',
'-evalt', '{x|x>0} <<: NATURAL',
'-evalt', '{x|x>0} <<: INTEGER',
'-evalf', '{x|x>0} <<: {x|x>2}',
'-evalt', '{x|x>0} <<: {x|x>-200}',
'-evalf', 'NATURAL1 <<: 0..10000',
'-evalf', '{x|x<20000} <<: 0..10000',
'-evalf', '{x|x>0} /<: NATURAL1',
'-evalt', '{x|x>=0} /<: NATURAL1',
'-evalf', '{x|x>0} /<: INTEGER',
'-evalf', '{x|x>0} /<: {x|x>-1}',
'-evalf', '{x|x>0} /<: {x|x>0}',
'-evalt', '{x|x>0} /<: {x|x>1}',
'-evalt', 'NATURAL1 /<: 0..10000',
'-evalt', '{x|x>=0} /<<: NATURAL',
'-evalt', '{x|x>=0} /<<: NATURAL1',
'-evalf', '{x|x>=0} /<<: INTEGER',
'-evalt', 'NATURAL /<<: NATURAL1',
'-evalf', '1..100 /<<: NATURAL1',
'-evalt', '0..100 /<<: NATURAL1',
'-evalt', '{x|x<20000} /<<: 0..10000',
'-evalt', 'NATURAL1 /<<: 0..10000',

'-evalf', '0..B <: 1..B & B:1..1000000',
'-evalf', '0..x <: 1..1000 & x>0',
'-evalf', '0..B <<: 1..B & B:1..1000000',
'-evalf', '0..x <<: 1..1000 & x>0',
'-evalt', '{x,y} /<<: NATURAL1 & x:2..3 & y:0..5',
'-evalf', '{x,y} /<<: NATURAL1 & x:2..3 & y:1..5',
'-evalt', '{x,y} <<: NATURAL1 & x:2..3 & y:4..5',
'-evalf', '{x,y} <<: NATURAL1 & x:-22..-3 & y:-4..0',
'-evalt', '%(x,y).( x:1..4 & y : {-1, 0, 1} | x + y)[INTEGER*INTEGER] /\\ {x|x<=3} = 0..3',
'-evalt', '{x,z|x:1..z & x:z..3} = {(1|->1),(2|->2),(3|->3)}' % just a check about interval propagation
%,
%'-evalt', 'x /\\ y /= {} & x<:1..10000 & y<:10000..20000',
%'-evalt', 'x /\\ y /= {} & x = {y|y : 1..n & y mod 2 = 0} & y = {v|v:1..n & v mod 17 = 0} & n = 10000',
%'-evalt', '{x|NATURAL1 /\\ 0..x /= {} & x< 20} = 1..19'
], 'Check constraint propagation of intersection with intervals').
cli_testcase(1167, [laws], ['../prob_examples/public_examples/EventBPrologPackages/Laws/generalized_union_inter_ctx.eventb', '-mc', 1250, '-nodead', '-p', 'DEFAULT_SETSIZE', 3, '-p', 'MAX_INITIALISATIONS', 100, '-strict', '-p', 'CLPFD', 'TRUE', '-assertions', '-cbc_assertions'], 'Check laws about generalized union/inter').
cli_testcase(1168, [b_test,smt_test], ['-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE',
'-evalt', '{y,f|f = %x.(x:1..10|x+y) & f(5)=1005} = {(1000|->[1001,1002,1003,1004,1005,1006,1007,1008,1009,1010])}'
], 'Check constraint propagation for partially known functions').
cli_testcase(1169, [b_test,cbc], ['-p', 'CLPFD', 'TRUE',
 '-p','MAX_INITIALISATIONS', 0, '-cbc_assertions',
'../prob_examples/public_examples/B/CBC/RULE_r967_1.mch', '-expcterr', 'cbc_assertions'
], 'Check ProB finds counter example for rule').
cli_testcase(1170, [b_test,cbc], ['-p', 'CLPFD', 'TRUE',
 '-p','MAX_INITIALISATIONS', 0, '-cbc_assertions_proof', '-p', 'DISPROVER_MODE', 'TRUE',
 '-expcterr', 'cbc_assertions_enumeration_warning',
'../prob_examples/public_examples/B/CBC/RULE_InSetLeavesXY_32_1.mch',
'../prob_examples/public_examples/B/CBC/RULE_r967_preuve_backward.mch'
], 'Check ProB finds counter example for rule').
cli_testcase(1171, [b_test,cbc], ['-p', 'CLPFD', 'TRUE',
 '-p','MAX_INITIALISATIONS', 0, '-cbc_assertions_proof', '-p', 'DISPROVER_MODE', 'TRUE',
 '-expcterr', 'cbc_assertions_enumeration_warning',
'../prob_examples/public_examples/EventBPrologPackages/Disprover/EnumWarningTests/DefSet_ShouldGenerateEnumWarning2.mch'
], 'Check ProB finds no counter example for rule and no proof').
cli_testcase(1172, [b_test,cbc], ['-p', 'CLPFD', 'TRUE',
 '-p','MAX_INITIALISATIONS', 0, '-cbc_assertions_proof', '-p', 'DISPROVER_MODE', 'TRUE',
'../prob_examples/public_examples/EventBPrologPackages/Disprover/EnumWarningTests/DefSet_NoEnumWarning1.mch'
], 'Check ProB finds no counter example for rule and no proof').

cli_testcase(1173, [b_test,cbc], ['-p', 'CLPFD', 'FALSE',
 '-p','MAX_INITIALISATIONS', 0, '-cbc_assertions_proof', '-p', 'DISPROVER_MODE', 'TRUE',
 '-expcterr', 'cbc_assertions_enumeration_warning',
'../prob_examples/public_examples/EventBPrologPackages/Disprover/EnumWarningTests/DefSet_ShouldGenerateEnumWarning1.mch'
], 'Check ProB finds no counter example for rule and no proof').
cli_testcase(1174, [b_test,cbc], ['-p', 'CLPFD', 'TRUE',
 '-p','MAX_INITIALISATIONS', 0, '-cbc_assertions_proof', '-p', 'DISPROVER_MODE', 'TRUE',
 '-expcterr', 'cbc_assertions_enumeration_warning',
'../prob_examples/public_examples/EventBPrologPackages/Disprover/EnumWarningTests/DefSet_ShouldGenerateEnumWarning1.mch'
], 'Check ProB finds no counter example for rule and no proof').
cli_testcase(1175, [b_test], ['-p', 'CLPFD', 'TRUE',
'-evalf', 'n: INTEGER &  {n} = r & r={n1} & (n1/=n)',
'-evalf', 'n: POW(INTEGER) &  {n} = r & r={n1} & (n1/=n)',
'-evalf', 'n: BOOL &  {n} = r & r={n1} & (n1/=n)',
'-evalf', 'n>0 &  {n} ={n1} & (n1/=n)',
'-evalf', 'f:POW(NATURAL) & f2 = f \\/ {p2} & p1:f & p1/:f2 & p1/=p2',
'-evalf', 'f:POW(NATURAL) & f2 = f \\/ {p2} & p1:f & p1/:f2',
'-evalf', 'p1:dom(f) & f2 = {p2} <<| f & p1 /: dom(f2) & p1/=p2 & f:NATURAL <->NATURAL',
'-evalt', 'p1:dom(f) & f2 = {p2} <<| f & p1 /: dom(f2) & p1=p2 & f:NATURAL <->NATURAL',
'-evalf', 'x /= y & s={x,y,z} & x/=z & y/=z & x:NATURAL1 & card(s)<3',
'-evalf', 'x:s & y:s & z:s & x /= y  & x/=z & y/=z & x:NATURAL1 & card(s)<3',
'-evalf', 'x /= y & x/=z & y/=z & s={x,y,z} & card(s)<3 & s<:STRING',
'-evalf', 'f:NATURAL +-> INTEGER & f2 = f <+ {p2|->1} & p1:dom(f) & p1/:dom(f2) & p1/=p2',
'-evalt', 'x /= "a" & (x /= "a" => y>1000000) & (y>1000000 => x = "b") & y<2000000',
'-evalt', 'x /= "a" & (x /= "a" => y>100000 & y<200000) & (y>100000 => x = "b")',
'-evalt', 'x /= "a" & ("a" /= x => y>100000 & y<200000) & (y>100000 => x = "b")',
'-evalt', 'x /= 33 & (x /= 33 => y>1000000) & (y>1000000 => x = 2133) & y<2000000',
'-evalt', 'x /= 1 & (1 /= x => y>100000 & y<200000) & (y>100000 => x = 2)',
'-evalf', 'x /= 1 & (1 /= x => y>100000 & y<200000) & (y>100000 => x = 1)'
], 'Check that a few inconsistencies can be detected without enumeration').
cli_testcase(1176, [por], ['../prob_examples/public_examples/B/Ivo/BenchmarksPOR/Peterson/em_abs.mch', '-mc', 10000, '-noinv', '-p', por, 'ample_sets', '-strict'], 'Model checking with partial order reduction (Without Coverage). Peterson\'s algorithm').
cli_testcase(1177, [por], ['../prob_examples/public_examples/B/Ivo/BenchmarksPOR/Peterson/em_abs.mch', '-mc', 10000, '-check_complete_operation_coverage', '-p', por, 'ample_sets', '-strict'], 'Model checking with partial order reduction (Operations Coverage Check). Peterson\'s algorithm').
cli_testcase(1178, [por], ['../prob_examples/public_examples/B/Ivo/BenchmarksPOR/Peterson/peterson_R.ref', '-mc', 10000, '-check_complete_operation_coverage', '-p', por, 'ample_sets', '-strict'], 'Model checking with partial order reduction (min...max coverage check). Peterson\'s algorithm').
cli_testcase(1179, [por], ['../prob_examples/public_examples/B/Ivo/BenchmarksPOR/Peterson/syspeterson.mch', '-noinv', '-mc', 10000, '-p', por, 'ample_sets', '-strict'], 'Model checking with partial order reduction (min...max coverage check). Peterson\'s algorithm').
cli_testcase(1180, [b_test], ['-p', 'CLPFD', 'TRUE',
'-evalt', '{x|#y.(y>x & (y mod 1000 = 0 or y<x+20)) & x:1..4} = 1..4',
'-evalt', '{x|#y.(y>x & (y mod 1000 = 0 or y<x+20)) & x:1..40} = 1..40',
'-evalf', '{x|#y.(y>x & (y mod 1000 = 0 or y<x+20)) & x:1..40} = 1..4',
'-evalf', '{x|#y.(y>x & (y mod 1000 = 0 or y<x+20)) & x:1..40} /= 1..40',
'-evalt', '{x|#y.(y>x) & x:1..100} /= {}',
'-evalf', '{x|#y.(y>x) & x:1..100} = {}',
'-evalt', 'card({x|#y.(y>x) & x:1..100}) = 100',
'-evalt', 'card({x,y| x>=60000 & x<=64999 & x mod 2 =1 & y=x+x}) = 2500',
'-evalt', 'card({x,y| x>=60000 & x<=64999 & x mod 2 =1 & y=x+x}) = 2500',
'-evalt', 'x = (1 .. 100000) - {4} & card(x) = 99999',
'-evalt', '{x| x:1..100 & {y|y:x..x+10} = {z|z>=x & z<x+11 & #v.(v>x)}} = 1..100',
'-evalt', '[0,0,0,0,0,0,0,0,0,1]:{x|x:seq(0..1) & 1:ran(x)}',
'-evalf', '[0,0,0,0,0,0,0,0,0,0]:{x|x:seq(0..1) & 1:ran(x)}',
'-evalt', '[0,0,1,1,0,1,0]:{x|x:seq(0..1) & card(x)>5}',
'-evalt', '{x| x:1..10 or x:{2**20,2**21} or x = 2**27} = {1,2,3,4,5,6,7,8,9,10,1048576,2097152,134217728}'
],'Check enum warnings properly dealt with').
cli_testcase(1181, [b_test], ['-p', 'CLPFD', 'TRUE', '-expcterr', 'eval_string',
'-evalt', '{x|x:seq(0..1) & card(x)>5} /= {x|x:seq(0..1) & card(x)>6}'
],'Check enum warnings generated').
cli_testcase(1182, [tickets], ['../prob_examples/public_examples/B/Tickets/Hansen6/PerfectNumber_v2.mch', '-t'], 'Check proper working of ProB for perfect number set comprehension').
cli_testcase(1183, [tickets], ['../prob_examples/public_examples/B/Tickets/Valerio3/POWER2.mch', '-init', '-strict', '-aa', '18', '0', '1'], 'Variation of test 274; check that enum warning generated').
cli_testcase(1184, [eventb_test,cbc], ['../prob_examples/examples/EventBPrologPackages/Advance_WP2/SubstationMeter_2703_2013.eventb',  '-strict', '-mc', 1000, '-p', 'CLPFD', 'TRUE', '-expcterr', virtual_time_out], 'Test multi-level animation (like 1052 but without finite restriction on events).').
cli_testcase(1185, [laws], ['../prob_examples/public_examples/EventBPrologPackages/Laws/LawsEventBSpecific.eventb', '-t', '-mc', '1000', '-nodead', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE', '-p', 'CLPFD', 'TRUE'], 'same as 290 (but no virtual time-out with CLPFD)').
cli_testcase(1186, [laws], ['../prob_examples/public_examples/EventBPrologPackages/Laws/finiteTest_ctx.eventb',
  '-mc', '1000', '-p', 'MAX_INITIALISATIONS', 30, '-nodead', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE', '-p', 'CLPFD', 'TRUE', '-cbc_assertions', '-assertions', '-p', 'PROOF_INFO', 'FALSE'], 'some checks with the finite operator').
cli_testcase(1187, [b_test,sequences], ['-evalt', 'x:seq({1}) & 55:dom(x)', '-evalf', 'x:seq({1}) & 55:dom(x) & card(x)=10', '-evalt', 'x:seq({1}) & 5:dom(x) & card(x)>6', '-p', 'CLPFD', 'FALSE', '-strict'
%, '-expcterr', eval_string_enum_warning
], 'Ensure proper enumeration of sequences (CLPFD=FALSE).').
cli_testcase(1188, [b_test,external], ['../prob_examples/public_examples/B/ExternalFunctions/Argv.mch', '-mc', '1000', '-nodead', '-strict', '-argv', 'one two three', '-cc', 5, 12 ], 'check the ARGV external function').
cli_testcase(1189, [b_test,tickets], ['../prob_examples/public_examples/B/Tickets/Hansen7/ClosureTest.mch', '-init', '-properties'], 'Check closure1 enumeration issue resolved').
cli_testcase(1190, [b_test], ['../prob_examples/public_examples/B/Implementations/ReverseArray.mch', '-init', '-mc', 1000, '-p', 'MAX_INITIALISATIONS', 100 ], 'Check simple example with WHILE loop').
cli_testcase(1191, [time_out_tests], ['../prob_examples/public_examples/B/PerformanceTests/SMTLIB/DTP_k2_n35_c175_s1.mch', '-strict', '-init', '-expcterr', 'setup_constants_fails'], 'Copy of 396 using CLPFD. This time, no time_out occurs.').
cli_testcase(1192, [eventb_test,records], ['../prob_examples/public_examples/EventBPrologPackages/Records/SampleThreeRecordMachine_mch.eventb', '-t', '-strict', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-t', '-cc', 4, 6], 'Check that record detection works with nested records').
cli_testcase(1193, [eventb_test,records], ['../prob_examples/public_examples/EventBPrologPackages/Records/IntBoolMachine_mch.eventb', '-t', '-strict', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-t', '-expcterr', virtual_time_out,
 '-evalt', 'make_BIBIRec((TRUE,55),(FALSE,66)) = ((TRUE,55),(FALSE,66))'], 'Check that record detection works with nested records').
cli_testcase(1194, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/platoon/platoon1_prob_simplified.eventb', '-t', '-strict', '-p', 'USE_RECORD_CONSTRUCTION', 'TRUE', '-p', 'MAX_OPERATIONS', 20,  '-p', 'MAX_INITIALISATIONS', 40, '-expcterr', virtual_time_out,
 '-evalt', 'sqrt(0)=0',
 '-evalt', 'sqrt(10)=3',
 '-evalt', 'sqrt(10000)=100',
 '-evalt', 'dist(make(10|->20) |-> make(10|->20)) = 0',
 '-evalt', 'dist(make(10|->20) |-> make(10|->30)) = 10',
 '-evalt', 'dist(make(10|->20) |-> make(0|->20)) = 10',
 '-evalt', 'dist(make(10|->20) |-> make(20|->20)) = 10',
 '-evalt', 'dist(make(10|->10) |-> make(20|->20)) = 14',
 '-evalt', 'dist(make(10|->10) |-> make(10010|->10010)) = 14142',
 '-evalt', 'xcoord(make(1024,2048)) = 1024',
 '-evalt', 'ycoord(make(1024,2048)) = 2048'
 ], 'Check that record detection + symbolic treatment work').
cli_testcase(1195, [performance_tests,b_test], ['../prob_examples/public_examples/B/Mathematical/GraphIso/CheckLargeGraphIsomorphismAsConstants_v2.mch',
 '-t', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'TIME_OUT', '5000'],
 'Test constraint solving for larger graph').
cli_testcase(1196, [b_test], ['../prob_examples/public_examples/B/Benchmarks/CarlaTravelAgencyErr.mch', '-t', '-strict',
   '-csvhist', 'card(session) |-> card(session_response) |-> card(user_hotel_bookings) |-> user_rental_bookings',
               '../prob_examples/public_examples/B/Benchmarks/CarlaTravelAgencyErr.csv'],
 'Test evaluating expression over history and saving to CSV file').
cli_testcase(1197, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/Tests/PartitionTest_ctx.eventb', '-init', '-assertions', '-strict'], 'Test partition border cases').
cli_testcase(1198, [b_test,infinite,total_function], [
 '-evalt', '(NATURAL * NATURAL)* {1024} : (NATURAL*NATURAL) --> NATURAL',
 '-evalt', 'NATURAL * {0} : NATURAL --> NATURAL',
 '-evalt', 'NATURAL * {0} : NATURAL +-> NATURAL',
 '-evalt', 'NATURAL1 * {0} : NATURAL +-> NATURAL',
 '-evalt', 'NATURAL1 * {0} : INTEGER +-> NATURAL',
 '-evalf', 'NATURAL * {0} : NATURAL --> NATURAL1',
 '-evalf', 'NATURAL * {0} : NATURAL +-> NATURAL1',
 '-evalf', 'NATURAL1 * {0} : NATURAL --> NATURAL',
 '-evalf', 'NATURAL * {0} : NATURAL1 --> NATURAL'
 ], 'Test cartesian product with singleton set recognised as total function').
cli_testcase(1199, [eventb_test,tickets,infinite, theories], ['../prob_examples/public_examples/EventBPrologPackages/Tickets/Kanso_Symbolic/C_Cart_ctx.eventb',
 '../prob_examples/public_examples/EventBPrologPackages/Tickets/Kanso_Symbolic/C_bijNAT_ctx.eventb',
 '../prob_examples/public_examples/EventBPrologPackages/Tickets/Kanso_Symbolic/C_symbolic_ctx.eventb',
 '../prob_examples/public_examples/EventBPrologPackages/Theory/ConstraintSolving_Demo_ctx.eventb',
 '-init','-assertions', '-strict'], 'Checking that various infinite functions can be defined and checked').
cli_testcase(1200, [b_test], ['../prob_examples/public_examples/B/Mathematical/MaxClique.mch',
 '-init','-assertions', '-strict'], 'Checking constraint solving capabilities').
cli_testcase(1201, [eventb_test, theories], ['../prob_examples/public_examples/EventBPrologPackages/Theory/SeqTest2_ctx.eventb', '-init', '-assertions', '-strict'], 'Test Theory Plugin support').
cli_testcase(1202, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Tickets/Kanso_ModularExp/ExpModFunctionCtxt_ctx.eventb',
 '../prob_examples/public_examples/EventBPrologPackages/Tickets/Kanso_ModularExp/ModularExponenation.mch', '-init', '-assertions', '-strict'], 'Test Recursive Function support').
cli_testcase(1203, [b_test,queens], ['../prob_examples/public_examples/B/Puzzles/PeaceableQueens/JustQueens_8_8.mch', '-init', '-strict'], 'A constraint solving example').
cli_testcase(1204, [eventb_test, theories], ['../prob_examples/public_examples/EventBPrologPackages/Theory/NaturalTest.eventb', '-mc', 10, '-nodead', '-cc', 8, 7, '-strict'], 'Checking the operators of the inductive theory of natural numbers').
cli_testcase(1205, [eventb_test, theories], ['../prob_examples/public_examples/EventBPrologPackages/Theory/RecOpTypeTest.eventb', '-assertions', '-strict'], 'Checking a reference to a theory\'s type parameter in a recursive operator definition').
cli_testcase(1206, [tickets], ['../prob_examples/public_examples/B/Tickets/PROB-320/Hierarchical_temp.mch', '-t',
'-expcterr', 'precondition_error'], 'Check precondition error raised and error message generation works. (Ticket PROB-320)').
cli_testcase(1207, [tickets], ['-p', 'TIME_OUT', 7000, '-p', 'CLPFD', 'TRUE',
'-evalf', '61*(x*x)+1 = y*y & x:1..1500000 & y>0'], 'Check that no instantiation error or other problem occurs').
cli_testcase(1208, [typechecker_test,unit], ['-typecheckertest','../prob_examples/public_examples/B/Typecheckertests/including.mch', '-strict'], 'Typechecker test when including other machines').
cli_testcase(1209, [typechecker_test,unit], ['-typecheckertest','../prob_examples/public_examples/B/Typecheckertests/definitions.mch', '-strict'], 'Typechecker test when using definitions').
cli_testcase(1210, [typechecker_test,unit,parser], ['-typecheckertest','../prob_examples/public_examples/B/Typecheckertests/refi.ref', '-strict'], 'Typechecker test when using refinement').
cli_testcase(1211, [csp_test], ['../prob_examples/public_examples/CSP/Tickets/Ivo/ListPattern.csp', '-t', '-strict'], 'Testing compiling of list patterns given as function parameters.').
cli_testcase(1212, [csp_test], ['../prob_examples/public_examples/CSP/ucsexamples/chapter15/cache.csp', '-assertions', '-strict'], 'Testing compiling of list patterns given as function parameters.').
cli_testcase(1213, [puzzles_test], ['../prob_examples/public_examples/B/Puzzles/EulerWay.mch', '-init'], 'Test that this puzzle can be solved quickly; used to take very long; solved in svn commit r12434').
cli_testcase(1214, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/OverridingBuiltinFunctions.csp', '-assertions', '-strict'], 'Testing if we can override built-in functions.').
cli_testcase(1215, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/OverridingBuiltinFunctionsLets.csp', '-assertions', '-strict'], 'Testing if we can override built-in functions.').
cli_testcase(1216, [b_test,cbc], [
 '-evalt', 'z:10001..10002 & {x,y} \\/ t   = {z}',
 '-evalt', 'z:10001..10002 & {x} \\/ {y}   = {z}',
 '-evalf', 'z:10001..10002 & {x,y} \\/ t   = {z} & y<10001',
 '-evalf', 'z:10001..10002 & {x,y} \\/ t   = {z} & t<:9000..10000 & t/= {}',
 '-evalf', 'z:10001..10002 & {x,y} \\/ t   = {z} & t:POW1(9000..10000)',
 '-evalt', 'z:10001..10002 & {x} \\/ {y,v}   = {z}',
 '-evalt', 'z:10001..10002 & {x,v} \\/ {y}   = {z}',
 '-evalt', 'z:10001..10002 & {x,v} \\/ t   = {z} & {a} \\/ {b} = t',
 '-evalt', 'union({{x},{y},{z}}) = {3333}', % works by ast_cleanup translating to union
 '-evalt', 'union({{x}}) = {3333}', % ditto
 '-evalt', 'union({{x},{y},{z},{w}}) = {v} & v:200001..300001 & w:300001..400001', % ditto
 '-evalt', 'union({S,V,W}) = 1..10 & S/\\V = {} & S/\\W ={} & V /\\ W = {} & card(S) < card(V) & card(V) < card(W)' % ditto + partition detection
 % TO DO: z:10001..10002 & {x,w} \\/ {y,v}   = {z}
 ], 'Testing propagation in union from result to arguments.').
cli_testcase(1217, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/Disprover/ProofTests/TestPrj_Nat_ctx.eventb', '-cbc_assertions',
 '-p', 'PROOF_INFO', 'FALSE', '-strict'], 'Testing if prj1|->prj2 theorems can be proven').
cli_testcase(1218, [b_test,sequences], [
 '-evalf', 'p:perm(5..20) & p(10)=21'
 ], 'Testing propagation for permutation sequence.').
cli_testcase(1219, [puzzles_test,queens], ['-p', 'CLPFD', 'TRUE', '../prob_examples/public_examples/B/Puzzles/NQueens40_perm.mch', '-init', '-properties', '-strict'], '').
cli_testcase(1220, [csp_test], ['../prob_examples/public_examples/CSP/other/Ivo/AssertionsLet.csp', '-csp_assertion', 'c?x -> c.x -> STOP [T= c?x:{1..3} -> c.x -> STOP', '-assertions', '-cc', 38, 41, '-strict'], '').
cli_testcase(1221, [ltl], ['../prob_examples/public_examples/CSP/Tickets/PROB-319/toy3.csp', '-ltllimit', 10000, '-ltlfile','../prob_examples/public_examples/CSP/Tickets/PROB-319/toyprop3c.ltl','-expcterr', ltl_counterexample, '-strict'], 'Regression test for ticket PROB-319: Unable to find a counter-example in the SCC').
 cli_testcase(1222, [private], ['-p', 'CLPFD', 'TRUE', '../prob_examples/examples/B/ClearSy/alloc_large.mch', '-init', '-p', 'TIME_OUT', '18500'],
 'Test that ProB finds isomorphism for large graph'). % increased time out from 9500
cli_testcase(1223, [b_test], [
 '-evalf', 'z:11..10002 & z2:10003..520004 & z3: -1000..10 & z4:-22222..-2000 &a: {z,z2} &  a:{z3,z4}',
 '-evalt', 'z:11..10002 & z2:10003..520004 & z3: -1000..11 & z4:-22222..-2000 &a: {z2,z} &  a:{z3,z4}',
 '-evalf', 'x <: 1..7 & y <: 1..7 & (x<:y <=> y<:x) & x/\\y = {3} & x\\/y = 1..7 & card(x) = card(y)+1',
 '-evalt', 'y:1..1000001 & f:100001..100005 --> 1..9000  & x:dom(f) & x:2..100003 & (x>5000 => y:100001..100002)',
 '-evalf', 'f: 1..10 --> 100..200 & x:dom(f~) & x>200',
 '-evalf', 'f: 10..200 --> 1..200000 & x|->y : f & (x>200 or x<10)',
 '-evalf', 'f: (BOOL*BOOL) --> 100..200000 & b|->x : f & (x>200000 or x<100)',
 '-evalf', 'f: 10001..10110 --> NATURAL & !x.(x:dom(f) => f(x):dom(f)) & g:20010..20020 --> BOOL & !x.(x:dom(f) => f(x):dom(g))',
 '-evalt', 'f : 1001..2001 --> 1900..3333 & x|->y : f & y<x',
 '-evalf', 'f : 1001..2001 --> 1900..3333 & x|->y : f & y+101<x',
 '-evalf', 'x:1..400 & y:1..400 & f: 2..390 --> 391..399 & x|->y :f & y|->x :f',
 '-evalt', 'r: 1001..1005 <-> 1000..1099 & x|->y : r & v|->w : r &  x+y = v+w+103 & card(r) = 3',
 '-evalt', 'card({r,x,y,v,w|r: 1001..1005 <-> 1000..1099 & x|->y : r & v|->w : r &  x+y = v+w+103 & card(r) = 2}) = 1',
 '-evalt', '{r,x,y,v,w|r: 1001..1005 <-> 1000..1099 & x|->y : r & v|->w : r &  x+y = v+w+103 & card(r) = 2} = {(((({(1001|->1000),(1005|->1099)}|->1005)|->1099)|->1001)|->1000)}',
 '-evalt', 'card({r,x,y,v,w|r: 1001..1005 <-> 1000..1099 & x|->y : r & v|->w : r &  x+y = v+w+103 & card(r) = 3}) = 498',
 '-evalt', 'card({r,x,y,v,w|r: 1001..1005 +-> 1000..1099 & x|->y : r & v|->w : r &  x+y = v+w+103 & card(r) = 3}) = 300',
 '-evalt', 's = {(1,rec(a:1)), (2,rec(a:2))} &  (1,rec(a:x)) : s',
 '-evalt', 'a = {rec(x:1,y:22),rec(x:2,y:44)} & {v,w|rec(x:v,y:w):a} = {(1,22),(2,44)}',
 '-evalt', 'a = {rec(x:(1,33),y:22),rec(x:(2,34),y:44),rec(x:(3,34),y:45)} & {v1,w|rec(x:(v1,34),y:w):a} = {(2|->44),(3|->45)}',
 '-evalf', 'a = {rec(x:(1,33),y:22),rec(x:(2,34),y:44),rec(x:(3,34),y:45)} & rec(x:(v1,v2),y:w):a & v1+v2+w <56',
 '-evalt', 'a = {rec(x:(1,33),y:22),rec(x:(2,34),y:44),rec(x:(3,34),y:45)} & rec(x:(v1,v2),y:w):a & v1+v2+w <57',
 '-evalt', 'ran({vv,ww,x|x=rec(x:ww,y:vv) & vv:1..n & ww:33..34})=r1 & n=50 &  rec(x:a,y:b):r1 & rec(x:a+1,y:b):r1 & b>a',
 '-evalt', 'a = {rec(x:(1,33),y:22,z:TRUE),rec(x:(2,34),y:44,z:TRUE),rec(x:(3,34),y:45,z:TRUE)} & rec(x:(v1,v2),y:w,z:bb):a & v1+v2+w <57',
 '-evalf', 'a = {rec(x:(1,33),y:22,z:FALSE),rec(x:(2,34),y:44,z:TRUE),rec(x:(3,34),y:45,z:TRUE)} & rec(x:(v1,v2),y:w,z:TRUE):a & v1+v2+w <57'
 ], 'Testing propagation for membership.').
cli_testcase(1224, [typechecker_test,tickets], ['-init','../prob_examples/public_examples/EventBPrologPackages/Tickets/PROB-323/PROB-323.eventb','-strict'], 'This machine should not throw a type error.').
cli_testcase(1225, [eventb_test,laws, theories], ['../prob_examples/public_examples/EventBPrologPackages/Theory/ListLaws_ctx.eventb',
 '../prob_examples/public_examples/EventBPrologPackages/Theory/PairLaws_ctx.eventb',
 '-assertions','-strict'], 'Various checks for sets involving freetypes (Lists Theory)').
cli_testcase(1226, [b_test,finite], [
 '-evalf', 's=({x|x mod 2 = 0 & x>200}) & s:FIN(s)'
 ,  '-expcterr', eval_string_enum_warning
 ], 'Check that finite warning produced').
cli_testcase(1227, [eventb_test,laws, theories], ['../prob_examples/public_examples/EventBPrologPackages/Theory/ListLawsMC_mch.eventb','-mc',500, '-bf', '-strict',
 '-expcterr', 'virtual_time_out'], 'Various checks for sets involving freetypes (Lists Theory)').
cli_testcase(1228, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/Disprover/EnumWarningTests/MammarLaleau_Failures_ReadDoors_inv8_selected.mch', '-cbc_assertions',
 '-p', 'PROOF_INFO', 'FALSE', '-strict'], 'Check no enum warning occurs when using bool with integer inside').
cli_testcase(1229, [eventb_test], ['../prob_examples/public_examples/EventBPrologPackages/Tickets/PROB-322/ExpressionTest.eventb','-init','-strict'], 'Checking usage of typeof-Operator').
cli_testcase(1230, [cbc], ['../prob_examples/public_examples/EventBPrologPackages/Disprover/EnumWarningTests/MammarLaleau_Cylinders_LockGearCylinder_inv4_selected.mch',
%'../prob_examples/public_examples/EventBPrologPackages/Disprover/EnumWarningTests/MammarLaleau_Doors_Make_GearExtended_inv2_all.mch',
'-cbc_assertions',
 '-p', 'PROOF_INFO', 'FALSE' ,'-p', 'DISPROVER_MODE', 'TRUE', '-strict'], 'Check no time-out occurs').
cli_testcase(1231, [b_test,fairness,ltl], ['../prob_examples/public_examples/B/LTL/Fairness/TestingFairness.mch', '-ltlassertions', '-strict'], 'Test fairness implementation.').
cli_testcase(1232, [b_test,fairness,ltl], ['../prob_examples/public_examples/B/LTL/Fairness/MutualExclusion.mch', '-ltlassertions', '-strict'], 'Test fairness implementation.').
cli_testcase(1233, [b_test,fairness,ltl], ['../prob_examples/public_examples/B/LTL/Fairness/WeakFairnessTest.mch', '-ltlassertions', '-strict'], 'Test fairness implementation.').
cli_testcase(1234, [csp_test,fairness,ltl], ['../prob_examples/public_examples/CSP/LTL/phils.csp', '-ltlassertions', '-strict'], 'Test fairness implementation.').
cli_testcase(1235, [csp_test,fairness,ltl], ['../prob_examples/public_examples/CSP/LTL/FairnessUnderscoreArgs.csp', '-ltlassertions', '-strict'], 'Test fairness.').
cli_testcase(1236, [b_test,fairness,ltl], ['../prob_examples/public_examples/B/LTL/Fairness/RecursiveSCCs.mch', '-ltlformula', 'SF(C) => GF {addr=4 or addr =5}', '-expcterr', ltl, '-strict'], 'Test fairness implementation for finding the right counter example.').
cli_testcase(1237, [b_test,fairness,ltl], ['../prob_examples/public_examples/B/LTL/Fairness/RecursiveSCCs.mch', '-ltlformula', 'SF(E) => GF {addr=4 or addr =5}', '-expcterr', ltl, '-strict'], 'Test fairness implementation for finding the right counter example.').
cli_testcase(1238, [b_test,fairness,ltl], ['../prob_examples/public_examples/B/LTL/Fairness/RecursiveSCCs.mch', '-ltlformula', 'SF(D) => GF {addr=4 or addr =5}', '-expcterr', ltl, '-strict'], 'Test fairness implementation for finding the right counter example.').
cli_testcase(1239, [b_test,fairness,ltl], ['../prob_examples/public_examples/B/LTL/Fairness/RecursiveSCCs.mch', '-ltlformula', 'SF(C) & SF(E) => GF {addr=4 or addr =5}', '-expcterr', ltl, '-strict'], 'Test fairness implementation for finding the right counter example.').
cli_testcase(1240, [b_test,fairness,ltl], ['../prob_examples/public_examples/B/LTL/Fairness/RecursiveSCCs.mch', '-ltlformula', '(SF(C) & SF(E)) or (SF(F) & SF(E)) => GF {addr = 4 or addr = 5}', '-expcterr', ltl, '-strict'], 'Test fairness implementation for finding the right counter example.').
cli_testcase(1241, [b_test,fairness,ltl], ['../prob_examples/public_examples/B/LTL/Fairness/RecursiveSCCs.mch', '-ltlformula', '(SF(C) & SF(E)) or (SF(F) & SF(E)) => GF {addr = 4 or addr = 5}', '-expcterr', ltl, '-strict'], 'Test fairness implementation for finding the right counter example.').
cli_testcase(1242, [csp_test,fairness,ltl], ['../prob_examples/public_examples/CSP/LTL/rabp_scc_issue.csp', '-ltlassertions', '-strict'], 'Test fairness.').
cli_testcase(1243, [csp_test,fairness,ltl], ['../prob_examples/public_examples/CSP/LTL/rabp_scc_issue.csp', '-ltlformula', 'SEF => G ([left.0] => F [right.0])', '-strict'], 'Test general fairness.').
cli_testcase(1244, [b_test, ltl], ['../prob_examples/public_examples/B/LTL/Fairness/TestingFairness.mch',
 '-ltlformulat', true,  '-ltlformulat', 'G{1=1}',
 '-ctlformulat', 'AG{1=1}', '-ctlformulaf', 'EF{1=2}',
 %'-ltlformulaf', false, % does not work !
 '-ltlformulaf', 'G{1=2}', '-strict'], 'Testing the trivial LTL formula true + other trivial ones.').
cli_testcase(1245, [b_test, ltl], ['../prob_examples/public_examples/B/LTL/Fairness/TestingFairness.mch', '-ltlformulat', false, '-expcterr', 'ltl', '-strict'], "Testing the trivial LTL formula false.").
cli_testcase(1246, [cbc], [
 '-evalf', 'x/0 = 1 & x:1..20',
 '-evalf', 'x:-2..2 &  10/x =1',
 '-evalt', 'x:-2..2 & 10/x = 5',
 '-p', 'DISPROVER_MODE', 'TRUE'], 'Testing no WD error generated in disprover mode.').
cli_testcase(1247, [cbc], [
 '-evalt', 'i>200 & (i mod 37 = 7)',
 '-evalt', 'i>200 & (i mod 37 = 1)',
 '-evalt', 'i>20 & (i mod 2 = 0)',
 '-evalf', 'i>200 & (i mod 37 = 7) & i < 229',
 '-evalt', '{i|i>200 & (i mod 50 = 10) & i<1000} = {210,260,310,360,410,460,510,560,610,660,710,760,810,860,910,960}',
 '-evalt', '{i|i>200 & (i mod 50 = 10) & i<1000 & i mod 100 /= 10} = {260,360,460,560,660,760,860,960}',
 '-evalt', 'x > 0 &  x mod 30 =0 & x mod 19 = 0',
 '-evalt', 'x > 0 &  x mod 50 =0 & x mod 61 = 0 & x mod 23 = 0',
 '-evalf', 'x > 0 &  x mod 50 =0 & x mod 61 = 0 & x mod 23 = 0 & x<70150'
 ], 'Testing improved constraint propagation for modulo.').
cli_testcase(1248, [tickets], ['../prob_examples/public_examples/B/Tickets/Disprover_InstantiationError/GraphColour_PO.mch', '-init', '-p', 'TIME_OUT', 300, '-p', 'DISPROVER_MODE', 'TRUE',
'-expcterr', setup_constants_fails, '-expcterr', time_out, '-strict'], 'Check that no instantiation error occurs').
cli_testcase(1249, [cbc,slot], ['../prob_examples/public_examples/B/PerformanceTests/SlotToolSolver_Inconsistency.mch', '-init', '-p', 'TIME_OUT', 2500, '-expcterr', setup_constants_fails, '-strict',
'-expcterr', no_det_value_stored_for_constant], 'Check that inconsistency detected quickly').
cli_testcase(1250, [cbc], [
 '-evalf', 'f:0..100 --> POW(1..100) & f(20) = {99} & !x.(x:dom(f) => !y.(y:f(x) => y /= 99))',
 '-evalt', 'f:0..100 --> POW(1..100) & f(20) = {98} & !x.(x:dom(f) => !y.(y:f(x) => y /= 99))',
 '-evalf', 'f:0..100 --> POW(1..100) & f(20) = {99} & !x.(x:dom(f) => !y.(y:f(x) & y>20 => y /= 99))'
 ], 'Check universal quantification inconsistency detected').
cli_testcase(1251, [cbc], ['../prob_examples/examples/RodinModels/Pacemaker/PO_M4_AATR_Decrease_Interval_inv2_INV.mch', '-init', '-p', 'TIME_OUT', 500, '-p', 'DISPROVER_MODE', 'TRUE', '-expcterr', setup_constants_fails, '-strict'], 'Check that inconsistency detected quickly').
cli_testcase(1252, [cbc,wd], ['../prob_examples/examples/RodinModels/Pacemaker/PO_M4_AATR_Decrease_Interval_inv2_INV.mch', '-init', '-p', 'TIME_OUT', 300, '-p', 'DISPROVER_MODE', 'FALSE', '-expcterr', setup_constants_fails, '-expcterr', time_out, '-expcterr', well_definedness_error, '-strict'], 'Check that division by zero detected.').
cli_testcase(1253, [tickets], ['../prob_examples/public_examples/B/Tickets/LambdaUnionEnumWarning/Seq1UnionFunction.mch',
 '-init', '-assertions', '-strict'], 'Check that symbolic function not expanded in rel. composition').
cli_testcase(1254, [cbc, tickets, chr], ['-p', 'CHR', 'TRUE',
 '-evalf', 'x> a & x: 33..a',
 '-evalt', 'x>=a & x>=33 & x<=a',
 '-evalf', 'x>a & x>=33 & x<=a'
 ], 'Check CHR detects inconsistencies').
cli_testcase(1255, [cbc], ['-p', 'DISPROVER_MODE', 'TRUE',
  '-evalf', 'z>0 & z mod z = 100 & x:{1,2,3} & not(x>0)'], 'Check no enum warning').
cli_testcase(1256, [tickets], [
'../prob_examples/public_examples/EventBPrologPackages/Tickets/Cansell_RingLead/ring_ctx.eventb',
  '-init'], 'Check no infinite loop when computing minimum cardinality').
cli_testcase(1257, [b_test,csp_test,ltl,fairness,cspb],['../prob_examples/public_examples/CSPB/Williams/ME.mch', '-csp-guide', '../prob_examples/public_examples/CSPB/Williams/ME.csp', '-ltlassertions', '-strict'],'Check ltl assertions with imposed fairness constraints in (ME.mch|| ME.csp).').
cli_testcase(1258, [eventb_test,ltl],['../prob_examples/public_examples/B/Ivo/BenchmarksPOR/Concurrent/fact_m_v3_mch.eventb', '-ltlformulat', 'F ([found0])', '-strict'],'Check an ltl formula on an Event-B model.').
cli_testcase(1259, [b_test,csp_test,ltl,fairness,cspb],['../prob_examples/public_examples/CSPB/Williams/pris.mch', '-csp-guide', '../prob_examples/public_examples/CSPB/Williams/pris.csp', '-ltlformulat', 'SF(In.p1) & SF(In.p2) & SF(In.p3) => F[Done]', '-strict'],'Check an LTL formula with imposed fairness constraints in (pris.mch|| pris.csp).').
cli_testcase(1260, [b_test],['../prob_examples/public_examples/B/NewSyntax/FreetypeIntList.mch','-p','MAXINT',10,'-t','-strict'],'Check rudimental freetype functionally').
cli_testcase(1261, [cbc], [
 '-evalt', '(x:{y|y:NATURAL & y mod 100 = 1} <=> v=0) & v=0 & x>1'
 ], 'Test reification of closures').
cli_testcase(1262, [b_test],['../prob_examples/public_examples/B/Tickets/Hansen8/EnumerateFunctionTests.mch','-mc', 1000, '-cc', 4,10,'-strict'],'Check union with full relations does not time out').
cli_testcase(1263, [b_test,tickets],['../prob_examples/public_examples/EventBPrologPackages/Tickets/CardTruthBody/DPLL_Backtrack_mch_not_symbolic.eventb', '-t', '-mc', 100, '-strict'],'Check function unit_resolve detected as symbolic without symbolic annotation').
cli_testcase(1264, [b_test,infinite], ['../prob_examples/public_examples/B/Mathematical/FormaleSprachen.mch',
 '-init','-assertions', '-strict'], 'Checking various operations work on infinite sets (e.g., complement set).').
cli_testcase(1265, [tickets], ['../prob_examples/public_examples/EventBPrologPackages/Tickets/MinCardInference/RegularGrammar_Ex1_ctx.eventb', '-init',
 '-evalt', 'card(Symbols)=4',
 '-evalt', 'card(NonTerminals)=2',
 '-evalt', 'card(Terminals)=2'], 'Check that minimum cardinality inference works').
cli_testcase(1266, [b_test],['../prob_examples/public_examples/B/Tickets/Hansen9/SimpleCSGGrammar2_SlowCLPFD.mch','-t','-strict','-p','CLPFD','TRUE'],'Check no time-out').
cli_testcase(1267, [cbc,sequences], [
 '-evalt', '{n|[11,22,3,4,5,6,7] /|\\ n = [11,22]} = {2}'
 , '-evalt', '{l|[2222]^l /|\\ 7 = l} = {[2222,2222,2222,2222,2222,2222,2222]}'
 , '-evalt', '{x,n|[1111,x,33,44] /|\\ n = [1111,2222] & n:0..4} = {(2222,2)}'
 , '-evalt', '{x,y,n|[1111,x,3333,4444,y,y] /|\\ n = [1111,2222,3333,4444,5555] & n:0..6} = {((2222|->5555)|->5)}'
 %,'-evalt', '{n,x,y|[11,22,3,4,5,6,7] /|\ n = [x,y]} = {((2|->11)|->22)}' % does not work yet TO DO
 ],
 'test sequence constraint solving').
cli_testcase(1268, [b_test,recursion,external], ['../prob_examples/public_examples/B/RecursiveFunctions/Factorial_TLAStyle.mch', '-init', '-assertions', '-strict', '-p', 'SYMBOLIC', 'FALSE'], 'Test that other recursive function works also with TLA style if-then-else.').
cli_testcase(1269, [tickets], [
 '-evalt', 'm={{1,2},{1,3}} & {k1,k2|k1:m & 1:k1 & k2:m & 1:k2 & k1/=k2} =r & card(r)=2',
 '-evalt', 'm={{1,2},{1,3},{2,3}} & {k1,k2|k1:m & 1:k1 & k2:m & 1:k2 & k1/=k2} =r & card(r)=2',
 '-evalt', 'm={{1,2},{1,3}} & {k1,k2|k1:m & 1:k1 & k2:m-{k1} & 1:k2} =r & card(r)=2',
 '-evalt', 'm={{1,2},{1,3},{2,3}} & {k1,k2|k1:m & 1:k1 & k2:m-{k1} & 1:k2} =r & card(r)=2'
 ], 'ensure bug in not_equal_object for sets remains fixed').
cli_testcase(1270, [laws], ['../prob_examples/public_examples/B/Laws/SetLawsPow2.mch', '-mc', '200', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE'], 'More laws about sets of sets').
cli_testcase(1271, [b_test],['../prob_examples/public_examples/B/Tickets/Hansen10/PrettyPrintBug.mch','-mc', 200, '-t', '-his', '../prob_examples/public_examples/B/Tickets/Hansen10/PrettyPrintBug_his.txt','-strict'],'Check no time-out').
cli_testcase(1272, [cbc], [
 '-evalf', 'card({x|x:1..10 & x*x<i}) = 0 & i>1',
 '-evalt', 'card({x|x>1 & x<2**e & x mod 2 = 0}) = 0',
 '-evalt', 'card({b|b:1..8 & b mod n = 1})=c & n:2..4',
 '-evalt', 'f: 1..n --> 0..1000 & f(1)=0 & card({i|i:1..n & f(i)>0}) = n/2 & n=100',
 '-evalt', 'n=20 & f:1..n --> BOOL & card({x|x:1..n & f(x)=TRUE})=n/2 & !y.(y:1..(n-1) => ( f(y)=TRUE => f(y+1)=TRUE))',
 '-evalt', 'n=100 & f:1..n --> BOOL & card({x|x:dom(f) & f(x)=TRUE})=50'
 ], 'ensure proper reification of card').
cli_testcase(1273, [cbc], [
 '-evalf', '{10,22} <: {x|x>10 & x mod 2 = 0}',
 '-evalt', '{100,22} <: {x|x>10 & x mod 2 = 0}',
 '-evalf', '{11,22} <: {x|x>10 & x mod 2 = 0}',
 '-evalt', 'card({S|S <<: 12..15 & (12:S => 15:S & 14 /:S) & (13:S <=> 12:S) & S/={}})=4',
 '-evalt', 'card({S|S <<: 12..14 & (12:S => 15:S & 14 /:S) & (13:S <=> 12:S) & S/={}}) =1',
 '-evalt', 'card({S|S <<: 12..19 & (12:S => 15:S & 14 /:S) & (13:S <=> 12:S) & S/={}})=79',
 '-evalt', 'card({S|S <<: 11..18 & (12:S => 15:S & 14 /:S) & (15:S <=> 12:S) & S/={}})=95',
 '-evalt', 'card({S|S <: 11..18 & (12:S => 15:S & 14 /:S) & (15:S <=> 12:S) & S/={}})=95'
 ], 'ensure no expansion in subset and proper computation of all solutions').
cli_testcase(1274, [tickets],['../prob_examples/public_examples/B/Tickets/Hansen11/RecNatFunLoop.mch', '-init', '-assertions', '-strict', '-vv'],'Check no infinite recursion').
cli_testcase(1275, [tickets,typechecker_test,let],['../prob_examples/public_examples/B/ErrorMachines/TestScopeLHSAssignmentLET.mch', '-init', '-strict', '-expcterr', type_error, '-expcterr', load_main_file],'LET introduces a duplicate variable').
cli_testcase(1276, [tickets,cbc],['../prob_examples/public_examples/B/Compilers/LLParsing.mch', '-t', '-strict'],'Check no enumeration warning due to existential quantifiers').
cli_testcase(1277, [tickets,typechecker_test],['../prob_examples/public_examples/B/Tickets/PROB-333/InvalidAssignmentBecomesSuch.mch', '-init', '-strict', '-expcterr', type_error, '-expcterr', load_main_file],'A becomes such assignment is not allowed to write to an read-only identifier').
cli_testcase(1278, [tickets,typechecker_test],['../prob_examples/public_examples/B/Tickets/PROB-333/InvalidAssignmentElementOf.mch', '-init', '-strict', '-expcterr', type_error, '-expcterr', load_main_file],'An element of assignment is not allowed to write to an read-only identifier').
cli_testcase(1279, [tickets,typechecker_test],['../prob_examples/public_examples/B/Tickets/PROB-333/InvalidAssignmentOpCall.mch', '-init', '-strict', '-expcterr', type_error, '-expcterr', load_main_file],'An operation call assignment is not allowed to write to an read-only identifier').
cli_testcase(1280, [tickets,typechecker_test],['../prob_examples/public_examples/B/Tickets/PROB-333/InvalidAssignmentToAnyVar.mch', '-init', '-strict', '-expcterr', type_error, '-expcterr', load_main_file],'Assignment to variables introduced by ANY are not allowed').
cli_testcase(1281, [tickets,typechecker_test,let],['../prob_examples/public_examples/B/Tickets/PROB-333/InvalidAssignmentToLetVar.mch', '-init', '-strict', '-expcterr', type_error, '-expcterr', load_main_file],'Assignment to variables introduced by LET are not allowed').
cli_testcase(1282, [tickets,let],['../prob_examples/public_examples/B/Tickets/PROB-333/M1.mch', '-init', '-strict'
, '-evalt', '#(b).(b:NAT)'
, '-evalt', '#(a).(a:NAT)'
, '-evalt', '!(a).(a:ID => card(ID-{a}) = 1)'
, '-evalt', '!(ab).(ab:ID => card(ID-{ab}) = 1)'
, '-evalt', ' #aa.(not(myTRUE))'
, '-evalt', '{aa|myTRUE} = {aa}'
],'Assignment to variables introduced by LET are not allowed').
cli_testcase(1283, [b_test,wd], ['../prob_examples/public_examples/B/ErrorMachines/NestedFunUpdate_Error.mch', '-nodead', '-mc', 10, '-strict',
 '-expcterr', 'well_definedness_error'], 'Ensure we find error in assignment.').
cli_testcase(1284, [eventb,tickets], ['../prob_examples/public_examples/EventBPrologPackages/Tickets/SlowLambdaFunctions/TestParIf2_v2_mch.eventb',
  '-p', 'MAX_OPERATIONS', 40,
  '-p', 'TIME_OUT', 1000,
  '-mc', 1000, '-expcterr', 'model_check_incomplete', '-cc', 193, 4369,
  '-strict'], 'Ensure performance acceptable for nested function applications.').
cli_testcase(1285, [cbc], ['../prob_examples/public_examples/B/Tickets/PartialFunCBCPerformance/Prove_simpler_ctx.eventb',
 '../prob_examples/public_examples/B/Tickets/PartialFunCBCPerformance/ProveCommutativityAxioms_ctx.eventb',
  '-p', 'TIME_OUT', 1000,   '-p', 'MAX_INITIALISATIONS', 90,
  '-init', '-mc', 100, '-nodead', '-strict'], 'Ensure we find solution to initalisation.').
cli_testcase(1286, [eventb_test,choose, theories],['../prob_examples/public_examples/EventBPrologPackages/Theory/TestHilbert.eventb','-assertions','-strict'], 'Test if the tagged CHOOSE operator in a theory works.').
cli_testcase(1287, [eventb_test, theories],['../prob_examples/public_examples/EventBPrologPackages/Theory/RecursiveSetFunctions_ctx.eventb','-assertions', '-init', '-strict',
 '-evalt', 'sumset(1..1000) = 500500'], 'Test that recursive functions with COND work and that we can use Theory operators in evalt.').
cli_testcase(1288, [tickets,smt_test,chr],['../prob_examples/examples/B/Siemens/RuleValidation/v7_4_trad_pour_m_leuschel-2012-05-04/OrderXY_78_1.mch',
'-p', 'CLPFD', 'TRUE', '-p', 'SMT', 'TRUE',  '-p', 'CHR', 'TRUE',  '-p', 'DISPROVER_MODE', 'TRUE', '-p', 'TIME_OUT', 100,
'-expcterr', cbc_assertions_time_out,
'-cbc_assertions','-strict'], 'Test Siemens Proof Rule.').
cli_testcase(1289, [tickets,records],['../prob_examples/public_examples/B/Tickets/Hansen12/RecordCallResidue.mch', '-init', '-assertions', '-strict'],'Check no call residue for records.').
cli_testcase(1290, [tickets],['../prob_examples/public_examples/B/Tickets/Hansen13/VirtualTimeout.mch', '-init', '-mc', 100, '-strict', '-expcterr', 'model_check_incomplete','-expcterr', 'time_out_for_invariant'],'Check incomplete model checking detected.').
cli_testcase(1291, [tickets,infinite,override],['../prob_examples/public_examples/B/Tickets/Hansen13/OverrideSymbolicTest.mch', '-t', '-strict'],'Check override substitution works symbolically.').
cli_testcase(1292, [b_test,case],['../prob_examples/examples/B/Satpathy/Teletext104/Teletext104_deterministic.mch', '-t', '-strict'],'Check this relatively large machine works.').
cli_testcase(1293, [b_test,tla],['../prob_examples/public_examples/TLA/Tests/TLC_Regression_Tests/test1.tla', '-mc', 100, '-strict'], 'TLC regression test: equality.').
cli_testcase(1294, [b_test,tla],['../prob_examples/public_examples/TLA/Tests/TLC_Regression_Tests/test2.tla', '-mc', 100, '-strict'], 'TLC regression test: functions.').
% cli_testcase(1295, [b_test,tla],['../prob_examples/public_examples/TLA/Tests/TLC_Regression_Tests/test3.tla', '-mc', 100, '-strict'], 'TLC regression test: function application.'). virtual timeout
cli_testcase(1296, [b_test,tla],['../prob_examples/public_examples/TLA/Tests/TLC_Regression_Tests/test4.tla', '-mc', 100, '-strict'], 'TLC regression test: sets.').
cli_testcase(1297, [b_test,tla],['../prob_examples/public_examples/TLA/Tests/TLC_Regression_Tests/test5.tla', '-mc', 100, '-strict'], 'TLC regression test: Cartesian Product.').
cli_testcase(1298, [b_test,tla],['../prob_examples/public_examples/TLA/Tests/TLC_Regression_Tests/test6.tla', '-mc', 100, '-strict'], 'TLC regression test: Propositional Logic.').
cli_testcase(1299, [b_test,tla],['../prob_examples/public_examples/TLA/Tests/TLC_Regression_Tests/test7.tla', '-mc', 100, '-strict'], 'TLC regression test: Choose operator.').
cli_testcase(1300, [cbc], [
 '-evalf', '[1,2,3,3,3,4] : seq({1, 2, 3}) - seq({1, 2})',
 '-evalt', '[1,2,3,3,3,4] /: seq({1, 2, 3}) - seq({1, 2})',
 '-evalf', '[1,2,2,2,2,1] : seq({1, 2, 3}) - seq({1, 2})',
 '-evalt', '[1,2,2,2,2,1] /: seq({1, 2, 3}) - seq({1, 2})',
 '-evalt', '[1,2,3,3,3,2] : seq({1, 2, 3}) - seq({1, 2})',
 '-evalf', '[1,2,3,3,3,2] /: seq({1, 2, 3}) - seq({1, 2})',
 '-evalt', '[1,2,2,2,2,1,v] : seq({1, 2, 3}) - seq({1, 2}) & v>1',
 '-evalt', '{x | x : NATURAL - {x|x>10}} = 0..10',
 '-evalt', '{x | x : NATURAL - NATURAL1} = 0..0',
 '-evalt', 'x /: NATURAL - {x|x>10} & x>0 & x<100',
 '-evalt', '11..99 = {x|x /: NATURAL - {x|x>10} & x>0 & x<100}',
 '-evalt', 'x : NATURAL1 /\\ {x|x>20 & x mod 2 = 0}',
 '-evalt', 'x /: NATURAL1 /\\ {x|x>20 & x mod 2 = 0}',
 '-evalt', '-11 /: NATURAL1 /\\ {x|x>20 & x mod 2 = 0}',
 '-evalt', '1024 : NATURAL1 /\\ {x|x>20 & x mod 2 = 0}',
 '-evalf', '1024 /: NATURAL1 /\\ {x|x>20 & x mod 2 = 0}',
 '-evalt', '1025 /: NATURAL1 /\\ {x|x>20 & x mod 2 = 0}',
 '-evalf', '1025 : NATURAL1 /\\ {x|x>20 & x mod 2 = 0}',
 '-evalt', '[1,2] : seq({1, 2, 3}) /\\ seq({0, 1, 2})',
 '-evalt', '[1,2,3] /: seq({1, 2, 3}) /\\ seq({0, 1, 2})',
 '-evalt', '[1,2,3,4] /: seq({1, 2, 3}) /\\ seq({0, 1, 2})',
 '-evalt', '[1,2,2,2,1,2] : seq({1, 2, 3}) /\\ seq({0, 1, 2})',
 '-evalt', 'seq({1, 2, 3}) /\\ seq({0, 1, 2}) = sq & [1,2,2,2,1] : sq',
 '-evalt', 'seq({1, 2, 3}) \\/ seq({0, 1, 2}) = sq & [1,2,3,3,2] : sq',
 '-evalf', 'seq({1, 2, 3}) \\/ seq({0, 1, 2}) = sq & [1,2,3,3,2] /: sq',
 '-evalt', 'seq({1, 2, 3}) \\/ seq({0, 1, 2}) = sq & [1,2,3,3,2,0] /: sq',
 '-evalf', 'seq({1, 2, 3}) \\/ seq({0, 1, 2}) = sq & [1,2,3,3,2,0] : sq',
 '-evalt', '{200,201,301,400} - {x|x>0 & x mod 100 = 0} = {201,301}',
 '-evalt', '{200,201,301,400} /\\ {x|x>0 & x mod 100 = 0} = {200,400}',
 '-evalt', '500 : ( {x|x>0 & x mod 100 = 0} - {200,400})',
 '-evalt', '400 /: ( {x|x>0 & x mod 100 = 0} - {200,400})',
 '-evalf', '400 : ( {x|x>0 & x mod 100 = 0} - {200,400})',
 '-evalt', 'abs = id(NATURAL) \\/ %x.(x<0|-x) & abs(2001)=2001 & abs(-2002)=2002', % added 23.3.2015
 '-evalt', 'f = id(NATURAL) /\\ %x.(x<100|-x) & f={0|->0}', % added 23.3.2015
 '-evalt', 'abs = %x.(x<0|-x) \\/ id(NATURAL) & abs(2001)=2001 & abs(-2002)=2002'  % added 23.3.2015
 ], 'ensure symbolic treatment of set difference and intersection membership').
cli_testcase(1301, [tickets,infinite,strings], [
 '-evalt', 'NATURAL1 /: {NATURAL,INTEGER,NATURAL,1..2,{}}',
 '-evalf', 'NATURAL1 : {NATURAL,INTEGER,NATURAL,1..2,{}}',
 '-evalt', 'NATURAL1 : {NATURAL,INTEGER,NATURAL1,1..2,{}}',
 '-evalt', '1..2 : {NATURAL,INTEGER,NATURAL1,1..2,{},1..3,{2}}',
 '-evalt', '{} : {NATURAL,INTEGER,NATURAL1,1..2,{},1..3,{2}}',
 '-evalt', '{NATURAL} <: {NATURAL,INTEGER,NATURAL1,1..2,{},1..3,{2}}',
 '-evalt', '{NATURAL} <<: {NATURAL,INTEGER,NATURAL1,1..2,{},1..3,{2}}',
 '-evalt', '{NATURAL,{},INTEGER} <<: {NATURAL,INTEGER,NATURAL1,1..2,{},1..3,{2}}',
 '-evalt', '{NATURAL,{},INTEGER,{1,2}} <<: {NATURAL,INTEGER,NATURAL1,1..2,{},1..3,{2}}',
 '-evalf', '{NATURAL,{},INTEGER,{1,3}} <<: {NATURAL,INTEGER,NATURAL1,1..2,{},1..3,{2}}',
 '-evalt', '7=card({NATURAL,INTEGER,NATURAL1,1..2,{},1..3,{2}})',
 '-evalt', '5=card({NATURAL1,INTEGER,NATURAL1,1..2,{},1..3,{1,2}})',
 '-evalt', '{NATURAL1,INTEGER,x} = r & {}:r',
 '-evalt', '{STRING,{},{"hallo"}} = s & card(s)=3 & {}:s & STRING:s',
 '-evalt', 'r = { {1..2},{NATURAL}, {INTEGER}, {{}}, {NATURAL1}, {NATURAL1,NATURAL} } & {x} : r & x=NATURAL',
 '-evalt', 'r = { {1..2},{NATURAL}, {INTEGER}, {{}}, {NATURAL1}, {NATURAL1,NATURAL} } & {x,y} : r & x=NATURAL & y=NATURAL1',
 '-evalf', 'r = { {1..2},{NATURAL}, {INTEGER}, {{}}, {NATURAL1}, {NATURAL1,INTEGER} } & {x,y} : r & x=NATURAL & y=NATURAL1',
 '-evalt', '{x|x>10 & x*x = 100} : {NATURAL,{},1..3,{2},INTEGER}',
 '-evalt', '{x|x>4 & x*x = 100} : {NATURAL,{},1..3,{10},INTEGER}',
 '-evalt', '{3,2,1} : {NATURAL1,{},1..3,{2},INTEGER}'
 ], 'test putting infinite global sets into AVL sets').
cli_testcase(1302,[private,data_validation,alstom],['../prob_examples/examples/B/Alstom/Ticket_July2_2014/Rule_TMS_Signal_0107.mch','-strict', '-t', '-pref_group', integer, int32, '-p', 'TIME_OUT', 15000 ], 'Ticket related to variables only read in while loop').
cli_testcase(1303,[private,data_validation,alstom],['../prob_examples/examples/B/Alstom/Ticket_July2_2014/Rule_DB_SIGAREA_0030.mch','-strict', '-t', '-pref_group', integer, int32, '-p', 'TIME_OUT', 15000, '-expcterr', 'invariant_violation' , '-expcterr', bmachine_static_checks, '-expcterrpos', type_error, 719, 48], 'Ticket related to local variables modified in while loop which are also constants of another machine (length)').
cli_testcase(1304, [eventb_test,laws, theories], ['../prob_examples/public_examples/EventBPrologPackages/Theory/ListLawsMC_inf_mch.eventb','-t', '-strict',
 '-expcterr', 'virtual_time_out'], 'Various checks for sets involving freetypes (Lists Theory; PROB-325) that no wd-error occurs').
cli_testcase(1305, [cbc], ['../prob_examples/public_examples/B/Puzzles/Euler_Project/Euler_Problem_009.mch', '-init', '-strict', '-p', 'CLPFD', 'TRUE', '-properties'], 'Test that this puzzle now solved also when encoded as machine').
cli_testcase(1306, [b_test],['../prob_examples/public_examples/B/Tickets/Hansen15/PostCorrespondence_MC.mch','-mc', 50, '-t', '-strict'],'Check no call residue').
cli_testcase(1307,[private,b_test,alstom],['../prob_examples/examples/B/Alstom/Ticket_July9_2014/Rule_TMS_Overlap_0062.mch','-strict', '-expcterr',  type_error, '-expcterr',  load_main_file], 'Ticket related to operation defined multiple times + raising of error messages').
cli_testcase(1308, [b_test,kodkod,'PROB-331'], ['../prob_examples/public_examples/B/Tickets/PROB-331/HadamardMatrixRel_ForKodkod.mch', '-init', '-strict', '-p','TIME_OUT',3000, '-p', 'KODKOD', 'TRUE', '-p', 'KODKOD_ONLY_FULL', 'FALSE'], 'Check that Kodkod finds solution.').
cli_testcase(1309,[proz],['../prob_examples/public_examples/Z/Tickets/PROB-343/RecursiveDatatype.fuzz','-p','DEFAULT_SETSIZE','2','-p','TIME_OUT','1000','-mc',1000,'-nodead','-cc',4,3,'-strict'],'Z Freetypes: Checking simple recursive definition').
cli_testcase(1310, [tickets,'PROB-335'], [
 '../prob_examples/public_examples/B/Tickets/PROB-335/RecordDetectionBugSimple.mch', % used to generate a virtual time-out (fixed July 13th 2014)
'../prob_examples/public_examples/B/Tickets/PROB-335/ExtremaFinding.mch', '-mc', 100, '-p', 'MAX_INITIALISATIONS', 20, '-nodead',
 '-init', '-assertions', '-strict'], 'Check that Record Detection does not mess up.').
cli_testcase(1311, [cbc], ['-evalt', 'a : NATURAL & b : NATURAL & c : NATURAL & a < b & b < c & a**2 + b ** 2 = c ** 2 & a + b + c = 1000 & result=PI(x).(x:{a,b,c}|x)'], 'Test that Euler Puzzle 9 can be solved using Eval console').
cli_testcase(1312, [cbc], ['../prob_examples/public_examples/B/Puzzles/Euler_Project/Euler_Problem_004.mch', '-init', '-strict', '-p', 'CLPFD', 'TRUE', '-properties', '-assertions','-p', 'TIME_OUT', 10000 ], 'Test that this puzzle can be solved'). % temporariliy set an increased time-out (Sept. 2014) TO DO: remove again !!
cli_testcase(1313, [tickets,'PROB-346',infinite],
  ['-evalt', 'x=prj1(INTEGER,INTEGER) & y=prj1(INTEGER,INTEGER) & x=y',
   '-evalt', 'x=prj1(INTEGER,POW(NATURAL*INTEGER)) & y=prj1(INTEGER,POW(NATURAL*INTEGER)) & x=y',
   '-evalt', 'prj1(NATURAL1,NATURAL1)(222,333)=222',
   '-evalt', 'prj2(NATURAL1,NATURAL1)(222,333)=333',
   '-evalt', 'dom(prj1(NATURAL,NATURAL1)) = NATURAL*NATURAL1',
   '-evalt', 'dom(prj2(NATURAL1,NATURAL1)) = NATURAL1*NATURAL1'
  ], 'Check that no infinite expansion occurs when comparing two prj1 functions + other prj1/prj2 tests').
cli_testcase(1314, [cbc], [
 '-evalt', ' s = [x,y,z] & rev(s) = {1|->v,v|->w,w|->v} & v>w',
 '-evalt', '{s,x,y,z,v,w|s = [x,y,z] & rev(s) = {1|->v,v|->w,w|->v} & v>w} = {((((([2,3,3]|->2)|->3)|->3)|->3)|->2)}',
 '-evalt', 'rev({x|->y}) = {v|->w} & w:NATURAL',
 '-evalt', 'card({s,x,y,z,v,w|s = [x,y,z] & rev(s) = {1|->v,v|->w,w|->v} }) = 2',
 '-evalt', '{s,x,y,z,v,w|s = [x,y,z] & rev(s) = {1|->v,v|->w,w|->v} } = {((((([2,3,2]|->2)|->3)|->2)|->2)|->3),((((([2,3,3]|->2)|->3)|->3)|->3)|->2)}',
 '-evalt' , '{v,w,x,y,z|{1|->v,v|->w,w|->v} = [x,y,z]} = {((((2|->3)|->2)|->3)|->2),((((3|->2)|->3)|->3)|->2)}',
 '-evalt' , '{v,w,x,y|{1|->v,v|->w,w|->v} = [x,y,x]} = {(((2|->3)|->2)|->3)}',
 '-evalt', 'min({1+10,x}) = max({1+10,x})',
 '-evalt', '{x|min({1+10,x}) = max({1+10,x})} = {11}'
 ], 'Testing constraint propagation for reverse and unification with set extensions.').
cli_testcase(1315, [cbc], [
 '-evalf', 'x>y & x=y',
 '-evalf', 'x+1 > y+1 & x=y',
 '-evalf', 'x+1 < y+1 & x=y',
 % require CHR: '-evalf', 'x<y & y<z & x=z', or '-evalf', 'x+1 < y & x=y'
 '-evalt', '!(x,y).(x:INTEGER & x>y => x/=y)'
 ], 'Check simple inconsistencies with <, > detected now').
cli_testcase(1316, [cbc], ['../prob_examples/public_examples/B/Puzzles/Euler_Project/Euler_Problem_067.mch', '-init', '-strict', '-p', 'CLPFD', 'TRUE', '-assertions','-p', 'TIME_OUT', 5000  %increased time-out for cobra
], 'Test challenging puzzle').
cli_testcase(1317, [csp_test,ltl], ['../prob_examples/public_examples/CSP/other/Ivo/Underscore_LTL.csp', '-ltlassertions', '-strict'], 'Checking underscore arguments in LTL formulas.').
cli_testcase(1318, [b_test], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/Demo/scheduler.mch', '-mc', '100', '-cc', '36', '156', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'COMPRESSION' , 'TRUE'], 'Test COMPRESSION for spec using sets').
cli_testcase(1319, [b_test], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/Demo/scheduler.mch',
   '-mc', '100', '-bf', '-cc', '36', '156', '-p', 'DEFAULT_SETSIZE', '3', '-save_state',
  '../prob_examples/public_examples/B/Demo/scheduler_bf_states.P'], 'Test -save_state works').
cli_testcase(1320, [cbc], [
 '-evalt', 's = {y|y>2 & y mod 33 /= 1} & r={x|x:30..100 & x /:s} & r={34,67,100}'], 'check b_not_test_closure_enum works').
cli_testcase(1321, [cbc,wd], [
 '-eval', '{} : (1..(10/0) +-> {22})',
 '-expcterr', well_definedness_error], 'check WD error detected for binary_in_definitely_true').
cli_testcase(1322, [cbc,wd], [
 '-eval', '{} : POW(1..(10/0))',
 '-expcterr', well_definedness_error], 'check WD error detected for unary_in_definitely_true').
cli_testcase(1323, [eventb,tickets,theories], ['../prob_examples/public_examples/EventBPrologPackages/Tickets/testSUM_Theory_agg_op_ctx.eventb',
  '-init', '-assertions',
  '-strict'], 'Ensure no variable clash with agg_op1/2.').
cli_testcase(1324, [b_test], [
 '-evalt', 'a & b',
 '-evalt', 'a=5',
 '-p', 'BOOL_AS_PREDICATE', 'TRUE'
 ], 'Extended formula syntax.').
cli_testcase(1325, [b_test], ['../prob_examples/public_examples/B/FeatureChecks/Refinement/RefDefSetAsEnumSet.mch',
  '-init', '-assertions',
  '-strict'], 'Ensure enumerated set declaration in refinement using all-diff with card detected.').
cli_testcase(1326, [b_test,cse, cse_test, strings], ['-p', 'CSE', 'TRUE', '-p', 'TIME_OUT', 150,
 '-evalf', 'x:1..500000 & y:1..500000 &  (x+y)**2 = 4+(x+y)**2',
 '-evalf', '(x*y*z) = (y*x*z)+10',
 '-evalf', '(x*y*z) = (y*z*x)+10',
 '-evalf', '(x*(z*y)) = ((y*x)*z)+10',
 '-evalf', '(x*y*z*z*x) = (z*x*y*x*z)+10',
 '-evalf', '(x*y*(z*z)*x) = (z*(x*y)*x*z)+10',
 '-evalf', '(x\\/y\\/v\\/x) /= (v\\/x\\/y) & x<:BOOL', % CSE does not fully detect LHS and RHS identical (because of idempotence)
 '-evalf', '(x\\/y\\/v\\/x) /= (v\\/x\\/x\\/y) & x<:BOOL',
 '-evalf', '(x\\/y\\/v\\/x) /= (v\\/x\\/x\\/y) & x<:STRING',
 '-evalf', '(x\\/y\\/v\\/x) <<: (v\\/x\\/x\\/y) & x<:STRING',
 '-evalf', '(x\\/y\\/v\\/x) /<: (v\\/x\\/x\\/y) & x<:STRING',
 '-evalt', '(x\\/y\\/v\\/x) <: (v\\/x\\/x\\/y) & x<:BOOL',
 '-evalt', '(x\\/y\\/v\\/x) /<<: (v\\/x\\/x\\/y) & x<:BOOL',
 '-evalt', '(x*y*(z*z)*x) = (z*(x*y)*x*z*x)-36 & x:0..2 & y:0..2 & z:0..3',
 '-evalt', '#x.((x>10+10) & (x>10+10 => b=TRUE) & x:1..200)',
 '-evalt', '{0|->1,4|->2}(x*2+x*2) > 1',
 '-evalt', '{0|->x*x, 1|->(x*x)+1}(v) = 9 & v:0..1 & x:0..10', % had WD error v:0..10
 '-evalt', '{xx,yy| xx:1..100 & yy>xx & yy<110 & xx|->yy:{yy,xx|xx:20..22 & yy:{xx-2,xx-3}}} = {(17|->20),(18|->20),(18|->21),(19|->21),(19|->22),(20|->22)}',
 '-evalt', '{xx,yy| xx:1..(90+5*2) & yy>xx & yy<100+(5*2) & xx|->yy:{yy,xx|xx:(10+5*2)..22 & yy:{xx-2,xx-3}}} = {(17|->20),(18|->20),(18|->21),(19|->21),(19|->22),(20|->22)}',
 '-evalt', '{xx,yy| xx:1..(90+5*2) & yy>xx & yy<100+(5*2) & xx-2 > 15 & xx-2 < 18 & xx|->yy:{yy,xx|xx:(10+5*2)..22 & yy:{xx-2,xx-3}}} = {(18|->20),(18|->21),(19|->21),(19|->22)}'
 ], 'Ensure CSE works').
cli_testcase(1327, [tickets], ['-p', 'CSE', 'TRUE', '-p', 'TIME_OUT', '150s',
 '-expcterr', eclipse_preference
 ], 'Ensure error in set preference discovered').
cli_testcase(1328, [cbc_refinement], ['../prob_examples/examples/RodinModels/Seiter_RefinementChecking/auth_0_mch.eventb', '-cbc_refinement'], '').
cli_testcase(1329, [cbc_refinement], ['../prob_examples/examples/RodinModels/Seiter_RefinementChecking/auth_1_mch.eventb', '-cbc_refinement'], '').
cli_testcase(1330, [cbc_refinement], ['../prob_examples/examples/RodinModels/Seiter_RefinementChecking/auth_2_mch.eventb', '-cbc_refinement', '-expcterr', cbc_refinement], '').
cli_testcase(1331, [tickets, 'PROB-356',slot], ['../prob_examples/public_examples/B/Tickets/PROB-356/SlotSolver.mch',
 '-init', '-p', 'TIME_OUT', '25000', '-eval',  'CHECK(ges, ger, 1)'
 ], 'Ensure no access record errors').
cli_testcase(1332, [tickets, 'PROB-355',slot], ['../prob_examples/public_examples/B/Tickets/PROB-355/SlotSolver.mch',
 '-init', '-p', 'TIME_OUT', '20000', '-eval',  's1 : COURSES & CHECK(s1, soz, 1)'
 ], 'Ensure no infinite loop').
cli_testcase(1333, [cbc_deadlock,cache], ['../prob_examples/public_examples/B/Benchmarks/Cruise_finite1.mch', '-cbc_deadlock', '-strict', '-mc', '100', '-cache', '../prob_examples/public_examples/cache/'], 'Check that test 315 works with cache').
cli_testcase(1334, [b_test,cse_test,codespeed], ['../prob_examples/public_examples/B/Benchmarks/Cruise_finite1.mch',  '-strict', '-model_check', '-p', 'CSE', 'TRUE', '--hash64', 203058745203935766, '-p', 'MAX_INITIALISATIONS', 5, '-cc', '1361', '25696'], 'Check CSE for Cruise Controller'). % the hash is only valid on 64-bit systems ( conditional skip added); adapted hash on 8.12.2015 for new variable order
cli_testcase(1335, [b_test], ['../prob_examples/public_examples/B/Benchmarks/Cruise_finite1.mch',  '-strict', '-mc', '20', '--hash64', 1095242179479036905, '-expcterr', 'hash'], 'Check that --hash option works').
cli_testcase(1336, [tickets, 'SLOT-24',slot,codespeed], ['../prob_examples/public_examples/B/Tickets/SLOT-24/SlotSolver.mch',
 '-init', '-evalt',  'CHECK(ges, ger, 1)', '-evalf', 'CHECK(rom, jap, 1)', '-main_assertions'
 ], 'Ensure no access record errors').
cli_testcase(1337, [parser,tickets, 'PARSERLIB-32'], ['../prob_examples/public_examples/B/Tickets/PARSERLIB-32/TestDefOrder.mch',
 '-mc', 100
 ], 'Ensure that definition order for predicate-definitions does not matter').
cli_testcase(1338, [private, 'SLOT-24', recursion,slot], ['../prob_examples/examples/B/SlotTool/SlotTool_19092014/SlotSolver.mch',
 '-init', '-assertions'
 ], 'Ensure complicated unsat core computations keep on working').
cli_testcase(1339, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/GenericUsageOfSquare.mch', '-plugin', 'units', '-strict', '-pp', '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/GenericUsageOfSquare_internal.mch'], 'Check that operations calls and definitions are parametrical regarding units').
cli_testcase(1340, [b_test,fairness,ltl], ['../prob_examples/public_examples/B/LTL/Fairness/RecursiveSCCs.mch', '-ltlformula', 'SF(C) & WF(D) => GF {addr=4 or addr =5}', '-expcterr', ltl, '-strict'], 'Test fairness implementation for finding the right counter example.').
cli_testcase(1341, [b_test], ['../prob_examples/public_examples/B/Demo/Bakery0.mch', '-bf', '-mc', 100, '--all_paths', '../prob_examples/public_examples/B/Demo/Bakery0_all_paths.xml', '-strict'], 'XML all path export.').
cli_testcase(1342, [b_test,fairness,ltl], ['../prob_examples/public_examples/B/LTL/Fairness/M_Events_with_Args.mch', '-ltlformula', 'SF(Evt) => F (not(e(Evt1)) & not(e(Evt2)))', '-expcterr', ltl, '-strict'], 'Test fairness implementation for finding the right counter example.').
cli_testcase(1343, [csp_test,assertions], ['../prob_examples/public_examples/CSP/other/Ivo/AlsoPatterns.csp', '-assertions', '-strict'], 'Testing enumerating of also patterns on channel data types.').
cli_testcase(1344, [csp_test,assertions], ['../prob_examples/public_examples/CSP/other/Ivo/AParallel.csp', '-assertions', '-strict'], 'Testing pattern matching for alphabetised parallel operators.').
cli_testcase(1345, [csp_test,assertions], ['../prob_examples/public_examples/CSP/other/Ivo/AParallel1.csp', '-assertions', '-strict'], 'Testing pattern matching for alphabetised parallel operators.').
cli_testcase(1346, [csp_test,assertions], ['../prob_examples/public_examples/CSP/other/Andreas_Hense/P2P/p2p_v3.csp', '-assertions', '-strict'], 'Testing assertions in the CSP specification (for this specification ProB performs better than FDR2).').
cli_testcase(1347, [csp_test], ['../prob_examples/public_examples/CSP/other/Andreas_Hense/ATM/ATM_v5.csp', '-t', '-strict'], '').
cli_testcase(1348, [csp_test,assertions,records], ['../prob_examples/public_examples/CSP/other/Ivo/EnumeratingDataTypes.csp', '-assertions', '-strict'], 'Enumerating partial data type records.').
cli_testcase(1349, [b_test,ltl], ['../prob_examples/public_examples/B/Ivo/Threads.mch', '-ltlassertions', '-strict'], 'Covering the new APs (deadlock(...),deterministic(...), and controller(...)) of the LTL interpreter.').
cli_testcase(1350, [cbc,tickets], ['../prob_examples/examples/EventBPrologPackages/Savary/R08_ByteArray_mch.eventb', '-cbc_sequence', 'aload_R07' ], 'Check cbc sequence on example with partially enumerated sets').
cli_testcase(1351, [tickets], ['../prob_examples/examples/B/ClearSy/GeneralElectric_DV/Ticket_CLPFD_Overflow/Machine_Bug3.mch', '-init', '-p', 'MAXINT', '2147483647', '-p', 'MININT', '-2147483648', '-p', 'TIME_OUT', '3500',
 '-expcterrpos', type_error, 716, 48,
 '-expecterr', 'bmachine_construction',
 '-expcterrpos', 'bmachine_static_checks', 716, 0    % added this 15.12.2015 as we now also generate warning for length
 ], 'Check abs DEFINITION error message works (and VAR hides constant)').
cli_testcase(1352, [tickets], ['../prob_examples/examples/B/ClearSy/GeneralElectric_DV/Ticket_CLPFD_Overflow/Machine_Bug3_renamed.mch',
  '-t', '-p', 'MAXINT', '2147483647', '-p', 'MININT', '-2147483648', '-p', 'TIME_OUT', '3500' ], 'Check no error in compiling While loop').
cli_testcase(1353, [b_test], [
 '-evalt', '{x|x>99988844111111222200 & x<99988844111111222300 & x mod 20 = 1} = res & 2000 /: res & 99988844111111222201 : res',
 '-eval', '{x|x>99988844111111222200 & x<99988844111111222300 & x mod 20 = 1}'
 , '-evalt', '{x|x>ll & x<99988844111111222300 & x mod 20 = 9} = res  & (ll=99988844111111222200 or ll = 99988844111111222201) & 99988844111111222209 : res'
 , '-evalt', '{x|x>ll & x<99988844111111222300 & x mod 20 = 2} = res & 2000 /: res & 99988844111111222202 : res & (ll=99988844111111222200 or ll = 99988844111111222201)'
 ], 'Check CLPFD Overflow is caught').
cli_testcase(1354, [ltl], ['../prob_examples/public_examples/B/Ivo/HDCP/HDCP0_mch_mch.eventb','-ltllimit', 10000,'-ltlfile','../prob_examples/public_examples/B/Ivo/HDCP/HDCP0.ltl', '-strict'], 'Regression test for checking LTL formulas.').
cli_testcase(1355, [ltl], ['../prob_examples/public_examples/B/Ivo/HDCP/HDCP1_mch_mch.eventb','-ltllimit', 10000,'-ltlfile','../prob_examples/public_examples/B/Ivo/HDCP/HDCP1.ltl', '-strict'], 'Regression test for checking LTL formulas.').
cli_testcase(1356, [cbc,cbc_tests], ['../prob_examples/public_examples/B/Benchmarks/Cruise_finite1.mch', '-cbc_tests', 4, '1=1', '../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_cbc_tests.xml', '-expcterr', 'cbc_tests'], 'Check that full coverage cannot be achieved').
cli_testcase(1357, [cbc,cbc_tests,enabling], ['../prob_examples/public_examples/B/Benchmarks/Cruise_finite1.mch', '-cbc_tests', 5, '1=1', '../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_cbc_tests.xml', '-feasibility_analysis'], 'Check that full coverage can be achieved').
cli_testcase(1358, [cbc,enabling], ['../prob_examples/examples/EventBPrologPackages/Savary/R08_ByteArray_enum_partition.eventb', '-feasibility_analysis_csv', '../prob_examples/examples/EventBPrologPackages/Savary/R08_ByteArray_enum_partition_feasibility.csv'], 'Check feasibility analysis works').
cli_testcase(1359, [cbc,tickets], ['../prob_examples/public_examples/EventBPrologPackages/Tickets/SeqTest/SeqTest_ctx.eventb', '../prob_examples/public_examples/EventBPrologPackages/Tickets/SeqTest/SeqTest3_ctx.eventb',
'../prob_examples/public_examples/EventBPrologPackages/Tickets/SeqTest/SortRecFunction_ctx.eventb',
 '-p', 'MAXINT', '3', '-t' ,'-assertions'], 'Check no enumeration warning').
cli_testcase(1360, [cbc,enabling], ['../prob_examples/public_examples/B/CBC/FeasibilityOperationCheck.mch',
 '-opterr', 'clpfd_overflow',
'-feasibility_analysis_csv', '../prob_examples/public_examples/B/CBC/FeasibilityOperationCheck_feasibility.csv'], 'Check feasibility analysis works').
cli_testcase(1361, [ctl,tickets], ['../prob_examples/public_examples/B/LTL/MammarFrappier/FlightSystem_corrected_3_3.mch', '-ctlformula', 'AG({Customers1|->Flights1 /: tickets} => EF({Customers1|->Flights1 : tickets}))',
 '-ltllimit', 100, '-expcterr', 'ctl_incomplete', '-expcterr', 'state_space',
  '-p', 'STRICT_RAISE_WARNINGS', 'TRUE'], 'Check that incomplete CTL model check detected').
cli_testcase(1362, [ctl,tickets], ['../prob_examples/public_examples/B/LTL/MammarFrappier/FlightSystem_corrected_3_3.mch',
  '-ctlformulat', 'AG({Customers1|->Flights1 /: tickets} => EF({Customers1|->Flights1 : tickets}))',
  '-p', 'STRICT_RAISE_WARNINGS', 'TRUE', '-cc', 2199, 15212], 'Check that CTL model check works without ltllimit').
cli_testcase(1363, [ctl], ['../prob_examples/public_examples/B/Simple/UpCounter.mch',
 '-p', 'MAXINT', 20,
  '-ctlformulaf', 'AG({c<mx})', '-his', '../prob_examples/public_examples/B/Simple/UpCounter_his.txt', '-his_option','show_states'
  ], 'Check that CTL model check works without ltllimit').
cli_testcase(1364, [ctl], ['../prob_examples/public_examples/B/Demo/scheduler.mch',
  '-p', 'DEFAULT_SETSIZE', '3',
  '-ctlformulaf', 'AG {active={} or ready = {} or waiting ={}}', '-bf', '-mc', 9999, '-his', '../prob_examples/public_examples/B/Demo/scheduler_his.txt', '-his_option','show_states'
  ], 'Check that CTL model check generates correct counter example').
cli_testcase(1365, [ctl], ['../prob_examples/public_examples/B/Demo/scheduler.mch',
  '-ctlformulat', 'AG({card(ready)>1} => EF {card(active)>0})',
  '-ctlformulat', 'AG EF {active = {}}',
  '-ctlformulat', 'EG {active = {}}',
  '-p', 'DEFAULT_SETSIZE', '3'
  ], 'Check CTL formulas').
cli_testcase(1366, [cbc, cbc_tests,bmc], ['../prob_examples/public_examples/B/SymbolicModelChecking/TimingExampleSimpler_v2.mch', '-bmc', 50, '-his', '../prob_examples/public_examples/B/SymbolicModelChecking/TimingExampleSimpler_v2_his.txt', '-his_option','show_states',
'-expcterr', 'invariant_violation'], 'Check BMC works').
cli_testcase(1367, [cbc,tickets,bmc], ['../prob_examples/public_examples/B/Tickets/OpCallModifiesInfo/OpCallModifiesInfo.mch', '-bmc', 5,
'-expcterr', 'invariant_violation'], 'Check Modifies Info correct for operation calls').
cli_testcase(1368, [cbc,tickets], ['../prob_examples/public_examples/B/Tickets/ComponentPartitioning/ManyPartialFunctions.mch', '-init',
 '-p', 'DEFAULT_SETSIZE', 5], 'Check component partition works with equality inlining').
cli_testcase(1369, [b_test,tla,tlc],['../prob_examples/public_examples/B/Simple/UpCounter.mch', '-mc_with_tlc'], 'TLC regression test').
cli_testcase(1370, [cbc,cbc_tests], ['../prob_examples/public_examples/B/TestCaseGeneration/Aymerick/CBC_M_mch.eventb', '-cbc_tests', 4, '', '', '-expcterr', 'cbc_tests'], 'Check that full coverage cannot be achieved and no exceptions').
cli_testcase(1371, [cbc,cbc_tests], ['../prob_examples/public_examples/B/TestCaseGeneration/Aymerick/CBC_M_mch.eventb', '-cbc_tests', 4, '', '', '-cbc_cover_match', 'c'], 'Check that cbc_cover_match option works').
cli_testcase(1372, [b_test], ['../prob_examples/public_examples/B/Mathematical/FormaleSprachen.mch',
'-p', 'SMT', 'TRUE',
 '-init','-assertions', '-strict'], 'Test 1264 but with SMT mode.').
cli_testcase(1373, [private,codespeed], ['../prob_examples/examples/B/Alstom/vesg_Aug11/gradient_train_altitude.mch',
 '-animate', 5, '-p', 'TIME_OUT', 300000000, '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647, '-p', 'CLPFD', 'FALSE'], 'Test Alstom example').
 cli_testcase(1374, [tickets], [
 '-evalf', 'n:0..4 & m=6 & not(((((card((1 .. n <-> 5 .. m) \\/ {{0 |-> 0}}) = 1 + card(1 .. n <-> 5 .. m) & card((1 .. n +-> 5 .. m) \\/ {{0 |-> 0}}) = 1 + card(1 .. n +-> 5 .. m)) & card((1 .. n --> 5 .. m) \\/ {{0 |-> 0}}) = 1 + card(1 .. n --> 5 .. m)) & card((1 .. n >+> 5 .. m) \\/ {{0 |-> 0}}) = 1 + card(1 .. n >+> 5 .. m)) & card((1 .. n >-> 5 .. m) \\/ {{0 |-> 0}}) = 1 + card(1 .. n >-> 5 .. m)) & card((1 .. n -->> 5 .. m) \\/ {{0 |-> 0}}) = 1 + card(1 .. n -->> 5 .. m))'],
  'Test that no choice points due to add_element').
cli_testcase(1375, [cbc, cbc_tests, tickets,bmc], ['../prob_examples/public_examples/B/SymbolicModelChecking/Counters/VerySimpleCounterWrong.mch', '-bmc', 6,
'-expcterr', 'invariant_violation'], 'Check BMC works').
cli_testcase(1376, [private,data_validation], ['../prob_examples/examples/B/Systerel/C578/2013/machines_24092013/02_001_adapted.mch',
 '-init' ,'-assertions', '-p', 'SYMBOLIC', 'TRUE', '-p', 'MEMO', 'TRUE', '-p', 'CLPFD', 'TRUE', '-p', 'TIME_OUT', 155000,
  '-p', 'ENUMERATE_INFINITE_TYPES', 'FALSE', '-p', 'EXPAND_FORALL_UPTO', 0 ], 'Check performance (previously part of 1162)').
cli_testcase(1377, [private,data_validation,cse_test], ['../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_2_ori.mch', '-animate_all', '-strict', '-cc', 49, 48, '-p', 'TIME_OUT', 35000, '-p', 'CLPFD', 'FALSE', '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647,
'-p', 'MAX_OPERATIONS', 1, '-his', '../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_2_ori.his', '-sptxt', '../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_2_ori.sptxt', '-expcterr', invariant_violation, '-p', 'CSE', 'TRUE'], 'Rule Test 4 of Alstom compilation project with CSE').  % Signal_Overlap_SDDBs is non-deterministic: hence MAX_OPERATIONS 1
cli_testcase(1378, [private,data_validation,cse_test], ['../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_SIGAREA_0024_ori.mch', '-animate_all', '-strict', '-cc', 40, 39, '-p', 'TIME_OUT', 35000, '-p', 'CLPFD', 'FALSE', '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647, '-his', '../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_SIGAREA_0024_ori.his', '-p', 'CSE', 'TRUE'], 'Rule Test 5 of Alstom compilation project with CSE').
cli_testcase(1379, [private,data_validation,cse_test], ['../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_Route_0001ori.mch', '-animate_all', '-strict', '-cc', 97, 96, '-p', 'TIME_OUT', 28000, '-p', 'CLPFD', 'FALSE', '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647, '-his', '../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_Route_0001ori.his', '-p', 'CSE', 'TRUE'], 'Rule Test 1 of Alstom compilation project with CSE').
cli_testcase(1380, [private,data_validation,alstom,cse_test], ['../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_ESA_TRIGGER_0009_ori.mch', '-strict', '-cc', 112, 111, '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647, '-t', '-p', 'CSE', 'TRUE'], 'Rule Test 2 of Alstom compilation project with CSE').
cli_testcase(1381, [tickets,cse_test], ['../prob_examples/public_examples/B/Tickets/CSE/BoolValue.mch',
'../prob_examples/public_examples/B/Tickets/CSE/ExplicitComputations_Reduced.mch',
 '-init', '-p', 'CSE', 'TRUE'], 'Check sharing of predicates works').
cli_testcase(1382, [cbc, disprover], ['../prob_examples/examples/B/Alstom/DisproverProject/prob13265_hyp_removed.mch',
'-cbc_assertions', '-p', 'MAX_INITIALISATIONS', 0], 'Check partitioning for cbc_assertions works').
cli_testcase(1383, [tickets, external], ['../prob_examples/examples/B/Alstom/Ticket_Dec08_2014/Segment_refAdhesionLevel_GER5.mch',
'-t', '-sptxt', '../prob_examples/examples/B/Alstom/Ticket_Dec08_2014/Segment_refAdhesionLevel_GER5_sptxt.txt'], 'Check join works in conjunction with perm with paritally instantiated lists').
 cli_testcase(1384, [cbc,sigma], [
 '-evalt', '{y|SIGMA(x).(x:1..10 & x mod 5 = 0|x+y) = 215} = {100}',
 '-evalt', '{y|PI(x).(x:1..10 & x mod 5 = 0|x+y) = 66} = {-16,1}',
 '-evalt', 'SIGMA(x).(x:{y} \\/ {z}|x) = 10 & y>z & z>0',
 '-evalt', '{y,z|SIGMA(x).(x:{y} \\/ {z}|x+3) = 10 & y>z & z>0} = {(3,1)}',
 '-evalt', 'SIGMA(i).(i:1..1000|i)=500500',
 '-evalt', 'SIGMA(i).(i:0..1000|i)=500500',
 '-evalt', 'SIGMA(i).(i:-2..1000|i)=500497',
 '-evalt', 'SIGMA(i).(i:-3..1000|i)=500494',
 '-evalt', 'SIGMA(i).(i:-2..-3|i)=0',
 '-evalt', '{y,m|PI(x).(x:1..3|x+y) = m & m:25..100} = {(2,60)}',
 '-evalt', 'x = [a,b,c,d,e,f,g,h,i,j,k,l,m,n,o] & SIGMA(i).(i:dom(x)|x(i))=124 & !j.(j:2..size(x) => x(j-1)<x(j)) & a>0',
 '-evalf', 's:seq(NATURAL1) & size(s)=220 & SIGMA(i).(i:1..size(s)|s(i)) = 200',
 '-evalf', 'x = SIGMA(i).(i:1..n|i) & x<0',
 '-evalt', 'x = SIGMA(i).(i:1..n|i) & n>1 & x = 5050',
 '-evalt', '{n|5050 = SIGMA(i).(i:1..n|i)} = {100}',
 '-evalf', 'x = SIGMA(i).(i:1..n|i) & n>100 & x < 5151',
 '-evalt', '{x,n|x = SIGMA(i).(i:-3..n|i) & n>1 & x < 5150 & x>5100} = {(5145,101)}',
 '-evalf', 's = [x,10,y] & PI(i).(i:dom(s)|s(i)) = r & x>0 & y>0 & r<10',
 '-evalt', 's<:11..15 & card(s)<3 & SIGMA(x).(x:s|x)=23'
  ], 'Check constraint propagation of SIGMA / PI work better now').
cli_testcase(1385, [cbc,sigma], ['../prob_examples/public_examples/B/Puzzles/Minesweeper.mch',
 '../prob_examples/public_examples/B/Puzzles/Minesweeper_v2.mch',
 %'../prob_examples/public_examples/B/Puzzles/Minesweeper_v3.mch', still time-out
 '../prob_examples/public_examples/B/Puzzles/Minesweeper_v4.mch',
 '-init', '-cbc_assertions'], 'Check this version with SIGMA can be solved efficiently').
 cli_testcase(1386, [cbc,atelierb], ['../prob_examples/examples/B/Alstom/DisproverProject/FalsePO1.mch',
'-cbc_assertions', '-expecterr', 'cbc_assertions', '-cbc_result_file', '../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result.txt'], 'Test Atelier-B Disprover interface').
 cli_testcase(1387, [cbc,atelierb,chr], ['../prob_examples/examples/B/Alstom/DisproverProject/TimeOutPO1.mch',
 '-p', 'CHR', 'TRUE', '-p', 'DISPROVER_MODE', 'TRUE',
 '-cbc_assertions',
 '-cbc_result_file', '../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result.txt'], 'Test Atelier-B Disprover interface').
 cli_testcase(1388, [cbc,atelierb,chr], ['../prob_examples/examples/B/Alstom/DisproverProject/TruePO1.mch',
 '-p', 'CHR', 'TRUE', '-p', 'DISPROVER_MODE', 'TRUE',
 '-cbc_assertions',
 '-cbc_result_file', '../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result.txt'], 'Test Atelier-B Disprover interface').
 cli_testcase(1389, [cbc,atelierb,chr], ['../prob_examples/examples/B/Alstom/DisproverProject/DeferredSetPO1.mch',
 '-p', 'CHR', 'TRUE', '-p', 'DISPROVER_MODE', 'TRUE',
 '-cbc_assertions',
 '-cbc_result_file', '../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result.txt'], 'Test Atelier-B Disprover interface').
 /* does not work yet:
cli_testcase(1382, [private, 'SLOT-24', recursion,slot,cse], ['../prob_examples/examples/B/SlotTool/SlotTool_19092014/SlotSolver.mch',
 '-init', '-assertions', '-p', 'CSE', 'TRUE'
 ], 'Ensure complicated unsat core computations keep on working with CSE TRUE').
*/
cli_testcase(1390, [tickets], [
 '-evalf', '({1}*INTEGER)=f & res = f[{1}] & res=MININT..MAXINT', '-expcterr', eval_string_enum_warning],
  'Check that we do not report this as true / false').
cli_testcase(1391, [tickets,proz], ['../prob_examples/public_examples/Z/Daniel/Praxis/network2.fuzz',
'../prob_examples/public_examples/Z/Daniel/scheduler.fuzz',
'../prob_examples/public_examples/Z/Daniel/SudokuSETS9.fuzz',
'../prob_examples/public_examples/Z/Daniel/watergate.fuzz',
'-t'],
 'Ensure no pending co-routines for instantiation of freetype value (network2)').
 cli_testcase(1392, [cbc,sigma], [
 '-evalt', 'r = {s1,s2| SIGMA(i).(i:s1|i) = SIGMA(j).(j:s2|j)} & {x,y}|->{10}:r & {6}|->{x}:r',
 '-evalt', 'r = {s1,s2| SIGMA(i).(i:s1|i) = SIGMA(j).(j:s2|j)} & {x,y|{x,y}|->{10}:r & {6}|->{x}:r} = {(6,4)}',
 '-evalf', '{x,y}=s & {v,w,z}=r & s<:r & r<:11..14 & SIGMA(i).(i:s|i) > SIGMA(j).(j:r|j)' % already worked before
 ], 'Check that we can find a solution : choice point enumerated before enumerating y').
cli_testcase(1393, [tickets,slot], ['../prob_examples/public_examples/B/Tickets/SLOT-PERFORMANCE3/WiWiSolver.mch',
 '-init', % '-v', % generates non-ground closure warnings
 '-evalt',  '{x,s,y| x:MAJORS & s : 1..6 & y = bool(CHECK_P_WP_SEMESTER(x, s))} = {((bwl_bachelor|->1)|->TRUE),((bwl_bachelor|->2)|->TRUE),((bwl_bachelor|->3)|->FALSE),((bwl_bachelor|->4)|->FALSE),((bwl_bachelor|->5)|->FALSE),((bwl_bachelor|->6)|->TRUE),((bwl_master|->1)|->FALSE),((bwl_master|->2)|->FALSE),((bwl_master|->3)|->FALSE),((bwl_master|->4)|->TRUE),((bwl_master|->5)|->TRUE),((bwl_master|->6)|->TRUE),((vwl_bachelor|->1)|->TRUE),((vwl_bachelor|->2)|->TRUE),((vwl_bachelor|->3)|->FALSE),((vwl_bachelor|->4)|->FALSE),((vwl_bachelor|->5)|->TRUE),((vwl_bachelor|->6)|->TRUE),((vwl_master|->1)|->FALSE),((vwl_master|->2)|->TRUE),((vwl_master|->3)|->FALSE),((vwl_master|->4)|->TRUE),((vwl_master|->5)|->TRUE),((vwl_master|->6)|->TRUE),((wichem_bachelor|->1)|->TRUE),((wichem_bachelor|->2)|->TRUE),((wichem_bachelor|->3)|->TRUE),((wichem_bachelor|->4)|->TRUE),((wichem_bachelor|->5)|->FALSE),((wichem_bachelor|->6)|->FALSE),((wichem_master|->1)|->TRUE),((wichem_master|->2)|->FALSE),((wichem_master|->3)|->TRUE),((wichem_master|->4)|->TRUE),((wichem_master|->5)|->TRUE),((wichem_master|->6)|->TRUE)}'
 ], 'Ensure no time-out due to total-function setup too late').
cli_testcase(1394, [tickets,alstom,private,codespeed], ['../prob_examples/examples/B/Alstom/ixl/lausanne.mch', '-t',
 '-p', 'TIME_OUT', 5000], 'Ensure no time-out due to rel_compose calling union without WF').
cli_testcase(1395, [cbc], ['../prob_examples/public_examples/B/SymmetryReduction/DefSetWithConstantsTest2.mch', '-init',
 '-expecterr', 'setup_constants_fails', '-p', 'TIME_OUT', '9000'], 'Ensure static symmetry redution works').
cli_testcase(1396, [cbc, symmtest], ['../prob_examples/public_examples/B/SymmetryReduction/DefSetWithConstantsTest3.mch', '-mc', 9999, '-nodead',
 '-cc', 19, 18 ], 'Ensure static symmetry redution works').
cli_testcase(1397, [cbc], ['../prob_examples/public_examples/B/Puzzles/HandshakeSimple_TotFun.mch' ], 'Ensure forall quantifier also uses contrapositive').
cli_testcase(1398, [cbc], ['../prob_examples/public_examples/B/SymmetryReduction/DefSetWithConstantsTest2.mch', '-cbc_sequence', '',
 '-expecterr', 'cbc_sequence_no_solution_found'], 'Ensure static symmetry redution works').

 cli_testcase(1399, [cbc], [
 '-evalf', 'v0 = (x0 - x2) & v1 = (x1 - x2) & v2 = (x0 - x1) & v0 <= 1 & v0 >= 0 & v1 <= 1 & v1 >= 0 & not(x0 = x1) & not(v2 = 1) & not(v2 = -1) & v2+v1=v0'
 ], 'Check that we can detect inconsistency without enum warning'). % requires  propagate_zero in clpfd_interface; variation of toroidal queens 2 with v2+v1=v0 added for CLPFD
 cli_testcase(1400, [cbc,disprover,chr], [
 '-evalf', 'v0 = (x0 - x2) & v1 = (x1 - x2) & v2 = (x0 - x1) & v0 <= 1 & v0 >= 0 & v1 <= 1 & v1 >= 0 & not(x0 = x1) & not(v2 = 1) & not(v2 = -1)',
 '-evalf' , 'not(v2=1) & v0 = (x0 - x2) & v1+x2 = (x1) & v2+x1 = (x0) & v0:0..1 & v1:0..1 & not(x0 = x1) & not(v2 = -1)',
 '-p', 'CHR', 'TRUE'
 ], 'Check that we can detect inconsistency without enum warning'). % requires  propagate_zero in clpfd_interface + IDL CHR rules; derived from toroidal queens 2 SMT benchmark

cli_testcase(1401, ['PROB-369',csp_test,cspb,'PROB-371','PROB-368',sequences], ['../prob_examples/public_examples/CSPB/Tickets/PROB-369/BufferSpec2.fuzz', '-csp-guide', '../prob_examples/public_examples/CSPB/Tickets/PROB-369/BufferSpec2.csp', '-mc', '100', '-strict', '-t',
 '-evalt', '-2 mod 3 = 1'], 'Check synchronisation with sequence type works and z minor mode set for modulo').
cli_testcase(1402, [cbc,atelierb,chr,disprover], ['../prob_examples/examples/B/Alstom/DisproverProject/TruePO1_Taut.mch',
 '-p', 'CHR', 'TRUE', '-p', 'DISPROVER_MODE', 'TRUE',
 '-cbc_assertions_tautology_proof',
 '-cbc_result_file', '../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result.txt'], 'Test Atelier-B Disprover interface').
cli_testcase(1403, [cbc,smt,chr], [
 '../prob_examples/public_examples/SMT/QF_IDL/queens_bench/n_queen/queen20-1.smt2',
 '../prob_examples/public_examples/SMT/QF_IDL/queens_bench/toroidal_bench/toroidal_queen2-1.smt2',
 '../prob_examples/public_examples/SMT/QF_IDL/queens_bench/toroidal_bench/toroidal_queen3-1.smt2',
 '../prob_examples/public_examples/SMT/QF_IDL/queens_bench/toroidal_bench/toroidal_queen5-1.smt2',
 '../prob_examples/public_examples/SMT/QF_IDL/queens_bench/super_queen/super_queen10-1.smt2',
 %'../prob_examples/public_examples/SMT/QF_IDL/queens_bench/super_queen/super_queen20-1.smt2',
 '-p', 'CHR', 'TRUE', '-p', 'SMT', 'TRUE'], 'Test SMT interface and solving').
%

cli_testcase(1404, [ltl], ['../prob_examples/public_examples/EventBPrologPackages/ABZ_Landing_Gear/Ref4_ControllerHandle_mch.eventb', '-mc', 99999, '-cc' , 8629, 41811,
 '-ltlformula', 'G (deterministic(con_stop_stimulate_extend_gear_valve,con_stimulate_extend_gear_valve,con_stop_stimulate_retract_gear_valve,con_stimulate_retract_gear_valve,con_stop_stimulate_close_door_valve,con_stimulate_close_door_valve,con_stop_stimulate_open_door_valve,con_stimulate_open_door_valve))'
 ], 'Check LTL pattern deterministic').
cli_testcase(1405, [dot], ['../prob_examples/public_examples/B/Simple/VerySimpleDFA.mch', '-bf', '-mc', 1000, '-strict', '-dotexpr', transition_diagram, 's', '../prob_examples/public_examples/B/Simple/VerySimpleDFA_transition_diagram.dot'],'Check Transition Diagram for Expression.').
cli_testcase(1406, [b_test,cbc,atelierb,disprover], ['-cbc_assertions_tautology_proof', '-p', 'DISPROVER_MODE', 'TRUE', '-p', 'ALLOW_UNTYPED_IDENTIFIERS', 'TRUE',
 '-expcterr', 'cbc_assertions_enumeration_warning',
'../prob_examples/examples/B/Alstom/DisproverProject/TypingIssues/prob17066_not_provable.mch'
], 'Check ProB finds no counter example for rule and no proof').
cli_testcase(1407, [b_test,cbc,atelierb,disprover], ['-cbc_assertions_tautology_proof', '-p', 'DISPROVER_MODE', 'TRUE', '-p', 'ALLOW_UNTYPED_IDENTIFIERS', 'TRUE',
'../prob_examples/examples/B/Alstom/DisproverProject/TypingIssues/prob17066_provable.mch'
], 'Check ProB finds no counter example for rule and no proof').

cli_testcase(1408, [cbc,sigma], [
 '-evalt', 'x+y+z+v=40 & x<y & y<z & z<v &x>0  & all = {x,y,z,v} & gen = {a1,a2,a3,r| a1\\/a2\\/a3=all & a1/\\a2={} & a2/\\a3={} & a1/\\a3={} & r=SIGMA(t1).(t1:a1|t1)-SIGMA(t2).(t2:a2|t2)} & 1..40 <: ran(gen)',  % Stones Puzzle
 '-evalt' , 'all = {1,2,4} & gen = {a1,a2,r| a1\\/a2=all & a1/\\a2={} & r=SIGMA(t1).(t1:a1|t1)+SIGMA(t2).(t2:a2|t2)}'
 ], 'Check that partition propagates well (no enum warning)').

cli_testcase(1409, [cbc,disprover], [
 '../prob_examples/public_examples/ProofObligations/GearDoor1_mch.pl',
 '../prob_examples/public_examples/ProofObligations/GearsDoorsHandle3_mch.pl',
 '../prob_examples/examples/ProofObligations/ABZ14_Mery/M2_mch.pl',
 '../prob_examples/examples/ProofObligations/ABZ14_Mery/M20_mch.pl',
 '../prob_examples/examples/ProofObligations/ABZ14_Mery/M3_mch.pl',
 '../prob_examples/examples/ProofObligations/ABZ14_Mery/M30_mch.pl',
 '../prob_examples/public_examples/ProofObligations/Abrial_ABZ_Devel3/m1_true.pl',
 '../prob_examples/public_examples/ProofObligations/Abrial_ABZ_Devel3/m2_mch.pl',
 '-p', 'DOUBLE_EVALUATION', 'FALSE'
 %,'../prob_examples/examples/ProofObligations/ABZ14_Mery/M4_mch.pl'
  ],'Check Disprover can prove all POs').
cli_testcase(1410, [cbc,disprover], [
 '../prob_examples/examples/ProofObligations/Abrial_Ch4_Brp/brp_1_true.pl',
 '../prob_examples/examples/ProofObligations/Abrial_Ch4_Brp/brp_3_true.pl',
 '-p', 'DOUBLE_EVALUATION', 'FALSE'
  ],'Check Disprover can prove all POs; this test also checks register predicate works better').
cli_testcase(1411, [cbc,disprover], [
 '../prob_examples/public_examples/ProofObligations/DisproverDefSetTest2_ctx.pl',
 '../prob_examples/public_examples/ProofObligations/DisproverDefSetTest3_ctx.pl',
 '../prob_examples/public_examples/ProofObligations/DisproverDefSetTest4_ctx.pl',
 '-check_disprover_result', 0,0,2,
 '-p', 'DOUBLE_EVALUATION', 'FALSE'
  ],'Check Disprover cannot prove po').
cli_testcase(1412, [cbc,random,sigma], [
 '-p', 'RANDOMISE_ENUMERATION_ORDER', 'TRUE',
 '-evalt', 'x:1..20 --> 1..100 & SIGMA(v).(v:1..20|x(v)) = 20*50'
 % without random we get x = [100,100,100,100,100,100,100,100,100,90,1,1,1,1,1,1,1,1,1,1]
 % with it: e.g., x = [46,29,99,5,2,61,100,49,46,23,8,61,37,100,99,29,98,8,5,95]
  ],'Check random enumeration').
cli_testcase(1413, [cbc,chr,disprover], [
 '-p', 'CHR', 'TRUE',
 '-evalf', 'xx>0 & xx+1 <= p & xx=p'
 %'-evalf', 'xx>0 & xx+2 <= p & xx=p'  % does not work yet
  ],'Check CHR propagation').
cli_testcase(1414, [b_test,cbc,atelierb,disprover], ['-cbc_assertions_tautology_proof', '-p', 'DISPROVER_MODE', 'TRUE',
  '-p', 'CHR', 'TRUE', '-p', 'CSE', 'TRUE',
%'../prob_examples/examples/B/Alstom/DisproverProject/ProvableByPP/prob9313.mch', % does not work yet
'../prob_examples/examples/B/Alstom/DisproverProject/ProvableByPP/prob9313_1.mch'
], 'Check ProB finds proof').
cli_testcase(1415, [cbc,random], [
 '-p', 'RANDOMISE_ENUMERATION_ORDER', 'TRUE',
 '../prob_examples/public_examples/B/FeatureChecks/RandomEnumeration.mch',
 '-init', '-assertions'
  ],'Check random enumeration').
cli_testcase(1416, [b_test], ['../prob_examples/public_examples/B/ErrorMachines/ImplicationPriority.mch','-expcterr',static_check_machine, '-init'], 'Test warning thrown for strange PROPERTIES clause.').
cli_testcase(1417, [cbc], ['../prob_examples/public_examples/B/PerformanceTests/SetMultipleSolutions.mch', '-mc', 100, '-nodead'], 'Test no max-reached warning (improvement to cardinality_as_int1 which avoids unordered enumeration).').
cli_testcase(1418, [cbc,sigma], [
'-evalf', 'card(s)=5 & card(r)=3 & SIGMA(x).(x:r|x)=SIGMA(y).(y:s|y) & s<:0..16 & r<:s',
'-evalt', 'card(x)=10 & !y.(y:x & y<50 => y+1:x) & x<:1..50',
'-evalt', '{x|card(x)=10 & !y.(y:x & y<50 => y+1:x) & x<:1..50} = {41..50}',
'-evalt', '{x|card(x)=50 & !y.(y:x & y<50 => y+1:x) & x<:1..50} = {1..50}',
'-evalt', '{x|card(x)=150 & !y.(y:x & y<50 => y+1:x) & x<:1..50} = {}',
'-evalt', 'card(x)=10 & !(y,z).(y|->z:x & y<50 => y+1|->z:x) & x<:(1..50)*{TRUE}',
'-evalt', 'card(x)=10 & !(y).(y:x & y\'f1<50 => rec(f1:y\'f1+1):x) & x<:struct(f1:1..50)',
'-evalt', 'card(x)=10 & !(y,z).(y|->z:x & y<50 => y+1|->z:x) & x<:(1..50)*{"A","B"}'
% slow:'-evalt', 'card(x)=10 & !(y).(y:x & y\'f1<50 => rec(f1:y\'f1+1,f2:y\'f2):x) & x<:struct(f1:1..50,f2:BOOL)'
% also slow: card(x)=10 & !(y).(y:x & prj1(INTEGER,BOOL)(y)<50 => (prj1(INTEGER,BOOL)(y)+1,TRUE):x) & x<:(1..50)*BOOL
% fast: card(x)=10 & !(y).(y:x & prj1(INTEGER,BOOL)(y)<50 => (prj1(INTEGER,BOOL)(y)+1,prj2(INTEGER,BOOL)(y)):x) & x<:(1..50)*BOOL
% slow card(x)=4 & !(y).(y:x & y'f2<50 => rec(f2:y'f2+1,f1:y'f1):x) & x<:struct(f2:1..50,f1:BOOL)
% slow : card(x)=10 & !(y,z).(z|->y:x => ( y<50 => z|->y+1:x)) & x<:BOOL*(1..50)
], 'Test this is solved quickly using new ordered_value check for card.').
cli_testcase(1419, [cbc], [ '-p', 'SMT', 'TRUE',
'-evalt', 'card(x)=1 & !(y).(y:x & y\'f1<50 => rec(f1:y\'f1+1,f2:y\'f2):x) & x<:struct(f1:1..50,f2:BOOL)',
'-evalt', 'card(x)=10 & !(y).(y:x & y\'f1<50 => rec(f1:y\'f1+1,f2:y\'f2):x) & x<:struct(f1:1..50,f2:{TRUE})',
'-evalt', 'card(x)=1 & !(y).(y:x & prj1(INTEGER,BOOL)(y)<50 => (prj1(INTEGER,BOOL)(y)+1,TRUE):x) & x<:(1..50)*BOOL'
], 'Test this is solved quickly using new ordered_value check for card & avoid bug in expand all quantifeir.').
cli_testcase(1420, [tickets,sigma], [
'-evalt', 'Z={((((1|->1)|->-59843)|->-57163)),((((2|->1)|->-57163)|->-56910))} & r=SIGMA(a,b,c,d).((a,b,c,d):Z|d-c)'
], 'Test issue with Sigma/Pi for multiple variables').
cli_testcase(1421, [symbolic_model_checker], ['../prob_examples/public_examples/B/SymbolicModelChecking/Counters/InfiniteAndFiniteCounterCorrect.mch', '-symbolic_model_check', 'ic3'], 'Contains Finite and Infinite Counter').
cli_testcase(1422, [symbolic_model_checker], ['../prob_examples/public_examples/B/SymbolicModelChecking/Counters/InfiniteAndFiniteCounterWrong.mch', '-symbolic_model_check', 'ic3', '-expcterr', 'invariant_violation'], 'Contains Finite and Infinite Counter + Invariant Violation').
cli_testcase(1423, [symbolic_model_checker], ['../prob_examples/public_examples/B/SymbolicModelChecking/Counters/InfiniteCounter.mch', '-symbolic_model_check', 'ic3'], 'Contains Infinite Counter').
cli_testcase(1424, [symbolic_model_checker], ['../prob_examples/public_examples/B/SymbolicModelChecking/Counters/VerySimpleCounter.mch', '-symbolic_model_check', 'ic3'], 'Simple Counter').
cli_testcase(1425, [symbolic_model_checker], ['../prob_examples/public_examples/B/SymbolicModelChecking/Counters/VerySimpleCounterWrong.mch', '-symbolic_model_check', 'ic3', '-expcterr', 'invariant_violation'], 'Simple Counter + Invariant Violation').
cli_testcase(1426, [symbolic_model_checker], ['../prob_examples/public_examples/B/SymbolicModelChecking/Counters/VerySimpleCounterWrongLargeBranching.mch', '-symbolic_model_check', 'ic3', '-expcterr', 'invariant_violation'],'Simple Counter with large branching factor').
cli_testcase(1427, [error_checks], ['../prob_examples/public_examples/B/ErrorMachines/InitTriesToAssignConstant.mch', '-t', '-strict', '-expcterr', 'type_error', '-expcterr', 'load_main_file'], 'Check that we detect when we assign a constant in the INITIALISATION').
cli_testcase(1428, [symbolic_model_checker], ['../prob_examples/public_examples/B/SymbolicModelChecking/Bakery0.mch', '-symbolic_model_check', 'ic3'], 'Bakery Model without Invariant Violation').
cli_testcase(1429, [symbolic_model_checker], ['../prob_examples/public_examples/B/SymbolicModelChecking/Bakery0Wrong.mch', '-symbolic_model_check', 'ic3', '-expcterr', 'invariant_violation'], 'Bakery Model with Invariant Violation').
cli_testcase(1430, [laws], ['../prob_examples/public_examples/B/Laws/SetLawsNatural.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE'], 'Mathematical Laws about Sets with Natural sets mixed').
cli_testcase(1431, [symbolic_model_checker], ['../prob_examples/public_examples/B/SymbolicModelChecking/TimingExampleSimpler_v2_VariablesEvenMoreLimited.mch', '-symbolic_model_check', 'ic3', '-expcterr', 'invariant_violation'], 'Invariant Violation, Variables limited to finite sets').
cli_testcase(1432, [symbolic_model_checker], ['../prob_examples/public_examples/B/SymbolicModelChecking/TimingExampleSimpler_v3_VariablesLimited.mch', '-symbolic_model_check', 'ic3'], 'No Invariant Violation, Variables limited to finite sets').
cli_testcase(1433, [symbolic_model_checker], ['../prob_examples/public_examples/B/SymbolicModelChecking/TimingExampleSimpler_v2.mch', '-symbolic_model_check', 'ic3', '-expcterr', 'model_check_incomplete'], 'Negative Test: Currently the solvers are too weak for this test.').
cli_testcase(1434, [symbolic_model_checker], ['../prob_examples/public_examples/B/SymbolicModelChecking/TimingExampleSimpler_v3.mch', '-symbolic_model_check', 'ic3', '-expcterr', 'model_check_incomplete'], 'Negative Test: Currently the solvers are too weak for this test.').
cli_testcase(1435, [cbc], ['../prob_examples/public_examples/B/CBC/GraphMining_RunningExample_Unsat.mch', '-init', '-expcterr', 'setup_constants_fails'], 'Test odered set symmetry breaking works with card(S)>Nr constraint.').
cli_testcase(1436, [cbc,disprover], [
 '../prob_examples/public_examples/ProofObligations/GearDoor1_mch.pl',
 '../prob_examples/public_examples/ProofObligations/GearsDoorsHandle3_mch.pl',
 '../prob_examples/examples/ProofObligations/ABZ14_Mery/M2_mch.pl',
 '../prob_examples/examples/ProofObligations/ABZ14_Mery/M20_mch.pl',
 '../prob_examples/examples/ProofObligations/ABZ14_Mery/M3_mch.pl',
 '../prob_examples/examples/ProofObligations/ABZ14_Mery/M30_mch.pl',
 '../prob_examples/public_examples/ProofObligations/Abrial_ABZ_Devel3/m1_true.pl',
 '../prob_examples/public_examples/ProofObligations/Abrial_ABZ_Devel3/m2_mch.pl',
 '../prob_examples/public_examples/ProofObligations/Abrial_ABZ_Devel2/m0_mch.pl',
 '../prob_examples/public_examples/ProofObligations/Abrial_ABZ_Devel2/m1_mch.pl',
 '../prob_examples/public_examples/ProofObligations/Abrial_ABZ_Devel2/m2_mch.pl',
 '../prob_examples/public_examples/ProofObligations/Abrial_ABZ_Devel1/LPN0_mch.pl',
 '../prob_examples/public_examples/ProofObligations/Abrial_ABZ_Devel1/LPN1_mch.pl',
 '../prob_examples/public_examples/ProofObligations/Abrial_ABZ_Devel1/LPN2_mch.pl',
 '-p', 'DOUBLE_EVALUATION', 'TRUE'
 %,'../prob_examples/examples/ProofObligations/ABZ14_Mery/M4_mch.pl'
  ],'Test 1409 with double check').
cli_testcase(1437, [cbc,disprover], [
 '../prob_examples/examples/ProofObligations/Abrial_Ch4_Brp/brp_1_true.pl',
 '../prob_examples/examples/ProofObligations/Abrial_Ch4_Brp/brp_3_true.pl',
 '-p', 'DOUBLE_EVALUATION', 'TRUE'
  ],'Test 1410 with double check').
cli_testcase(1438, [cbc,disprover], [
 '../prob_examples/public_examples/ProofObligations/DisproverDefSetTest2_ctx.pl',
 '../prob_examples/public_examples/ProofObligations/DisproverDefSetTest3_ctx.pl',
 '../prob_examples/public_examples/ProofObligations/DisproverDefSetTest4_ctx.pl',
 '-check_disprover_result', 0,0,2,
 '-p', 'DOUBLE_EVALUATION', 'TRUE'
  ],'Test 1411 with double check').
cli_testcase(1439, [cbc], ['../prob_examples/public_examples/B/Tickets/LambdaInverse/LambdaInverseBehaviour.mch', '-init', '-sptxt', '../prob_examples/public_examples/B/Tickets/LambdaInverse/LambdaInverseBehaviour.sptxt'], 'Test no lambda closure in result state.').
cli_testcase(1440, [cbc,disprover], [
 '../prob_examples/public_examples/ProofObligations/InconsistenHyp_ctx.pl',
 '-p', 'DOUBLE_EVALUATION', 'TRUE', '-p', 'STRICT_RAISE_WARNINGS', 'TRUE',
 '-expcterr', 'disprover_inconsistent_hypotheses'
  ],'Test 1411 with double check').
cli_testcase(1441, [disprover,s21], [
 '../prob_examples/public_examples/ProofObligations/S21/train_0_mch.pl',
 '-check_disprover_result', 11,0,24
  ],'S21 proof obligations').
cli_testcase(1442, [disprover,s21], [
 '../prob_examples/public_examples/ProofObligations/S21/train_1_mch.pl',
 '-check_disprover_result', 15,0,28
  ],'S21 proof obligations').
cli_testcase(1443, [disprover,s21], [
 '../prob_examples/public_examples/ProofObligations/S21/train_2_mch.pl',
 '-check_disprover_result', 0,0,6
  ],'S21 proof obligations').
 cli_testcase(1444, [disprover,s21], [
  '../prob_examples/public_examples/ProofObligations/S21/train_3_mch.pl',
  '-check_disprover_result', 11,0,0,
 '-expcterr', 'disprover_inconsistent_hypotheses'
   ],'S21 proof obligations').
cli_testcase(1445, [disprover,s21], [
  '../prob_examples/public_examples/ProofObligations/S21/train_4_mch.pl',
  '-check_disprover_result', 1,0,0,
 '-expcterr', 'disprover_inconsistent_hypotheses'
   ],'S21 proof obligations').
cli_testcase(1446, [disprover,s21], [
 '../prob_examples/public_examples/ProofObligations/S21/train_5_mch.pl',
 '-check_disprover_result', 79,0,0,
 '-expcterr', 'disprover_inconsistent_hypotheses'
  ],'S21 proof obligations').
cli_testcase(1447, [disprover,s21], [
 '../prob_examples/public_examples/ProofObligations/S21/train_ctx0_ctx.pl',
 '-check_disprover_result', 1,10,6,
 '-timeout', 2000 % timeout per PO
  ],'S21 proof obligations').
cli_testcase(1448, [disprover,s21], [
 '../prob_examples/public_examples/ProofObligations/S21/train_ctx1_ctx.pl',
 '-check_disprover_result', 1,0,2
  ],'S21 proof obligations').
cli_testcase(1449, [disprover,s21], [
 '../prob_examples/public_examples/ProofObligations/S21/train_ctx2_ctx.pl',
 '-check_disprover_result', 7,0,0,
 '-expcterr', 'disprover_inconsistent_hypotheses'
  ],'S21 proof obligations').
cli_testcase(1450, [disprover,s21], [
 '../prob_examples/public_examples/ProofObligations/S21/train_ctx5_ctx.pl',
 '-check_disprover_result', 0,0,0
  ],'S21 proof obligations').
 cli_testcase(1451, [cbc,atelierb,chr], ['../prob_examples/examples/B/Alstom/DisproverProject/TrueContradiction.mch',
 '-p', 'CHR', 'TRUE', '-p', 'DISPROVER_MODE', 'TRUE',
 '-cbc_assertions', '-cbc_option', contradiction_check, '-cbc_option', tautology_check , '-cbc_option', unsat_core,
 '-cbc_result_file', '../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result.txt',
 '-strict', '-expcterr', contradiction_in_hypotheses], 'Test Atelier-B Disprover interface').
cli_testcase(1452, [disprover], [
 '../prob_examples/public_examples/ProofObligations/Mammar_Laleau_ABZ/Cylinders_mch.pl',
 '-timeout', 2500,
 '-check_disprover_result', 37,0,0
  ],'ABZ Mammar Laleau - requires exists expansion').
cli_testcase(1453, [disprover], [
 '../prob_examples/public_examples/ProofObligations/Mammar_Laleau_ABZ/Gears_mch.pl',
 %'../prob_examples/public_examples/ProofObligations/Mammar_Laleau_ABZ/Doors_mch.pl',
 '../prob_examples/public_examples/ProofObligations/Mammar_Laleau_ABZ/GearsIntermediateStates_mch.pl',
 '../prob_examples/public_examples/ProofObligations/Mammar_Laleau_ABZ/Failures_mch.pl' % requires good reification of partial_function test
  ],'ABZ Mammar Laleau POs should all be proven, test closure reification').
cli_testcase(1454, [disprover,cbc], ['-p','SMT','TRUE',
'-evalf', 'not(handler_ind  : dom({FALSE |-> d,TRUE |-> c}) &{FALSE |-> a,TRUE |-> b} : BOOL +-> INTEGER) & handler_ind : BOOL & d:INTEGER', % other small test
'-evalf', 'not(handler_ind  : dom({FALSE |-> d,TRUE |-> c}) &{FALSE |-> a,TRUE |-> b} : BOOL --> INTEGER) & handler_ind : BOOL & d:INTEGER', % other small test
'-evalf', 'not(f<+{po |-> {TRUE |-> currentTime + 12,FALSE |-> 16}(po)} : BOOL --> NATURAL) &  f : BOOL --> NATURAL&po=TRUE & currentTime:NATURAL', % this checks not_total_function improvement
'-evalf', 'not(f <+ {po |-> {TRUE |-> currentTime + 12,FALSE |-> 16}(po)} : BOOL +-> NATURAL) &  f : BOOL +-> NATURAL & currentTime>0', % this checks not_partial_function improvement
'-evalf', 'not(BOOL * {currentTime + 4} : BOOL --> NATURAL) & currentTime >0' % this checks not_subset improvement for not_total_function
  ],'Check reification of x:NATURAL and choice point in not_total_function and not_partial_function').
 cli_testcase(1455, [cbc,atelierb,disprover], ['-p', 'CHR', 'TRUE', '-p', 'DISPROVER_MODE', 'TRUE',
 '-evalf', 'not(#(i0).(i0 : INTEGER & ((tdc <+ {i |-> TRUE}) <+ {i0 |-> TRUE} = {1,2,3} * {TRUE} & door_closing = TRUE => door_closed = FALSE))) & (door_closing = TRUE => (door_open = FALSE & door_closed = FALSE))'
 ], 'Test splitting of existential quantifier').
 % some laws also checked in CHR+DISPROVER mode:
cli_testcase(1456, [laws,rel_fnc,cbc,disprover], ['../prob_examples/public_examples/B/Laws/ExplicitComputations.mch', '-t', '-mc', '100000', '-nodead', '-c', '-strict', '-p', 'CLPFD', 'TRUE', '-p', 'TIME_OUT', '9000',
 '-p', 'DISPROVER_MODE', 'TRUE'], 'Many explicit computations which check that all operators work as expected on various concrete data (CLPFD+DISPROVER mode).').
cli_testcase(1457, [laws,chr], ['../prob_examples/public_examples/B/Laws/BoolLaws.mch', '-mc', '100000', '-nodead', '-c', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE',
'-p', 'DISPROVER_MODE', 'TRUE', '-p', 'CHR', 'TRUE'], 'Various laws about BOOL datatype and boolean connectives.').
cli_testcase(1458, [laws,chr], ['../prob_examples/public_examples/B/Laws/ArithmeticLaws.mch', '-mc', '100000', '-c', '-nodead', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE',
'-p', 'DISPROVER_MODE', 'TRUE', '-p', 'CHR', 'TRUE'], 'Various arithmetic laws.').
cli_testcase(1459, [laws,chr], ['../prob_examples/public_examples/B/Laws/NatRangeLaws.mch', '-mc', '100000', '-c', '-nodead', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE',
'-p', 'DISPROVER_MODE', 'TRUE', '-p', 'CHR', 'TRUE'], 'Laws about intervals etc.').
cli_testcase(1460, [laws,rel_fnc,chr], ['../prob_examples/public_examples/B/Laws/RelLaws.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE',
'-p', 'DISPROVER_MODE', 'TRUE', '-p', 'CHR', 'TRUE'], 'Use model checking to check a variety of laws on relations.').
cli_testcase(1461, [laws,chr], ['../prob_examples/public_examples/B/Laws/SetLaws.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'FALSE',
'-p', 'DISPROVER_MODE', 'TRUE', '-p', 'CHR', 'TRUE'], 'Mathematical Laws about Sets').
 cli_testcase(1462, [cbc,atelierb,disprover], ['-p', 'CHR', 'TRUE', '-p', 'DISPROVER_MODE', 'TRUE',
 '-evalf', 'not(!(i0).(i0 : {1,2,3} => (three_door_open <+ {i |-> FALSE})(i0) = FALSE or three_door_closed(i0) = FALSE)) & three_door_open : {1,2,3} --> BOOL & three_door_closed : {1,2,3} --> BOOL & ( door_open = TRUE => three_door_open = {1,2,3} * {TRUE}) & ( door_closed = TRUE => three_door_closed = {1,2,3} * {TRUE}) & !(i).(i : {1,2,3} => three_door_open(i) = FALSE or three_door_closed(i) = FALSE) & (door_open = FALSE or door_closed = FALSE)',
 '-evalf', 'not( (gears_rtr = TRUE <=> three_gears_rtr <+ {i |-> TRUE} = {1,2,3} * {TRUE} )) & ( gears_rtr = TRUE <=> three_gears_rtr = {1,2,3} * {TRUE}) & i : {1,2,3} & three_gears_ext(i) = FALSE & three_gears_rtr(i) = FALSE & (three_gears_rtr <+ {i |-> TRUE}) /= {1,2,3} * {TRUE} '
 ], 'Test improved treatment of function overwrite').
cli_testcase(1463, [disprover], [
 '../prob_examples/public_examples/ProofObligations/Abrial_ABZ_Devel2/m6_mch.pl',
 '-check_disprover_result', 102,0,1
  ],'Requires function overwrite + better sharing of negated preds').
cli_testcase(1464, [cbc,cse, cse_test,sigma], [ '-p', 'CSE', 'TRUE',
 '-evalf', 'x+y >100 & x+y <99',
 '-evalt', '(x+y)+(x+y)=(x+y)',
 '-evalt', '((x+y)+z)*(x+y)+((x+z)+y) + (x+z) = 0',
 '-evalt', '!x.(x:NATURAL &x+z<10 => x+z<11) & z=2',
 '-evalt', 'SIGMA(x).(x:NATURAL &x+z<4 | x+z)=5 & z=2',
 '-evalt', 'PI(x).(x:NATURAL &x+z<4 | x+z)=6 & z=2',
 '-evalt', '(z>1 => #x.(x:NATURAL &x+z<10 & x+z<11)) & z=2',
 '-evalt', '%(x).(x:NATURAL1 &x+z<4 | x+z)=[3] & z=2',
 '-evalt', '{x,r | x:NATURAL1 &x+z<4 & r= x+z}=[3] & z=2',
 '-evalt', '{v|v+(x+x) : 0..10 & v:0..5 & v+(x+x) < 3}={0} & x+x : 0..4',
 '-evalt', 'f:BOOL +-> 0..10 & (f/={} => f(TRUE)>5) & (FALSE:dom(f)=> f(TRUE)>5)',
 '-evalt', 'f:BOOL +-> 0..10 & (f/={} => f(TRUE)>5) <=> b=TRUE & (FALSE:dom(f)=> f(TRUE)>5) <=> b=FALSE',
 '-evalt', 'f:1..3+->1..3 & (x:dom(f) => #r.(r=f(x) & r>2 & r<10))<=>b=TRUE & x:3..4 & f={}',
 '-evalt', 'f:BOOL +-> 0..10 & (f/={} => (f(TRUE)>5 <=> c=TRUE)) <=> b=TRUE & (FALSE:dom(f)=> (f(TRUE)>5<=>c=FALSE)) <=> b=FALSE & c=FALSE',
 '-evalf', 'a: 1..6 --> {0,1,5} & b: 1..3 --> {30,210,21} & SIGMA(i).(i:1..0|a(i))+SIGMA(j).(j:1..3|b(i))=20 & r=b(i)'
], 'Some simple CSE tests').

cli_testcase(1465,[cse, cse_test],['-p', 'CSE', 'TRUE', '-p', 'MAXINT', 127, '-p', 'MININT', -128,

  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/plavis-MemoryLoad_SP_55.prob',
  '-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/qsee-TransmitMemoryDumpOk21_SP_6.prob'
  % ,'-evalt_file', '../prob_examples/examples/Setlog/prob-ttf/plavis-TransData_SP_13.prob' %% does not work yet
  ], 'Test from 1003').
cli_testcase(1466,[cse,tickets], ['-p', 'CSE', 'TRUE',
  '../prob_examples/public_examples/ProofObligations/Abrial_ABZ_Devel3/m3_act0_inv2_as_b.mch',
  '-init'
], 'Test WD issue does not prevent from finding solution').

cli_testcase(1467, [disprover,pacemaker], [
 '../prob_examples/examples/ProofObligations/Pacemaker_1_15March2011/M0_AOO_mch.pl',
 '-check_disprover_result', 27,0,1
  ],'Pacemaker proof obligations').
cli_testcase(1468, [por,enabling], ['../prob_examples/public_examples/B/Ivo/DepEnabling.mch', '-mc', '1000000', '-noinv', '-p', 'por', ample_sets, '-expcterr', deadlock, '-strict'], 'Testing whether the deadlock state is found when model checking with POR.').
cli_testcase(1469, [por,enabling], ['../prob_examples/public_examples/B/Ivo/DepEnabling.mch', '-mc', '1000000', '-noinv', '-p', 'por', ample_sets, '-expcterr', deadlock, '-strict'], 'Testing whether the deadlock state is found when model checking with POR.').
cli_testcase(1470, [csp_test,fairness,ltl], ['../prob_examples/public_examples/CSP/LTL/mbuff.csp', '-ltlformula', 'SF(snd_mess.t1.d1) & SF(mess.t1.d1) & SF(rcv_mess.t1.d1) & SF(right.t1.d1) => G ([left.t1.d1] => F [right.t1.d1])', '-check_complete', '-cc', '1406', '4063', '-strict'], 'Test fairness implementation.').
cli_testcase(1471, [csp_test,fairness,ltl], ['../prob_examples/public_examples/CSP/LTL/mbuff.csp', '-ltlformula', 'SEF => G ([left.t1.d1] => F [right.t1.d1])', '-check_complete', '-cc', '1406', '4063', '-strict'], 'Test fairness implementation.').
cli_testcase(1472, [csp_test,fairness,ltl], ['../prob_examples/public_examples/CSP/LTL/peterson.csp', '-ltlformula', 'WF(p2enter) & WF(p2critical) & WF(p2leave) & WF(p2resetflag2) & WF(p2set1turn) & WF(p1set2turn) & WF(p1gettrueflag2) & WF(p2gettrueflag2) & WF(p2gettrueflag1) & WF(p1get1turn) & WF(p2get2turn) & WF(p1enter) => G ([p1setflag1] => F [p1enter])', '-check_complete', '-cc', 57, 113, '-strict'], 'Test fairness implementation.').
cli_testcase(1473, [csp_test,fairness,ltl], ['../prob_examples/public_examples/CSP/LTL/peterson.csp', '-ltlformula', 'WEF => G ([p1setflag1] => F [p1enter])', '-check_complete', '-cc', 57, 113, '-strict'], 'Test fairness implementation (skipped until next ProB parser release).').
cli_testcase(1474, [csp_test,fairness,ltl], ['../prob_examples/public_examples/CSP/LTL/pris.csp', '-ltlformula', 'SF(in.A) & SF(in.B) & SF(in.C) => F[done]', '-cc', 48, 61, '-strict'], 'Test fairness implementation.').
cli_testcase(1475, [csp_test,fairness,ltl], ['../prob_examples/public_examples/CSP/LTL/pris.csp', '-ltlformula', 'SEF => F[done]', '-cc', 48, 61, '-strict'], 'Test fairness implementation (skipped until next ProB parser release).').
cli_testcase(1476, [csp_test,fairness,ltl], ['../prob_examples/public_examples/CSP/LTL/scheduler.csp', '-ltlformula', 'WF(ready.5) & WF(ready.4) & WF(ready.3) & WF(ready.2) & WF(leave.1) => G ([enter.1] => F [leave.1])', '-cc', 3194, 12046, '-strict'], 'Test fairness implementation.').
cli_testcase(1477, [csp_test,fairness,ltl], ['../prob_examples/public_examples/CSP/LTL/scheduler.csp', '-ltlformula', 'WEF => G ([enter.1] => F [leave.1])', '-cc', 3194, 12046, '-strict'], 'Test fairness implementation (skipped until next ProB parser release).').
cli_testcase(1478, [cbc,cse, cse_test,sigma], [ '-p', 'CSE', 'FALSE', '-p', 'SMT', 'TRUE',
 '-evalf', 'a: 1..6 --> {0,1,5} & b: 1..3 --> {30,210,21} & SIGMA(i).(i:1..0|a(i))+SIGMA(j).(j:1..3|b(i))=20 & r=b(i)'
], 'Test new FDSET collection for function application').
cli_testcase(1479, [csp_test,fairness,ltl], ['../prob_examples/public_examples/CSP/LTL/scheduler.csp', '-ltlformula', 'SF(ready.5) & SF(ready.4) & SF(ready.3) & SF(ready.2) & SF(leave.1) => G ([enter.1] => F [leave.1])', '-strict'], 'Test fairness implementation (fix of the memory exhaustion issue).').
cli_testcase(1480, [csp_test,fairness,ltl], ['../prob_examples/public_examples/CSP/LTL/scheduler.csp', '-ltlformula', 'SF(ready.5) & SF(ready.4) & SF(ready.3) & SF(ready.2) & SF(delete.1) => G ([new.1] => F [delete.1])', '-strict'], 'Test fairness implementation (fix of the memory exhaustion issue).').
cli_testcase(1481, [csp_test,fairness,ltl], ['../prob_examples/public_examples/CSP/LTL/phils.csp', '-ltlformula', 'SF(sits._) => G ([sits.0] => F [eats.0])', '-expcterr', ltl, '-strict'], 'Testing fairness check to fail when using _ as argument in order to set a group of events (e.g. sits._) as fairness constraints.').
cli_testcase(1482, [tickets,override], ['../prob_examples/public_examples/EventBPrologPackages/Tests/FunOverrideTest.mch', '-mc', 99, '-strict'], 'Test f(x) := E substitution deals with f being a relation.').
cli_testcase(1483, [laws,sequences], ['../prob_examples/public_examples/B/Laws/SeqLaws.mch', '-mc', '500', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE'], 'Laws about sequences').
cli_testcase(1484, [cbc,tickets], [
 '-evalt', 'res=x*x+4*x+14',
 '-evalt', 'x*x+4*x+14=res'
 ], 'Check no CLPFD overflow').
cli_testcase(1485, [tickets], [
 '-evalt', 'x = -(x) & y={x,2} & z={x,3}',
 '-evalt', 'x = -v & v=2+w & w=10 & y={x,2} & z={x,3}'
 ], 'Check no error in predicate partitioning and equality replacement').
cli_testcase(1486, [b_test,override], ['-t', '-strict', '../prob_examples/public_examples/B/FeatureChecks/OverrideRelationAssignment.mch'], 'Check function override assignment works for relation').
cli_testcase(1487, [records,cbc], [
 '-evalf', 'r:struct(f1:11..20,f2:12..20) & r\'f1+10 = r\'f2',
 '-evalt', 'r:struct(f1:11..20,f2:12..30) & r\'f1+10 = r\'f2',
 '-evalt', 'r:struct(f1:11..20,f2:12..30) & r\'f1+10 = r\'f2 & s:struct(f1:15..20,f2:1..100) & s=r'
 ], 'A few REPL tests with records').
cli_testcase(1488, [cbc], [
 '-evalf', 'A /= {} & A <<: {x} & x:INTEGER', % NotSubsetoOfSingleton
 '-evalf', '{(x,y)} <: R & x /: dom(R) & R : INTEGER <-> INTEGER', % SingletonNotSubsetRel
 '-evalf', 'X /\\ Y = Z & X = {} & Z /= {} & X<:INTEGER', % dinters(X,Y,Z) & X = {} & Z neq {} --> false
 '-evalf', 'R:INTEGER<->INTEGER & A={} & Q= A <| R & dom(Q)=D & D /= {}', %DresEmpty2 % is_rel(R) & A = {} & dres(A,R,Q) & dom(Q,D) & D neq {}
 '-evalf', 'R:INTEGER<->INTEGER & R /= {} & dom(R)=D & D <<: {X}', % DomNotSubsetOfSingleton % is_rel(R) & R neq {} & dom(R,D) & ssubset(D,{X})
 '-evalf', 'R:INTEGER<->INTEGER & R /= {} & dom(R)=D & A=D & A <| R=Q & R /\\ Q = {}', % DresCapEqual
 '-evalf', 'R:INTEGER<->INTEGER & R /= {} & dom(R)=D & A=D & A <| R = {}', % DresEqual
 '-evalf', 'R:INTEGER<->INTEGER & dom(R)=D & X /: D & dom({(X,Y),B}) = DS & D=DS & Y:INTEGER', % SingletonMappletNotInDom % is_rel(R) & dom(R,D) & X nin D & dom({[X,Y] / B},DS) & D = DS
 % This does not work R:INTEGER<->INTEGER & dom(R)=D & X /: D & dom({(X,Y),B,B2}) = DS & D=DS & Y:INTEGER
 '-p', 'CLPFD', 'TRUE'
 ], 'Check that ProB can handle Setlog TTF filtering rules').
cli_testcase(1489, [b_test,proz],['../prob_examples/examples/Z/Cristia/birthdayBook/zspec.fuzz', '-t', '-p', 'DEFAULT_SETSIZE', 3, '-strict'],'ProZ test (new Init/init schema detection)').
cli_testcase(1490, [cbc_tests,cbc,proz],['../prob_examples/examples/Z/Cristia/birthdayBook/zspec.fuzz',
 '-cbc_tests', 2 ,'', '', '-cbc_cover_all'],'Check that we can cover all events').
cli_testcase(1491, [eventb_test,infinite,override], ['../prob_examples/public_examples/EventBPrologPackages/FeatureTests/InfiniteChanges_mch.eventb', '-t'], 'Test that we can override infinite functions.').
cli_testcase(1492, [tickets,alstom,random], ['../prob_examples/examples/B/Alstom/ixl_Apr_2015/scheduler_i.mch', '-animate', 10, '-p', 'CLPFD', 'FALSE', '-p', 'RANDOMISE_ENUMERATION_ORDER', 'TRUE'], 'Test that random enumeration works without CLPFD.').
cli_testcase(1493, [pge_fast], ['../prob_examples/public_examples/B/Ivo/SkippingComplexGuardsEvaluation_mch_mx5000.eventb', '-mc', 1000000, '-p', 'pge', full, '-get_coverage_information', '../prob_examples/public_examples/B/Ivo/SkippingComplexGuardsEvaluation_mch_mx5000_Coverage.txt', '-strict'], 'Pathological test case for the PGE optimisation: 0 guard evaluations should be skipped.').
cli_testcase(1494, [pge_fast], ['../prob_examples/public_examples/B/Ivo/NoDisablings_mx5.eventb', '-mc', 1000000, '-p', 'pge', full, '-get_coverage_information', '../prob_examples/public_examples/B/Ivo/NoDisablings_mx5_Coverage.txt', '-strict'], 'Pathological test case for the PGE optimisation: fixed number of guard evaluations should be skipped.').
cli_testcase(1495, [pge_fast,enabling], ['../prob_examples/public_examples/B/Ivo/PGE_Deadlock.mch', '-mc', 1000000, '-noinv', '-p', 'pge', full, '-expcterr', deadlock, '-strict'], 'Testing whether the deadlock state is discovered when -noinv is set (invariant should not be used for the enabling analysis when model checking just for deadlock freedom).').
cli_testcase(1496, [pge_fast], ['../prob_examples/public_examples/B/Ivo/PGE_Deadlock.mch', '-mc', 1000000, '-p', 'pge', full, '-expcterr', invariant_violation, '-strict'], 'Testing whether the invariant violation is discovered.').
cli_testcase(1497, [pge_fast], ['../prob_examples/public_examples/B/Ivo/PGE_Deadlock.mch', '-ltlformulaf', 'G not deadlock', '-p', pge, full, '-strict'], 'Testing whether the deadlock state is discovered when using PGE while LTL model checking.').
 cli_testcase(1498, [cbc,queens], [
 '-evalt', '{x| x:1..3 -->> 5..7 & x(2)=6} = {{(1|->5),(2|->6),(3|->7)},{(1|->7),(2|->6),(3|->5)}}',
 '-evalt', '{x| x:1..3 -->> 5..7 & x(2)>x(3)} = {{(1|->7),(2|->6),(3|->5)},{(1|->5),(2|->7),(3|->6)},{(1|->6),(2|->7),(3|->5)}}',
 '-evalt', '{x| x:1..3 -->> 5..8 & x(2)>x(3)}={}',
 '-evalt', 'n = 50 & queens : 1..n -->> 1..n &!(q1,q2).(q1:1..n & q2:2..n & q2>q1 => queens(q1)+q2-q1 /= queens(q2) & queens(q1)-q2+q1 /= queens(q2))',
 '-evalt', '{x| x:1..3 +->> 5..7 & x(2)=6} = {{(1|->5),(2|->6),(3|->7)},{(1|->7),(2|->6),(3|->5)}}',
 '-evalt', '{x| x:1..3 +->> 5..7 & x(2)>x(3)} = {{(1|->7),(2|->6),(3|->5)},{(1|->5),(2|->7),(3|->6)},{(1|->6),(2|->7),(3|->5)}}',
 '-evalt', '{x| x:1..3 +->> 5..8 & x(2)>x(3)}={}',
 '-evalt', 'n = 50 & queens : 1..n +->> 1..n &!(q1,q2).(q1:1..n & q2:2..n & q2>q1 => queens(q1)+q2-q1 /= queens(q2) & queens(q1)-q2+q1 /= queens(q2))',
 '-evalt', 'n = 50 & queens : 1..n >->> 1..n &!(q1,q2).(q1:1..n & q2:2..n & q2>q1 => queens(q1)+q2-q1 /= queens(q2) & queens(q1)-q2+q1 /= queens(q2))' % now added also total_bijection
 ], 'Check -->> / +->> optimization works correctly (test new treatment to treat like >-> when domain card = range card; now also >->>)').
cli_testcase(1499, [cbc,cse], ['../prob_examples/public_examples/B/Special/ConstrainedBasedChecking/mesiSet.mch', '-cbc', 'all', '-strict', '-expcterr', 'cbc', '-p', 'CSE', 'TRUE'], 'Test 402 with CSE enabled; nonvar types occur').
cli_testcase(1500, [cse], ['../prob_examples/public_examples/B/PerformanceTests/CommonSubexpressions/Sequence_CSE_Test2.mch', '-mc', '1002', '-nodead', '-bf', '-hash64', 25104873554326101, '-p', 'CSE', 'TRUE'], 'Test CSE with Sequence Substitutions').
cli_testcase(1501, [cse], ['../prob_examples/public_examples/B/PerformanceTests/CommonSubexpressions/WhileLoop_CSE_Test1.mch', '-mc', '500', '-nodead', '-bf', '-hash64', 137772507059734089, '-p', 'CSE', 'TRUE'], 'Test CSE with While Substitutions').
cli_testcase(1502, [cse], ['../prob_examples/public_examples/B/PerformanceTests/CommonSubexpressions/WhileLoop_CSE_Test2.mch', '-mc', '500', '-nodead', '-bf', '-hash64', 642472838688353119, '-p', 'CSE', 'TRUE'], 'Test CSE with While Substitutions').
cli_testcase(1503, [cse], ['../prob_examples/public_examples/B/PerformanceTests/CommonSubexpressions/WhileLoop_CSE_Test3.mch', '-mc', '500', '-nodead', '-bf', '-hash64', 642472838688353119, '-p', 'CSE', 'TRUE'], 'Test CSE with While Substitutions').
cli_testcase(1504, [cse], ['../prob_examples/public_examples/B/PerformanceTests/CommonSubexpressions/WhileLoop_CSE_Test5.mch', '-mc', '500', '-nodead', '-bf', '-hash64', 571377810868391599, '-p', 'CSE', 'TRUE'], 'Test CSE with While Substitutions').
cli_testcase(1505, [smt_solver_integration], [
  '-evalt', ':z3 x = 1',
  '-evalt', ':cvc4 x = 1',
  '-evalf', ':z3 x = 1 & x > 2',
  '-evalf', ':cvc4 x = 1 & x > 2'
 ], 'some simple constraints that should be solved by cvc4 / z3').
cli_testcase(1506, [smt_solver_integration], [
  '-evalt', ':z3 r:struct(f1:11..20,f2:12..30) & r\'f1+10 = r\'f2',
  '-evalt', ':cvc4 r:struct(f1:11..20,f2:12..30) & r\'f1+10 = r\'f2',
  '-evalf', ':z3 r:struct(f1:11..20,f2:12..20) & r\'f1+10 = r\'f2',
  '-evalf', ':cvc4 r:struct(f1:11..20,f2:12..20) & r\'f1+10 = r\'f2',
  '-evalt', ':z3 r:struct(f1:11..20,f2:12..30) & r\'f1+10 = r\'f2 & s:struct(f1:15..20,f2:1..100) & s=r',
  '-evalt', ':cvc4 r:struct(f1:11..20,f2:12..30) & r\'f1+10 = r\'f2 & s:struct(f1:15..20,f2:1..100) & s=r',
  '-evalt', ':z3 v1:struct(lderb:{TRUE},mspue:{FALSE},fhhxr:{TRUE})',
  '-evalt', ':cvc4 v1:struct(lderb:{TRUE},mspue:{FALSE},fhhxr:{TRUE})'
 ], 'records support by smt solvers').
cli_testcase(1507, [smt_solver_integration], [
  '-evalt', ':cvc4 test = \"asdfasdfasdF\"'
 ], 'string support by smt solvers').
cli_testcase(1508, [smt_solver_integration], [
  '-evalf', ':cvc x>0 & x+1 <= p & x=p',
  '-evalf', ':cvc x>0 & x+2 <= p & x=p',
  '-evalf', ':z3 x>0 & x+1 <= p & x=p',
  '-evalf', ':z3 x>0 & x+2 <= p & x=p',
  '-evalf', ':cvc not(Pace_Int - sp >= 0) & (sp > 0 & sp < Pace_Int)', % taken from pacemaker disprover testcase
  '-evalf', ':z3 not(Pace_Int - sp >= 0) & (sp > 0 & sp < Pace_Int)' % taken from pacemaker disprover testcase
 ], 'check that smt solvers would be able to replace the chr code in some cases').
cli_testcase(1509, [smt_solver_integration], [
  '-evalt', ':cvc4 {x} <: {999}',
  '-evalt', ':z3 {x} <: {999}',
  '-evalt', ':cvc4 {x} <: {999,1000} & x > 999',
  '-evalt', ':z3 {x} <: {999,1000} & x > 999',
  '-evalt', ':z3 x = {1,2} /\\ {3,4}',
  '-evalt', ':z3 x={(1,2),(3,4),(4,5),(6,7),(8,9)} & y|->v : x & y|->w : x',
  '-evalt', ':cvc4 x={(1,2),(3,4),(4,5),(6,7),(8,9)} & y|->v : x & y|->w : x'
 ], 'simple set constraints for smt solvers').
cli_testcase(1510, [smt_solver_integration], [
  '-evalt', ':z3 x : {1,2} --> NAT',
  '-evalf', ':z3 x : {1,2} --> NAT & x = {(1,1)}',
  '-evalt', ':z3 x : {1,2} --> NAT & x = {(1,1),(2,1)}'
 ], 'functions for z3 (in theory supported by cvc4, but often unknown)').
cli_testcase(1511, [smt_solver_integration], [
  '-evalf', ':cvc s = min({1}) & s=2',
  '-evalf', ':z3 s = min({1}) & s=2',
  '-evalf', ':cvc s = min({1,2}) & s>2',
  '-evalf', ':z3 s = min({1,2}) & s>2'
], 'check that smt solvers can falsify min / max').
cli_testcase(1512, [cse, ltl], ['../prob_examples/public_examples/B/LTL/Fairness/TestingFairness.mch',
 '-p', 'CSE', 'TRUE', '-p', 'CSE_SUBST', 'TRUE',
 '-ltlformulat', true,  '-ltlformulat', 'G{1=1}',
 '-ctlformulat', 'AG{1=1}', '-ctlformulaf', 'EF{1=2}',
 %'-ltlformulaf', false, % does not work !
 '-ltlformulaf', 'G{1=2}', '-strict'], 'Test 1244 with CSE_SUBST, checks top-level PRE detected inside lazy_let_subst.').
cli_testcase(1513, [smt_solver_integration], [
  '-evalt', ':z3 x=dom({(1,2)})'
 ], 'domain and range computation using z3').
cli_testcase(1514, [cse,case],['../prob_examples/examples/B/Satpathy/Teletext104/Teletext104_deterministic.mch', '-t',
 '-p', 'CSE', 'TRUE', '-p', 'CSE_SUBST', 'TRUE'
 ],'Test 1292 with CSE_SUBST: Check this relatively large machine works.').
cli_testcase(1515, [cse], [
 '-p', 'CSE', 'TRUE', '-p', 'CSE_SUBST', 'TRUE',
  '-evalt', '(x>1 => 10/x>2) & #v.(v:10..20 & (x>2 => 10/x=v))'
 ], 'check that no pending co-routines or wd problem').
cli_testcase(1516, [smt_solver_integration], [
  '-evalf', ':z3 x = {(1,5),(2,6),(3,7)} & i = x[{1,2}] & 7:i',
  '-evalt', ':z3 x = {(1,5),(2,6),(3,7)} & i = x[{1,2}]'
 ], 'check that smt solvers can falsify image (at least z3 can)').
 cli_testcase(1517, [smt_solver_integration], [
   '-evalf', ':z3 x = 1 .. 4 & z = x /\\ 5 ..3 & z /= {}',
   '-evalf', ':cvc x = 1 .. 4 & z = x /\\ 5 ..3 & z /= {}',
   '-evalt', ':z3 x = 1 .. 4 & z = x /\\ 5 .. 3'
   %'-evalt', ':cvc x = 1 .. 4 & z = x /\\ 5 .. 3' removed for now
  ], 'intervals for smt solvers').
cli_testcase(1518, [cbc], [
  '-evalt', 'x = %i.(i:1..10|i+i) & x[{j}]={8}',
  '-evalt', 'ran({x,j|x = {1|->2,2|->4, 4|->8, 5|->8} & x[{j}]={8}})=4..5',
  '-evalt', 'ran({x,j|x = {1|->2,23|->4, 44|->8, 55|->8} & x[{j}]={8}})={44,55}',
  '-evalt', 'x = %i.(i:1001..1100|i+i) & x[{j}]={2100}',
  '-evalt', 'x = %i.(i:1..10|i+i) & x[{5,j}]={16,10}',
  '-evalf', 'x = %i.(i:1..10|i+i) & x[{5,j}]={16,11}'
 ], 'check that we propagate/instantiate for image').
cli_testcase(1519, [b_test,external,records,strings], ['../prob_examples/public_examples/B/ExternalFunctions/LibraryStrings.mch',
 '-evalt', '{rec(sa:"1.101"),rec(sa:"1.102")} =x & x<: (struct(sa :{s| s : STRING & size(STRING_SPLIT(s,"."))=2}))',
 '-evalf', '{rec(sa:"1.101"),rec(sa:"1.102.3")} =x & x<: (struct(sa :{s| s : STRING & size(STRING_SPLIT(s,"."))=2}))'
 ], 'Check that no infinite expansion occurs here.').
cli_testcase(1520, [b_test,unicode,atelierb],
   ['../prob_examples/public_examples/B/Tickets/UnicodeSupport/t1_2_atelierb.mch', '-t'
 ],'Test Atelier-B unicode support.').
cli_testcase(1521, [b_test,external,records,infinite],
['../prob_examples/public_examples/B/Tickets/SymbolicSubArgs/SymbolicSubArg.mch',
'-p', 'ENUMERATE_INFINITE_TYPES', 'FALSE', % test also passes without this
 '-assertions'
 ], 'Check that no infinite expansion occurs here.').
cli_testcase(1522, [b_test,external,records,infinite],
['../prob_examples/public_examples/B/Tickets/SymbolicSubArgs/SymbolicSubArg.mch',
 '../prob_examples/public_examples/B/Tickets/SymbolicSubArgs/NonSymbolicSubArg.mch',
 '../prob_examples/public_examples/B/Tickets/SymbolicSubArgs/NonSymbolicSubArgDEF.mch',
 '-assertions'
 ], 'Like test 1521, but a few virtual time outs occur for NonSymbolic files.').
cli_testcase(1523, [smt_solver_integration], [
   '-evalt', ':z3 x = %i.(i:1..10|i+i) & x[{j}]={8}',
   '-evalt', ':z3 x = %i.(i:1..10|i+i) & x[{5,j}]={16,10}',
   '-evalf', ':z3 x = %i.(i:1..10|i+i) & x[{5,j}]={16,11}'
  ], 'parts of test 1518 using smt solvers').
cli_testcase(1524, [tickets,'PROB-381'], [
'../prob_examples/public_examples/B/Tickets/PROB-381/DuplicateFieldInRecord.mch',
'../prob_examples/public_examples/B/Tickets/PROB-381/DuplicateFieldInRecordType.mch', '-init',
'-expcterr', 'type_error', '-expcterr', 'load_main_file'], 'Check multiple use of record field detected. (Ticket PROB-381)').
cli_testcase(1525, [cbc], [
 '-evalf', 'x:S & y:S & y/:T & S<:NATURAL & T=S'
 ], 'ensure above is detected FALSE wo enum warning; requires better reification for FD equality').
cli_testcase(1526, [cbc], [
 '-evalt', 'f = {(1|->2),(2|->3),(3|->4),(4|->5),(5|->6),(6|->7),(7|->8),(8|->9),(9|->10),(10|->8)} & x = f[x] & x /= {}',
 '-evalt', 'f = %x.(x:1..9|x+1) \\/ {10|->8} & x = f[x] & x /= {}',
 '-evalt', 'f = %x.(x:1..n|x+1) \\/ {n+1|->8} & x = f[x] & x /= {} & n=10'
 ], 'ensure above solved without enum warning (better propagation for image)').
cli_testcase(1527, [smt_solver_integration], [
   '-evalt', ':z3 x:1..400 & y:1..400 & f: 280..290 --> 290..299 & x|->y :f & y|->x :f'
 ], 'check that a model with a complicated representation can be translated back from z3 to prob').
cli_testcase(1528, [cbc], [
 '-evalt', 'x = ["A","G","G"] & z = ["A","G","T","G","G"] & n = min({size(x),size(z)}) & m :1..n & xi: 1..m >-> 1..size(x) & zi: 1..m >-> 1..size(z) & !i.(i:1..m => x(xi(i)) = z(zi(i))) & !j.(j:1..(m-1) => xi(j)<xi(j+1) & zi(j)<zi(j+1)) & m>2 & sol1 = (xi;x) & sol2 = (zi;z)',
 '-evalt', 'x = ["A","G","G"] & z = ["A","G","T","G","G"] & n = min({size(x),size(z)}) & m :1..n & xi: 1..m >-> 1..size(x) & zi: 1..m >-> 1..size(z) & !j.(j:1..(m-1) => xi(j)<xi(j+1) & zi(j)<zi(j+1)) & m>2 & (xi;x) = (zi;z)',
 '-evalt', 'x = ["A","G","G","Z","G"] & z = ["A","G","T","G","G","T","G"] & n = min({size(x),size(z)}) & m :1..n & xi: 1..m >-> 1..size(x) & zi: 1..m >-> 1..size(z) & !j.(j:1..(m-1) => xi(j)<xi(j+1) & zi(j)<zi(j+1)) & m>3 & (xi;x) = (zi;z)',
 '-evalf', 'x = ["A","G","G"] & z = ["A","G","T","G","G"] & n = min({size(x),size(z)}) & m :1..n & xi: 1..m >-> 1..size(x) & zi: 1..m >-> 1..size(z) & !i.(i:1..m => x(xi(i)) = z(zi(i))) & !j.(j:1..(m-1) => xi(j)<xi(j+1) & zi(j)<zi(j+1)) & m>3 & sol1 = (xi;x) & sol2 = (zi;z)' ], 'find longest common substring example').
cli_testcase(1529, [csp_test], ['../prob_examples/public_examples/CSP/simple/verysimple.csp',
 '-csp_assertion', 'MAIN |= LTL: \"G ([b.blue.blue] => X [c.blue.blue])\"',
 '-csp_assertion', 'MAIN |= LTL: \"G ([i.2] => X [j.3])\"',
 '-csp_assertion', 'MAIN |= LTL: \"G ([a.red] => X e(b.red._))\"',
 '-csp_assertion', 'MAIN |= CTL: \"EGAF e(a.red)\"' , '-strict'
 ], 'Check LTL and CTL assertions (all are satisfied by the MAIN process).').
cli_testcase(1530, [csp_test], ['../prob_examples/public_examples/CSP/simple/verysimple.csp',
 '-csp_assertion', 'MAIN |= LTL: \"G ([i.2] => X [j.4])\"', '-expcterr', model_check_fails, '-strict'
 ], 'Check LTL assertion (check should fail).').
cli_testcase(1531, [csp_test], ['../prob_examples/public_examples/CSP/simple/verysimple.csp',
 '-csp_assertion', 'MAIN |= CTL: \"AGAF e(a.red)\"', '-strict'
 ], 'Check CTL assertion (check should not fail).').
cli_testcase(1532, [cbc], ['../prob_examples/public_examples/B/Tickets/FiveWayTraffic/FiveWay_new_mch.eventb',
 '-mc', 1000, '-strict','-cc', '29', '65',  '--hash64', 971801032035978952
 ], 'Check no time-out for setup-constants (relational image for singleton set).').
cli_testcase(1533, [cbc,disprover], [
 '../prob_examples/public_examples/B/Tickets/FiveWayTraffic/TwoWay_mch.pl',
 '../prob_examples/public_examples/B/Tickets/FiveWayTraffic/Traffic_Context2_prob_ctx.pl'
 %, '../prob_examples/public_examples/B/Tickets/FiveWayTraffic/FiveWay_new_prob_mch.pl' % requires higher-timeout
  ],'Check Disprover can prove all POs').
cli_testcase(1534, [cbc], [
   '-evalt', 'card({x,y| { x \\/ y  } <: {{1} \\/ {2}}}) = 9' % { x \/ y  } <: {{1} \/ {2}} from Grieskamp Zeta paper
 ], 'check set comprehenion can be computed without enum warning (union constraint propagation)').
cli_testcase(1535, [cbc], [
    '-evalt', '{x,y| x = {1,2} &  x \\/ y  = {1,2,3} & 1:y } = {({1,2}|->{1,3}),({1,2}|->{1,2,3})}'
   ,'-evalt', 'card({x,y| x<:1..3 & y<:1..3 & ( x \\/ y)   =  (1..3)  }) = 27'
   ,'-evalt', 'card({x,y| { x \\/ y }  <:  {{1} \\/ {2} \\/ {3}}  }) = 27'
   ,'-evalt', '{x,y| x = {1,2} &  x \\/ y  = 1..5 & 1:y } = {({1,2}|->{1,2,3,4,5}),({1,2}|->{1,3,4,5})}'
   ,'-evalt', '{x,y| x = {1,2} &  x \\/ y  = 1..5  } = {({1,2}|->{1,2,3,4,5}),({1,2}|->{1,3,4,5}),({1,2}|->{2,3,4,5}),({1,2}|->{3,4,5})}'
   ,'-evalt', 'card({x,y|   x \\/ y  = 1..5  }) = 243'
 ], 'Check kernel bug in add_element and copy_list_skeleton fixed').
cli_testcase(1536, [cbc], [
     '-evalt', 'sqr = %x.(x>=0|x*x) & (sqr)[{xx}] = {2500}'
   , '-evalt', 'sqr = %x.(x>=0&x<52|x*x) & (sqr)[{xx}] = {2500}'
   , '-evalt', 'sqr = %x.(x>=0|x*x) & (sqr)[{xx,vv}] = {2500}'
 ], 'check no enum warning (image propagation)').
cli_testcase(1537, [ticket,fuzzer,laws], [
     '-evalf', 'x:0..5 & (-2) ** x = (8)'
   , '-evalt', 'x:0..5 & (-2) ** x = -8'
   , '-evalt', '{x|x>0 & (-2) ** x = -8} = {3}'
   , '-evalf', 'x:0..3 & (-1) ** x = (-46)'
   , '-evalf', 'x:-18..-11 & -1 ** x = -46'
   , '-evalt', '{x|x:0..5 & (-1) ** x = -1} = {1,3,5}'
   , '-evalt', '{x|x:0..5 & (-1) ** x = 1} = {0,2,4}'
   , '-evalt', '{x|x:0..5 & (1) ** x = 1} = 0..5'
  % , '-evalf', 'x=2 & x**y <1 & y>0' % check propagation
   , '-evalf', 'x**1 /= x' % check propagation
   , '-evalf', 'x**0 /= 1' % check propagation
   , '-evalf', '1**x /= 1 & x>0' % check propagation
 ], 'check no instantiation error when base is 1 or -1').
cli_testcase(1538, [kodkod], [
     '-evalt', ':kodkod 2+3=x' % check no warning because nothing for kodkod to solve
    ,'-evalt', ':kodkod G = %x.(x:1..n|(x+1) mod n) & A\\/B = dom(G)\\/ran(G) & A/\\B={} & G[A]<:B & G[B]<:A & n=15 & 1:A' % bipartite graph detection
 ], 'check no warning and :kodkod works').
cli_testcase(1539, [laws,tla],['../prob_examples/public_examples/TLC/Laws/NegDivision.tla', '-mc', 1000, '-strict'], 'Test that the division semantics of TLA preserved.').
cli_testcase(1540, [laws,proz,sequences],['../prob_examples/public_examples/Z/ZLive/IntegerTests.fuzz',
 '../prob_examples/public_examples/Z/ZLive/SetTests.fuzz',
 '../prob_examples/public_examples/Z/ZLive/SequenceTests.fuzz',
 '../prob_examples/public_examples/Z/ZLive/RelationTests.fuzz',
 '../prob_examples/public_examples/Z/ZLive/FreeTypeTests.fuzz',
 '../prob_examples/public_examples/Z/ZLive/MiscTests.fuzz',
  '-mc', 1000, '-nodead', '-strict'], 'Test Integer Z laws and that the division semantics of Z is preserved.').
cli_testcase(1541, [cbc], [
 '-evalt', '2 : ran({y|y : INTEGER * INTEGER & #(x).(y = x |-> x + 2 & x  : NATURAL)})',
 '-evalt', '-2 /: ran({y|y : INTEGER * INTEGER & #(x).(y = x |-> x + 2 & x  : NATURAL)})',
 '-evalt', '1 /: ran({y|y : INTEGER * INTEGER & #(x).(y = x |-> x + 2 & x  : NATURAL)})',
 '-evalt', '-10..1 /\\ ran({y|y : INTEGER * INTEGER & #(x).(y = x |-> x + 2 & x  : NATURAL)}) = {}',
 '-evalt', '-10..10 /\\ ran({y|y : INTEGER * INTEGER & #(x).(y = x |-> x + 2 & x  : NATURAL)}) = 2..10',
 '-evalf', '-2 : ran({y|y : INTEGER * INTEGER & #(x).(y = x |-> x + 2 & x  : NATURAL)})',
 '-evalt', '2:ran({x|x:BOOL*BOOL*INTEGER & #(b,z).( x = (b,b,z) & z>1 & z<10)})',
 '-evalt', '22:ran({x|x:BOOL*BOOL*INTEGER & #(b,z).( x = (b,b,z) & z>1)})',
 '-evalt', '1/:ran({x|x:BOOL*BOOL*INTEGER & #(b,z).( x = (b,b,z) & z>1 )})',
 '-evalt', '1 : dom({y|y : INTEGER * INTEGER & #(x).(y = x |-> x + 2 & x  : NATURAL)})',
 '-evalt', '(TRUE,TRUE):dom({x|x:BOOL*BOOL*INTEGER & #(b,z).( x = (b,b,z) & z>1 )})',
 '-evalf', '(TRUE,FALSE):dom({x|x:BOOL*BOOL*INTEGER & #(b,z).( x = (b,b,z) & z>1 )})'
 ],'ensure proper symbolic treatment for domain/range').
cli_testcase(1542, [cbc], [
 '-evalf', '{x|x>100 & x mod 102 = 2} = {}',
 '-evalf', '%x.(x>100 & x mod 102 = 2|x*x) = {}',
 '-evalt', '{x|x>100 & x mod 102 = 2} /\\ {x| x<100 & x mod 102 =2} = {}',
 '-evalt', '%x.(x>100 & x mod 102 = 2|x*x) /= {}',
 '-evalt', '{x|x>100 & x mod 102 = 2} /= {}',
 '-evalt', '{x|x*x < i & x>2} ={} & i>2',
 '-evalt', '{x|x*x < i & x>2} /= {} & i>9',
 '-evalt', '{x|x<20} /= {}',
 '-evalf', '{x|x<20} = {}'
 ], 'symbolic treatment of closure emptiness').
cli_testcase(1543, [tickets],
 ['../prob_examples/public_examples/B/Tickets/Beauger2/bug_not_labels_corrected.mch', '-aa', 0,3,0,
  '-ppf', '../prob_examples/public_examples/B/Tickets/Beauger2/bug_not_labels_corrected_ppf.mch'
  ], 'Check that labels correctly assigned.').
cli_testcase(1544, [b_test,trees],['../prob_examples/public_examples/B/Tester/TreeExample.mch',
 '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE', '-assertions', '-init', '-strict'], 'Test that the Atelier-B tree operators supported.').
cli_testcase(1545, [laws], ['../prob_examples/public_examples/B/Laws/ArithmeticLaws.mch', '-mc', '100000', '-c', '-nodead', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE',
'-p', 'NORMALIZE_AST', 'TRUE'], 'Various arithmetic laws with AST Normalization.').
cli_testcase(1546, [cbc,trees,sequences], [
 '-evalt', '{si|si : dom((%u.(u : seq(INTEGER)|[1] ^ u) ; {[1] |-> 33,[1,2] |-> 55})) & si/=[]} = {[2]}',
 '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '-evalt', 'arity({[]|->22,[1]|->33,[1,1]|->44,[1,2]|->55},[1])=2'
 ], 'ensure proper constraint solving/propagation for sequence concatenation').
cli_testcase(1547, [laws,trees,sequences], ['../prob_examples/public_examples/B/Laws/TreeLaws.mch',
 '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE', '-mc', '100', '-nogoal', '-nodead', '-strict', '-p', 'CLPFD', 'TRUE'], 'Laws about trees').
cli_testcase(1548, [b_test,fairness,ltl], ['../prob_examples/public_examples/B/Ivo/Tickets/Test2.mch', '-ltlformula', 'WEF => not (GF {addr=2} & GF [C])', '-expcterr', ltl, '-strict'], 'Test fairness implementation for finding the right counter example.').
cli_testcase(1549, [b_test,fairness,ltl], ['../prob_examples/public_examples/B/Ivo/Tickets/Test2.mch', '-ltlformula', 'SEF => not (GF {addr=2} & GF [C])', '-strict'], '').
cli_testcase(1550, [cbc], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '-evalt', 'not((y > 0 & y * y > 20) <=> (y * y > 25 & y > 0))',
 '-evalf', 'not((y > 0 & y * y > 20) <=> (y * y > 21 & y > 0))'
 ], 'test constraint solving for <=>').
cli_testcase(1551, [cbc], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '-evalt', '{b|[a,b,c](2)=333} =res & a : res & b:res'
 ], 'test constraint solving and compilation of closures (apply sequence extension)').
cli_testcase(1552, [cbc,trees], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '-evalt', 'a = {(1, {([]|->2)} ), (2, const(1, [a(1)]))}',
 '-evalt', 'a = {1 |-> {[] |-> 2},   2 |-> (dom({pi,ff,p|((pi : seq(INTEGER) & ff : INTEGER) & p : seq(INTEGER)) & (p |-> ff : [a(1)](1) & pi = 1 -> p)}) \\/ {[] |-> 1})}',
 '-evalt', 'a : 1..2 --> tree({1,2}) & a = {(2, bin(2)), (1, const(1, [a(2)]))}'
 ], 'test constraint solving and compilation of closures (function application)').
cli_testcase(1553, [csp_test], [
 '../prob_examples/public_examples/CSP/Tickets/Ivo/RenamedSymbol.csp', '-assertions', '-strict'],
 '').
cli_testcase(1554, [tickets,while],['../prob_examples/public_examples/B/Tickets/Hansen18_While/Loop.mch',
 '../prob_examples/public_examples/B/Tickets/Hansen18_While/LoopBy2.mch',
 '../prob_examples/public_examples/B/Tickets/Hansen18_While/LoopByDouble.mch',
 '../prob_examples/public_examples/B/Tickets/Hansen18_While/LoopByTwiceSeqComp.mch',
 '../prob_examples/public_examples/B/Tickets/Hansen18_While/LoopByClash.mch',
 '-t', '-mc', 100, '-strict'],
 'Check compilation of operation calls inside while loop.').
cli_testcase(1555, [tickets,cbc,slot,trees],[
  '../prob_examples/public_examples/B/Tickets/Schneider2_Trees/NewSolver_v2.mch',
  '../prob_examples/public_examples/B/Tickets/Schneider2_Trees/NewSolver_v3.mch',
 '-evalt', 'CHOOSE_MODULES("bk-phi-H-2013", solution)'], 'Test constraint solving with trees').
cli_testcase(1556, [cbc],[ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
  '../prob_examples/public_examples/EventBPrologPackages/Laws/Mengen_Gesetze_PowBool_ctx.eventb',
  '../prob_examples/public_examples/EventBPrologPackages/Laws/Mengen_Gesetze_FiniteInt_ctx.eventb', '-v',
 '-cbc_assertion', 'Gesetze:assoc1', '-cbc_assertion', 'Gesetze:assoc2',
 '-cbc_assertion', 'Gesetze:comm1', '-cbc_assertion', 'Gesetze:comm2',
 '-cbc_assertion', 'Gesetze:DeMorgan1', '-cbc_assertion', 'Gesetze:DeMorgan2',
 '-p', 'PROOF_INFO', 'FALSE'], 'Test cbc assertion with label works').
cli_testcase(1557, [laws],[ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
  '../prob_examples/public_examples/EventBPrologPackages/Laws/Mengen_Gesetze_Bool_ctx.eventb',
  '../prob_examples/public_examples/EventBPrologPackages/Laws/KuratowskiPairDefinition_Enum_ctx.eventb',
  %'../prob_examples/public_examples/EventBPrologPackages/Laws/KuratowskiPairDefinition_Int_ctx.eventb',
  %'../prob_examples/public_examples/EventBPrologPackages/Laws/Mengen_Gesetze_PowBool_ctx.eventb', % tales 24 seconds
 '-cbc_assertions', '-p', 'PROOF_INFO', 'FALSE'], 'Test laws over sets using cbc').
cli_testcase(1558, [cbc],[ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '-evalt', 'f: 1..40 --> BOOL & card({x|x|->TRUE:((1..40)*BOOL) & x|->TRUE :f}) : 10..11', % did not work before
 '-evalt', 'f: 1..40 --> BOOL & card({x|x:1..40 & x|->TRUE :f}) : 10..11'
 ], 'Test reification of closure for cardinality').
cli_testcase(1559, [cbc,tickets],[ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '../prob_examples/public_examples/EventBPrologPackages/Tickets/EinsteinPuzzle/UnsatCore_einstein_ctx_v2.mch', '-init',
 '-evalt', 't: Menschen >-> Getraenke & Daene |-> Tee : t',
 '-evalt', 't: Menschen -->> Getraenke & Daene |-> Tee : t',
 '-evalt', 't: Menschen --> Getraenke & Daene |-> Tee : t',
 '-evalt', 't: Menschen >->> Getraenke & Daene |-> Tee : t'
 ], 'Test reification of closure for cardinality').
cli_testcase(1560, [cbc], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '-evalf', 'n>=0 & n1>0 & {n1} \\/ {n} = r & r={n} & not(n1=n)',
 '-evalf', '{n} /<: {n,n1} & n:INTEGER',
 '-evalf', '0..B <: 1..B & B:1..1000000',
 '-evalf', '0..x <: 1..1000 & x>0',
 '-evalt', '{x|x>0} <: NATURAL1',
 '-evalf', 'x:seq1(0..1) & x:seq1(2..3)',
 '-evalt', 'a..b = {x} & x:100..1002',
 '-evalt', 'x <: (1..100000000 \\/ {-1}) & x={}',
 '-evalt', 'card(iseq1(1 .. 3) \\/ {{}}) = 16',
 '-evalt', 'bool(!x.(x<:1..10 => card(x) <11))',
 '-evalt', 'r:INTEGER<->INTEGER & dom(r)=1..4',
 '-evalt', 'r:INTEGER<->INTEGER & ran(r)=2..20 & dom(r) = 1..10',
 '-evalf', '{x|x>20} = {1,y,5,x,4}',
 '-evalf', '{x|x>20} = y & 1:y',
 '-evalt', '{x|x>20} = y & 21:y',
 '-evalf', '2..b = {1,y,5,x,4}',
 '-evalt', '{X| X /\\ {1} = {} & X \\/ {1} = 1..20} = {2..20}',
 '-evalt', 'a..b = {x} & x:100..1002',
 '-eval', '{x|[1,2,3,4,5,6] |>> {x} = [1,2,3,4,5]}',
 '-eval', '{x|[33] |>> {x} = []}',
 '-evalt', 'all = [S,E,N,D, M,O,R, Y] & {S,E,N,D, M,O,R, Y} <: 0..9 & S >0 & M >0 &   card(ran(all)) = size(all) &   S*1000 + E*100 + N*10 + D +   M*1000 + O*100 + R*10 + E = M*10000 + O*1000 + N*100 + E*10 + Y'
 ], 'test a few constraints from the ProB-TO-DO-List').
cli_testcase(1561, [cbc], [
 '-evalf', 'not(#x.(x>100))',
 '-evalt', '1..b = {1,y,5,x,4}',
 '-evalt', 'a..b = {1,y,5,x,4}'
 ], 'test a few constraints from the ProB-TO-DO-List (with enum warning)').
cli_testcase(1562, [slot,trees,parser],[
  '../prob_examples/public_examples/B/Tickets/Schneider5_Trees/Solver.mch',
  '-p', 'TIME_OUT', 35000,
 '-evalt', '%(x).(x : dom(course_module_combinations) & card(course_module_combinations(x)) > 0|1)={("bk-phi-H-2013"|->1)}',
 '-evalt', '%(x).(x : dom(course_module_combinations) & course_module_combinations(x) /= {} |1)={("bk-phi-H-2013"|->1)}'
 % '-evalt', '%(x).(x : dom(course_module_combinations) & card(course_module_combinations(x)) > 1|1)={("bk-phi-H-2013"|->1)}' % still causes problem
 ], 'Test no complicated set-unification with backtracking').
cli_testcase(1563, [parser,tickets], [
   '../prob_examples/public_examples/B/Tickets/CommentsOneLine/OneLineComments.mch', '-t'
   ], 'Test parser can deal with one line comments').
cli_testcase(1564, [kodkod], [
   '-evalf', 'x:0..3 & y:0..3 & (x*y)+7<0',
   '-evalf', ':kodkod x:0..3 & y:0..3 & (x*y)+7<0'
   ], 'Test Kodkod bug fixed').
cli_testcase(1565, [fairness], [
    '../prob_examples/public_examples/B/LTL/Fairness/RecursiveSCCs2.mch', '-ltlformula', 'SF(E) & WF(C) => F {addr = 4 or addr = 5}', '-expcterr', 'ltl', '-strict'], 'Test finding the right fairness loop.').
cli_testcase(1566, [fairness], [
    '../prob_examples/public_examples/B/LTL/Fairness/RecursiveSCCs2.mch', '-ltlformula', 'SF(C) & SF(E) & SF(G) => F {addr = 4 or addr = 5}', '-expcterr', 'ltl', '-strict'], 'Test finding the right fairness loop.').
cli_testcase(1567, [fairness], [
    '../prob_examples/public_examples/B/LTL/Fairness/RecursiveSCCs2.mch', '-ltlformula', 'SEF => F {addr = 4 or addr = 5}', '-strict'], 'Testing the fairness algorithm.').
cli_testcase(1568, [fairness], [
    '../prob_examples/public_examples/B/LTL/Fairness/RecursiveSCCs3.mch', '-ltlformula', 'SF(E) & SF(D) => F { addr = 8}', '-expcterr', 'ltl', '-strict'], 'Test finding the right fairness loop.').
cli_testcase(1569, [fairness], [
    '../prob_examples/public_examples/B/LTL/Fairness/RecursiveSCCs3.mch', '-ltlformula', 'SF(B) & SF(G) & SF(E) & SF(D) => F { addr = 8}', '-expcterr', 'ltl', '-strict'], 'Test finding the right fairness loop.').
cli_testcase(1570, [fairness], [
    '../prob_examples/public_examples/B/LTL/Fairness/RecursiveSCCs3.mch', '-ltlformula', 'WEF => F { addr = 8}', '-expcterr', 'ltl', '-strict'], 'Test finding the right fairness loop.').
cli_testcase(1571, [fairness], [
    '../prob_examples/public_examples/B/LTL/Fairness/RecursiveSCCs3.mch', '-ltlformula', 'SEF => F { addr = 8}', '-strict'], 'Testing the fairness checking algorithm.').
cli_testcase(1572, [fairness], [
    '../prob_examples/public_examples/CSP/LTL/Fairness/bully.csp', '-csp_assertion', 'Network |= LTL: \"WEF => G ([fail.2] => F [coordinator.1.2])\"', '-expcterr', 'model_check_fails', '-strict'], 'Test finding the right fairness loop.').
cli_testcase(1573, [fairness], [
    '../prob_examples/public_examples/CSP/LTL/Fairness/bully.csp', '-csp_assertion', 'Network |= LTL: \"SEF => G ([fail.2] => F [coordinator.1.2])\"', '-expcterr', 'model_check_fails', '-strict'], 'Test finding the right fairness loop.').
cli_testcase(1574, [chr], ['-p', 'CHR', 'TRUE', '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
   '-evalt', '{x,y,s|x..y <: 11..18  & card(x..y)=s & x=y} = {((11|->11)|->1),((12|->12)|->1),((13|->13)|->1),((14|->14)|->1),((15|->15)|->1),((16|->16)|->1),((17|->17)|->1),((18|->18)|->1)}',
   '-evalt', '{x,y|x..y <: 11..18  & card(x..y)=5 & x<y} = {(11|->15),(12|->16),(13|->17),(14|->18)}'
   ], 'Test CHR interval propagation').
cli_testcase(1575, [external, b_test], [
    '../prob_examples/public_examples/B/ExternalFunctions/CSV_Reader.mch',
    '../prob_examples/public_examples/B/ExternalFunctions/CSV/ChemicalElements.mch', '-t'], 'test CSV reader.').
cli_testcase(1576, [chr], ['-p', 'CHR', 'TRUE',
   '-evalt', 'x+y=100 & x>y',
   '-evalt', 'x=100-y & x>y'
   ], 'Test stronger CHR IDL propagation').
cli_testcase(1577, [cbc], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
   '-evalt', 'x = NATURAL1 \\/ -3..-1  & 11:x & -3:x & 0 /: x',
   '-evalt', '-3..-1 \\/ NATURAL1 = x  & 11:x & -3:x & 0 /: x',
   '-evalt', 'x = NATURAL1 \\/ {-1}  & 11:x & -1:x & 0 /: x',
   '-evalt', 'x = NATURAL1 \\/ {-1,-3}  & 11:x & -1:x & -3:x & 0 /: x',
   '-evalt', 'x = {-1,-33} \\/ NATURAL1  & 11:x & -1:x & -33:x & 0 /: x'
   ], 'Test better symbolic treatment of union').
cli_testcase(1578, [cbc,bmc], ['../prob_examples/public_examples/B/CBC/BinarySearchEvents_Overflow_v2.mch', '-bmc', 5,
'-expcterr', 'invariant_violation'], 'Check error found using bmc').
cli_testcase(1579, [cbc,bmc], ['../prob_examples/public_examples/B/CBC/BinarySearchEvents_Overflow_v3.mch', '-bmc', 10,
'-expcterr', 'invariant_violation'], 'Check error found using bmc').
cli_testcase(1580, [symbolic_model_checker], ['../prob_examples/public_examples/B/CBC/BinarySearchEvents_Overflow_v2.mch', '-symbolic_model_check', 'bmc',
'-expcterr', 'invariant_violation'], 'Check error found using BMC').
cli_testcase(1581, [symbolic_model_checker], ['../prob_examples/public_examples/B/CBC/BinarySearchEvents_Overflow_v2.mch', '-symbolic_model_check', 'kinduction',
'-expcterr', 'invariant_violation'], 'Check error found using k-Induction').
cli_testcase(1582, [symbolic_model_checker], ['../prob_examples/public_examples/B/CBC/BinarySearchEvents_Overflow.mch', '-symbolic_model_check', 'ic3',
'-expcterr', 'invariant_violation'], 'Check error found using IC3').
cli_testcase(1583, [symbolic_model_checker], ['../prob_examples/public_examples/B/CBC/BinarySearchEvents_Overflow.mch', '-symbolic_model_check', 'ctigar',
'-expcterr', 'invariant_violation'], 'Check error found using CTIGAR').
cli_testcase(1584, [symbolic_model_checker], ['../prob_examples/public_examples/B/CBC/BinarySearchEvents_Overflow.mch', '-symbolic_model_check', 'bmc',
'-expcterr', 'invariant_violation'], 'Check error found using BMC').
cli_testcase(1585, [symbolic_model_checker], ['../prob_examples/public_examples/B/CBC/BinarySearchEvents_Overflow.mch', '-symbolic_model_check', 'kinduction',
'-expcterr', 'invariant_violation'], 'Check error found using k-Induction').
cli_testcase(1586, [cbc,wd,error_checks], ['-p', 'TRY_FIND_ABORT', 'TRUE',
   '-eval', 'r = {1|->2, 2|-> 1/0}(1)',
   '-expcterr', 'well_definedness_error'
   ], 'Test better symbolic treatment of union').
cli_testcase(1587, [cbc], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
   '-evalt', 'f: 1..100 --> 10..20 & x:10..20 & f2=f & x2=x+1 & not(f2:1..100 --> 10..20 & x2:11..20)',
   '-evalf', 'f: 1..100 --> 10..20 & x:10..20 & f2=f & x2=x+1 & not(f2:1..100 --> 10..20 & x2:11..21)'
   ], 'Test inlining of id1=id2 equalities and subsequent detection of identical predicates').
cli_testcase(1588, [cbc], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
   '-evalt', 'f: 1..100 --> 10..20 & x:10..20 & f2=f & x2=x+1 & not(f2:1..100 --> 10..20 & x2:11..20)',
   '-evalf', 'f: 1..100 --> 10..20 & x:10..20 & f2=f & x2=x+1 & not(f2:1..100 --> 10..20 & x2:11..21)'
   ], 'Test inlining of id1=id2 equalities and subsequent detection of identical predicates').
cli_testcase(1589, [cbc], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
   '../prob_examples/public_examples/B/Tickets/ForallDomainExpansion/UnsatCore.mch',
   '-init'
   ],'Test that forall with domain is expanded').
cli_testcase(1590, [cbc,bmc], [
   '../prob_examples/public_examples/B/Tickets/ForallDomainExpansion/binarySearchFail_impl_prob_mch.eventb',
   '-bmc', 10, '-expcterr', 'invariant_violation'
   ],'Test that forall with domain is expanded and BMC finds problem').

cli_testcase(1591, [tla,tlc],['../prob_examples/public_examples/B/SymmetryReduction/Peterson.mch', '-mc_with_tlc'], 'TLC regression test').
cli_testcase(1592, [tla,tlc],['../prob_examples/public_examples/B/SymmetryReduction/Peterson_err.mch', '-mc_with_tlc'
%, '-expcterr', 'invariant_violation' % currently there is no feedback to probcli about invariant violations found
], 'TLC regression test').
cli_testcase(1593, [puzzles],['../prob_examples/public_examples/B/Puzzles/Bridges.mch', '-init'
], 'Bridges Puzzle').
cli_testcase(1594, [cbc,sigma], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
   '-evalt', 'SIGMA(x).(x:M|30) = 3000 & M<:1..100',
   '-evalt', '{M|SIGMA(x).(x:M|30) = 300 & M<:1..10} = {1..10}',
   '-evalt', '{M|SIGMA(x).(x:M|30) = 270 & M<:1..10}=res & card(res)=10',
   % this is slow: {M|SIGMA(x).(x:M|30) = 570 & M<:1..20}=res & card(res)=20; 17 seconds
   % this does not work: {M|SIGMA(x).(x:M|30) = 3000 & M<:1..100}
   '-evalt', 'SIGMA(x).(x:M|cp(x)) = 3000 & M=1..100 & cp:M --> {0,15,30}',
   '-evalt', 'M=1..20 & {cp|SIGMA(x).(x:M|cp(x)) = 570 & cp:M --> {0,30}}=res & card(res)=20',
   '-evalt', 'M=1..100 & {cp|SIGMA(x).(x:M|cp(x)) = 3000 & cp:M --> {0,15,30}}=res & card(res)=1',
   '-evalt', 'PI(x).(x:{-1} \\/ 1..20|x) = -2432902008176640000',
   '-evalt', 'PI(x).(x:M|20) = 400 & M<:1..10',
   '-evalf', 'SIGMA(x).(x:M|20) = 50 & M<:1..10',
   '-evalf', 'PI(x).(x:M|20) = 500 & M<:1..10'
   ], 'Test SIGMA and PI').
cli_testcase(1595, [tickets,eventb_test],['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
'../prob_examples/public_examples/EventBPrologPackages/Tickets/TheoremInGuards/TestM1_mch.eventb',
'../prob_examples/public_examples/EventBPrologPackages/Tickets/TheoremInGuards/VM_1_mch.eventb',
'-t', '-mc', 100 ], 'Test theorem in guards can use dropped abstract variables').
cli_testcase(1596, [tickets,eventb_test],['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
'../prob_examples/public_examples/EventBPrologPackages/Tickets/TheoremInGuards/TestM1_error_mch.eventb',
'-mc', 100, '-expcterr', 'event_error:teste:invalid_theorem_in_guard'
 ], 'Test theorem in guards can use dropped abstract variables and error found').
cli_testcase(1597, [tickets,tla],['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
'../prob_examples/public_examples/B/Tickets/ExistsEnumeration/MC_AlternatingBit_Internal.mch',
'-t', '-mc', 500 ], 'Test no problem with LoseAck action and existential quantifier').
cli_testcase(1598, [tickets,tla],['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
'../prob_examples/public_examples/TLA/TLA_Distribution/FIFO/MCInnerFIFO.mch', '-scope', 'card(q)<=qLen',
 '-p', 'MAX_INITIALISATIONS', 40,
 '-mc', 6000, '-cc', 5810, 9661 ,  '--hash64', 523977694511884201], 'Test SCOPE works correctly').
cli_testcase(1599, [b_test], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '-expcterr', 'bmachine_static_checks',
'../prob_examples/public_examples/B/Tester/ExistentialGlobalSetIDTest.mch', '-mc', '1000',
 '-evalt', '#(ID).( cc /: ID & dom({x,y|x=1 & y:ID}) = {})'
 ], 'Check that local identifier ID does not confuse optimiser').
cli_testcase(1600, [tickets,private,quantified], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '../prob_examples/examples/B/Tickets/MemoryError_bugly.mch', '-init', '-assertions'
 ], 'Check that UNION and dom(dom()) memory consumption issue solved').
cli_testcase(1601, [b_test, implementation, refinement], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
'../prob_examples/examples/B/ClearSy/turbomeca/pld_i.imp','-t'
 ], 'Check implementation machine').
cli_testcase(1602, [b_test], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '../prob_examples/public_examples/B/SchneiderBook/Chapter17/Array.mch','-t'
 ], 'Check Schneider example').
cli_testcase(1603, [alstom,private], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '../prob_examples/examples/B/Alstom/exemple7.mch',
 '../prob_examples/examples/B/Alstom/exemple7_occuper_zone.mch',
 '-t'
 ], 'Check Alstom machines').
cli_testcase(1604, [eventb_test,private], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '../prob_examples/examples/RodinModels/PaulSimon/ca_m01_mch.eventb',
 '-p', 'MAX_INITIALISATIONS', '30', '-p', 'MAX_OPERATIONS', '30',
 '-p', 'DEFAULT_SETSIZE', '3',
 '-t'
 ], 'Check Paul Simon complicated EventB machine').
cli_testcase(1605, [cbc, infinite], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '-evalt', 'f = %x.(x:INTEGER|0) & f : INTEGER --> {0}',
 '-evalt', 'f = %x.(x:INTEGER|0) & f : INTEGER +-> {0}',
 '-evalt', 'f = %x.(x:INTEGER|0) & f : INTEGER +-> {u}',
 '-evalt', 'f = %x.(x:INTEGER|0) & {u|f : INTEGER +-> {u}} = {0}',
 '-evalt', 'f = %x.(x:INTEGER|0) & {u|f : INTEGER --> {u}} = {0}',
 '-evalt', 'f = %x.(x:INTEGER|0) & f : INTEGER +-> NATURAL',
 '-evalt', 'f = %x.(x:INTEGER|0) & f : INTEGER --> NATURAL',
 '-evalt', 'f = %x.(x:INTEGER|0) & f : INTEGER +-> 0..2',
 '-evalf', 'f = %x.(x:INTEGER|0) & f : INTEGER +-> 1..2',
 '-evalf', 'f = %x.(x:INTEGER|0) & f : INTEGER --> 1..2',
 '-evalt', 'f = %x.(x:INTEGER|1000) & f : INTEGER +-> 999..1002',
 '-evalt', 'f = %x.(x:INTEGER|1000) & f : INTEGER --> 999..1002',
 '-evalt', 'f = INTEGER*{1000} & f : INTEGER --> 999..1002',
 '-evalt', 'f = %x.(x:INTEGER|1000) & f = INTEGER*{1000}',
 '-evalf', 'f = %x.(x:INTEGER|1000) & f = INTEGER*{1001}',
 '-evalf', 'f = %x.(x:INTEGER|0) & f : INTEGER +-> NATURAL1',
 '-evalt', 'f = %x.(x:INTEGER|0) & f : INTEGER -->> {0}',
 '-evalf', 'f = %x.(x:INTEGER|0) & f : INTEGER -->> {0,1}',
 '-evalt', 'f = %x.(x/=0|1000/x) & f : INTEGER +-> INTEGER & f(500)=2',
 '-evalf', 'f = %x.(x/=0|1000/x) & f : INTEGER --> INTEGER',
 '-evalt', '{x|x/=0} = INTEGER \\ {0}',
 '-evalf', '{x|x/=0} = INTEGER \\ {1}',
 '-evalf', '{x|x/=0} /= INTEGER \\ {0}',
 '-evalt', '{x|x/=0} /= INTEGER \\ {1}',
 '-evalf', '{x|x/=0} = INTEGER',
 '-evalt', '{x|x/=0} /= INTEGER',
 '-evalf', 'INTEGER={x|x/=0}',
 '-evalt', 'INTEGER/={x|x/=0}',
 '-evalt', '{b|INTEGER={x|x/=0} <=> b=TRUE}={FALSE}',
 '-evalt', 'NATURAL/={x|x/=0}',
 '-evalt', 'NATURAL1/={x|x/=0}',
 '-evalt', '{x|x/=0} /= NATURAL1',
 '-evalf', '{x|x/=0} = NATURAL'
 ], 'Check that %x.(x:P|E) is detected as cartesian product closure for dom_range_for_specific_closure').
 cli_testcase(1606, [cbc], [
   '-evalt', 'n=10000000 & x : (1..n) \\ {5000000}',
   '-evalt', 'x:(1000..a) \\ {b}',
   '-evalt', 'x: 1000..b /\\ 1500..c',
   '-evalt', 'x:1000..2000 & x /: b..2000',
   '-evalt', 'x:(1000..2000) \\ (b..2000)',
   '-evalt', 'x:(1000..a) \\ (b..a)',
   '-evalt', '{x,y| x:200..205 & y : 101..190 & x /: y..y+100} = {(202|->101),(203|->101),(203|->102),(204|->101),(204|->102),(204|->103),(205|->101),(205|->102),(205|->103),(205|->104)}',
   '-evalt', 'card({x,y| x:200..295 & y : 101..190 & x /: y..y+100})=4455'
   ], 'Test symbolic treatment of set-difference membership').
cli_testcase(1607, [tickets], ['../prob_examples/public_examples/B/Tester/TestAnyChoice.mch',
 '-t', '-strict'], 'Test that definition can be used at top-level of forall quantifier').
cli_testcase(1608, [symmtest,proz], ['../prob_examples/public_examples/Z/Daniel/Derrick_ICFEM06.fuzz', '-strict',
'-mc', '10000', '-cc', '29', '149' ], 'Test min card inference works.').
cli_testcase(1609, [cbc,tickets], [ '-p', 'MAXINT', 2147483648,
   '-evalt', 'x : NAT1 +-> {1,2,3} & (n:NAT & 1..n = dom(x)) & x /= {}'
   ], 'Test that pf does not expand NAT1').
cli_testcase(1610, [cbc], [  '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
   '-evalt', 'f=[{1},{3},{6,2,4,8},{1,2,3}] & x:1..4 & y=f(x) & card(y)>3',
   '-evalt', '{f,x,y|f=[{1},{3},{6,2,4,8},{1,2,3}] & x:1..4 & y=f(x) & card(y)>3} = {(([{1},{3},{2,4,6,8},{1,2,3}]|->3)|->{2,4,6,8})}'
   ], 'Test cardinality restriction').
cli_testcase(1611, [b_test,sequences], ['-evalt', 'x:seq({1}) & 55:dom(x)', '-evalf', 'x:seq({1}) & 55:dom(x) & card(x)=10', '-evalt', 'x:seq({1}) & 5:dom(x) & card(x)>6', '-p', 'CLPFD', 'TRUE', '-strict'
%, '-expcterr', eval_string_enum_warning
], 'Ensure proper enumeration of sequences (CLPFD=TRUE).'). % variation of test 1187
cli_testcase(1612, [eventb_test, theories], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE', '-assertions',
'../prob_examples/public_examples/EventBPrologPackages/Theory/Sequences/TestConcat_ctx.eventb',
'../prob_examples/public_examples/EventBPrologPackages/Theory/Sequences/TestMySequences_ctx_ok.eventb',
'../prob_examples/public_examples/EventBPrologPackages/Theory/Sequences/TestMySequences_ctx_ok2.eventb'], 'Check Sequence Theory Mapping.').
cli_testcase(1613, [eventb_test, theories], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
'-assertions', '-expecterr', 'check_assertions',
'../prob_examples/public_examples/EventBPrologPackages/Theory/Sequences/TestMySequences_ctx_err.eventb'],
'Check Sequence Theory Mapping leads to error (due to wrong mapping).').
cli_testcase(1614, [eventb_test, theories], [  '-assertions',
'../prob_examples/public_examples/EventBPrologPackages/Theory/Sequences/MySequences_ctx.eventb'], 'Check Sequence Theory in context.').
cli_testcase(1615, [cbc, cbc_tests,bmc], ['../prob_examples/public_examples/B/SymbolicModelChecking/Counters/CounterArticle.mch', '-bmc', 4,
'-expcterr', 'invariant_violation'], 'Check BMC works').
cli_testcase(1616, [symbolic_model_checker], ['../prob_examples/public_examples/B/SymbolicModelChecking/Counters/CounterArticle64_ok.mch', '-symbolic_model_check', 'kinduction'], 'Check kinduction works for this example').
cli_testcase(1617, [bmc,tickets], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '../prob_examples/public_examples/EventBPrologPackages/Tickets/bmc_unsat_axioms/f_m0.eventb',
 '../prob_examples/public_examples/EventBPrologPackages/Tickets/bmc_unsat_axioms/f_m1.eventb',
'-bmc', 10], 'Check no counter example reported').
cli_testcase(1618, [cbc], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '../prob_examples/public_examples/B/Mathematical/ArgumentationAsSets.mch',
 '../prob_examples/public_examples/EventBPrologPackages/Mathematical/Arguments_Example_ctx.eventb',
  '-assertions'], 'Check Argumentation Theory example').
cli_testcase(1619, [b_test, infinite,total_function], ['-evalt', 'x mod 100000 >= 99999',
 '-evalt', 'x : NATURAL --> {1,2,3}',
 '-evalt', 'x : NATURAL --> {1,2,3} & x(7) = 3',
 '-evalf', 'x : NATURAL --> {1,2,3} & x(7) = 3 & x(7) = 2',
 '-evalt', 'not(x : INTEGER --> {1,2})'
], 'Ensure symbolic treatment of (infinite) total functions.').
cli_testcase(1620, [b_test,external,choose], ['../prob_examples/public_examples/B/ExternalFunctions/ChooseExternal.mch',  '-assertions'], 'Test choose operator.').
cli_testcase(1621, [b_test,let], [
 '../prob_examples/public_examples/B/Tester/LetTest_IllegalCyclic.mch',
 '../prob_examples/public_examples/B/Tester/LetTest_IllegalNondet.mch',
 '../prob_examples/public_examples/B/Tester/LetTest_IllegalNotAllDefined.mch',
 '../prob_examples/public_examples/B/Tester/LetTest_IllegalNotVarLHS.mch',
 '../prob_examples/public_examples/B/Tester/LetTest_IllegalVar.mch',
  '-expcterr','type_error', '-expcterr', 'load_main_file'], 'Check illegal LETs detected.').
cli_testcase(1622, [b_test,recursion], ['../prob_examples/public_examples/B/FunctionalProgramming/Factorial_NewSyntax.mch',
 '../prob_examples/public_examples/B/FunctionalProgramming/SeqNaturalFoldr_NewSyntax.mch',
 '../prob_examples/public_examples/B/RecursiveFunctions/SortSet_NewSyntax.mch',
 '-t', '-strict'], 'Test new IF-THEN-ELSE / LET Syntax for expressions.').
cli_testcase(1623, [b_test], [
 '-evalf', 'z < first([z]) & 0<z',
 '-evalf', 'z < first([z])',
 '-evalf', 'z < last([z]) & 0<z',
 %'-evalf', 'z < first([z,z]) & 0<z', % currently requires DISPROVER_MODE
 %'-evalf', 'z < first(v) & v=[z] & 0<z',
 '-evalf', 'z > first([z]) & z>0'
], 'Ensure ast_cleanup removes first/last.').
cli_testcase(1624, [cbc], [
 '-evalt', 'x:1..n --> BOOL & card({y|y:1..n & y|->FALSE:x})=cf & cf:15..26 & card({y|y:1..n & y|->TRUE:x})=ct &ct:22..25 & ct+cf=n &  x(n/2)=FALSE & x(n/4)=FALSE & n=40',
 '-evalt', 'x:1..n --> BOOL & card({y|y:1..n & y|->FALSE:x})=cf & cf:15..26 & card({y|y:1..n & y|->TRUE:x})=ct &ct:22..25 & x(n/2)=FALSE & x(n/4)=FALSE & n=40',
 '-evalt', 'x:1..n --> BOOL & card({y|y:1..n & y|->FALSE:x})=cf & cf:15..26 & card({y|y:1..n & y|->TRUE:x})=ct &ct:22..25 & ct+cf=n &  x(n/2)=TRUE & x(n/4)=TRUE & n=40',
  '-evalt', 'x:1..n --> BOOL & card({y|y:1..n & y|->FALSE:x})=cf & cf:5..16 & card({y|y:1..n & y|->TRUE:x})=ct &ct:27..29 & ct+cf=n &  x(n/2)=FALSE & x(n/4)=FALSE & n=40 & x(3*n/4)=TRUE',
 '-evalf', 'x:1..n --> BOOL & card({y|y:1..n & y|->FALSE:x})=cf & cf:15..19 & card({y|y:1..n & y|->TRUE:x})=ct &ct:15..20 & ct+cf=n &  x(n/2)=FALSE & x(n/4)=FALSE & n=40',
 % x:1..n --> BOOL & card({y|y:1..n & y|->FALSE:x})=cf & cf:15..19 & card({y|y:1..n & y|->TRUE:x})=ct &ct:15..20 &  x(n/2)=FALSE & x(n/4)=FALSE & n=40  % this is not yet detected as failing quickly
 '-evalt', 'x:1..n --> BOOL & card(x~[{FALSE}])=cf & cf:15..26 & card(x~[{TRUE}])=ct & ct:22..25 &  x(n/2)=FALSE & x(n/4)=FALSE & ct+cf=n & n=40',
 % '-evalt', 'x:1..n --> BOOL & card(x~[{FALSE}]):15..26 & card(x~[{TRUE}]):22..25 &  x(n/2)=FALSE & x(n/4)=FALSE & n=40' does not work yet
 '-evalt', 's=1..20 & x <: s & card(x)=cx & cx:3..4 & y <: s & card(y)=cy & cy:3..4 & x/\\y={} & cx+cy=7 & (9:x or 9:y) & (18:y or 10:x)', % this test does not really exercise reification yet
 '-evalt', 's=1..20 & x: s-->BOOL & card({t|t:1..20 & t|->TRUE:x})=cx & cx:3..4 & y:s-->BOOL & card({t|t:1..20 & t|->TRUE:y})=cy & cy:3..4 & x~[{TRUE}]/\\y~[{TRUE}]={} & cx+cy=7 & (9|->TRUE:x or 9|->TRUE:y) & (18|->TRUE:y or 10|->TRUE:x)', % THIS ONE IS FASTER: reason t:1..20  TO DO: try and get rid of need for this !!
 '-evalt', 's=1..n & x<:s & card(x):15..26 & n/2/:x & n/4/:x & n=40' % another test; not requiring reification
], 'Test cardinality reification propagates through check_finite_card.').
cli_testcase(1625, [cbc], [
 '-evalt', 'i=2..x & card(i):10..9122110 & x > 9121000'], 'Check card of interval propagates').
cli_testcase(1626, [cbc,records], [
 '-evalt', '{v|v:0..5 & card({x,y|x|->y:{(2,2),(2,4)} & x+y>v}):2..3}=0..3',
 '-evalt', '{c,v,z|card({x|rec(a:x):v & x:(z..(z+3))})=c & v={rec(a:2),rec(a:4)} & c>1} = {((2|->{rec(a:2),rec(a:4)})|->1),((2|->{rec(a:2),rec(a:4)})|->2)}'
 ], 'A few card tests; test reification involving records').
cli_testcase(1627, [cbc], [
 '-evalt', 's=1..20 & x: s-->BOOL & card({t|t|->TRUE:x})=cx & cx:3..4 & y:s-->BOOL & card({t|t|->TRUE:y})=cy & cy:3..4 & x~[{TRUE}]/\\y~[{TRUE}]={} & cx+cy=7 & (9|->TRUE:x or 9|->TRUE:y) & (18|->TRUE:y or 10|->TRUE:x)', % this counter part was much slower !  due to missing t:1..20 constraint; now solved
 '-evalt', 'x:1..n --> BOOL & card({y|y|->FALSE:x}):15..26 & card({y|y|->TRUE:x}):12..25 & x(n/2)=FALSE & x(n/4)=FALSE & n=40',
 '-evalt', 's=1..20 & x: s-->BOOL & card({t|t|->TRUE:x}):18..19',
 '-evalt', 's=1..20 & x: s-->BOOL & card({t|t|->FALSE:x}):18..19',
 '-evalt', 's=1..130 & x: s-->BOOL & card({t|t|->TRUE:x}):128..129',
 '-evalt', 's=1..258 & x: s-->BOOL & card({t|t|->FALSE:x})=card({t|t|->TRUE:x})',
 '-evalt', 's:1..n --> 1..n & card({x|x|->x:s})=n & n=50',
 '-evalt', 'card({s,n|s:1..n --> 1..n & card({x|x|->x:s})=n & n=50})=1', % non-linear pattern
  '-evalt', 's: 1..20 --> (BOOL*(1..20)) & card({x|x|->(TRUE|->x):s})=10 & card({x|x|->(FALSE|->x):s})=10',
  '-evalt', 's: 1..20 --> (BOOL*(1..20)) & card({x|x|->(TRUE|->x):s})=10 & card({x|x|->(FALSE|->x+1):s})=10', % non-linear pattern with addition
  '-evalt', 's: 1..20 --> (BOOL*(1..20)) & card({x|x|->(FALSE|->x):s})=10 & card({x|x|->(TRUE|->x-1):s})=10' % non-linear pattern with minus
 ], 'Check card propagates without need for t:1..20 constraint').
cli_testcase(1628, [cbc,tickets], [
  '-evalt', 'card({f|f : 1..20 --> POW(7..11) & !x.(x:1..20 => card(f(x)) = 4 & {7,8,9} <: f(x)) &  !x.(x:1..10 => f(x) /= f(x+10)) & !x.(x:1..9 => f(10+x) /= f(1+x)) & 10: f(1) })=1',
 % {f|f : 1..20 --> POW(7..11) & !x.(x:1..20 => card(f(x)) = 4 & {7,8,9} <: f(x)) &  !x.(x:1..10 => f(x) /= f(x+10)) & !x.(x:1..9 => f(10+x) /= f(1+x)) & 10: f(1) & printf("~nf(1)=~w~n",f(1)) & !i.(i:2..20 => printf("f(i)=~w~n",(i,f(i)))) }
  '-evalt', '1=card({f|f : 1..100 --> POW(7..11) & !x.(x:1..100 => card(f(x)) = 4 & {7,8,9} <: f(x)) &  !x.(x:1..90 => f(x) /= f(x+10)) & !x.(x:1..9 => f(90+x) /= f(1+x)) & 10: f(1)})'
  ], 'A variation of test 34; computing all solutions').
cli_testcase(1629, [cbc,queens],
['../prob_examples/public_examples/B/Tickets/CardReify1/MaxQueensNoCheck_v5.mch',
 '-t', '-mc', 10], 'Check reification of card works with dom of total function').
cli_testcase(1630, [cbc], [
 '-evalt', 'f:1..20 >->> 1..20 & card({i|i:1..20 & f~(i)=i})=4', %this also worked before
 '-evalt', 'f:1..20 >->> 1..20 & card({i|i:ran(f) & f~(i)=i})=4'
 ], 'Check reification of card with range').
cli_testcase(1631, [cbc], ['../prob_examples/public_examples/B/Puzzles/Queens/NBishopsSets.mch', '-init'], 'See that ProB can solve this puzzle').
cli_testcase(1632, [cbc,while], ['../prob_examples/public_examples/B/Tickets/Hansen23_WhilePerformance/WhileSlow_CartProduct.mch', '-t',
  '-p', 'TIME_OUT', 15000 ], % increased time-out for jenkins
 'See that performance of while loop with IF with existential quantifier is ok').
cli_testcase(1633, [kodkod], [
 '-evalf', ':kodkod args<:{1,2,3,4} & x:args & x<2 & x/=1',
 '-evalf', ':kodkod args<:{2,4,5,9,16,32} & x:args & x<4 & x/=2', % used to generate java.lang.IllegalArgumentException: too many atoms (32 <= 2^6) for too few powers of 2 (4< 6),
 '-evalt', ':kodkod args<:{2,4,5,9,16,32} & x:args & x<4' % used to generate java.lang.IllegalArgumentException: too many atoms (32 <= 2^6) for too few powers of 2 (4< 6)
% '-evalf', ':kodkod args={1,2,3} & atts : args <-> args & dom(atts) = args & x = card(atts)&x < 2' %% TO DO: fix KODKOD
], 'check kodkod integer ranges are correctly computed, prevent regression').
cli_testcase(1634, [b_test,external], [
'../prob_examples/public_examples/B/ExternalFunctions/MAXIMIZE_Test.mch',
'../prob_examples/public_examples/B/Puzzles/Queens/MaxQueensNoCheck_MAXIMIZE.mch',
'../prob_examples/public_examples/B/Mathematical/Argumentation/Example3_MAXIMIZE.mch',
 '-assertions'],
 'Check MAXIMIZE external function').
cli_testcase(1635, [private,tickets,codespeed], ['../prob_examples/examples/B/Alstom/DataValidationProject/Tickets/Burdy3/Rule_DB_General_0006.mch', '-t',
   %'-p','SHOW_EVENTB_ANY_VALUES','TRUE', '-p', 'MAX_OPERATIONS', 1000,
   '-p', 'MAXINT', 2147483647, '-p', 'MININT', -2147483647, '-p', 'TIME_OUT', 5000, '-p','CLPFD', 'TRUE'
   ],
 'Codespeed test; test that cartesian product / concat_sequence performance issue fixed'). % AlstomBurdy_Rule6_CLP
cli_testcase(1636, [b_test,tickets,wd], [
'../prob_examples/examples/B/ClearSy/GeneralElectric_DV/Ticket_Coudour1/rule.mch',
 '-expcterr', well_definedness_error,
 '-mc', 10, '-nodead'],
 'Check WD error found').
cli_testcase(1637, [b_test,tickets], [
'../prob_examples/examples/B/ClearSy/GeneralElectric_DV/Ticket_Coudour2/fonction_totale.mch',
 '-t', '-expcterr', setup_constants_inconsistent,
 '-properties', '-expcterr', check_properties],
 'Check partial setup constants possible despite inconsistency').
cli_testcase(1638, [b_test,tickets], [
'../prob_examples/examples/B/ClearSy/GeneralElectric_DV/Ticket_Coudour2/fonction_totale_v2.mch',
 '-animate_all', '-expcterr', setup_constants_fails],
 'Check inconsistent properties detected by animate_all').
cli_testcase(1639, [b_test], ['-check_java_version'], 'Check check_java_version works').
cli_testcase(1640, [tickets], ['-evalt', '{}/:STRING --> BOOL',
  '-evalt', '{"ab"|->TRUE }/:STRING --> BOOL',
  '-evalt', '{"ab"|->TRUE,"ab"|->FALSE }/:STRING --> BOOL',
  '-evalt', 'not({TRUE|->"STRING2"} : BOOL >+>> STRING)',
  '-evalt', 'not({TRUE|->"STRING2"} : BOOL +->> STRING)'
  ], 'Check bug in not_total_function fixed').
cli_testcase(1641, [laws,rel_fnc], ['../prob_examples/public_examples/B/Laws/FunLawsStrings.mch', '-mc', '500', '-c', '-nogoal', '-nodead', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE'], 'Use model checking to check a variety of laws on functions over infinite base domain.').
cli_testcase(1642, [private,tickets,operation_reuse], [
'../prob_examples/examples/B/ClearSy/GeneralElectric_DV/Ticket_Coudour3/rule1.mch',
 '-t', '-p', 'OPERATION_REUSE', 'TRUE',
 '-expcterr', invariant_violation],
 'Check operation reuse problem fixed').
cli_testcase(1643, [b_test,operation_reuse], ['../prob_examples/public_examples/B/Benchmarks/CSM.mch', '-mc', '1000', '-strict',
  '-p', 'OPERATION_REUSE', 'TRUE','-check_complete', '-cc', '341', '1229'], 'Test 13 with operation reuse').
cli_testcase(1644, [b_test,operation_reuse], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/SymmetryReduction/scheduler1.ref', '-mc', '150', '-cc', '145', '447',
'-p', 'OPERATION_REUSE', 'TRUE', '-p', 'DEFAULT_SETSIZE', '3', '-p', 'SYMMETRY_MODE', 'off', '-strict'], 'Test 73 with operation reuse').
cli_testcase(1645, [b_test,operation_reuse], ['-p', 'MAXINT', '5', '../prob_examples/public_examples/B/Demo/DiningCryptographers.mch', '-mc', '100', '-cc', '65', '96', '-p', 'MAX_INITIALISATIONS', '100', '-p', 'OPERATION_REUSE', 'TRUE','-p', 'SYMMETRY_MODE', 'off', '-strict'], 'Test 106 with operation reuse').
cli_testcase(1646, [cbc,operation_reuse], ['../prob_examples/public_examples/B/Tickets/FiveWayTraffic/FiveWay_new_mch.eventb',
 '-mc', 1000, '-strict','-p', 'OPERATION_REUSE', 'TRUE','-cc', '29', '65',  '--hash64', 971801032035978952
 ], 'Test 1532 with operation reuse.').
cli_testcase(1647, [b_test,csp_test,pge,cspb],['../prob_examples/public_examples/CSPB/Williams/ME.mch', '-csp-guide', '../prob_examples/public_examples/CSPB/Williams/ME.csp', '-mc', 10000, '-p', 'pge', disabled, '-strict'],'Testing PGE optimisation for (ME.mch||ME.csp).').
cli_testcase(1648, [b_test,csp_test,pge,cspb],['../prob_examples/public_examples/CSPB/Ivo/verysimple.mch', '-csp-guide', '../prob_examples/public_examples/CSPB/Ivo/verysimple.csp', '-mc', 10000, '-bf', '-p', 'pge', disabled, '-expcterr', deadlock, '-strict'],'Finding the deadlock because of the impossibility to execute the b operation (verysimple.mch||verysimple.csp).').
cli_testcase(1649, [cbc], [
 '-evalf', 'occupies : Pigeons >-> Holes & Pigeons=1..100 & Holes=1..99',
 '-evalt', 'occupies : Pigeons >-> Holes & Pigeons=2..100 & Holes=1..99',
 '-evalt', 'id(NATURAL) : NATURAL >-> NATURAL', % also check infinite functions work
 '-evalt', 'id(NATURAL1) : NATURAL1 >-> NATURAL1', % also check infinite functions work
 '-evalt', 'id(NATURAL1) : NATURAL1 >-> NATURAL',
 '-evalf', 'id(NATURAL1) : NATURAL >-> NATURAL',
 '-evalf', 'id(NATURAL) : NATURAL1 >-> NATURAL',
 '-evalf', 'id(NATURAL) : NATURAL >-> NATURAL1',
 '-evalt', 'id(NATURAL1) /: NATURAL >-> NATURAL'
], 'Ensure card. check for total_injection works').
cli_testcase(1650, [parser,tickets], [
'../prob_examples/public_examples/B/Tester/SyntaxErrors/MissingSemicolon.mch', '-v',
  '-expcterrpos', parse_error, 13, 2, '-expcterr', 'load_main_file'],
 'Check parser correctly reports syntax error and position').
cli_testcase(1651, [b_test,execute], ['../prob_examples/public_examples/B/Simple/UpCounter.mch', '-init','-execute_all',
 '-animate_stats', '-strict', '-goal', 'c=20', '-p', 'MAXINT', '20'], 'Check -execute_all.').
cli_testcase(1652, [parser,tickets], [
'-eval', '2>1 & & 3>2',
  '-expcterrpos', parse_machine_predicate_error, 1, 6 ],
 'Check parser correctly reports syntax error and position for eval').
cli_testcase(1653, [b_test,tickets], [
 '../prob_examples/public_examples/B/Tester/SelectMultipleTrueCases.mch',
  '-t', '-mc', 100, '-nogoal', '-cc', 4, 10, '--hash64', 357695998652715541 ],
 'Select with multiple true branches dealt with').
cli_testcase(1654, [parser,tickets,'PARSERLIB-44'], [
 '../prob_examples/public_examples/B/Tickets/PARSERLIB-44/BinPacking_UnicodeMinus.mch',
  '-init' ],
 'Check parser correctly parses unicode minus used for -->').
cli_testcase(1655, [laws,records], ['../prob_examples/public_examples/B/NewSyntax/RecSetLaws.mch', '-mc', '1000', '-nogoal', '-nodead', '-c', '-strict','-p', 'DOUBLE_EVALUATION', 'TRUE', '-t'], 'Various laws about records.').
cli_testcase(1656, [tickets,records,private], ['../prob_examples/examples/B/Tickets/record_projection_bugly.mch', '-assertions'], 'Check performance.').
cli_testcase(1657, [eventb_test,tickets,'PROB-351'], [
 '../prob_examples/examples/EventBPrologPackages/Advance_WP2/v5_Aug2014/ex_mch.eventb',
  '-t' ],
 'Check COMMS__ComputeLeastCostRoutes efficient (avoiding re-calculation of same assignment; see PROB-351)').
cli_testcase(1658, [parser,tickets], [
 '../prob_examples/public_examples/B/Tickets/Hansen25_OpNamingWarning/Naming_VAR.mch',
  '-expcterrpos', type_error, 5, 6
  ],
 'Check type checker reports id clash in VAR').
cli_testcase(1659, [parser,tickets], [
'-eval', ' 2  =  TRUE',
  '-expcterrpos', type_expression_error, 1, 7 ],
 'Check type checker correctly reports error and position in eval').
cli_testcase(1660, [parser,tickets], [
 '../prob_examples/public_examples/B/Tickets/Hansen25_OpNamingWarning/ReadingUninitialisedVar.mch', '-animate', 10,
  '-expcterrpos', store, 6, 12 ,  '-expcterrpos', identifier_not_found, 6, 12, '-nodead' ],
 'Check interpreter reports unitialised var and that -nodead works for -animate').
cli_testcase(1661, [eventb_test,tickets], [
 '../prob_examples/examples/EventBPrologPackages/Advance_WP2/v5_Aug2014/ex_mch.eventb',
  '-mc', 10 , '-expcterr', 'time_out_for_assertions'],
 'Check that virtual-time_out in assertion checking caught').
cli_testcase(1662, [b_test,csp_test,cspb], ['../prob_examples/public_examples/CSPB/FeatureTests/AssertionViolation.mch',
'-csp-guide', '../prob_examples/public_examples/CSPB/FeatureTests/AssertionViolation.csp',
'-mc', '1000', '-expcterr', assertion_violation, '-nodead'], 'Test that static assertion violation found').
cli_testcase(1663, [b_test,csp_test,cspb], ['../prob_examples/public_examples/CSPB/FeatureTests/AssertionViolationDynamic.mch',
'-csp-guide', '../prob_examples/public_examples/CSPB/FeatureTests/AssertionViolationDynamic.csp',
'-mc', '1000', '-expcterr', assertion_violation], 'Test that dynamic assertion violation found').
cli_testcase(1664, [b_test,csp_test,cspb], ['../prob_examples/public_examples/CSPB/FeatureTests/AssertionViolationDynamic.mch',
'-csp-guide', '../prob_examples/public_examples/CSPB/FeatureTests/AssertionViolationDynamicShort.csp',
'-mc', '1000', '-nodead'], 'Test that no dynamic assertion violation found').
cli_testcase(1665, [plugin_units], ['../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/Records.mch',
  '-plugin', 'units', '-strict',
  '-pp', '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/Records_internal.mch'],
  'Check that records are supported in unit plugin').
cli_testcase(1666, [kodkod,tickets,random], [
 '../prob_examples/public_examples/B/Tickets/BendispostoTutoren/tuts_cando_only.mch', '-init', '-assertions',
  '-p', 'RANDOMISE_ENUMERATION_ORDER', 'TRUE', '-p', 'KODKOD', 'TRUE' ],
 'Check that Randomise Enumeration order and Kodkod work together').
cli_testcase(1667, [cbc,tickets,kodkod,random], [
 '../prob_examples/public_examples/B/Tickets/BendispostoTutoren/tuts_working.mch', '-init', '-assertions',
  '-p', 'RANDOMISE_ENUMERATION_ORDER', 'TRUE', '-p', 'KODKOD', 'TRUE' ],
 'Check that Randomise Enumeration order and Kodkod work together').
cli_testcase(1668, [cbc,random], [
 '../prob_examples/public_examples/B/Tickets/BendispostoTutoren/tuts_working.mch', '-init', '-assertions',
  '-p', 'RANDOMISE_ENUMERATION_ORDER', 'TRUE', '-p', 'KODKOD', 'FALSE' ],
 'Check that Randomise Enumeration order and Kodkod work together').
cli_testcase(1669, [tickets,cbc], ['-p', 'MAXINT', 3,
 '-evalt', 'card(perm(NAT))=24',
 '-evalt', 'perm(NAT) /= {}',
 '-evalf', 'perm(NAT) = {}',
 '-evalt', 'perm(NAT) ={[1,0,2,3],[1,0,3,2],[2,0,1,3],[2,0,3,1],[3,0,1,2],[3,0,2,1],[0,1,2,3],[0,1,3,2],[2,1,0,3],[2,1,3,0],[3,1,0,2],[3,1,2,0],[0,2,1,3],[0,2,3,1],[1,2,0,3],[1,2,3,0],[3,2,0,1],[3,2,1,0],[0,3,1,2],[0,3,2,1],[1,3,0,2],[1,3,2,0],[2,3,0,1],[2,3,1,0]}',
 '-evalt', 'perm(NAT) = perm(0..MAXINT)',
 '-evalt', 'perm(NAT1) = perm(1..MAXINT)',
 '-evalt', 'card(perm(INT))=120',
 '-evalt', 'card(iseq(NAT)) = 65'
 ],
 'Check that perm(NAT)={} bug fixed').
cli_testcase(1670, [b_test], ['-t', '-expcterr', trace_checking_fails,
 '../prob_examples/public_examples/B/FeatureChecks/DEFINITIONS/SETPREF.mch', '-p', 'MAXINT', '22', '-card', 'ID', '2'], 'check trace checking fails').
cli_testcase(1671, [tickets, 'PARSERLIB-47', parser], ['-t', '--model-check',
  '../prob_examples/public_examples/B/FeatureChecks/DollarSymbol/ValidMachines/AssignByPredicateWithDEFTransitive.mch',
  '../prob_examples/public_examples/B/FeatureChecks/DollarSymbol/ValidMachines/AssignByPredicateWithDEF.mch'],
   'check $0 can be used within definitions').
cli_testcase(1672, [tickets, 'PARSERLIB-47', parser], [
  '../prob_examples/public_examples/B/FeatureChecks/DollarSymbol/ErrorMachines/DollarInInvariant.mch',
  '-expcterrpos', type_error, 2, 21, '-expcterr', load_main_file],
   'check $0 cannot be used in invariant').
cli_testcase(1673, [tickets, parser], [
  '../prob_examples/public_examples/B/ErrorMachines/IllegalSeesIncludes/CyclicM1.mch',
  '-expcterr', parse_error, '-expcterr', load_main_file],
   'check cyclic sees detected and does not lead to parser exception').
cli_testcase(1674, [pragmas], ['-init',
  '../prob_examples/public_examples/B/Pragmas/FilePragma.mch'],
   'check @file "Path" pragma works').
cli_testcase(1675, [b_test,mcm_tests], ['../prob_examples/public_examples/B/TestCaseGeneration/MCM_Tests/SimpleTestModel.mch', '-mcm_tests', '20', '1000', '1=1', '../prob_examples/public_examples/B/TestCaseGeneration/MCM_Tests/SimpleTestModel_mcm_tests.xml', '-strict'],'State based test generation with non-deterministic initialisation').
cli_testcase(1676, [b_test,mcm_tests], ['../prob_examples/examples/EventBPrologPackages/Soton/MCM_Tests/FENCEM7_mch.eventb',
  '-mcm_tests', '10', '4000', '1=1',
       '../prob_examples/examples/EventBPrologPackages/Soton/MCM_Tests/tests.xml',
  '-mcm_cover', 'IssueLoad IssueStore ObserveLoadAfterStore',
  '-p', 'INTERNAL_ARGUMENT_PREFIX', 'p',
 '-strict'],'State based test generation with internal operation arguments').
cli_testcase(1677, ['PROB-412',tickets,while], [
 '../prob_examples/public_examples/B/Tickets/PROB-412/WhileVariantNotDecreased.mch', '-animate', 10,
  '-expcterr', animate, '-expcterrpos', while_variant_error, 24, 16 ],
 'Check variant not decreasing error found and position ok').
cli_testcase(1678, [while,tickets], [
 '../prob_examples/public_examples/B/Tickets/Hansen24_WhileSeesBug/While.mch',
 '../prob_examples/public_examples/B/Tickets/Hansen24_WhileSeesBug/Runner.mch',
 '-t', '-mc', 1000 ],
 'Check bug with while loop compilation because of incorrect read/modifies info (not renamed) fixed').
cli_testcase(1679, [b_test], [
 '../prob_examples/public_examples/B/Tester/ParityFunction.mch',
 '../prob_examples/public_examples/B/Mathematical/Sieve_WithMin.mch',
 '../prob_examples/public_examples/B/Benchmarks/Chapter13/Booksrr.ref',
 '../prob_examples/public_examples/B/Benchmarks/Chapter_10/Safes.mch',
 '-t'],
 'A few trace tests from the Tcl/Tk Regression Test suite').
cli_testcase(1680, [b_test], [
 '../prob_examples/examples//B/Siemens/TestString.mch',
 '-expcterr', 'virtual_time_out',
 '-t'],
 'A few trace tests from the Tcl/Tk Regression Test suite').
cli_testcase(1681, [b_test], [
 '../prob_examples/examples/B/ClearSy/control/M1.ref',
 %'../prob_examples/examples/B/ClearSy/control/M1_i.imp',  % bmachine_static_checks: Operation result "status" has the same name as a variable or constant in operation "UTILS.nextState".
 '-t'],
 'Check no exists body warning').
cli_testcase(1682, [tickets,private], [
 '../prob_examples/examples//B/Alstom/ssp/tms_ssp_0004.mch',
 '-t'],
 'Check no time-out').
cli_testcase(1683, [cbc], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '-evalt', '(A=TRUE) <=> (X:Y..Z) & Y:1..10 & Z:0..2 & X >12 & (A=FALSE => X:15..16)'
 ], 'check interval reification').
cli_testcase(1684, [cbc], [ '-p', 'CLPFD', 'FALSE', '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '-evalt', '(A=TRUE) <=> (X:Y..Z) & Y:1..10 & Z:0..2 & X >12 & (A=FALSE => X:15..16)'
 ], 'check interval reification').
cli_testcase(1685, [cbc], [
 '-evalt', 'NATURAL1 : x & (x = { {22}, {33}} or x={ NATURAL1 })',
 '-evalf', 'NATURAL1 : x & x = { {22}, {33}}',
 '-evalt', '{x|x>30 & x mod 3 =0} /\\ {x|x<50 & x mod 2 =0} = {36,42,48}'
 ], 'a few checks').

cli_testcase(1686, [cbc,laws,finite], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
  '-evalt', 'x:FIN(x) & x=POW(POW(1..9))',
  '-evalf', 'x/:FIN(x) & x=POW(POW(1..9))',
  '-evalt', 'x:FIN(x) & x=FIN(POW(1..10))',
  '-evalt', 'x:FIN(x) & x=POW(POW(1..10))',
  '-evalf', 'x/:FIN(x) & x=POW(POW(1..10))',
  '-evalt', 'x:FIN(x) & x=POW(POW(POW(1..10)))',
  '-evalf', 'x:FIN(x) & x=POW(POW(POW(NATURAL1)))',
  '-evalt', 'x/:FIN(x) & x=POW(POW(POW(NATURAL1)))',
  '-evalf', 'x:FIN(x) & x=FIN(POW(POW(NATURAL1)))',
  '-evalt', 'x/:FIN(x) & x=FIN(POW(POW(NATURAL1)))',
  '-evalf', 'x:FIN(x) & x=FIN(INTEGER)',
  '-evalt', 'x/:FIN(x) & x=FIN(INTEGER)',
  '-evalt', 'x/:FIN(x) & x=POW(INTEGER)',
  '-evalf', 'x:FIN(x) & x = %x.(x:NATURAL1|1)',
  '-evalt', 'x:FIN(x) & x = %x.(x:POW(POW(1..11))|1)',
%  '-evalt', 'x/:FIN(x) & x = %x.(x:POW(POW(1..11))|1)', % computes x, time-out
  '-evalt', 'x:FIN(x) & x = (3..2)*INTEGER',
  '-evalt', 'x:FIN(x) & x = INTEGER*(3..2)',
  '-evalf', 'x:FIN(x) & x = INTEGER*(2..3)',
  '-evalt', 'x:FIN(x) & x = POW(POW(1..10))*(2..3)',
  '-evalt', 'x:FIN(x) & x = POW(POW(1..11))*(2..3)',
  '-evalt', 'x:FIN(x) & x = POW(POW(1..11))*{2,4}',
  '-evalt', 'x:FIN(x) & x = POW(POW(1..11))*(4..3)',
  '-evalt', 'x:FIN(x) & x = NATURAL1*(4..3)',
  '-evalf', 'x:FIN(x) & x = NATURAL1*{2,4}',
  '-evalf', 'x:FIN(x) & x = (1..2)*NATURAL1',
  '-evalf', 'x:FIN(x) & x = INTEGER*NATURAL1',
  '-evalt', 'x:FIN(x) & x=iseq(1..10)',
  '-evalf', 'x:FIN(x) & x=iseq(NATURAL)',
  '-evalt', 'x:FIN(x) & x=perm(POW(1..9))',
  '-evalf', 'x:FIN(x) & x= {x|x : INTEGER & x < 3}',
  '-evalt', 'x:FIN(x) & x=iseq(POW(POW(1..11)))',
  '-evalt', 'x:FIN(x) & x=iseq1(POW(POW(1..11)))',
  '-evalf', 'x:FIN(x) & x=seq(POW(POW(1..11)))',
  '-evalf', 'x:FIN(x) & x=seq1(POW(POW(1..11)))',
  '-evalt', 'x:FIN(x) & x=seq(4..3)',
  '-evalt', 'x:FIN(x) & x=seq1(4..3)',
  '-evalt', 'x:FIN(x) & x=id(POW(POW(1..11)))',
  '-evalf', 'x:FIN(x) & x=perm(NATURAL)',
  '-evalt', 'x:FIN(x) & x=perm(POW(POW(1..11)))',
  '-evalf', 'x/:FIN(x) & x=perm(POW(POW(1..11)))'
  ], 'a few checks about finite').
cli_testcase(1687, [tickets,parser], [
 '../prob_examples/public_examples/B/Tickets/Hansen27_NestedMchErr/M1.mch',
  '-expcterr', load_main_file, '-expcterrpos', parse_error, 2, 5 ],
 'Check error found and position ok').
cli_testcase(1688, [cbc,laws,finite], [ '-p', 'DISPROVER_MODE', 'TRUE',
  '-eval', 'x/:FIN(x) & x=(POW(1..11) +-> POW(1..9))',
%  '-eval', 'x:FIN(x) & x=perm(POW(1..9))',
  '-expcterr', 'eval_string_time_out'
  ], 'a few checks about finite; ideally we want -evalt').  % TO DO: fix this issue
cli_testcase(1689, ['PROB-411',tickets,kodkod], [
  '-evalf', ':kodkod args={1,2,3} & atts : args <-> args & dom(atts) = args & x = card(atts)&x < 2'
  ], 'check card interval analysis bug fixed').
cli_testcase(1690, [tickets], [
'../prob_examples/public_examples/B/Tickets/Hansen21_Sees/AllGroups.mch',
  '-t', '-expcterr', 'invariant_violation'
  ], 'check sees bug remains fixed').
cli_testcase(1691, [tickets,records], [
'../prob_examples/public_examples/B/Tickets/RecordPartiallyTyped/Knowledge_loop.mch',
  '-init'
  ], 'check that record type bug fixed, no type generated: record([field(a,b)|any])').
cli_testcase(1692, [tickets,quantified], [  '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
  '-evalt', 'UNION(i).(i:1..2|{i|->i}) ^ UNION(j).(j:1..3|{j|->j}) = [1,2,1,2,3]',
  '-evalt', 'UNION(i).(i:1..2|{i|->i}) ^ {} = [1,2]',
  '-evalf', 'UNION(i).(i:1..2|{i|->i}) ^ UNION(j).(j:1..3|{j|->j}) = [1,2,3,4,5]'
  ], 'ensure both UNIONs are translated').
cli_testcase(1693, [tickets,let], [
 '-eval', 'i = 20 & (LET i BE i= i+1 IN i END) =res',
 '-expcterr', type_expression_error
  ], 'ensure error in nested LET expression detected').
cli_testcase(1694, [tickets,let], [
 '-eval', '(LET i,j BE i=i+10 & j=i+20 IN i+j END)',
 '-expcterr', type_expression_error
  ], 'ensure error in nested LET expression detected').
cli_testcase(1695, [tickets,let], [
 '-eval', '(LET i,j BE i=10 & j=i+20 IN i+j END)',
 '-expcterr', type_expression_error
  ], 'ensure error in nested LET expression detected').
cli_testcase(1696, [tickets,let], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '-evalt', 'i=1 & r=LET i BE i=10 IN LET j BE j=i+20 IN i+j END END & r=40',
% '-evalt', 'k = 1..n & n:2..3 & (LET i,j BE i={10}\\/k & j= i \\/ i IN res=i \\/ j END)',
% '-evalt', '(LET i,j BE i=10 & j=i+20 IN i+j END) =40',
 '-evalt', 'i = 100 & (LET i,j BE i={10} & j= 20 IN (LET j BE j=i \\/ i IN i \\/ j END) END) = {10}',
 '-evalt', 'x=1 & y=2 & #(x,y).(x=30 & y=x+x & res=y+x) & res=90'
  ], 'ensure nested LET predicate and expression work').
cli_testcase(1697, [cbc], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
  '-evalt', '{y,n|{x|x:1..n} = {y}} = {1|->1}',
  '-evalt', 'x:{y|y>0 & y mod 100 = 0} & x:{z|z<1000 & z mod 500 = 0}',
  '-evalt', '{x|x:1..n} = {y}'
  ], 'check improved set comprehension propagation').
cli_testcase(1698, [cbc], [
  '-evalt', 'x:c1 & c1={y|y>0 & y mod 100 = 0} & x:c2 & c2= {z|z<1000 & z mod 500 = 0}',
  % these now did not work before:
 '-evalt', 'p = {x|x>100 & x<n & x mod 2 =0} & p = {102,104}',
 '-evalt', 'p = {x|x>100 & x<n & x mod 2 =0} & p = {102,104,106,108}',
 '-evalt', 'p = {x|x>100 & x<n & x mod 2 =0} & p = {102,104,106,108} & n=110',
 '-evalf', 'p = {x|x>100 & x<n & x mod 2 =0} & p = {102,104,106,108,113}'
% '-evalf', 'p = {x|x>100 & x<n & x mod 2 =0} & p = {102,104,106,108,112}' % TO DO: does not work yet
% '-evalt', 'p = {x|x>100 & x<n & x mod 2 =0} & {102} <: p' % TO DO: nor this
  ], 'check improved set comprehension propagation').
cli_testcase(1699, [private,csp_test,cspb], ['../prob_examples/examples/CSPB/Treharne/CSPB-SwanseaPlanModel/Interlocking.mch',
'-csp-guide', '../prob_examples/examples/CSPB/Treharne/CSPB-SwanseaPlanModel/Ctrl.csp',
'-mc', '40', '-bf', '-cc', 148, 687, '--hash64', 906866094249096570], 'Test CSP||B model').
cli_testcase(1700, [private,csp_test,cspb], ['../prob_examples/public_examples/B/Ivo/BenchmarksFairness/ProB/DPhil/dphil_ltl6.csp',
'-mc', '3000', '-bf', '-cc', 1764, 7125, '--hash64', 419761536083491376], 'Test Dining Philosophers state space (CSP normalisation)').
cli_testcase(1701, [csp_test,cspb], ['../prob_examples/public_examples/CSPB/Counter.mch',
'-csp-guide', '../prob_examples/public_examples/CSPB/Counter.csp',
'-mc', '40', '-bf', '-cc', 15, 15, '--hash64', 987057609118852348], 'Test CSP||B model').

cli_testcase(1702,[cbc,enabling],[
'-enabling_analysis_csv','../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_enabling_analysis.csv',
'../prob_examples/public_examples/B/Benchmarks/Cruise_finite1.mch',
'-p', 'TIME_OUT', 5000
], 'Check CSV output of Cruise Control').
cli_testcase(1703,[cbc,enabling],[
'-enabling_analysis_csv','../prob_examples/public_examples/EventBPrologPackages/Advance/CAN_Bus/CB3FSMM_mch_v2_enabling_analysis.csv',
'../prob_examples/public_examples/EventBPrologPackages/Advance/CAN_Bus/CB3FSMM_mch_v2.eventb',
'-p', 'TIME_OUT', 5000
], 'Check CSV output of CAN Bus Control').
cli_testcase(1704, [tickets], ['../prob_examples/public_examples/B/Tickets/Space InPath/phonebook7.mch', '-t', '-mc', '100', '-strict'], 'check space in path works').
cli_testcase(1705, [tla,tlc], ['../prob_examples/public_examples/EventBPrologPackages/Tickets/ExportClassicalB/TestLeftAsVar_mch.eventb',
'../prob_examples/public_examples/EventBPrologPackages/Tickets/Tester/UseBKeywords.eventb', '-mc_with_tlc'], 'check TLC for Event-B model with classical B keywords').

cli_testcase(1706, [b_test, infinite,total_function], [
 '-evalt', 'f : NATURAL --> STRING',
 '-evalt', 'f : NATURAL --> POW(INTEGER)',
 '-evalt', 'f : STRING --> {11,22,33} & f("abc")>22 & f("xyz") < 22',
 '-evalt', 'f : STRING --> STRING & f("abc")="ABC" & f("xyz") = "XYZ"'
], 'Ensure symbolic treatment of (infinite) total functions with complicated range types.').

cli_testcase(1707, [b_test,cruise,operation_reuse], ['../prob_examples/public_examples/B/Benchmarks/Cruise_finite1.mch', '-mc', '1400', '-cc', '1361', '25696', '-strict', '--hash64', 203058745203935766, '-p', 'MAX_INITIALISATIONS', 5,
'-p', 'OPERATION_REUSE', full], 'Complete Model Check of Cruise Controller').
cli_testcase(1708, [private,tickets,operation_reuse], [
'../prob_examples/examples/B/ClearSy/GeneralElectric_DV/Ticket_Coudour3/rule1.mch',
 '-t', '-p', 'OPERATION_REUSE', 'full',
 '-expcterr', invariant_violation],
 'Check operation reuse problem fixed').
cli_testcase(1709, [b_test,operation_reuse], ['../prob_examples/public_examples/B/Benchmarks/CSM.mch', '-mc', '1000', '-strict',
  '-p', 'OPERATION_REUSE', 'full','-check_complete', '-cc', '341', '1229'], 'Test 13 with operation reuse').

cli_testcase(1710, [cbc], [
 '-evalt', 'x = {1|->2, 2|->3, 3|->4} & r={y|(y,3):x} & r={2}',
 '-evalt', 'x = {1|->2, 2|->3, 3|->4, 22|->3} & r={y|(y,3):x} & r={2,22}',
 '-evalt', 'x = {1|->2, 2|->3, 3|->4} & r={y|(y,5):x} & r={}'
], 'Test new compilation of set membership.').
cli_testcase(1711, [tickets], ['../prob_examples/public_examples/B/ErrorMachines/ParseErrors/DefinitionInQuantifier2.mch',
 '-expcterrpos', type_error, 16, 18,
 '-expcterrpos', bmachine_construction, 12, 1, '-expcterr', 'load_main_file'
 ], 'Check illegal DEFINITION args detected').
cli_testcase(1712, [b_test], ['../prob_examples/public_examples/B/Tester/DefGoal.mch',
 '-execute_all', '-nogoal', '-check_goal'
 ], 'Check -check_goal works with -execute_all').
cli_testcase(1713, [b_test,external], ['../prob_examples/public_examples/B/ExternalFunctions/TestLibraryMeta.mch', '-model-check', '-strict'], 'Check that external functions on state space work.').
cli_testcase(1714, [cbc], [
 '-evalf', 'r = {x,y,z|x:{1,4,7,10,13,16,19,21,24,27,30,33,36,39,42,45,48}&y=x+1&z=y+1} & (x |-> y) |-> z : r & y : 10 .. 13 & z < 11', % takes Kodkod about 2 seconds to solve
 '-evalt', 'r = {x,y,z|x:{1,4,7,10,13,16,19,21,24,27,30,33,36,39,42,45,48}&y=x+1&z=y+1} & (x |-> y) |-> z : r & y : 10 .. 13 & z > 11', % takes Kodkod about 2 seconds, z3 90 ms, ProB 10 ms
  '-evalt', 'r = {x,y,z|x:{1,4,7,10,13,16,19,21,24,27,30,33,36,39,42,45,48}&y=x+1&z=y+1} & {x,y,z|(x |-> y) |-> z : r & y : 10 .. 13 & z >11} = {(10,11,12)}'
], 'Test new clpfd_table module.'). % tests also pass with prior version of ProB
 cli_testcase(1715, [tickets,codespeed], [
 '../prob_examples/examples/B/Alstom/Boucles_Jul11/essai2_boucle1.mch',
  '-t', '-strict',
  '-expcterr', 'invariant_violation',
   '-p', 'MAXINT', 10000000, '-p', 'MININT', -10000000, '-p', 'TIME_OUT', 8000],
   'Test performance (codespeed test Alstom_essai2_boucle1) + tests watched-literal function application').
cli_testcase(1716, [tickets,slot], ['../prob_examples/public_examples/B/Tickets/SLOT-PERFORMANCE6/Solver_v2.mch',
 '-init', % '-v', % generates non-ground closure warnings
 '-evalt', 'CHECK({"BK-GER-H-2013"}, cmcc, choiceUnits, choiceGroups)', % this worked before
 '-evalt', 'CHECK_wr({"BK-GES-H-2013"}, cmcc, choiceUnits, choiceGroups,{mod47,mod46, mod45, mod44, mod43},{au76,au77,au79,au80,au81},UNITS)', % this also worked before
 '-evalt',  'CHECK_wr({"BK-GES-H-2013"}, cmcc, choiceUnits, choiceGroups,{mod47,mod46, mod45, mod44, mod43},{au76,au77,au79,au80,au81,au82,au83,au84},{unit182,unit183,unit184,unit185,unit186,unit187,unit188,unit189,unit190,unit191,unit192,unit193,unit194,unit234,unit267,unit268} )' % this did NOT work before
% '-evalt', 'CHECK({"BK-GES-H-2013"}, cmcc, choiceUnits, choiceGroups)' % does not yet work
 ], 'Ensure no time-out and that new element/3 constraint for function application and table constraint work').
cli_testcase(1717, [cbc], [
 '-evalf', 'f = {aa|->(1,aa), x|->(2,x+1), y|->(3,y+1)} & x:1..2 & aa:0..1 & y:4..5 & r = f(v) & v:3..4 & x+y<v',
 '-evalf', 'f = {aa|->(1,aa), x|->(2,x+1), y|->(3,y+1)} & x:1..2 & aa:0..1 & y:3..5 & r = f(v) & v:2..4 & prj1(INTEGER,INTEGER)(r) /: 2..3',
 '-evalt', 'f = {aa|->(1,aa), x|->(2,x+1), y|->(3,y+1)} & x:1..2 & aa:0..1 & y:3..5 & r = f(v) & v:2..4 & prj1(INTEGER,INTEGER)(r) =2',
 '-evalf', 'f = {aa|->rec(p1:1,p2:aa), x|->rec(p1:2,p2:x+1), y|->rec(p1:3,p2:y+1)} & x:1..2 & aa:0..1 & y:3..5 & r = f(v) & v:2..4 & r\'p1 /: 2..3',
 '-evalt', 'f = {aa|->rec(p1:1,p2:aa), x|->rec(p1:2,p2:x+1), y|->rec(p1:3,p2:y+1)} & x:1..2 & aa:0..1 & y:3..5 & r = f(v) & v:2..4 & r\'p1 =2'
 ], 'Exercise new element/3 function application treatment').
cli_testcase(1718, [cbc], ['../prob_examples/public_examples/B/Demo/Lift_inf.mch', '-cbc_sequence', 'dec;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc;inc', '-strict', '-goal', 'floor=40', '-check_goal'], 'Check cbc_sequence for longer sequence').
cli_testcase(1719, [smt_solver_integration], [
  '-p','SMT_SUPPORTED_INTERPRETER','TRUE',
  '-evalt', 'x = 1',
  '-evalf', 'x = 1 & x > 2',
  '-evalf', 'x>y & y>x',
  '-evalt', 'r:struct(f1:11..20,f2:12..30) & r\'f1+10 = r\'f2',
  '-evalf', 'r:struct(f1:11..20,f2:12..20) & r\'f1+10 = r\'f2',
  '-evalt', 'r:struct(f1:11..20,f2:12..30) & r\'f1+10 = r\'f2 & s:struct(f1:15..20,f2:1..100) & s=r',
  '-evalt', 'v1:struct(lderb:{TRUE},mspue:{FALSE},fhhxr:{TRUE})',
  '-evalf', 'x>0 & x+2 <= p & x=p',
  '-evalf', 'not(Pace_Int - sp >= 0) & (sp > 0 & sp < Pace_Int)',
  '-evalt', '{x} <: {999}',
  '-evalt', '{x} <: {999,1000} & x > 999',
  '-evalt', 'x = {1,2} /\\ {3,4}',
  '-evalt', 'x = {(1,2),(3,4),(4,5),(6,7),(8,9)} & y|->v : x & y|->w : x',
  '-evalt', 'x : {1,2} --> NAT',
  '-evalf', 'x : {1,2} --> NAT & x = {(1,1)}',
  '-evalt', 'x : {1,2} --> NAT & x = {(1,1),(2,1)}',
  '-evalf', 's = min({1}) & s=2',
  '-evalf', 's = min({1,2}) & s>2',
  '-evalt', 'x = %i.(i:1..10|i+i) & x[{j}]={8}',
  '-evalt', 'x = %i.(i:1..10|i+i) & x[{5,j}]={16,10}',
  '-evalf', 'x = %i.(i:1..10|i+i) & x[{5,j}]={16,11}',
  '-evalt', 'x:1..10 & #y.(y:1..10 & y>x) & #y.(y:1..10 & y<x)'
 ], 'some constraints that should be solved by the integrated ProB / SMT solver').
cli_testcase(1720, [cbc], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
  '-evalt', '{x|x:1..n} = {y,y+1}',
  '-evalt', '1..n = {y,y+1}',
  '-evalt', '1..n = r & r={y,y+1}',
  '-evalt', '{n,y|#r.(1001..n = r & r={y,y+1})} = {(1002,1001)}',
  '-evalf', '1001..n = r & r={y,y+2}',
  '-evalt', '1001..n = {y,y+2,y+1} & n<20000' % TO DO: get rid of necessity for n<20000
 ], 'test improved constraint propagation for in_interval and frozen_dif').
cli_testcase(1721, [cbc], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
  '-evalf', 'X:INTEGER & X*Y /= X*Y2 & Y2=Y',
  '-evalf', 'X:INTEGER & X+Y /= Y2+X & Y2=Y'
 ], 'test improved clpfd_neq_expr').
cli_testcase(1722, [cbc,cse], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE', '-p', 'CSE', 'TRUE',
  '-evalf', 'not(#(x,y,z).((x+1)*y = 1000 & (x+1)*z=y & z=(x+1)  ))', % already worked before
  '-evalt', 'not(#(x,y,z).((x+1)*y = 1001 & (x+1)*z=y & z=(x+1)  ))', % already worked before
  '-evalt', 'not(#(X,Y,Z,Z1,Z2,Z3,Z4).(X:INTEGER & X*Y=Z1 & Z1*Z = Z2 & Z*X = Z3 & Z3*Y = Z4 & Z2 /= Z4))',
  '-evalf', 'not(not(#(X,Y,Z,Z1,Z2,Z3,Z4).(X:INTEGER & X*Y=Z1 & Z1*Z = Z2 & Z*X = Z3 & Z3*Y = Z4 & Z2 /= Z4)))',
  '-evalf', '#(X,Y,Z).(X:INTEGER & #(Z1,Z2,Z3,Z4).(X*Y=Z1 & Z1*Z = Z2 & Z*X = Z3 & Z3*Y = Z4 & Z2 /= Z4))'
 ], 'test single usage identifier inlining').
cli_testcase(1723, [cbc], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
  '-evalf', 'X:INTEGER & Y*X /= X*Y2 & Y2=Y',
  '-evalf', 'X:INTEGER & X*Y*(Z+V) /= (V+Z)*X*Y',
  '-evalf', 'X:INTEGER & X+Y+(Z*V) /= (V*Z)+X+Y',
  '-evalt', 'not(#(X,Y,Z,Z1,Z2,Z3,Z4).(X:INTEGER & X*Y=Z1 & Z1*Z = Z2 & Z*X = Z3 & Z3*Y = Z4 & Z2 /= Z4))',
  '-evalf', 'x:INTEGER & (x-(x+a*b)) /= (y-(b*a+x)) & y=x', % added later
  '-evalf', 'x:INTEGER & (-(x+a*b)) /= (-(b*a+y)) & y=x' % added later
 ], 'test improved commutativity detection in clpfd_neq_expr').
cli_testcase(1724, [tickets], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '-execute_all',  '-check_goal',
 '../prob_examples/public_examples/B/Tester/CheckGoal.mch'],'test check_goal with executed_all').
cli_testcase(1725, [tickets, b_test], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
 '-assertions',  '-init',
 '../prob_examples/public_examples/B/Tester/Visibility/M2_xx.mch'],'test visibility in assertions clause').
cli_testcase(1726, [b_test,external], ['../prob_examples/public_examples/B/ExternalFunctions/XML_Reader.mch', '-t', '-assertions',
'-evalt', 'READ_XML_FROM_STRING("<h3>hallo</h3>") = [rec(attributes:{},element:"h3",meta:{("xmlLineNumber"|->"1")},pId:0,recId:1),  rec(attributes:{("text"|->"hallo")},element:"CText",meta:{("xmlLineNumber"|->  "1")},pId:1,recId:2)]',
'-strict'], 'Check that external function to read XML works.').
cli_testcase(1727, [b_test,ltl,safety], ['../prob_examples/public_examples/B/LTL/SafetyMC/BeverageVendingMachine.mch', '-ltlformulaf', 'G not [bget]', '-p', 'use_safety_ltl_model_checker', true, '-strict'],'Check that safety model checker does not produce incomplete counter examples when transition propositions appear in the safety ltl formula.').
cli_testcase(1728, [error_checks], [ '-strict',
 '-expcterrpos', 'bmachine_static_checks', 10, 15,
 '../prob_examples/public_examples/B/Tester/InitialisationChecks/IllegalOrderOfInits.mch'],'test detection of illegal initialisations').
cli_testcase(1729, [error_checks], [ '-strict',
 '-expcterrpos', 'bmachine_static_checks', 10, 25,
 '../prob_examples/public_examples/B/Tester/InitialisationChecks/IllegalParallelInit.mch'],'test detection of illegal initialisations').
cli_testcase(1730, [error_checks], [ '-strict',
 '-expcterrpos', 'bmachine_static_checks', 14, 2,
 '../prob_examples/public_examples/B/Tester/InitialisationChecks/IllegalChoiceOrderInit.mch'],'test detection of illegal initialisations').
cli_testcase(1731, [cbc,infinite], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
  '-p', 'ENUMERATE_INFINITE_TYPES', 'FALSE',
  '-evalt', 'f=%x.(x:iseq(struct(a:STRING))|x) & f([rec(a:"a")])=[rec(a:"a")]',
  '-evalt', 'f=%x.(x:iseq(struct(a:STRING,b:BOOL))|x) & f([rec(a:"a",b:TRUE)])=[rec(a:"a",b:xx)] & xx=TRUE',
  '-evalt', 'f=%x.(x:iseq(struct(a:NATURAL,b:BOOL))|x) & f([rec(a:222,b:TRUE)])=[rec(a:222,b:xx)] & xx=TRUE',
  '-evalt', 'f=%x.(x:iseq1(struct(a:NATURAL,b:BOOL))|x) & f([rec(a:222,b:TRUE)])=[rec(a:222,b:xx)] & xx=TRUE',
  '-evalt', 'f=%x.(x:seq1(struct(a:NATURAL,b:BOOL))|x) & f([rec(a:222,b:TRUE)])=[rec(a:222,b:xx)] & xx=TRUE',
  '-evalt', 'f=%x.(x:seq(struct(a:NATURAL,b:BOOL))|x) & f([rec(a:222,b:TRUE)])=[rec(a:222,b:xx)] & xx=TRUE',
  '-evalt', 'f=%x.(x:iseq(struct(a:seq1(NATURAL),b:BOOL))|x) & f([rec(a:[222],b:TRUE)])=[rec(a:[222],b:xx)] & xx=TRUE',
  '-evalf', 'f=%x.(x:iseq(struct(a:STRING))|x) & f([rec(a:"a")])=[rec(a:"b")]'],
  'test detection of infinite function without virtual time-outs').
cli_testcase(1732, [tickets], [
  '../prob_examples/public_examples/B/Tickets/Hansen31_VirtualTimeout/VirtualTimeout.mch', '-mc', 1000, '-expcterr','model_check_incomplete','-expcterr','virtual_time_out'], 'Check that we detect that model-checking incomplete.').
cli_testcase(1733, [tickets,execute], [ % '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
  '../prob_examples/public_examples/B/Tickets/Hansen31_VirtualTimeout/VirtualTimeout.mch', '-execute', 1000,
   '-animate_stats',
    '-expcterr','virtual_time_out_execute'], 'Check that we detect that execute incomplete.').
cli_testcase(1734, [tickets,animate], [ % '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
  '../prob_examples/public_examples/B/Tickets/Hansen31_VirtualTimeout/VirtualTimeout.mch', '-animate', 5,
   '-animate_stats',
    '-expcterr','virtual_time_out'], 'Check that we detect that execute incomplete.').
cli_testcase(1735, [cbc], ['../prob_examples/public_examples/B/Tester/InitialisationChecks/InitCBCCheck.mch', '-cbc_sequence', 'Goal', '-strict'], 'Check cbc_sequence with non-det. INITIALISATION').
cli_testcase(1736, [error_checks,b_test], ['../prob_examples/public_examples/B/Tester/InitialisationChecks/InitIF_Fails.mch', '-init', '-expcterrpos', initialisation_fails, 13,17, '-strict'], 'Check INITIALISATION error inside IF located').
cli_testcase(1737, [cbc,b_test,quantified], ['../prob_examples/public_examples/B/PerformanceTests/UNION_CardGT.mch', '-t'], 'test_closure for lambda closures avoid interference between domain and range expression').
cli_testcase(1738, [external,b_test], ['../prob_examples/public_examples/B/ExternalFunctions/XML/utf8_data.mch', '-assertions'], 'test READ_XML with UTF-8 encoding').
cli_testcase(1739, [cbc], ['../prob_examples/examples/B/Alloy/req2_v3_not_conj_check.mch',
   '../prob_examples/examples/B/Alloy/req2_v3_not_conj_check2.mch', % version with larger CONTEXT set, more difficult
    '-t', '-p', 'TIME_OUT', '9500'], 'test not conjunction treatment of kernel').
cli_testcase(1740, [cbc], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
  '-p', 'ENUMERATE_INFINITE_TYPES', 'FALSE',
  '-evalt', '#x.(x:1..10000 & x mod 22 = 11) or b=TRUE',
  '-evalt', '(#x.(x:1..10000 & x mod 22 = 11) or b=TRUE) => b=FALSE',
  '-evalt', '(#x.(x:1..10000 & x mod 22 = 22) or b=TRUE) => b=FALSE',
  '-evalf', '(b=FALSE => (#x.(x:1..10000 & x mod 22 = 22) or b=TRUE)) & (b=TRUE => (#x.(x:1..10000 & x mod 22 = 22) or b=FALSE))'
  ],
  'test reification of RHS of disjunction (or)').
cli_testcase(1741, [ctl], ['../prob_examples/public_examples/B/Ivo/CTLIssue.mch', '-ctlformulat', 'AF {x=4}'], 'Testing bug-fix caused by an unnecessary cut.').
cli_testcase(1742, [csp_test], ['../prob_examples/public_examples/CSP/simple/verysimple.csp',
  '-ltlformulat', 'GF e(a.red)', '-strict'
 ], 'Test check should show that verysimple fulfills the LTL formula, which is equivalent to the CTL formula AGAF e(a.red).').
cli_testcase(1743, [cbc,infinite], ['-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
   '-evalt', '{3} <| %x.(x:NATURAL|x+1) = {3|->4}',
   '-evalf', '{3} <| %x.(x:NATURAL|x+1) = {a,b} & a/=b',
   '-evalt', '2..4 = ran((1..3) <| %x.(x:NATURAL|x+1))',
   '-evalt', '1..3 = dom(%x.(x:NATURAL|x+1)|>(2..4))',
   '-evalt', '(1..9) <| pred = %x.(x:1..9|x-1)',
   '-evalt', '(1..9) <| succ = %x.(x:1..9|x+1)',
   '-evalt', '{n} <| succ = {15|->x} & n<20'
   ], 'Test better symbolic treatment of union').
cli_testcase(1744, [cbc,infinite], [
   '-evalt', '4|->b : {3} <<| %x.(x:NATURAL|x+1)',
   '-evalf', '4|->b : (%x.(x:NATURAL|x+1) |>> (1..9))',
   '-evalt', '14|->b : (%x.(x:NATURAL|x+1) |>> (1..9))',
   '-evalt', '{b|14|->b : (%x.(x:NATURAL|x+1) |>> (1..9))} = {15}',
   '-evalt', 'x = {-1,-33} \\/ NATURAL1  & 11:x & -1:x & -33:x & 0 /: x'
   ], 'Test better symbolic treatment of union').
 cli_testcase(1745, [tickets,codespeed], [
 '../prob_examples/public_examples/B/Puzzles/EulerWay.mch',
 %'../prob_examples/public_examples/B/Puzzles/EulerWayDisjunction.mch', % this does not work yet; disjoin_true change makes this version too slow
 '-init', '-p', 'TIME_OUT', 9500
],
   'EulerWay Codespeed Test').
 cli_testcase(1746, [cbc], [
 '../prob_examples/public_examples/B/CBC/Feasibility_baloadR07.mch',
 '-init'
],
   'Check reification of existential quantifier with 27 cases not problematic').
 cli_testcase(1747, [codespeed], [
 '../prob_examples/public_examples/B/Demo/SortByPermutation_v2.mch',
 '-t', '-p', 'TIME_OUT', 9500],
   'SortByPermutation_v2 Codespeed Test').
 cli_testcase(1748, [codespeed], [
 '../prob_examples/public_examples/B/PerformanceTests/ModelChecking/CountToMAXINT.mch',
 '-mc', 500000, '-p', 'MAXINT', 500000, '-nodead'],
   'MC_CountToMAXINT_50000 Codespeed Test').
cli_testcase(1749, [codespeed], ['../prob_examples/public_examples/EventBPrologPackages/SET_Game/SET_GAM_Sym_NoSet20_mch.eventb', '-mc', '1000', '-p', 'SYMMETRY_MODE', 'hash', '-p', 'TIME_OUT', '7000', '-p', 'CLPFD', 'TRUE', '-df', '-goal', 'n=18', '-p', 'MAX_OPERATIONS', '20', '-strict', '-expcterr', 'goal_found', '-p', 'CSE', 'TRUE'], 'Test 395 with CSE').
 cli_testcase(1750, [alstom,codespeed], [
 '../prob_examples/examples/B/Alstom/axl3/axl3.mch',
 '-t', '-p', 'CLPFD', 'TRUE'],
   'axl3 Codespeed Test').
cli_testcase(1751, [alstom,codespeed], [
 '../prob_examples/examples/B/Alstom/cbtc/actions_cbtc.mch',
 '-t', '-p', 'CLPFD', 'TRUE'],
   'actions_cbtc Codespeed Test').
cli_testcase(1752, [private,codespeed], [
 '../prob_examples/examples/B/Alstom/DataValidationProject/Rule_TMS_Route_0006.mch',
 '-p', 'MAXINT', 10000000, '-p', 'MININT', -10000000, '-p', 'TIME_OUT', '60000',
 '-p', 'SHOW_EVENTB_ANY_VALUES', 'true',
 '-mc', 1000, '-df', '-nodead', '-expcterr', 'invariant_violation',
 '-p', 'CLPFD', 'FALSE'],
   'ClearSy_RuleTMS_R0006 Codespeed Test').
cli_testcase(1753, [codespeed], [
 '../prob_examples/examples/B/Alstom/DataValidationProject/Rule_TMS_Route_0006.mch',
 '-p', 'MAXINT', 10000000, '-p', 'MININT', -10000000, '-p', 'TIME_OUT', '60000',
 '-p', 'SHOW_EVENTB_ANY_VALUES', 'true',
 '-mc', 1000, '-df', '-nodead', '-expcterr', 'invariant_violation',
 '-p', 'CLPFD', 'TRUE'],
   'ClearSy_RuleTMS_R0006_CLPFD Codespeed Test').
 cli_testcase(1754, [cbc,sigma], [
 '../prob_examples/public_examples/B/CBC/Tests/TestSIGMA.mch',
 '-init', '-assertions'
],
   'Check reification of IF-THEN-ELSE in SIGMA + WD detection of total avl function').
cli_testcase(1755, [tickets,quantified], [  '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
  '-evalt', 'UNION(x).(x:1..4|{x+y}) = 12..15',
  '-evalf', 'UNION(x).(x:1..4|{x+y}) = 12..16',
  '-evalt', '{y|UNION(x).(x:1..4|{x+y}) = 12..15} = {11}',
  '-evalt', '{n|UNION(x).(x:1..10|{x}) = 1..n} = {10}'
  % UNION(x).(x:1..n|{x}) = 1..10  generates warning
  % {n|UNION(x).(x:1..n|{x}) = 1..0} = {0}  generates virtual time-out
  ], 'ensure new UNION treatment with better propagation').
cli_testcase(1756, [tickets], [
 '../prob_examples/public_examples/B/Tester/SetSetupTests/CardDefSetConstraint.mch',
 '-init', '-assertions'
],
   'Check deferred set size inference improved').
cli_testcase(1757, [cbc], [  '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
  '-evalt', '{x,y|IF x:0..3 THEN y=2 ELSE 1=0 END} = {(0|->2),(1|->2),(2|->2),(3|->2)}',
  '-evalt', '{x,y|IF x/:0..3 THEN 1=0 ELSE y=2 END} = {(0|->2),(1|->2),(2|->2),(3|->2)}'
  ], 'exercise => FALSE AST simplification').
cli_testcase(1758, [b_test,ltl,safety], [
 '../prob_examples/public_examples/B/LTL/SafetyMC/BeverageVendingMachine.mch',
 '-ltlformulaf', 'G not [bget]',
 '-p', 'use_safety_ltl_model_checker', true, '-strict'
 ],
 'Check that safety model checker does not produce incomplete counter examples when transition propositions appear in the safety ltl formula.').
cli_testcase(1759, [b_test,ltl,safety], [
 '../prob_examples/public_examples/B/LTL/SafetyMC/BeverageVendingMachine.mch',
 '-ltlformulaf', 'G not ({nsoda = 0} => X [sget])',
 '-p', 'use_safety_ltl_model_checker', true, '-strict'
 ],
 'Check that safety model checker does not produce incomplete counter examples when transition propositions appear in the safety ltl formula.').
cli_testcase(1760, [b_test,ltl,safety], [
 '../prob_examples/public_examples/B/LTL/SafetyMC/BeverageVendingMachine.mch',
 '-ltlformulaf', '{paid=FALSE} or ({nbeer=1} & {nsoda =1}) & (({nbeer = 0}) R (X [bget]))',
 '-p', 'use_safety_ltl_model_checker', true, '-strict'
 ],'Check that safety model checker does not produce incomplete counter examples when transition propositions appear in the safety ltl formula.').
cli_testcase(1761, [b_test,ltl,safety], [
 '../prob_examples/public_examples/B/LTL/SafetyMC/BeverageVendingMachine.mch',
 '-ltlassertions', '-p', 'use_safety_ltl_model_checker', true, '-strict'
 ],'Checking that the safety LTL model checker succeeds on each assertions.').
cli_testcase(1762, [b_test,ltl,safety], [
 '../prob_examples/public_examples/B/LTL/SafetyMC/BeverageVendingMachine.mch',
 '-ltlformulaf', 'F  [ret_coin] => {paid=TRUE}', '-p', 'use_safety_ltl_model_checker', true, '-strict'
 ],'Check that safety model checker does not produce incomplete counter examples when transition propositions appear in the safety ltl formula.').
cli_testcase(1763, [cbc,slot,tickets,wd], ['../prob_examples/public_examples/B/Tickets/SLOT-WD-ISSUE/Solver.mch',
 '-assertions',
 '-evalt', 'UNSAT_CORE_SESSIONS({group3, group4}) = {session3,session4}'
   ], 'Check that wd issue with let predicates solved (expressions not guarded)').
cli_testcase(1764, [laws], [ '-p', 'STRICT_RAISE_ENUM_WARNINGS', 'TRUE',
  '../prob_examples/public_examples/B/Laws/BoolLaws.mch',
  '../prob_examples/public_examples/B/Laws/ArithmeticLaws.mch',
  '../prob_examples/public_examples/B/Laws/SetLaws.mch',
  '../prob_examples/public_examples/B/Laws/SetLawsNat.mch',
'-mc', '500', '-c', '-nogoal', '-nodead', '-strict', '-p', 'OPTIMIZE_AST', 'FALSE'], 'Mathematical Laws with AST optimization disabled').
cli_testcase(1765, [cbc], ['../prob_examples/examples/B/AWE_JASP/v7/JavaBC0.mch',
  '-t', '-mc', '200',
  '-nogoal', '-noass'
   ], 'Check Java Bytecode interpreter').
cli_testcase(1766, [private,case], ['../prob_examples/public_examples/B/Tester/ForAllExpansionEnumWarning.mch',
  '-init', '-aa', 0,0,1
   ], 'Check forall not expanded').
cli_testcase(1767, [b_test,case], ['../prob_examples/public_examples/B/Tester/CaseStmt_v2.mch', '-t', '-strict'],'Check case statement with multiple values for same case').
cli_testcase(1768, [b_test,case], ['../prob_examples/public_examples/B/Tester/CaseStmt_Illegal1.mch', '-init',
 '-expcterrpos', type_error, 24, 10, '-expcterr', load_main_file,
 '-strict'],'Check case statement with else').
cli_testcase(1769, ['PROB-411',tickets,kodkod], [ '-p', 'KODKOD_RAISE_WARNINGS', 'TRUE',
  '-evalt', ':kodkod card({a,b|a:BOOL & b:1..k})=4 & k:0..2',
  '-evalf', ':kodkod card({a,b|a:BOOL & b:1..k})=3 & k:0..10',
  '-evalt', ':kodkod card({a,b,c})=1 & {a,b,c}<:BOOL',
  '-evalt', ':kodkod card({a,b,c})=2 & {a,b,c}<:1..10 & a+b+c=20'
  ], 'check card interval analysis succeeds').
cli_testcase(1770,[cbc,enabling],[
'-enabling_analysis_csv','../prob_examples/public_examples/B/CBC/Enabling/ChoiceByPredicate_enabling_analysis.csv',
'../prob_examples/public_examples/B/CBC/Enabling/ChoiceByPredicate.mch',
'-p', 'TIME_OUT', 1000
], 'Check CSV enabling output for becomes_such').
cli_testcase(1771,[cbc,enabling],[
'-enabling_analysis_csv','../prob_examples/public_examples/B/CBC/Enabling/OpParameterTests_enabling_analysis.csv',
'../prob_examples/public_examples/B/CBC/Enabling/OpParameterTests.mch',
'-p', 'TIME_OUT', 2000
], 'Check CSV enabling output for variety of substitutions').
cli_testcase(1772, [por,ltl], ['../prob_examples/public_examples/B/Ivo/Threads51.mch', '-ltlformulat', 'GF {v1=n}',
  '-p', por, ample_sets], 'Check LTL property with partial order reduction.').
cli_testcase(1773, [por,ltl], ['../prob_examples/public_examples/B/Ivo/Threads51.mch', '-ltlformulat', 'G not {v1>n+1}',
  '-p', use_safety_ltl_model_checker, true, '-p', por, ample_sets], 'Check LTL property with partial order reduction.').
cli_testcase(1774, [por,ltl], ['../prob_examples/public_examples/B/Benchmarks/CAN_BUS_tlc.mch', 
  '-ltlformulat', 'G not deadlock', '-p', por, ample_sets, '-p', pge, full
  ], 'Check LTL property with partial order reduction.').
cli_testcase(1775, [por,ltl], ['../prob_examples/public_examples/B/Benchmarks/CAN_BUS_tlc.mch', 
  '-ltlformulat', 'G not deadlock', '-p', por, ample_sets, '-p', pge, full,
  '-p', use_safety_ltl_model_checker, true
  ], 'Check LTL property with partial order reduction.').

/*
cli_testcase(1693, [tickets], [
  %'-evalt', '{x|x:1..n} = {y,y+1}' % does not work yet
  ], 'test comprehension set propagation').
*/


% /Users/leuschel/git_root/prob_examples/public_examples/B/Simple/UpCounter.mch -ctlformula 'AG {c<mx}' -p MAXINT 10
% store counter example

% sicstus -l src/test_runner.pl --goal "id(690-694)."
% INSERT NEW TESTCASES ABOVE THIS LINE; cli_testcase is sorted in ascending order
% ----------------------------------------------------------------------

% ----------------------------------------------------------------------
% ----------------------------------------------------------------------

% ----------------------------------------------------------------------
% IF A NEW TESTCASE REQUIRES CHECKING OUTPUT, ADD NEW LINE(S) BELOW
% cli_testcase_diff_check_output is sorted in descending order

cli_testcase_diff_check_output(1771,'../prob_examples/public_examples/B/CBC/Enabling/OpParameterTests_enabling_analysis.csv', '../prob_examples/public_examples/B/CBC/Enabling/OpParameterTests_enabling_analysis_saved.csv').
cli_testcase_diff_check_output(1770,'../prob_examples/public_examples/B/CBC/Enabling/ChoiceByPredicate_enabling_analysis.csv', '../prob_examples/public_examples/B/CBC/Enabling/ChoiceByPredicate_enabling_analysis_saved.csv').
cli_testcase_diff_check_output(1703,'../prob_examples/public_examples/EventBPrologPackages/Advance/CAN_Bus/CB3FSMM_mch_v2_enabling_analysis.csv', '../prob_examples/public_examples/EventBPrologPackages/Advance/CAN_Bus/CB3FSMM_mch_v2_enabling_analysis_saved.csv').
cli_testcase_diff_check_output(1702,'../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_enabling_analysis.csv', '../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_enabling_analysis_saved.csv').
cli_testcase_diff_check_output(1676,'../prob_examples/examples/EventBPrologPackages/Soton/MCM_Tests/tests.xml', '../prob_examples/examples/EventBPrologPackages/Soton/MCM_Tests/FENCEM7_mcm_tests_1675_saved.xml').
cli_testcase_diff_check_output(1675,'../prob_examples/public_examples/B/TestCaseGeneration/MCM_Tests/SimpleTestModel_mcm_tests.xml', '../prob_examples/public_examples/B/TestCaseGeneration/MCM_Tests/SimpleTestModel_mcm_tests_saved.xml').
cli_testcase_diff_check_output(1665,'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/Records_internal.mch',
                               '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/Records_internal_saved.mch').
cli_testcase_diff_check_output(1543,'../prob_examples/public_examples/B/Tickets/Beauger2/bug_not_labels_corrected_ppf.mch','../prob_examples/public_examples/B/Tickets/Beauger2/bug_not_labels_corrected_ppf_saved.mch').
cli_testcase_diff_check_output(1494,'../prob_examples/public_examples/B/Ivo/NoDisablings_mx5_Coverage.txt','../prob_examples/public_examples/B/Ivo/NoDisablings_mx5_Coverage_saved.txt').
cli_testcase_diff_check_output(1493,'../prob_examples/public_examples/B/Ivo/SkippingComplexGuardsEvaluation_mch_mx5000_Coverage.txt','../prob_examples/public_examples/B/Ivo/SkippingComplexGuardsEvaluation_mch_mx5000_Coverage_saved.txt').
cli_testcase_diff_check_output(1451,'../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result.txt','../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result_true.txt').
cli_testcase_diff_check_output(1439,'../prob_examples/public_examples/B/Tickets/LambdaInverse/LambdaInverseBehaviour.sptxt','../prob_examples/public_examples/B/Tickets/LambdaInverse/LambdaInverseBehaviour_saved.sptxt').
cli_testcase_diff_check_output(1405,'../prob_examples/public_examples/B/Simple/VerySimpleDFA_transition_diagram.dot','../prob_examples/public_examples/B/Simple/VerySimpleDFA_transition_diagram_saved.dot').
cli_testcase_diff_check_output(1402,'../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result.txt','../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result_true.txt').
cli_testcase_diff_check_output(1389,'../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result.txt','../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result_defset.txt').
cli_testcase_diff_check_output(1388,'../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result.txt','../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result_true.txt').
cli_testcase_diff_check_output(1387,'../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result.txt','../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result_enumwarning.txt').
cli_testcase_diff_check_output(1386,'../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result.txt','../prob_examples/examples/B/Alstom/DisproverProject/ProB_Result_false.txt').
cli_testcase_diff_check_output(1383,'../prob_examples/examples/B/Alstom/Ticket_Dec08_2014/Segment_refAdhesionLevel_GER5_sptxt.txt','../prob_examples/examples/B/Alstom/Ticket_Dec08_2014/Segment_refAdhesionLevel_GER5_sptxt_saved.txt').
cli_testcase_diff_check_output(1378,'../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_SIGAREA_0024_ori.his','../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_SIGAREA_0024_ori_saved.his').
cli_testcase_diff_check_output(1377,'../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_2_ori.his','../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_2_ori_saved.his').
cli_testcase_diff_check_output(1377,'../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_2_ori.sptxt','../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_2_ori_saved.sptxt').
% the provided counter-example varies for test 1366:
%cli_testcase_diff_check_output(1366,'../prob_examples/public_examples/B/SymbolicModelChecking/TimingExampleSimpler_v2_his.txt',
%                                    '../prob_examples/public_examples/B/SymbolicModelChecking/TimingExampleSimpler_v2_his_saved.txt').
cli_testcase_diff_check_output(1364,'../prob_examples/public_examples/B/Demo/scheduler_his.txt',
                                    '../prob_examples/public_examples/B/Demo/scheduler_his_ctl_saved.txt').
cli_testcase_diff_check_output(1363,'../prob_examples/public_examples/B/Simple/UpCounter_his.txt',
                                    '../prob_examples/public_examples/B/Simple/UpCounter_his_ctl20_saved.txt').
cli_testcase_diff_check_output(1360,'../prob_examples/public_examples/B/CBC/FeasibilityOperationCheck_feasibility.csv',
                               '../prob_examples/public_examples/B/CBC/FeasibilityOperationCheck_feasibility_saved.csv').
cli_testcase_diff_check_output(1358,'../prob_examples/examples/EventBPrologPackages/Savary/R08_ByteArray_enum_partition_feasibility.csv',
                               '../prob_examples/examples/EventBPrologPackages/Savary/R08_ByteArray_enum_partition_feasibility_saved.csv').
cli_testcase_diff_check_output(1357,'../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_cbc_tests.xml',
                               '../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_cbc_tests_saved.xml').
cli_testcase_diff_check_output(1341,'../prob_examples/public_examples/B/Demo/Bakery0_all_paths.xml',
                               '../prob_examples/public_examples/B/Demo/Bakery0_all_paths_saved.xml').
cli_testcase_diff_check_output(1339,'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/GenericUsageOfSquare_internal.mch',
                               '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/GenericUsageOfSquare_internal_saved.mch').
cli_testcase_diff_check_output(1319,'../prob_examples/public_examples/B/Demo/scheduler_bf_states.P','../prob_examples/public_examples/B/Demo/scheduler_bf_saved_states.P').
cli_testcase_diff_check_output(1271,'../prob_examples/public_examples/B/Tickets/Hansen10/PrettyPrintBug_his.txt','../prob_examples/public_examples/B/Tickets/Hansen10/PrettyPrintBug_his_saved.txt').
cli_testcase_diff_check_output(1196,'../prob_examples/public_examples/B/Benchmarks/CarlaTravelAgencyErr.csv','../prob_examples/public_examples/B/Benchmarks/CarlaTravelAgencyErr_saved.csv').
cli_testcase_diff_check_output(1148,'../prob_examples/public_examples/B/Ivo/BenchmarksPOR/Peterson/peterson_R_Coverage_POR.txt','../prob_examples/public_examples/B/Ivo/BenchmarksPOR/Peterson/peterson_R_Coverage_saved.txt').
cli_testcase_diff_check_output(1147,'../prob_examples/public_examples/B/Ivo/BenchmarksPOR/Peterson/em_abs_Coverage_POR.txt','../prob_examples/public_examples/B/Ivo/BenchmarksPOR/Peterson/em_abs_Coverage_saved.txt').
cli_testcase_diff_check_output(1144,'../prob_examples/public_examples/B/Ivo/BenchmarksPOR/BPEL2B/Model_Machine_4_NoDeadlock_v5_Coverage_POR.txt','../prob_examples/public_examples/B/Ivo/BenchmarksPOR/BPEL2B/Model_Machine_4_NoDeadlock_v5_Coverage_saved.txt').
cli_testcase_diff_check_output(1119,'../prob_examples/public_examples/CSPB/Tickets/Bauer/machine_his.txt','../prob_examples/public_examples/CSPB/Tickets/Bauer/machine_his_saved.txt').
cli_testcase_diff_check_output(1116,'../prob_examples/public_examples/EventBPrologPackages/Tests/UnicodeIds_spdot.dot','../prob_examples/public_examples/EventBPrologPackages/Tests/UnicodeIds_spdot_saved.dot').
cli_testcase_diff_check_output(1045,'../prob_examples/examples/RodinModels/Deploy/Bosch/CruiseControl/CrsCtl_m4_Coverage_pge.txt','../prob_examples/examples/RodinModels/Deploy/Bosch/CruiseControl/CrsCtl_m4_Coverage.txt').
cli_testcase_diff_check_output(1044,'../prob_examples/public_examples/B/EventB/EventB_Projekt/lift_solution_Coverage_pge.txt','../prob_examples/public_examples/B/EventB/EventB_Projekt/lift_solution_Coverage.txt').
cli_testcase_diff_check_output(1043,'../prob_examples/public_examples/B/Benchmarks/CAN_BUS_tlc.cov','../prob_examples/public_examples/B/Benchmarks/CAN_BUS_tlc_saved.cov').
cli_testcase_diff_check_output(1033,'../prob_examples/public_examples/B/Benchmarks/scheduler_transition_diagram.dot','../prob_examples/public_examples/B/Benchmarks/scheduler_transition_diagram_saved.dot').
cli_testcase_diff_check_output(989,'../prob_examples/public_examples/EventBPrologPackages/PerformanceTests/LargeConstants_ctx_internal.mch','../prob_examples/public_examples/EventBPrologPackages/PerformanceTests/LargeConstants_ctx_internal_saved.mch').
cli_testcase_diff_check_output(988,'../prob_examples/public_examples/CSP/simple/Copy_spmdot.dot','../prob_examples/public_examples/CSP/simple/Copy_spmdot_saved.dot').
cli_testcase_diff_check_output(988,'../prob_examples/public_examples/CSP/simple/Copy_spddot.dot','../prob_examples/public_examples/CSP/simple/Copy_spddot_saved.dot').
cli_testcase_diff_check_output(987,'../prob_examples/public_examples/EventBPrologPackages/Advance/CAN_Bus/CB3FSMM_mch_v2_por.cov','../prob_examples/public_examples/EventBPrologPackages/Advance/CAN_Bus/CB3FSMM_mch_v2.cov').
cli_testcase_diff_check_output(977,'../prob_examples/public_examples/B/Ivo/BridgeTransitions_por.cov','../prob_examples/public_examples/B/Ivo/BridgeTransitions.cov').
cli_testcase_diff_check_output(982,'../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_1_ori.his','../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_1_ori_saved.his').
cli_testcase_diff_check_output(981,'../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_Route_0001ori.his','../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_Route_0001ori_saved.his').
cli_testcase_diff_check_output(980,'../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_SIGAREA_0024_ori.his','../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_DB_SIGAREA_0024_ori_saved.his').
cli_testcase_diff_check_output(979,'../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_2_ori.his','../prob_examples/examples/B/Alstom/CompilatonProject/Regles/Rule_TMS_Overlap_0017_2_ori_saved.his').
cli_testcase_diff_check_output(957,'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/DifferentUnitNotations_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/DifferentUnitNotations_internal_saved.mch').
cli_testcase_diff_check_output(944,'../prob_examples/public_examples/Eval/evalout.dot','../prob_examples/public_examples/Eval/evalout_saved.dot').
cli_testcase_diff_check_output(942,'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/Case_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/Case_internal_saved.mch').
cli_testcase_diff_check_output(894,'../prob_examples/public_examples/B/Daniel/minienabling_analysis.csv',
                               '../prob_examples/public_examples/B/Daniel/minienabling_analysis_saved.csv').
cli_testcase_diff_check_output(881,'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/Choice_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/Choice_internal_saved.mch').
cli_testcase_diff_check_output(865,'../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/TestingUnicodeFeatures_no_unicode.csp',
'../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/TestingUnicodeFeatures_no_unicode_saved.csp').
cli_testcase_diff_check_output(864,'../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/TestingUnicodeFeatures_unicode.csp',
'../prob_examples/public_examples/CSP/mydemos/OtherCoverageTests/TestingUnicodeFeatures_unicode_saved.csp').
cli_testcase_diff_check_output(808,'../prob_examples/public_examples/B/Benchmarks/CarlaTravelAgencyErr_state_graph.dot',
'../prob_examples/public_examples/B/Benchmarks/CarlaTravelAgencyErr_state_graph_saved.dot').
cli_testcase_diff_check_output(767,'../prob_examples/examples/B/Siemens/BAC9_ZC/pas_as_env_inv_ld_sptxt.txt',
'../prob_examples/examples/B/Siemens/BAC9_ZC/pas_as_env_inv_ld_sptxt_saved.txt').
cli_testcase_diff_check_output(707,'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m3_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m3_internal_saved.mch').
cli_testcase_diff_check_output(706,'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m2_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m2_internal_saved.mch').
cli_testcase_diff_check_output(705,'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m1_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m1_internal_saved.mch').
cli_testcase_diff_check_output(704,'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m0_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_nuclear/C_m0_internal_saved.mch').
cli_testcase_diff_check_output(701,'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_train/T_m1_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_train/T_m1_internal_saved.mch').
cli_testcase_diff_check_output(700,'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_train/T_m0_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/Abrial_Hybrid/hybrid_train/T_m0_internal_saved.mch').
cli_testcase_diff_check_output(696,'../prob_examples/public_examples/B/PragmasUnits/InferenceTests/InvolvedConstraintUnitsUnsat_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/InferenceTests/InvolvedConstraintUnitsUnsat_internal_saved.mch').
cli_testcase_diff_check_output(695,'../prob_examples/public_examples/B/PragmasUnits/InferenceTests/InvolvedConstraintUnits_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/InferenceTests/InvolvedConstraintUnits_internal_saved.mch').
cli_testcase_diff_check_output(686,'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/SequenceOperations_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/SequenceOperations_internal_saved.mch').
cli_testcase_diff_check_output(683,'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/AdditionConversion_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/AdditionConversion_internal_saved.mch').
cli_testcase_diff_check_output(681,'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/IntegerOperations_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/IntegerOperations_internal_saved.mch').
cli_testcase_diff_check_output(680,'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/FuelLevel/fuel_i_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/FuelLevel/fuel_i_internal_saved.mch').
cli_testcase_diff_check_output(679,'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/FuelLevel/fuel0_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/FuelLevel/fuel0_internal_saved.mch').
cli_testcase_diff_check_output(678,'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/FuelLevel/ctx_i_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/FuelLevel/ctx_i_internal_saved.mch').
cli_testcase_diff_check_output(677,'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/FuelLevel/ctx_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/FuelLevel/ctx_internal_saved.mch').
cli_testcase_diff_check_output(676,'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/UserDefinedAlias_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/UserDefinedAlias_internal_saved.mch').
cli_testcase_diff_check_output(675,'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/IfTypeAndClash_internal.mch',
'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/IfTypeAndClash_internal_saved.mch').
cli_testcase_diff_check_output(673,'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/IrregularConversions_internal.mch',
 '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/IrregularConversions_internal_saved.mch').
cli_testcase_diff_check_output(667,'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/System_internal.mch',
 '../prob_examples/public_examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/System_internal_saved.mch').
cli_testcase_diff_check_output(666,'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/TrafficLight_internal.mch',
 '../prob_examples/public_examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/TrafficLight_internal_saved.mch').
cli_testcase_diff_check_output(665,'../prob_examples/public_examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/Car_internal.mch',
 '../prob_examples/public_examples/B/PragmasUnits/CaseStudies/TrafficLightWarning/Car_internal_saved.mch').
cli_testcase_diff_check_output(662,'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/KnownExponent_internal.mch',
 '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/KnownExponent_internal_saved.mch').
cli_testcase_diff_check_output(661,'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/MultiplicationConversion_internal.mch',
 '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/MultiplicationConversion_internal_saved.mch').
cli_testcase_diff_check_output(657,'../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/SimpleCorrectUsage_internal.mch',
 '../prob_examples/public_examples/B/PragmasUnits/InternalRepresentationTests/SimpleCorrectUsage_internal_saved.mch').
cli_testcase_diff_check_output(634,'../prob_examples/public_examples/B/Kodkod/Loop_statespace.dot',
 '../prob_examples/public_examples/B/Kodkod/Loop_statespace_saved.dot').
cli_testcase_diff_check_output(616,'../prob_examples/public_examples/B/Tester/Partition_internal.mch',
 '../prob_examples/public_examples/B/Tester/Partition_internal_saved.mch').
cli_testcase_diff_check_output(613,'../prob_examples/public_examples/B/Pragmas/AssertionPragmas_internal.mch',
 '../prob_examples/public_examples/B/Pragmas/AssertionPragmas_internal_saved.mch').
cli_testcase_diff_check_output(540,'../prob_examples/public_examples/B/TestCaseGeneration/SimpleCounter_results.xml',
 '../prob_examples/public_examples/B/TestCaseGeneration/SimpleCounter_results_saved.xml').

cli_testcase_diff_check_output(514,'../prob_examples/public_examples/B/Mathematical/Voting/dot_output/NoCondorcetWinnerKodkod_A0_true.dot','../prob_examples/public_examples/B/Mathematical/Voting/dot_output/NoCondorcetWinnerKodkod_A0_true_saved.dot').
cli_testcase_diff_check_output(514,'../prob_examples/public_examples/B/Mathematical/Voting/dot_output/NoCondorcetWinnerKodkod_A1_true.dot','../prob_examples/public_examples/B/Mathematical/Voting/dot_output/NoCondorcetWinnerKodkod_A1_true_saved.dot').
cli_testcase_diff_check_output(514,'../prob_examples/public_examples/B/Mathematical/Voting/dot_output/NoCondorcetWinnerKodkod_A2_true.dot','../prob_examples/public_examples/B/Mathematical/Voting/dot_output/NoCondorcetWinnerKodkod_A2_true_saved.dot').
cli_testcase_diff_check_output(514,'../prob_examples/public_examples/B/Mathematical/Voting/dot_output/NoCondorcetWinnerKodkod_A3_true.dot','../prob_examples/public_examples/B/Mathematical/Voting/dot_output/NoCondorcetWinnerKodkod_A3_true_saved.dot').


cli_testcase_diff_check_output(584,'../prob_examples/public_examples/TLA/ErrorTests/GraphIso_tla.mch', '../prob_examples/public_examples/TLA/ErrorTests/GraphIso_tla_saved.mch').
cli_testcase_diff_check_output(563,'../prob_examples/public_examples/B/Daniel/TestM_testcases.xml', '../prob_examples/public_examples/B/Daniel/TestM_testcases_saved.xml').
cli_testcase_diff_check_output(562,'../prob_examples/public_examples/EventBPrologPackages/Pitesti/bank_testfile.xml','../prob_examples/public_examples/EventBPrologPackages/Pitesti/bank_testfile_saved.xml').
cli_testcase_diff_check_output(515,'../prob_examples/public_examples/B/Demo/scheduler_spdot.dot','../prob_examples/public_examples/B/Demo/scheduler_spdot_saved.dot').
cli_testcase_diff_check_output(515,'../prob_examples/public_examples/B/Demo/scheduler_spmdot.dot','../prob_examples/public_examples/B/Demo/scheduler_spmdot_saved.dot').
cli_testcase_diff_check_output(515,'../prob_examples/public_examples/B/Demo/scheduler_spddot.dot','../prob_examples/public_examples/B/Demo/scheduler_spddot_saved.dot').
cli_testcase_diff_check_output(510,'../prob_examples/public_examples/B/Laws/RelLaws_internal.mch','../prob_examples/public_examples/B/Laws/RelLaws_internal_saved.mch').
cli_testcase_diff_check_output(500,'../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_pp.mch','../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_prettyprintedf.mch').
cli_testcase_diff_check_output(499,'../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_pp.mch','../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_prettyprintedf.mch').
cli_testcase_diff_check_output(498,'../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_pp.mch','../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_pp_saved.mch').
cli_testcase_diff_check_output(497,'../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_pp.mch','../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_prettyprinted.mch').
cli_testcase_diff_check_output(451,'../prob_examples/public_examples/B/Simple/UpCounter_his.txt','../prob_examples/public_examples/B/Simple/UpCounter_his_saved.txt').
cli_testcase_diff_check_output(452,'../prob_examples/public_examples/B/Simple/UpCounter_his.txt','../prob_examples/public_examples/B/Simple/UpCounter_his_saved2.txt').
cli_testcase_diff_check_output(453,'../prob_examples/public_examples/B/Simple/UpCounter_his.txt','../prob_examples/public_examples/B/Simple/UpCounter_his_saved3.txt').
cli_testcase_diff_check_output(454,'../prob_examples/public_examples/B/Simple/UpCounter_his.txt','../prob_examples/public_examples/B/Simple/UpCounter_his_saved4.txt').
cli_testcase_diff_check_output(455,'../prob_examples/public_examples/B/Simple/UpCounter_sptxt.txt','../prob_examples/public_examples/B/Simple/UpCounter_sptxt_saved.txt').
cli_testcase_diff_check_output(461,'../prob_examples/public_examples/B/Daniel/rel_fnc_his.txt','../prob_examples/public_examples/B/Daniel/rel_fnc_his_saved.txt').

cli_testcase_diff_check_output(462,'../prob_examples/public_examples/B/Simple/UpCounter_his.txt','../prob_examples/public_examples/B/Simple/UpCounter_his_saved3.txt').
cli_testcase_diff_check_output(463,'../prob_examples/public_examples/B/CBC/SimpleDeadlock1_his.txt','../prob_examples/public_examples/B/CBC/SimpleDeadlock1_his_saved.txt'). % FAILS with old write_history
cli_testcase_diff_check_output(464,'../prob_examples/public_examples/B/CBC/ProcessIDStarvation_InvErr_his.txt','../prob_examples/public_examples/B/CBC/ProcessIDStarvation_InvErr_his_saved.txt').
cli_testcase_diff_check_output(466,'../prob_examples/public_examples/B/PerformanceTests/WhileLoopBench_his.txt','../prob_examples/public_examples/B/PerformanceTests/WhileLoopBench_saved_his.txt').
cli_testcase_diff_check_output(467,'../prob_examples/public_examples/B/FeatureChecks/IFTHENELSE_his.txt','../prob_examples/public_examples/B/FeatureChecks/IFTHENELSE_saved_his.txt').
cli_testcase_diff_check_output(468,'../prob_examples/public_examples/B/Puzzles/Farmer_sptxt.txt','../prob_examples/public_examples/B/Puzzles/Farmer_saved_sptxt.txt').
cli_testcase_diff_check_output(468,'../prob_examples/public_examples/B/Puzzles/Farmer_his.txt','../prob_examples/public_examples/B/Puzzles/Farmer_saved_his.txt').
cli_testcase_diff_check_output(469,'../prob_examples/public_examples/B/PerformanceTests/LargeSets/LargeExplicitSet_sptxt.txt','../prob_examples/public_examples/B/PerformanceTests/LargeSets/LargeExplicitSet_saved_sptxt.txt').
cli_testcase_diff_check_output(470,'../prob_examples/public_examples/EventBPrologPackages/Abrial_Train_Ch17/train_4_mch_lukas_his.txt','../prob_examples/public_examples/EventBPrologPackages/Abrial_Train_Ch17/train_4_mch_lukas_saved_his.txt').
cli_testcase_diff_check_output(470,'../prob_examples/public_examples/EventBPrologPackages/Abrial_Train_Ch17/train_4_mch_lukas_sptxt.txt','../prob_examples/public_examples/EventBPrologPackages/Abrial_Train_Ch17/train_4_mch_lukas_saved_sptxt.txt').
cli_testcase_diff_check_output(471,'../prob_examples/public_examples/B/Benchmarks/CarlaTravelAgencyErr_sptxt.txt','../prob_examples/public_examples/B/Benchmarks/CarlaTravelAgencyErr_saved_sptxt.txt').
cli_testcase_diff_check_output(471,'../prob_examples/public_examples/B/Benchmarks/CarlaTravelAgencyErr_his.txt','../prob_examples/public_examples/B/Benchmarks/CarlaTravelAgencyErr_saved_his.txt').
cli_testcase_diff_check_output(472,'../prob_examples/public_examples/B/SchneiderBook/Chapter12/TownsRRR_sptxt.txt','../prob_examples/public_examples/B/SchneiderBook/Chapter12/TownsRRR_saved_sptxt.txt').
cli_testcase_diff_check_output(472,'../prob_examples/public_examples/B/SchneiderBook/Chapter12/TownsRRR_his.txt','../prob_examples/public_examples/B/SchneiderBook/Chapter12/TownsRRR_saved_his.txt').
cli_testcase_diff_check_output(473,'../prob_examples/public_examples/B/NewSyntax/SET_Game_Rec_sptxt.txt','../prob_examples/public_examples/B/NewSyntax/SET_Game_Rec_saved_sptxt.txt').
cli_testcase_diff_check_output(473,'../prob_examples/public_examples/B/NewSyntax/SET_Game_Rec_his.txt','../prob_examples/public_examples/B/NewSyntax/SET_Game_Rec_saved_his.txt').
cli_testcase_diff_check_output(474,'../prob_examples/public_examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A0_true.dot','../prob_examples/public_examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A0_true_saved.dot').
cli_testcase_diff_check_output(474,'../prob_examples/public_examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A1_false.dot','../prob_examples/public_examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A1_false_saved.dot').
cli_testcase_diff_check_output(474,'../prob_examples/public_examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A2_false.dot','../prob_examples/public_examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A2_false_saved.dot').
cli_testcase_diff_check_output(474,'../prob_examples/public_examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A3_false.dot','../prob_examples/public_examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A3_false_saved.dot').
cli_testcase_diff_check_output(474,'../prob_examples/public_examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A4_false.dot','../prob_examples/public_examples/B/ExpressionViewer/dot_output/TotFunAnalyseTest_A4_false_saved.dot').
cli_testcase_diff_check_output(476,'../prob_examples/public_examples/B/Tickets/Mejia1/TestInit_statespace.dot','../prob_examples/public_examples/B/Tickets/Mejia1/TestInit_statespace_saved.dot').
cli_testcase_diff_check_output(479,'../prob_examples/public_examples/B/Puzzles/WhoKilledAgatha_statespace.dot','../prob_examples/public_examples/B/Puzzles/WhoKilledAgatha_statespace_saved.dot').
cli_testcase_diff_check_output(480,'../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_statespace.dot','../prob_examples/public_examples/B/Benchmarks/Cruise_finite1_statespace_saved.dot').
cli_testcase_diff_check_output(383,'../prob_examples/public_examples/B/Puzzles/SudokuHexAsConstant_sptxt.txt','../prob_examples/public_examples/B/Puzzles/SudokuHexAsConstant_sptxt_saved.txt').

% specify that a testcase output file should not be deleted before running the tests
cli_testcase_do_not_delete(584,'../prob_examples/public_examples/TLA/ErrorTests/GraphIso_tla.mch').

% declaration of additinoal testcase files; in order to be able to generate archives
%  of standalone tests
extra_testcase_file(305,'../prob_examples/public_examples/B/Laws/REPL_UNIT_TESTS0.def').
extra_testcase_file(306,'../prob_examples/public_examples/B/Laws/REPL_UNIT_TESTS1.def').
extra_testcase_file(307,'../prob_examples/public_examples/B/Laws/REPL_UNIT_TESTS.def').
extra_testcase_file(465,'../prob_examples/public_examples/B/Simple/UpCounter_saved10.P').
extra_testcase_file(472,'../prob_examples/public_examples/B/SchneiderBook/Chapter12/TownsRR.ref').
extra_testcase_file(472,'../prob_examples/public_examples/B/SchneiderBook/Chapter12/TownsR.ref').
extra_testcase_file(472,'../prob_examples/public_examples/B/SchneiderBook/Chapter12/Towns.mch').
extra_testcase_file(475,'../prob_examples/public_examples/B/Laws/REPL_UNIT_TESTS2.def').

% use
%    sicstus -l src/test_runner.pl --goal "run_last_test, halt ; halt."
% or
%    make runlast
% to just run last test case
% use
%    sicstus -l src/test_runner.pl --goal "id(ID), halt ; halt."
% to run a specific test
